#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include "shared.h"
#include "fmath\fmath.h"
#include "3d.h"
#include "version.h"
#include "compiled.h"

char typenames[4][20]={"WORLD","3DOBJECT","LIGHTSOURCE","CAMERA"};

camera3DE *cam;
light3DE *lsrc;
object3DE *world;
int nbfv,totalfaces=0,pseudo3DE;
depth *zdepth;

void init_3de(void){ //  caller APRES avoir loader TOUT les objets de la scne
   zdepth=(depth *)malloc(totalfaces*sizeof(depth));
   cam=malloc(sizeof(camera3DE));
   cam->type=CAMERA;
   lsrc=malloc(sizeof(light3DE));
   lsrc->type=LIGHT;
   world=init_object(0,0,0,0,0,MASTER);
   if (zdepth==NULL||cam==NULL||lsrc==NULL||world==NULL)
       out("init_3de(): out of memory",2);
}

void out(char *msg,int action){
   printf("%s\n",msg);
   switch(action){
       case 1: getch(); break;
       case 2: exit(1);
   }
}

void printinfo_3de(void){
   printf("3DE v%1i.%02i compiled on %s at %s\n",_3DE_VERHI,_3DE_VERLO,WCC_DATE,WCC_TIME);
}

object3DE *init_object(int nbpt,int nbfc,int px,int py,int pz,int type){
   object3DE *o;
   int i;
   o=(object3DE *)malloc(sizeof(object3DE));
   if (o==NULL) return(NULL);
   setmem(o,0,sizeof(object3DE));
   o->type=type;
   o->nbpt=nbpt;
   o->nbfc=nbfc;
   if (nbpt&&nbfc){
       o->pts=(vect *)malloc(o->nbpt*sizeof(vect));
       o->ptsrot=(vect *)malloc(o->nbpt*sizeof(vect));
       o->proj=(vect2D *)malloc(o->nbpt*sizeof(vect2D));
       o->pnorms=(vect *)malloc(o->nbpt*sizeof(vect));
       o->pnormsrot=(vect *)malloc(o->nbpt*sizeof(vect));
       o->norms=(vect *)malloc(o->nbfc*sizeof(vect));
       o->normsrot=(vect *)malloc(o->nbfc*sizeof(vect));
       o->fc=(face3DE *)malloc(o->nbfc*sizeof(face3DE));
       if (o->pts==NULL||o->ptsrot==NULL||o->norms==NULL||o->normsrot==NULL||
           o->pnorms==NULL||o->pnormsrot==NULL) return(NULL);
       setmem(o->fc,0,o->nbfc*sizeof(face3DE));
   }        
   setvector(o->pos,px,py,pz);
   setmatrix(&o->sysaxe.matrice,65536,0,0,0,65536,0,0,0,65536);
// stuff link
   o->level=0;
   o->nbchilds=0;
   o->child=0;
   o->parent=NULL;
   for (i=0;i<256;i++) o->childs[i]=NULL;
   setvector(o->link,0,0,0);
   return(o);
}

void calc_normals(object3DE *o){
   int i,j,i1,i2,i3,nb;
   vect v1,v2,v3,mo;
   for (i=0;i<o->nbfc;i++){
       i1=o->fc[i].i1; i2=o->fc[i].i2; i3=o->fc[i].i3;
       subvector(o->pts[i1],o->pts[i2],v1);
       vectornormalize(v1);
       subvector(o->pts[i1],o->pts[i3],v2);
       vectornormalize(v2);
       vectorcross(v1,v2,v3);
       vectornormalize(v3);
       copyvector(v3,o->norms[i]);
   }
   for (i=0;i<o->nbpt;i++){
       mo[0]=mo[1]=mo[2]=nb=0;
       for (j=0;j<o->nbfc;j++){
           if (o->fc[j].i1==i){
               addvector(mo,o->norms[j],mo);
               nb++;
           }
           if (o->fc[j].i2==i){
               addvector(mo,o->norms[j],mo);
               nb++;
           }
           if (o->fc[j].i3==i){
               addvector(mo,o->norms[j],mo);
               nb++;
           }
       }
       if (nb==0) nb=1;
       mo[0]/=nb;
       mo[1]/=nb;
       mo[2]/=nb;
       vectornormalize(mo);
       copyvector(mo,o->pnorms[i]);
   }
}

int sinx,siny,sinz,cosx,cosy,cosz;

void rotation(vect p){
   int tmpx,tmpy,tmpz;
   tmpx=fixmul(cosy,p[0])-fixmul(siny,p[2]);
   tmpz=fixmul(cosy,p[2])+fixmul(siny,p[0]);
   tmpy=fixmul(cosz,p[1])-fixmul(sinz,tmpx);
   p[0]=fixmul(cosz,tmpx)+fixmul(sinz,p[1]);
   p[1]=fixmul(cosx,tmpy)+fixmul(sinx,tmpz);
   p[2]=fixmul(cosx,tmpz)-fixmul(sinx,tmpy);
}

void calc_camera_matrix(void){
   vect rpos;
   
   sinx=-sinus(cam->dir[0]); cosx=-cosin(cam->dir[0]);
   siny=-sinus(cam->dir[1]); cosy=-cosin(cam->dir[1]);
   sinz=-sinus(cam->dir[2]); cosz=-cosin(cam->dir[2]);
   subvector(cam->odir,cam->dir,cam->rdir);
   copyvector(cam->dir,cam->odir);
   setvector(cam->xform.matrice[0],65536,0,0); // rduire (muls par 0)
   setvector(cam->xform.matrice[1],0,65536,0);
   setvector(cam->xform.matrice[2],0,0,65536);
   rotation(cam->xform.matrice[0]);
   rotation(cam->xform.matrice[1]);
   rotation(cam->xform.matrice[2]);
   
// rotation de la light-source
   subvector(lsrc->pos,cam->pos,rpos);
   mulmatrixvec(cam->xform.matrice,rpos,lsrc->posrot);
   if (lsrc->posrot[2]>256){
       lsrc->proj[0]=(scrX>>1)+(fixmuldiv(lsrc->posrot[0],65536,lsrc->posrot[2])>>8);
       lsrc->proj[1]=(scrY>>1)+(fixmuldiv(lsrc->posrot[1],65536,lsrc->posrot[2])>>8);
       zdepth[++nbfv].o=lsrc;
       zdepth[nbfv].zbuf=fixdiv(65536,lsrc->posrot[2]);
   }    
}

#define VISIBLE \
 (((o->proj[i3][0]-o->proj[i1][0])*(o->proj[i1][1]-o->proj[i2][1]))<\
  ((o->proj[i2][0]-o->proj[i1][0])*(o->proj[i1][1]-o->proj[i3][1])))

#define OUTOFSCREEN \
 ((((o->proj[i1][0]<clipx1)&&(o->proj[i2][0]<clipx1)&&(o->proj[i3][0]<clipx1))||\
   ((o->proj[i1][1]<clipy1)&&(o->proj[i2][1]<clipy1)&&(o->proj[i3][1]<clipy1))||\
   ((o->proj[i1][0]>clipx2)&&(o->proj[i2][0]>clipx2)&&(o->proj[i3][0]>clipx2))||\
   ((o->proj[i1][1]>clipy2)&&(o->proj[i2][1]>clipy2)&&(o->proj[i3][1]>clipy2))||\
   ((o->ptsrot[i1][2]<=256)||(o->ptsrot[i2][2]<=256)||(o->ptsrot[i3][2]<=256))))
   
void rotate(object3DE *o){
   int i,x,y,i1,i2,i3,thisnode,c;
   vect rpos,rposl;
   affine local2camera;
   object3DE *p;

   if (o->level>1){
       p=o->parent;
       c=o->child;
       copyvector(p->pos,o->pos);
       copyvector(p->dir,o->dir);
   }
   thisnode=0;
   while (thisnode<o->nbchilds) rotate(o->childs[thisnode++]);

// rotation de l'objet
   subvector(o->odir,o->dir,o->rdir);
   copyvector(o->dir,o->odir);
   rotatebase(o->sysaxe.matrice,0,sinus(o->rdir[0]),cosin(o->rdir[0]));
   rotatebase(o->sysaxe.matrice,1,sinus(o->rdir[1]),cosin(o->rdir[1]));
   rotatebase(o->sysaxe.matrice,2,sinus(o->rdir[2]),cosin(o->rdir[2]));
   orthonormalizematrix(o->sysaxe.matrice);

// rotation de la camra   
   subvector(o->pos,cam->pos,rpos);
   sinx=sinus(cam->dir[0]); cosx=cosin(cam->dir[0]);
   siny=sinus(cam->dir[1]); cosy=cosin(cam->dir[1]);
   sinz=sinus(-cam->dir[2]); cosz=cosin(-cam->dir[2]);
   rotation(rpos);

// cration de la matrice camera<-local (CAMERA*LOCAL)
   mulmatrix(cam->xform.matrice,o->sysaxe.matrice,local2camera.matrice);
   copyvector(rpos,local2camera.vector);
   affine_xform(local2camera,o->link,rposl);
   addvector(rpos,rposl,rpos);
   copyvector(rposl,local2camera.vector);
   
   for (i=0;i<o->nbpt;i++){
       affine_xform(local2camera,o->pts[i],o->ptsrot[i]);
       if (o->ptsrot[i][2]<=256) continue;
       o->proj[i][0]=(scrX>>1)+(fixmuldiv(o->ptsrot[i][0],65536,o->ptsrot[i][2])>>8);
       o->proj[i][1]=(scrY>>1)+(fixmuldiv(o->ptsrot[i][1],65536,o->ptsrot[i][2])>>8);
       if (pseudo3DE) mulmatrixvec(o->sysaxe.matrice,o->pnorms[i],o->pnormsrot[i]);
   }
   for (i=0;i<o->nbfc;i++){
       i1=o->fc[i].i1; i2=o->fc[i].i2; i3=o->fc[i].i3;
       if (!VISIBLE||OUTOFSCREEN) continue;
       zdepth[++nbfv].o=o;
       zdepth[nbfv].fn=i;
       // 21845==65536*0.333333333 donc une division par 3
       zdepth[nbfv].zbuf=fixdiv(65536,fixmul(21845,(o->ptsrot[i1][2]+o->ptsrot[i2][2]+o->ptsrot[i3][2])));
       mulmatrixvec(o->sysaxe.matrice,o->norms[i],o->normsrot[i]);
   }
}

void quicksort(int l,int r){
   int i,j,x,y;
   void *o;
   i=l; j=r; x=zdepth[(l+r)>>1].zbuf;
   do{
       while (zdepth[i].zbuf<x) i++;
       while (x<zdepth[j].zbuf) j--;
       if (i<=j){
           y=zdepth[i].zbuf; zdepth[i].zbuf=zdepth[j].zbuf; zdepth[j].zbuf=y;
           y=zdepth[i].fn; zdepth[i].fn=zdepth[j].fn; zdepth[j].fn=y;
           o=zdepth[i].o; zdepth[i].o=zdepth[j].o; zdepth[j].o=o;
           i++; j--;
       }
   } while (i<=j);
   if (l<j) quicksort(l,j);
   if (i<r) quicksort(i,r);
}

void adjust_object(object3DE *o){
   int i,px,gx,py,gy,pz,gz;
   vect centre={0,0,0};
   // centrer l'objet
   for (i=0;i<o->nbpt;i++) addvector(centre,o->pts[i],centre);
   centre[0]/=o->nbpt;
   centre[1]/=o->nbpt;
   centre[2]/=o->nbpt;
   for (i=0;i<o->nbpt;i++) subvector(o->pts[i],centre,o->pts[i]);
   o->pos[0]=centre[0];
   o->pos[1]=centre[1];
   o->pos[2]=centre[2];
}

//**************************************************************************//
// stuff pour le renderer                                                   //
//**************************************************************************//

void smooth(byte *w,int h,int v){
   byte *ww;
   int i;
   ww=(byte *)malloc(h*v);
   if (ww==NULL) return;
   movemem(w,ww,h*v);
   for (i=h+1;i<(h*v)-h-1;i++){
       w[i]=(ww[i-1]+ww[i+1]+ww[i-256]+ww[i+256]+ww[i-255]+
             ww[i+255]+ww[i-257]+ww[i+257])>>3;
   }
   free(ww);
}      

void set_phong_palette(int SHINE,double SPEC,int R0,int G0,int B0,byte *pal){
   int i,c,r,g,b;
   double u;
   for (i=0;i<256;i++){
       u=SPEC*63.0*exp(SHINE*log(sin(0.5*pi*i/256.0)+0.001));
       if ((r=(i*R0/256)+u)>63) r=63;
       if ((g=(i*G0/256)+u)>63) g=63;
       if ((b=(i*B0/256)+u)>63) b=63;
       *pal++=r; *pal++=g; *pal++=b;
  }
}

void set_VESA_phonglookup(int SHINE,double SPEC,byte *p){
   int i,c,r,g,b;
   double u;
   for (i=0;i<256;i++){
       u=(SPEC*63.0*exp(SHINE*log(sin(0.5*pi*i/256.0)+0.001)));
       if (u>63) u=63;
       if (u<0) u=0;
       *p++=u;
  }
}

void set_phong_stable(int SHINE,double SPEC,int ambiant,byte *pal,byte *light){
   int i,j,c,r,g,b;
   double u,level,ir,ig,ib;
   for (j=0;j<256;j++){
       for (i=0;i<256;i++){
           level=(double)i/256;
           ir=pal[j*3]*level;
           ig=pal[j*3+1]*level;
           ib=pal[j*3+2]*level;
           u=(ambiant/4)+SPEC*63.0*exp(SHINE*log(sin(0.5*3.141516*i/256.0)+0.001));
           if ((r=(ir+u))>63) r=63;
           if ((g=(ig+u))>63) g=63;
           if ((b=(ib+u))>63) b=63;
           light[j*256+i]=remap[r+(g<<6)+(b<<12)];
       }    
   }
}

void set_phong_lookup(byte *specular,byte *bmpspec){
   int i,j,k;
   for (j=0;j<256;j++)
       for (i=0;i<256;i++){
           specular[j*256+i]=255-sqrt(((128-i)*(128-i))+((128-j)*(128-j)));
           k=255-(sqrt(((128-i)*(128-i))+((128-j)*(128-j)))*2);
           if (k<0) k=0;
           bmpspec[j*256+i]=k;
       }    
}

//**************************************************************************//
// stuff pour loader fichiers .ASC .ROB                                     //
//**************************************************************************//

int readstr(FILE *f,char *string){
   return fscanf(f,"%s",string);
}

int gotostr(FILE *f,char *string){
   int found=0;
   char str[128];
   while(!found){
       if (readstr(f,str)==EOF) return(EOF);
       if (strcmp(string,str)==0) found=1;
   }
   return(0);
}

int getfloat(FILE *f,double *number){
   char str[128];
   int result;
   result=fscanf(f,"%s",str);
   *number=strtod(str,NULL);
   return(result);
}

int getint(FILE *f,int *number){
   return fscanf(f,"%d",number);
}

object3DE *loadASC(char *filename){
   int *tmp_u,*tmp_v;
   object3DE *o;
   char numstr[128];
   int i,mapped=1,p1,p2,p3,mem=0,nbpts,nbfc,i1,i2,i3;
   double x,y,z,u=0,v=0;
   FILE *f;

   repstosd(numstr,0,128/4);
   f=fopen(filename,"rt");
   if (f==NULL) return(NULL);
   if (gotostr(f,"Mapped")==EOF) mapped=0;
   fseek(f,0,SEEK_SET);
   if (gotostr(f,"Vertices:")==EOF) return(0);
   getint(f,&nbpts);
   gotostr(f,"Faces:");
   getint(f,&nbfc);
   o=init_object(nbpts,nbfc,0,0,0,OBJECT);
   if (o==NULL) return(NULL);
   tmp_u=malloc(nbpts*sizeof(int));
   tmp_v=malloc(nbpts*sizeof(int));
   
   totalfaces+=nbfc;

   gotostr(f,"Vertex"); gotostr(f,"list:");
   for (i=0;i<nbpts;i++){
       gotostr(f,"Vertex");
       sprintf(numstr,"%d:",i);
       gotostr(f,numstr);
       readstr(f,numstr); x=strtod(numstr+2,NULL);
       readstr(f,numstr); y=strtod(numstr+2,NULL);
       readstr(f,numstr); z=strtod(numstr+2,NULL);
       if (mapped){
           readstr(f,numstr); u=strtod(numstr+2,NULL);
           readstr(f,numstr); v=strtod(numstr+2,NULL);
       } else {
           u=rand()/32768.0; // hahahahahahahahaha!!!
           v=rand()/32768.0;
       }
       o->pts[i][0]=x*256;  // 8 bit de prcision pour les points
       o->pts[i][1]=y*256;
       o->pts[i][2]=z*256;
       tmp_u[i]=u*256;      // u&v coords 0 bit de prcision
       tmp_v[i]=v*256;
   }
   gotostr(f,"Face"); gotostr(f,"list:");
   for (i=0;i<nbfc;i++){
       gotostr(f,"Face");
       sprintf(numstr,"%d:",i);
       gotostr(f,numstr);
       readstr(f,numstr); p1=atoi(numstr+2);
       readstr(f,numstr); p2=atoi(numstr+2);
       readstr(f,numstr); p3=atoi(numstr+2);
       o->fc[i].i1=p1;
       o->fc[i].i2=p2;
       o->fc[i].i3=p3;
   }
   fclose(f);
   adjust_object(o);
   calc_normals(o);
   for (i=0;i<nbfc;i++){
       i1=o->fc[i].i1;
       i2=o->fc[i].i2;
       i3=o->fc[i].i3;
       o->fc[i].u1=tmp_u[i1];
       o->fc[i].v1=tmp_v[i1];
       o->fc[i].u2=tmp_u[i2];
       o->fc[i].v2=tmp_v[i2];
       o->fc[i].u3=tmp_u[i3];
       o->fc[i].v3=tmp_v[i3];
   }
   free(tmp_u);
   free(tmp_v);
   return(o);
}

object3DE *loadROB(char *filename){
   int i,nbpts,nbfc,uvcoord,type=0;
   char sign[4];
   object3DE *o;
   vect pt;
   FILE *f;
   f=fopen(filename,"rb");
   if (f==NULL) return(NULL);
   fread(sign,4,1,f);
   if (sign[0]!='R'||sign[1]!='O'||sign[2]!='B'||sign[3]!=26){
       printf("Bad file format (ROB)\n");
       fclose(f);
       return(NULL);
   }
   while (!type) fread(&type,4,1,f);
   fread(&nbpts,4,1,f);
   o=(object3DE *)malloc(sizeof(object3DE));
   if (o==NULL) return(NULL);
   setmem(o,0,sizeof(object3DE));
   o->type=OBJECT;
   o->nbpt=nbpts;
   o->pts=(vect *)malloc(o->nbpt*sizeof(vect));
   o->ptsrot=(vect *)malloc(o->nbpt*sizeof(vect));
   o->proj=(vect2D *)malloc(o->nbpt*sizeof(vect2D));
   o->pnorms=(vect *)malloc(o->nbpt*sizeof(vect));
   o->pnormsrot=(vect *)malloc(o->nbpt*sizeof(vect));
   if (o->pts==NULL||o->ptsrot==NULL||o->pnorms==NULL||o->pnormsrot==NULL||o->proj==NULL)
       return(NULL);

   for (i=0;i<nbpts;i++){
       fread(pt,12,1,f);
       o->pts[i][0]=pt[0]*256;
       o->pts[i][1]=pt[1]*256;
       o->pts[i][2]=pt[2]*256;
   }    
   fread(&nbfc,4,1,f);
   
   totalfaces+=nbfc;

   o->nbfc=nbfc;
   o->norms=(vect *)malloc(o->nbfc*sizeof(vect));
   o->normsrot=(vect *)malloc(o->nbfc*sizeof(vect));
   o->fc=(face3DE *)malloc(o->nbfc*sizeof(face3DE));
   if (o->norms==NULL||o->normsrot==NULL||o->fc==NULL)
       return(NULL);

   for (i=0;i<nbfc;i++){
       fread(&o->fc[i].i1,4,1,f);
       fread(&o->fc[i].i2,4,1,f);
       fread(&o->fc[i].i3,4,1,f);
       fread(&uvcoord,4,1,f);
       fread(&uvcoord,4,1,f); o->fc[i].u1=uvcoord;
       fread(&uvcoord,4,1,f); o->fc[i].v1=uvcoord;
       fread(&uvcoord,4,1,f); o->fc[i].u2=uvcoord;
       fread(&uvcoord,4,1,f); o->fc[i].v2=uvcoord;
       fread(&uvcoord,4,1,f); o->fc[i].u3=uvcoord;
       fread(&uvcoord,4,1,f); o->fc[i].v3=uvcoord;
       fread(&uvcoord,4,1,f);
   }
   fclose(f);
   o->nbpt-=8;
   adjust_object(o);
   calc_normals(o);
   setmatrix(&o->sysaxe.matrice,65536,0,0,0,65536,0,0,0,65536);
   return(o);
}
