/*
    3DE v2
    16.16 fixed point math, 24.8 fixed point coords
*/

#ifndef _3D_H
#define _3D_H

#define MASTER 0        // object root de l'arbre
#define OBJECT 1        // objet 3d
#define LIGHT 2         // light source
#define CAMERA 3        // camera

extern char typenames[4][20];

typedef int vect2D[2];

typedef struct{
   int i1,i2,i3;                        // vertices
   byte u1,v1,u2,v2,u3,v3;              // mapping coords [0..256]
   byte *texture;                       // texture
   byte *bumpmap;                       // bump map
   void (*method)();                    // rendering method
} face3DE;

typedef struct{
   int type;
   int nbpt,nbfc;                       // nombre de points et de faces
   vect pos,dir,odir,rdir;              // position de l'objet et direction
   vect *pts,*ptsrot;                   // points
   vect *norms,*normsrot;               // normales
   vect *pnorms,*pnormsrot;             // pseudo-normales
   vect2D *proj;                        // points projets
   face3DE *fc;                         // faces
   affine sysaxe;                       // systme d'axes local
//-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=// linkage des objets
   int level;                           // level dans l'arbre de linkage
   int nbchilds;                        // nombre de childs
   int child;                           // no du child sur le parent
   void *parent,*childs[1024];          // parent et childs (max. 1024)
   vect link;                           // position du link sur le parent
} object3DE;

typedef struct{
   int zbuf,fn;
   void *o;                             // pointeur sur un objet
} depth;

typedef struct{
   int type;
   vect pos,posrot,dir,odir,rdir;
   vect2D proj;
   affine xform;
} camera3DE;

typedef struct{
   int type;
   byte r,g,b;
   vect pos,posrot,dir,odir,rdir;
   vect2D proj;
   affine xform;
} light3DE;

extern camera3DE *cam;
extern light3DE *lsrc;
extern object3DE *world;
extern int pseudo3DE;
extern depth *zdepth;
extern int nbfv,totalfaces;
extern byte *remap;
// stuff renderer
extern int scrX,scrY,clipx1,clipy1,clipx2,clipy2;

// proc 3d
void out(char *msg,int action);
void init_3de(void);
void printinfo_3de(void);
object3DE *init_object(int nbpt,int nbfc,int px,int py,int pz,int type);
void calc_normals(object3DE *o);
void calc_camera_matrix(void);
void rotate(object3DE *o);

// proc renderer
void smooth(byte *w,int h,int v);
void set_phong_palette(int SHINE,double SPEC,int R0,int G0,int B0,byte *pal);
void set_phong_stable(int SHINE,double SPEC,int ambiant,byte *pal,byte *light);
void set_phong_lookup(byte *specular,byte *bmpspec);
void set_VESA_phonglookup(int SHINE,double SPEC,byte *p);

// proc files
object3DE *loadASC(char *filename);
object3DE *loadROB(char *filename);

#endif
