/*---------------------------------------------------------------------------
                      THE CHAOS ENGINE - 3DENGINE.C

                            Main 3d routines
---------------------------------------------------------------------------*/

#include <stdio.h>
#include <math.h>
#include <string.h>
#include <search.h>

#include "3dengine.h"
#include "math.h"
#include "gfx.h"

object_3d *objects_p[MAX_OBJECTS];
poly_3d **polylist_pp=NULL;

vertice_3d view_point={0,0,0};
vector_3d light_source;
dir view_dir={0,0,0};
int ambience, polys_in_list, objects_in_list, polygons_drawn, total_polys;

extern float camera_matrix[3][3];
extern float pmulx,pmuly,hfov_delta,vfov_delta;
extern int xcenter,ycenter;
extern rgb palette[256];

// Initialize the 3DEngine

int Init_3DEngine(void)
{
int i;

Init_Gfx();                          // Initialize GfxEngine
Init_Math();                         // Generate lockup tables etc...
polygons_drawn=objects_in_list=0;
for (i=0; i<64; i++) {
   palette[i].r=palette[i].g=palette[i].b=63-(char)(53*cos(((i*90)/63)*D2R));
   palette[i+64].r=63-(char)(53*cos(((i*90)/63)*D2R));
   palette[i+128].b=63-(char)(53*cos(((i*90)/63)*D2R));
   palette[i+192].r=palette[i+192].b=63-(char)(53*cos(((i*90)/63)*D2R));

}
   palette[184].r=palette[184].g=20;
   palette[185].r=palette[185].g=25;

   palette[186].r=palette[186].g=30;
   palette[187].r=palette[187].g=39;
   palette[188].r=palette[188].g=48;   //-(char)(63*cos(((i*90)/63)*D2R));
   palette[189].r=palette[189].g=56; //-(char)(63*cos(((i*90)/63)*D2R));
   palette[190].r=palette[190].g=60; //-(char)(63*cos(((i*90)/63)*D2R));
   palette[191].r=palette[191].g=palette[191].b=63; //-(char)(63*cos(((i*90)/63)*D2R));
//   palette[0].r=palette[0].g=palette[0].b=16;
return(0);
}

// Close down the 3DEngine

void Close_3DEngine(void)
{
Close_Gfx();
}

/*---------------------------------------------------------------------------
                              Object functions
---------------------------------------------------------------------------*/

// Add an object to object list

void Add_Object(object_3d *obj_p)
{
objects_p[objects_in_list++]=obj_p;
}

// Remove an object

void Delete_Object(int obj)
{
int i;
free(objects_p[i]);
for (i=obj; i<--objects_in_list; i++) objects_p[i]=objects_p[i+1];
}

// Free all objects in list

void Free_Objects(void)
{
int i;
for (i=0; i<objects_in_list; i++) {
   if (objects_p[i]->transformed_p!=NULL) free(objects_p[i]->transformed_p);
   if (objects_p[i]->vertices_p!=NULL) free(objects_p[i]->vertices_p);
   free(objects_p[i]);
}
}

// Returns address of an object

object_3d *Object_Address(int obj)
{
return(objects_p[obj]);
}

void Object_Off(object_3d *obj)
{
obj->flags&=~OBJ_ON;
}

void Object_On(object_3d *obj)
{
obj->flags|=OBJ_ON;
}

/*---------------------------------------------------------------------------
                     Misc functions to setup things
---------------------------------------------------------------------------*/

void Set_ViewPoint(vertice_3d *vp)
{
view_point=*vp;
}

vertice_3d Get_ViewPoint(void)
{
return(view_point);
}

void Set_ViewDirection(dir *direction)
{
view_dir=*direction;
}

dir Get_ViewDirection(void)
{
return(view_dir);
}

void Set_Ambience(int amb)
{
ambience=amb;
}

void Set_LightSource(vector_3d *light)
{
light_source=*light;
Normalize_Vector(&light_source);
}

/*---------------------------------------------------------------------------
                         Remove invisible objects.
---------------------------------------------------------------------------*/

void Remove_Hidden_Objects(void)
{
int o;
object_3d *obj;
vertice_3d obj_pos,pos;
float hdist,vdist;

for (o=0; o<objects_in_list; o++) {
   obj=objects_p[o];

   obj_pos.x=obj->world.x-view_point.x;
   obj_pos.y=obj->world.y-view_point.y;
   obj_pos.z=obj->world.z-view_point.z;

   pos.x = (obj_pos.x * camera_matrix[0][0]) + (obj_pos.y * camera_matrix[0][1])
         + (obj_pos.z * camera_matrix[0][2]);
   pos.y = (obj_pos.x * camera_matrix[1][0]) + (obj_pos.y * camera_matrix[1][1])
         + (obj_pos.z * camera_matrix[1][2]);
   pos.z = (obj_pos.x * camera_matrix[2][0]) + (obj_pos.y * camera_matrix[2][1])
         + (obj_pos.z * camera_matrix[2][2]);

   obj->flags &= ~(OBJ_VISIBLE | OBJ_CLIPPED);

   // Test far and near clipping planes
   if ((pos.z > (NEAR_CLIP-obj->max_radius)) && (pos.z < FAR_CLIP)) {
      // Test left and right clipping planes.
      if ((((pos.z * hfov_delta) + obj->max_radius) > pos.x) &&
      (((pos.z * -hfov_delta) - obj->max_radius) < pos.x) &&
      (((pos.z * vfov_delta) + obj->max_radius) > pos.y) &&
      (((pos.z * -vfov_delta) - obj->max_radius) < pos.y)) {

         obj->flags |= OBJ_VISIBLE;

         // Does object have to be z-clipped?
         if (fabs(pos.z-NEAR_CLIP) <= obj->max_radius) obj->flags |= OBJ_CLIPPED;
      }
   }
}
}

/*---------------------------------------------------------------------------
                       Check Polygons for Z-clip
---------------------------------------------------------------------------*/

void ZClip_Objects(void)
{
int o,p,i;
char visible;
object_3d *obj_p;

for (o=0; o<objects_in_list; o++) {
   obj_p=objects_p[o];
   if (obj_p->flags & OBJ_CLIPPED) {
      for (p=0; p<obj_p->polys; p++) {
         // Check all polygon vertices
         visible=FALSE;
         for (i=0; i<obj_p->polys_p[p].points; i++) {
            if (obj_p->transformed_p[obj_p->polys_p[p].pindxs[i]].z <= NEAR_CLIP)
               obj_p->polys_p[p].flags|=POLY_CLIPPED;
            else visible=TRUE;
         }
         if (!visible) obj_p->polys_p[p].flags&=~(POLY_VISIBLE | POLY_CLIPPED);
      }
   }
}
}


/*---------------------------------------------------------------------------
                      Project all vertices in objects.
---------------------------------------------------------------------------*/

void Project_Vertices(void)
{
int o,v;
float z;
object_3d *obj_p;
vertice_3d *trans_p;

for (o=0; o<objects_in_list; o++) {
   obj_p=objects_p[o];
   if (obj_p->flags & OBJ_VISIBLE) {
      trans_p=obj_p->transformed_p;
      for (v=0; v<obj_p->vertices; v++) {
         z=trans_p->z;
         obj_p->projected_p[v].x=xcenter+(int)((trans_p->x*pmulx)/z);
         obj_p->projected_p[v].y=ycenter-(int)((trans_p->y*pmuly)/z);
         trans_p++;
      }
   }
}
}


/*---------------------------------------------------------------------------
                         Remove invisible backfaces.
---------------------------------------------------------------------------*/

void Remove_BackFaces(void)
{
int i,p, o, ofs_1, ofs_2, ofs_3, flags, shade, i1,i2,i3;
float dp;
int x1,x2,x3,y1,y2,y3;
vector_3d u, v, face_normal, sight;
object_3d *obj;
poly_3d *poly_p;

for (o=0; o<objects_in_list; o++) {
   obj=objects_p[o];

   if ((obj->flags & OBJ_VISIBLE) && (obj->flags & OBJ_ON)) {

      for (p=0; p<obj->polys; p++) {

         poly_p=&(obj->polys_p[p]);

         if (poly_p->flags & POLY_TWOSIDED) continue;
  
         if (!(poly_p->flags & POLY_CLIPPED)) {

            if (!(poly_p->flags & POLY_VISIBLE)) continue;

            i=obj->polys_p[p].pindxs[0];
            x1=10000+obj->projected_p[i].x;
            y1=10000+obj->projected_p[i].y;

            i=obj->polys_p[p].pindxs[1];
            x2=10000+obj->projected_p[i].x;
            y2=10000+obj->projected_p[i].y;

            i=obj->polys_p[p].pindxs[2];
            x3=10000+obj->projected_p[i].x;
            y3=10000+obj->projected_p[i].y;

            if (((x2-x1)*(y3-y1)-(y2-y1)*(x3-x1))<0)
               obj->polys_p[p].flags|=POLY_VISIBLE;
            else obj->polys_p[p].flags&=~POLY_VISIBLE;
         } else {

            i1 = poly_p->pindxs[0];  // Get 3 offsets
            i2 = poly_p->pindxs[1];
            i3 = poly_p->pindxs[2];

            // Get vector 1
            u.x = obj->vertices_p[i2].x - obj->vertices_p[i1].x;
            u.y = obj->vertices_p[i2].y - obj->vertices_p[i1].y;
            u.z = obj->vertices_p[i2].z - obj->vertices_p[i1].z;

            // Get vector 2
            v.x = obj->vertices_p[i3].x - obj->vertices_p[i1].x;
            v.y = obj->vertices_p[i3].y - obj->vertices_p[i1].y;
            v.z = obj->vertices_p[i3].z - obj->vertices_p[i1].z;

            // Calculate surface normal
            Cross_Product(&v, &u, &face_normal);

            sight.x = view_point.x - (obj->vertices_p[i1].x + obj->world.x);
            sight.y = view_point.y - (obj->vertices_p[i1].y + obj->world.y);
            sight.z = view_point.z - (obj->vertices_p[i1].z + obj->world.z);

            dp=Dot_Product(&face_normal, &sight);
            if (dp>0) poly_p->flags|=POLY_VISIBLE;
            else poly_p->flags&=~(POLY_VISIBLE | POLY_CLIPPED);

         }
      }
   }
}
}

/*---------------------------------------------------------------------------
                           Polygon list functions.
---------------------------------------------------------------------------*/

// Collect polygons and add them to the polylist

void Make_PolyList(void)
{
int poly,o,p,c,i1,i2,i3,shade;
object_3d *obj;
vector_3d u,v,face_normal;
poly_3d *poly_p;
float dp;

polys_in_list=0;
for (o=0; o<objects_in_list; o++) {
   obj=objects_p[o];
   if ((obj->flags & OBJ_VISIBLE) && (obj->flags & OBJ_ON)) {

      // Clear vertex color table

      if (obj->flags & OBJ_GOURAUD)
         memset(obj->vertex_colors_p, 192, obj->vertices);

      // Loop trough all polygons in object

      for (poly=0; poly<obj->polys; poly++) {

         poly_p=&(obj->polys_p[poly]);
         if (poly_p->flags & POLY_VISIBLE) {
            polylist_pp[polys_in_list++]=poly_p;

            // Calculate face color if lambertshaded

            if (poly_p->flags & POLY_LAMBERTSHADED) {
               obj=(object_3d *)(poly_p->homeobj_p);

               i1 = poly_p->pindxs[0];  // Get 3 offsets
               i2 = poly_p->pindxs[1];
               i3 = poly_p->pindxs[2];

               // Get vector 1
               u.x = obj->vertices_p[i2].x - obj->vertices_p[i1].x;
               u.y = obj->vertices_p[i2].y - obj->vertices_p[i1].y;
               u.z = obj->vertices_p[i2].z - obj->vertices_p[i1].z;

               // Get vector 2
               v.x = obj->vertices_p[i3].x - obj->vertices_p[i1].x;
               v.y = obj->vertices_p[i3].y - obj->vertices_p[i1].y;
               v.z = obj->vertices_p[i3].z - obj->vertices_p[i1].z;

               // Calculate surface normal
               Cross_Product(&v, &u, &face_normal);
               // Do lambert shading.
               dp=Dot_Product(&face_normal, &light_source);
               if (dp>0) {
                  shade = ambience + (long)((NROF_SHADES * dp) / poly_p->normal_len);
                  if (shade>NROF_SHADES) shade = NROF_SHADES;
                  poly_p->shade = poly_p->color + shade;
               }
               else poly_p->shade = poly_p->color + ambience;

            } else poly_p->shade = poly_p->color;



            // Calculate vertex colors if gouraudshaded

            if (poly_p->flags & POLY_GOURAUDSHADED) {
               obj=(object_3d *)(poly_p->homeobj_p);
               for (p=0; p<poly_p->points; p++) {
                  if ((obj->vertex_colors_p[poly_p->pindxs[p]])==192) {
                     dp=Dot_Product(&(obj->vnormals_p[poly_p->pindxs[p]]), &light_source);
                     c=ambience+(int)(63*dp);
                     if (c<1) c=1; else if (c>63) c=63;
                     obj->vertex_colors_p[poly_p->pindxs[p]]=c;
                  }
               }
            }
         } else poly_p->flags|=POLY_VISIBLE;    // Reset visible flag
      }
   }
}
polygons_drawn+=polys_in_list;
}

int Compare(poly_3d **p1_pp, poly_3d **p2_pp)
{
poly_3d *p1_p,*p2_p;

p1_p=(poly_3d *)*p1_pp;
p2_p=(poly_3d *)*p2_pp;

if (p1_p->average_z>p2_p->average_z) return(-1);
else if (p1_p->average_z<p2_p->average_z) return(1);
else return(0);
}

// Depth sort polylist

void Sort_PolyList(void)
{
int bubble,p,pts,indx;
float sum,maxz;
poly_3d *tmp;
object_3d *obj;

for (p=0; p<polys_in_list; p++) {
   sum=0;
   maxz=-100000;
   obj=(object_3d *)polylist_pp[p]->homeobj_p;
   for (pts=0; pts<polylist_pp[p]->points; pts++) {
      indx=polylist_pp[p]->pindxs[pts];
      sum+=obj->transformed_p[indx].z;
      //if (maxz<obj->transformed_p[indx].z) maxz=obj->transformed_p[indx].z;
   }
   //polylist_pp[p]->average_z=(int)(maxz);
   polylist_pp[p]->average_z=(int)(sum/pts);
}
/*
do {
   bubble=FALSE;
   for (p=0; p<(polys_in_list-1); p++) {
      if (polylist_pp[p]->average_z<polylist_pp[p+1]->average_z) {
         tmp=polylist_pp[p];
         polylist_pp[p]=polylist_pp[p+1];
         polylist_pp[p+1]=tmp;
         bubble=TRUE;
      }
   }
} while (bubble); */

qsort(polylist_pp, polys_in_list, 4, Compare);
}


// Render polygons in polylist

void Draw_PolyList(void)
{
int p;

for (p=0; p<polys_in_list; p++) {
   if (polylist_pp[p]->flags & POLY_GOURAUDSHADED) {
      if ((polylist_pp[p]->flags & POLY_TRIANGLE) && (!(polylist_pp[p]->flags & POLY_CLIPPED)))
         Draw_Gouraud_Triangle(polylist_pp[p]);
      else Draw_Gouraud_Polygon(polylist_pp[p]);

   } else {
      if (polylist_pp[p]->flags & POLY_LAMBERTSHADED)
         Draw_Solid_Polygon(polylist_pp[p]);

   }
   if (polylist_pp[p]->flags & POLY_TEXTUREMAPPED) {
      if ((polylist_pp[p]->flags & POLY_TRIANGLE) && (!(polylist_pp[p]->flags & POLY_CLIPPED)))
         Draw_Texture_Triangle(polylist_pp[p]);
      else Draw_Texture_Polygon(polylist_pp[p]);

   }
}
}

