#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "stuff.h"
#include "fmath\fmath.h"
#include "3de\3d.h"
#include "3dsrdr.h"

int nbobj,nbcam,nblig;
object3DE *ob[4096];

#pragma pack(2)
typedef struct {
    word    id;
    dword   len;
} TChunkHeader, *PChunkHeader;
#pragma pack()

// ------------------------------------

    // Forward declaration.
void ChunkReader(FILE *f, int ind, long p);

void SkipReader (FILE *f, int ind, long p) {
}

void RGBFReader (FILE *f, int ind, long p) {
    float c[3];
    if (fread(&c, sizeof(c), 1, f) != 1) return;
    //printf("%*s    Red: %f, Green: %f, Blue: %f\n", ind, "", c[0], c[1], c[2]);
}

void RGBBReader (FILE *f, int ind, long p) {
    byte c[3];
    if (fread(&c, sizeof(c), 1, f) != 1) return;
    //printf("%*s    Red: %d, Green: %d, Blue: %d\n", ind, "", c[0], c[1], c[2]);
}

void ASCIIZReader (FILE *f, int ind, long p) {
    int c;

        // Read ASCIIZ name
    while ( (c = fgetc(f)) != EOF && c != '\0'){
        //putchar(c);
    }    
    //printf("\"\n");
}

void ObjBlockReader (FILE *f, int ind, long p) {
    int c;

        // Read ASCIIZ object name
    //printf("%*sObject name \"", ind, "");
    ASCIIZReader(f, ind, p);
        // Read rest of chunks inside this one.
    ChunkReader(f, ind, p);
}

void VertListReader (FILE *f, int ind, long p) {
    word nv,pp;
    float c[3];

    if (fread(&nv, sizeof(nv), 1, f) != 1) return;
    //printf("%*sVertices: %d\n", ind, "", nv);

    ob[nbobj]->nbpt=nv;
    ob[nbobj]->pts=(vect *)malloc(ob[nbobj]->nbpt*sizeof(vect));
    ob[nbobj]->ptsrot=(vect *)malloc(ob[nbobj]->nbpt*sizeof(vect));
    ob[nbobj]->proj=(vect2D *)malloc(ob[nbobj]->nbpt*sizeof(vect2D));
    ob[nbobj]->pnorms=(vect *)malloc(ob[nbobj]->nbpt*sizeof(vect));
    ob[nbobj]->pnormsrot=(vect *)malloc(ob[nbobj]->nbpt*sizeof(vect));
    
    pp=0;
    while (nv-- > 0) {
        if (fread(&c, sizeof(c), 1, f) != 1) return;
        ob[nbobj]->pts[pp][0]=c[0]*256;
        ob[nbobj]->pts[pp][1]=c[1]*256;
        ob[nbobj]->pts[pp][2]=c[2]*256;
        pp++;
        //printf("%*s    X: %f, Y: %f, Z: %f\n", ind, "", c[0], c[1], c[2]);
    }
}

void FaceListReader (FILE *f, int ind, long p) {
    word nv,pp;
    word c[3];
    word flags;

    if (fread(&nv, sizeof(nv), 1, f) != 1) return;
    //printf("%*sFaces: %d\n", ind, "", nv);

    ob[nbobj]->nbfc=nv;
    ob[nbobj]->norms=(vect *)malloc(ob[nbobj]->nbfc*sizeof(vect));
    ob[nbobj]->normsrot=(vect *)malloc(ob[nbobj]->nbfc*sizeof(vect));
    ob[nbobj]->fc=(face3DE *)malloc(ob[nbobj]->nbfc*sizeof(face3DE));
    setmem(ob[nbobj]->fc,0,ob[nbobj]->nbfc*sizeof(face3DE));

    pp=0;
    while (nv-- > 0) {
        if (fread(&c, sizeof(c), 1, f) != 1) return;
        if (fread(&flags, sizeof(flags), 1, f) != 1) return;
        ob[nbobj]->fc[pp].i1=c[0];
        ob[nbobj]->fc[pp].i2=c[1];
        ob[nbobj]->fc[pp].i3=c[2];
        pp++;
        //printf("%*s  A %d, B %d, C %d, 0x%X:",
        //    ind, "", c[0], c[1], c[2], flags);
        //printf(" AB %d BC %d CA %d UWrap %d VWrap %d\n",
        //       (flags & 0x04) != 0, (flags & 0x02) != 0, (flags & 0x01) != 0,
        //       (flags & 0x08) != 0, (flags & 0x10) != 0);
    }
        // Read rest of chunks inside this one.
    ChunkReader(f, ind, p);
}

void FaceMatReader (FILE *f, int ind, long p) {
    int c;
    word n, nf;

        // Read ASCIIZ material name
    //printf("%*sMaterial name for faces: \"", ind, "");
    ASCIIZReader(f, ind, p);

    if (fread(&n, sizeof(n), 1, f) != 1) return;
    //printf("%*sFaces with this material: %d\n", ind, "", n);
    while (n-- > 0) {
        if (fread(&nf, sizeof(nf), 1, f) != 1) return;
        //printf("%*s    Face %d\n",
        //       ind, "", nf);
    }
}

void MapListReader (FILE *f, int ind, long p) {
    word nv;
    float c[2];

    if (fread(&nv, sizeof(nv), 1, f) != 1) return;
    //printf("%*sVertices: %d\n", ind, "", nv);
    while (nv-- > 0) {
        if (fread(&c, sizeof(c), 1, f) != 1) return;
       // printf("%*s    U: %f, V: %f\n", ind, "", c[0], c[1]);
    }
}

void SmooListReader (FILE *f, int ind, long p) {
    dword s;
    int i;

    while (ftell(f) < p) {
        if (fread(&s, sizeof(s), 1, f) != 1) return;
        //printf("%*sSmoothing groups: ", ind, "");
        for (i = 0; i < 32; i++){
            if (s & (1 << i)){
                //printf("%d, ", i + 1);
            }    
        //printf("\n");
        }
    }
}

void TrMatrixReader(FILE *f, int ind, long p) {
    float rot[9];
    float trans[3];
    if (fread(&rot, sizeof(rot), 1, f) != 1) return;
    //printf("%*sRotation matrix:\n", ind, "");
    //printf("%*s    %f, %f, %f\n", ind, "", rot[0], rot[1], rot[2]);
    //printf("%*s    %f, %f, %f\n", ind, "", rot[3], rot[4], rot[5]);
    //printf("%*s    %f, %f, %f\n", ind, "", rot[6], rot[7], rot[8]);
    if (fread(&trans, sizeof(trans), 1, f) != 1) return;
    //printf("%*sTranslation matrix: %f, %f, %f\n",
    //       ind, "", trans[0], trans[1], trans[2]);
}

void LightReader(FILE *f, int ind, long p) {
    float c[3];
    if (fread(&c, sizeof(c), 1, f) != 1) return;
    //printf("%*s    X: %f, Y: %f, Z: %f\n", ind, "", c[0], c[1], c[2]);
        // Read rest of chunks inside this one.
    ChunkReader(f, ind, p);
}

void SpotLightReader(FILE *f, int ind, long p) {
    float c[5];
    if (fread(&c, sizeof(c), 1, f) != 1) return;
    //printf("%*s    Target X: %f, Y: %f, Z: %f; Hotspot %f, Falloff %f\n",
    //       ind, "", c[0], c[1], c[2], c[3], c[4]);
}
 
void CameraReader(FILE *f, int ind, long p) {
    float c[8];
    if (fread(&c, sizeof(c), 1, f) != 1) return;
    //printf("%*s    Position: X: %f, Y: %f, Z: %f\n", ind, "", c[0], c[1], c[2]);
    //printf("%*s    Target: X: %f, Y: %f, Z: %f\n", ind, "", c[3], c[4], c[5]);
    //printf("%*s    Bank: %f, Lens: %f\n", ind, "", c[6], c[7]);
}

void MatNameReader (FILE *f, int ind, long p) {
    int c;

        // Read ASCIIZ object name
    //printf("%*sMaterial name \"", ind, "");
    ASCIIZReader(f, ind, p);
}

void MapFileReader(FILE *f, int ind, long p) {
    int c;

        // Read ASCIIZ filename
    //printf("%*sMap filename \"", ind, "");
    ASCIIZReader(f, ind, p);
}

void FramesReader(FILE *f, int ind, long p) {
    dword c[2];
    if (fread(&c, sizeof(c), 1, f) != 1) return;
    //printf("%*s    Start: %ld, End: %ld\n",
    //       ind, "", c[0], c[1]);
}

void TrackObjNameReader(FILE *f, int ind, long p) {
    int c;
    word w[2];
    word parent;

        // Read ASCIIZ name
    //printf("%*sTrack object name \"", ind, "");
    ASCIIZReader(f, ind, p);
    if (fread(&w, sizeof(w), 1, f) != 1) return;
    if (fread(&parent, sizeof(parent), 1, f) != 1) return;
    //printf("%*sObject name data: Flags 0x%X, 0x%X, Parent %d\n",
    //       ind, "", w[0], w[1], parent);
}

void PivotPointReader(FILE *f, int ind, long p) {
    float pos[3];

    if (fread(&pos, sizeof(pos), 1, f) != 1) return;
    //printf("%*s  Pivot at X: %f, Y: %f, Z: %f\n",
    //       ind, "",
    //       pos[0], pos[1], pos[2]);
}

    /* Key info flags for position, rotation and scaling:
        Until I know the meaning of each bit in flags I assume all mean
        a following float data.
    */

        // NOTE THIS IS NOT A CHUNK, but A PART OF SEVERAL CHUNKS
void SplineFlagsReader(FILE *f, int ind, word flags) {
    int i;
    float dat;

    for (i = 0; i < 16; i++) {
        static const char *flagnames[] = {
            "Tension",
            "Continuity",
            "Bias",
            "Ease To",
            "Ease From",
        };
        if (flags & (1 << i)) {
            if (fread(&dat, sizeof(dat), 1, f) != 1) return;
            if (i < sizeof(flagnames)/sizeof(*flagnames)){
                //printf("%*s             %-15s = %f\n",
                //       ind, "", flagnames[i], dat);
            } else {
                //printf("%*s             %-15s = %f\n",
                //       ind, "", "Unknown", dat);
            }    
        }
    }
}

void TrackPosReader(FILE *f, int ind, long p) {
    word n, nf;
    float pos[3];
    word unkown;
    word flags;

    fseek(f, 10, SEEK_CUR);
    if (fread(&n, sizeof(n), 1, f) != 1) return;
    //printf("%*sPosition keys: %d\n", ind, "", n);
    fseek(f, 2, SEEK_CUR);
    while (n-- > 0) {
        if (fread(&nf, sizeof(nf), 1, f) != 1) return;
        if (fread(&unkown, sizeof(unkown), 1, f) != 1) return;
        if (fread(&flags, sizeof(flags), 1, f) != 1) return;
        //printf("%*s  Frame %3d: Flags 0x%X\n", ind, "", nf, flags);
        SplineFlagsReader(f, ind, flags);
        if (fread(&pos, sizeof(pos), 1, f) != 1) return;
        //printf("%*s             X: %f, Y: %f, Z: %f\n",
        //       ind, "", pos[0], pos[1], pos[2]);
    }
}

void TrackRotReader(FILE *f, int ind, long p) {
    word n, nf;
    float pos[4];
    word unkown;
    word flags;

    fseek(f, 10, SEEK_CUR);
    if (fread(&n, sizeof(n), 1, f) != 1) return;
    //printf("%*sRotation keys: %d\n", ind, "", n);
    fseek(f, 2, SEEK_CUR);
    while (n-- > 0) {
        if (fread(&nf, sizeof(nf), 1, f) != 1) return;
        if (fread(&unkown, sizeof(unkown), 1, f) != 1) return;
        if (fread(&flags, sizeof(flags), 1, f) != 1) return;
        //printf("%*s  Frame %3d: Flags 0x%X\n", ind, "", nf, flags);
        SplineFlagsReader(f, ind, flags);
        if (fread(&pos, sizeof(pos), 1, f) != 1) return;
        //printf("%*s             Angle: %f, X: %f, Y: %f, Z: %f\n",
        //       ind, "", pos[0]*180.0/3.141592687, pos[1], pos[2], pos[3]);
    }
}

void TrackScaleReader(FILE *f, int ind, long p) {
    word n, nf;
    float pos[3];
    word unkown;
    word flags;

    fseek(f, 10, SEEK_CUR);
    if (fread(&n, sizeof(n), 1, f) != 1) return;
    //printf("%*sScale keys: %d\n", ind, "", n);
    fseek(f, 2, SEEK_CUR);
    while (n-- > 0) {
        if (fread(&nf, sizeof(nf), 1, f) != 1) return;
        if (fread(&unkown, sizeof(unkown), 1, f) != 1) return;
        if (fread(&flags, sizeof(flags), 1, f) != 1) return;
        //printf("%*s  Frame %3d: Flags 0x%X\n", ind, "", nf, flags);
        SplineFlagsReader(f, ind, flags);
        if (fread(&pos, sizeof(pos), 1, f) != 1) return;
        //printf("%*s            X: %f, Y: %f, Z: %f\n",
        //       ind, "", pos[0], pos[1], pos[2]);
    }
}

void ObjNumberReader(FILE *f, int ind, long p) {
    word n;

    if (fread(&n, sizeof(n), 1, f) != 1) return;
    //printf("%*sObject number: %d\n", ind, "", n);
}


// ------------------------------------

struct {
    word id;
    const char *name;
    void (*func)(FILE *f, int ind, long p);
} ChunkNames[] = {
    {CHUNK_RGBF,        "RGB float",        RGBFReader},
    {CHUNK_RGBB,        "RGB byte",         RGBBReader},

    {CHUNK_PRJ,         "Project",          NULL},
    {CHUNK_MLI,         "Material Library", NULL},

    {CHUNK_MAIN,        "Main",             NULL},
    {CHUNK_OBJMESH,     "Object Mesh",      NULL},
    {CHUNK_BKGCOLOR,    "Background color", NULL},
    {CHUNK_AMBCOLOR,    "Ambient color",    NULL},
    {CHUNK_OBJBLOCK,    "Object Block",     ObjBlockReader},
    {CHUNK_TRIMESH,     "Tri-Mesh",         NULL},
    {CHUNK_VERTLIST,    "Vertex list",      VertListReader},
    {CHUNK_FACELIST,    "Face list",        FaceListReader},
    {CHUNK_FACEMAT,     "Face material",    FaceMatReader},
    {CHUNK_MAPLIST,     "Mappings list",    MapListReader},
    {CHUNK_SMOOLIST,    "Smoothings",       SmooListReader},
    {CHUNK_TRMATRIX,    "Matrix",           TrMatrixReader},
    {CHUNK_LIGHT,       "Light",            LightReader},
    {CHUNK_SPOTLIGHT,   "Spotlight",        SpotLightReader},
    {CHUNK_CAMERA,      "Camera",           CameraReader},
    {CHUNK_HIERARCHY,   "Hierarchy",        NULL},

    {CHUNK_VIEWPORT,    "Viewport info",    NULL},
    {CHUNK_MATERIAL,    "Material",         NULL},
    {CHUNK_MATNAME,     "Material name",    MatNameReader},
    {CHUNK_AMBIENT,     "Ambient color",    NULL},
    {CHUNK_DIFFUSE,     "Diffuse color",    NULL},
    {CHUNK_SPECULAR,    "Specular color",   NULL},
    {CHUNK_TEXTURE,     "Texture map",      NULL},
    {CHUNK_BUMPMAP,     "Bump map",         NULL},
    {CHUNK_MAPFILE,     "Map filename",     MapFileReader},

    {CHUNK_KEYFRAMER,   "Keyframer data",   NULL},
    {CHUNK_AMBIENTKEY,  "Ambient key",      NULL},
    {CHUNK_TRACKINFO,   "Track info",       NULL},
    {CHUNK_FRAMES,      "Frames",           FramesReader},
    {CHUNK_TRACKOBJNAME,"Track Obj. Name",  TrackObjNameReader},
    {CHUNK_TRACKPIVOT,  "Pivot point",      PivotPointReader},
    {CHUNK_TRACKPOS,    "Position keys",    TrackPosReader},
    {CHUNK_TRACKROTATE, "Rotation keys",    TrackRotReader},
    {CHUNK_TRACKSCALE,  "Scale keys",       TrackScaleReader},
    {CHUNK_OBJNUMBER,   "Object number",    ObjNumberReader},

    {CHUNK_TRACKCAMERA, "Camera track",             NULL},
    {CHUNK_TRACKCAMTGT, "Camera target track",      NULL},
    {CHUNK_TRACKLIGHT,  "Pointlight track",         NULL},
    {CHUNK_TRACKLIGTGT, "Pointlight target track",  NULL},
    {CHUNK_TRACKSPOTL,  "Spotlight track",          NULL},
    {CHUNK_TRACKFOV,    "FOV track",                NULL},
    {CHUNK_TRACKROLL,   "Roll track",               NULL},
};

int FindChunk(word id) {
    int i;
    for (i = 0; i < sizeof(ChunkNames)/sizeof(ChunkNames[0]); i++)
        if (id == ChunkNames[i].id)
            return i;
    return -1;
}

// ------------------------------------

void ChunkReader(FILE *f, int ind, long p) {
    TChunkHeader h;
    int n,i;
    long pc;

    while (ftell(f) < p) {
        pc = ftell(f);
        if (fread(&h, sizeof(h), 1, f) != 1) return;
        if (h.len == 0) return;
        n = FindChunk(h.id);
        if (n < 0) {
            fseek(f, pc + h.len, SEEK_SET);
        } else {
            //printf("%*sChunk type \"%s\", offset 0x%lX, size %d bytes\n",
            //       ind, "", ChunkNames[n].name, pc, h.len);
            switch (ChunkNames[n].id){
                case   CHUNK_LIGHT: nblig++; break;
                case  CHUNK_CAMERA: nbcam++; break;
                case CHUNK_TRIMESH: if (ob[nbobj]!=NULL){
                                        adjust_object(ob[nbobj]);
                                        calc_normals(ob[nbobj]);
                                        setmatrix(&ob[nbobj]->sysaxe.matrice,65536,0,0,0,65536,0,0,0,65536);
                                        ob[nbobj]->level=0;
                                        ob[nbobj]->nbchilds=0;
                                        ob[nbobj]->child=0;
                                        ob[nbobj]->parent=NULL;
                                        for (i=0;i<256;i++) ob[nbobj]->childs[i]=NULL;
                                        setvector(ob[nbobj]->link,0,0,0);
                                    }
                                    nbobj++;
                                    ob[nbobj]=(object3DE *)malloc(sizeof(object3DE));
                                    setmem(ob[nbobj],0,sizeof(object3DE));
                                    ob[nbobj]->type=OBJECT;
                                    break;
            }
            pc = pc + h.len;
            if (ChunkNames[n].func != NULL)
                ChunkNames[n].func(f, ind + 2, pc);
            else
                ChunkReader(f, ind + 2, pc);
            fseek(f, pc, SEEK_SET);
        }
        if (ferror(f))
            break;
    }
}

void load3DS(char *filename) {
    FILE *f;
    long p;
    
    nbobj=-1;
    nbcam=-1;
    nblig=-1;

    f = fopen(filename, "rb");
    if (f == NULL) {
        printf("Can't open 3DS file!\n");
        exit(1);
    }

        // Find file size.
    fseek(f, 0, SEEK_END);
    p = ftell(f);
    fseek(f, 0, SEEK_SET);
        // Go!
    ChunkReader(f, 0, p);
    printf("loaded %i objects %i cameras %i lights\n",nbobj+1,nbcam+1,nblig+1);
}
