/*---------------------------------------------------------------------------
                           THE CHAOS ENGINE - 3DW.C

                  Routines to load and initialize 3dw files
---------------------------------------------------------------------------*/

#include <stdio.h>
#include <fcntl.h>
#include <stdlib.h>
#include <io.h>
#include <math.h>

#include "3dengine.h"
#include "math.h"
#include "3dw.h"

extern poly_3d **polylist_pp;

char used_textures[1024], **textures_pp;
char *texturebase_p=NULL;
FILE *file;
TDW_header tdwhead;
level leveldata;

/*---------------------------------------------------------------------------
                        Open and prepare a 3DW file
---------------------------------------------------------------------------*/

int Init_3DW(char *name)
{
int i;
texture_header texhead;
object_header objhead;
level_header levhead;

file=NULL;
textures_pp=NULL;

if (NULL==(file=fopen(name, "rb"))) {
   printf("Can't open 3DW file '%s'\n", name);
   return(NULL);
}

if (NULL==(textures_pp=(char **)malloc(4096))) {
   printf("Not enough memory for texture table\n");
   fclose(file);
   return(NULL);
}

fread(&tdwhead, sizeof(tdwhead), 1, file);
printf("Examining 3DWfile '%s'...\n",name);
printf(" %d Texures at offset %d\n", tdwhead.textures, tdwhead.textures_start);
printf(" %d Objects at offset %d\n", tdwhead.objects, tdwhead.objects_start);
printf(" %d Levels at offset %d\n", tdwhead.levels, tdwhead.levels_start);
return(TRUE);
}

/*---------------------------------------------------------------------------
                         Close opened 3dw file
---------------------------------------------------------------------------*/

void Close_3DW(void)
{
if (textures_pp!=NULL) free(textures_pp);
if (file!=NULL) fclose(file);
}

/*---------------------------------------------------------------------------
                        Load and initialize a level
---------------------------------------------------------------------------*/

level *Load_Level(int level)
{
char *texture_p,*load_p;
int i,x,y,total_polys,textures,texbufsize;
object_3d *obj_p;
texture_header texhead;
level_header levhead;

// Clear texture-flags
memset(used_textures, 0, 1024);

fseek(file, tdwhead.levels_start, SEEK_SET);
fread(&levhead, sizeof(levhead), 1, file);
printf("Loading Level '%s'...\n", levhead.name);

leveldata.objects=levhead.objects;
//printf(" Objects: %i\n",levhead.objects);
leveldata.initpos=levhead.initpos;
leveldata.initdir=levhead.initdir;

Set_ViewPoint(&levhead.initpos);
Set_ViewDirection(&levhead.initdir);
Set_LightSource(&levhead.lightsrc);
Set_Ambience(levhead.ambience);
//SetPalette(levhead.palette);

// Load level objects

total_polys=0;
for (i=0; i<levhead.objects; i++) {
   obj_p=Load_Object();
   if (obj_p==NULL) {
      printf("Load_Level : Load object error\n");
      return(NULL);
   }
   Add_Object(obj_p);
   total_polys+=obj_p->polys;
}

textures=0;
for (i=0; i<tdwhead.textures; i++)
   if (used_textures[i]) textures++;

texbufsize=(textures*TEXTURE_SIZE+256)*TEXTURE_SIZE;

if (texbufsize) {
   if (NULL==(texturebase_p=(char *)malloc(65535+texbufsize))) {
      printf("Load_Level : Not enough memory for textures.\n");
      return(NULL);
   }
//   printf("Texture Base=%d\n",texturebase_p);

   // Align texture buffer to 64K
   texture_p=(char *)((unsigned long)(texturebase_p + 65535) & 0xffff0000);

   fseek(file, tdwhead.textures_start, SEEK_SET);

   // Load used textures

   for (i=0; i<tdwhead.textures; i++) {
      fread(&texhead, sizeof(texhead), 1, file);
      if (used_textures[i]) {
         if (texhead.xsize!=TEXTURE_SIZE || texhead.ysize!=TEXTURE_SIZE) {
            printf("Load_Level : Illegal texture size.");
            free(texturebase_p);
            return(NULL);
         }
         load_p=texture_p;
//         printf("Loading Texture %d\n",i);

         for (y=0; y<TEXTURE_SIZE; y++) {
            fread(load_p, TEXTURE_SIZE, 1, file);
            load_p+=256;
         }
         textures_pp[i]=texture_p;
         texture_p+=TEXTURE_SIZE;
         if (i & (255/TEXTURE_SIZE)==(255/TEXTURE_SIZE))
            texture_p+=256*TEXTURE_SIZE-256;
      }
   }

}

// Allocate memory for polygonlist

printf("Total number of polygons: %d\n", total_polys);
polylist_pp=NULL;
if (NULL==(polylist_pp=(poly_3d **)malloc(total_polys*sizeof(poly_3d *)))) {
   printf("Not Enough memory for polylist.\n");
   return(NULL);
}
return(&leveldata);
}

void Close_Level(void)
{
if (texturebase_p!=NULL) free(texturebase_p);
if (polylist_pp!=NULL) free(polylist_pp);
Free_Objects();
}


/*---------------------------------------------------------------------------
                      Load and initialize a 3DW object
---------------------------------------------------------------------------*/

object_3d *Load_Object(void)
{
char gouraud;
int i,startpos,p;
vertice_3d *tmp_p;
object_3d *obj_p;
level_object levhead;
object_header head;
facehead face;

// Allocate memory for object structure

if (NULL==(obj_p=(object_3d *)calloc(sizeof(object_3d), 1))) {
   printf("Minne!");      
   return(NULL);
}

fread(&levhead, sizeof(levhead), 1, file);     // Read the header
startpos=ftell(file);                          // Save file pos

obj_p->flags=levhead.flags;                    // Copy header data
obj_p->world=levhead.initpos;
obj_p->angle=levhead.initdir;
obj_p->rotation=levhead.initrot;

fseek(file, levhead.filepos, SEEK_SET);        //Read objectdata header
fread(&head, sizeof(head), 1, file);

obj_p->vertices=head.vertices;                 // Copy object info
obj_p->polys=head.faces;
obj_p->max_radius=head.max_radius;
obj_p->average_radius=head.average_radius;

//printf("Vertices %d, Faces %d\n", head.vertices, head.faces);

//printf("Rot: %d, %d, %d\n", obj_p->rot.x, obj_p->rot.y, obj_p->rot.z);
//printf("Vertices : %i Polys : %i\n", obj_p->vertices, obj_p->polys);

// Allocate memory for faces and transformed/projected vertices

if (!(obj_p->transformed_p=(vertice_3d *)calloc(head.vertices*(sizeof(vertice_3d) + sizeof(pix)) + head.faces * sizeof(poly_3d), 1)))
   return(NULL);
obj_p->polys_p=(poly_3d *)(obj_p->transformed_p + head.vertices);
obj_p->projected_p=(pix *)(obj_p->polys_p + head.faces);

// Allocate memory for vertices

if (!(obj_p->vertices_p=(vertice_3d *)malloc(head.vertices*sizeof(vertice_3d)))) {
   printf("Memory\n");
   return(NULL);
}

// Read the vertices and copy them to transformed_p...

fread (obj_p->vertices_p, sizeof(vertice_3d), obj_p->vertices, file);
memcpy(obj_p->transformed_p, obj_p->vertices_p, obj_p->vertices*sizeof(vertice_3d));

// Read faces...

gouraud=FALSE;
for (i=0; i<obj_p->polys; i++) {
   fread (&face, sizeof(facehead), 1, file);
   obj_p->polys_p[i].homeobj_p=(object_3d *)obj_p;
   obj_p->polys_p[i].flags=face.flags | POLY_VISIBLE;
   if (face.flags & POLY_GOURAUDSHADED) gouraud=TRUE;
   obj_p->polys_p[i].tmapindx=face.tmapindx;
   used_textures[face.tmapindx]=TRUE;
   obj_p->polys_p[i].points=face.points;
   if (face.points==3) obj_p->polys_p[i].flags|=POLY_TRIANGLE;
   obj_p->polys_p[i].color=face.color;
   obj_p->polys_p[i].shade=face.color;

   // Read vertex indexes
   fread(obj_p->polys_p[i].pindxs, face.points, 2, file);

}

// Scale object if required

if (levhead.modifyflags & MODIFY_SCALE) {
   for (i=0; i<head.vertices; i++) {
      obj_p->vertices_p[i].x*=levhead.object_scale;
      obj_p->vertices_p[i].y*=levhead.object_scale;
      obj_p->vertices_p[i].z*=levhead.object_scale;
   }
   obj_p->max_radius*=levhead.object_scale;
   obj_p->average_radius*=levhead.object_scale;
}

// Assign face attributes

if (levhead.modifyflags & MODIFY_POLYFLAGS) {
   for (i=0; i<obj_p->polys; i++) obj_p->polys_p[i].flags=levhead.polyflags;
   if (levhead.polyflags & POLY_GOURAUDSHADED) gouraud=TRUE;
}
if (levhead.modifyflags & MODIFY_POLYCOLOR)
   for (i=0; i<obj_p->polys; i++) obj_p->polys_p[i].color=levhead.polycolor;

if (levhead.modifyflags & MODIFY_POLYTEXTURE) {
   for (i=0; i<obj_p->polys; i++) obj_p->polys_p[i].tmapindx=levhead.polytexture;
   used_textures[levhead.polytexture]=TRUE;
}

if (levhead.modifyflags & MODIFY_COLORPATTERN) {
   p=0;
   for (i=0; i<obj_p->polys; i++) {
      obj_p->polys_p[i].color=levhead.colorpattern[p++];
      if (p>=levhead.colors) p=0;
   }
}

// Allocate memory for gouraud stuff if needed

if (gouraud) {
   if (NULL==(tmp_p=(vertice_3d *)realloc(obj_p->vertices_p, 2*head.vertices*sizeof(vertice_3d)+head.vertices)))
      return(NULL);
   obj_p->vertices_p=tmp_p;
   obj_p->vnormals_p=(tmp_p+head.vertices);
   obj_p->vertex_colors_p=(char *)(obj_p->vnormals_p+head.vertices);
   obj_p->normals=obj_p->vertices;
   obj_p->flags |= OBJ_GOURAUD;
}

obj_p->flags|=OBJ_LOADED;

if (!Init_Object(obj_p)) {
   printf("Load_Object : Init Objects Error\n");
   return(NULL);
}

fseek(file, startpos, SEEK_SET);
return(obj_p);
}

/*---------------------------------------------------------------------------
        Initialize object, precalculate gouraudshading normals etc
---------------------------------------------------------------------------*/

int Init_Object(object_3d *obj_p)
{
int i,p,p1,p2,p3;
vector_3d u,v,normal,*fnormals_p;
poly_3d *poly_p;

// Allocate temporary space for face-normals

if (NULL==(fnormals_p=(vector_3d *)calloc(obj_p->polys*sizeof(vector_3d), 1))) return(NULL);

// Calculate the surface normal for all faces

for (i=0; i<obj_p->polys; i++) {

   // Get offsets to three corners

   p1=obj_p->polys_p[i].pindxs[0];
   p2=obj_p->polys_p[i].pindxs[1];
   p3=obj_p->polys_p[i].pindxs[2];

   // Get two vectors

   u.x = obj_p->vertices_p[p1].x - obj_p->vertices_p[p2].x;
   u.y = obj_p->vertices_p[p1].y - obj_p->vertices_p[p2].y;
   u.z = obj_p->vertices_p[p1].z - obj_p->vertices_p[p2].z;

   v.x = obj_p->vertices_p[p1].x - obj_p->vertices_p[p3].x;
   v.y = obj_p->vertices_p[p1].y - obj_p->vertices_p[p3].y;
   v.z = obj_p->vertices_p[p1].z - obj_p->vertices_p[p3].z;

   // Calculate the surface-normal

   Cross_Product(&v, &u, &normal);
   fnormals_p[i]=normal;

   // Calculate the length of the vector and save it.

   obj_p->polys_p[i].normal_len=Vector_Magnitude(&normal);
}

if (obj_p->flags & OBJ_GOURAUD) {

   // Clear vertex-normal buffer

   memset(obj_p->vnormals_p, 0, obj_p->vertices * sizeof(vertice_3d));

   // Precalculate vertex-normals for gouraud shading.

   for (i=0; i<obj_p->polys; i++) {
      poly_p=&(obj_p->polys_p[i]);
      for (p=0; p<poly_p->points; p++) {
         obj_p->vnormals_p[poly_p->pindxs[p]].x+=fnormals_p[i].x;
         obj_p->vnormals_p[poly_p->pindxs[p]].y+=fnormals_p[i].y;
         obj_p->vnormals_p[poly_p->pindxs[p]].z+=fnormals_p[i].z;
      }
   }

   // Normalise vertex normals

   for (i=0; i<obj_p->vertices; i++) {
      Normalize_Vector(&(obj_p->vnormals_p[i]));
    //printf("Normal len : %f\n",Vector_Magnitude(&(obj_p->vertices_p[i+obj_p->vertices])));
   }

}

free(fnormals_p);
return(TRUE);
}
