/*---------------------------------------------------------------------------
                         THE CHAOS ENGINE - 3DW.H

---------------------------------------------------------------------------*/

#define ID_3DW 0x57443300

#define MODIFY_POLYFLAGS 1
#define MODIFY_POLYCOLOR 2
#define MODIFY_POLYTEXTURE 4
#define MODIFY_OBJSCALE 8
#define MODIFY_SCALE 16
#define MODIFY_COLORPATTERN 32

//-- 3DW file format structures ---------------------------------------------

#pragma pack (push);
#pragma pack (1);           // No structure member alignement

typedef struct {
   int id,flags;
   int textures_start,objects_start,levels_start;
   int textures,objects,levels;
} TDW_header;

typedef struct {
   short xsize,ysize,bytes_per_pix,flags;
} texture_header;

typedef struct {
   char name[32];
   int objects,ambience,flags,size,texturesize;
   vertice_3d initpos;
   vector_3d lightsrc;
   dir initdir;
   char palette[768];
} level_header;

typedef struct {
   int vertices,faces,size;
   float max_radius, average_radius;
} object_header;

typedef struct {
   int filepos,flags;
   short modifyflags,polyflags,polycolor,polytexture;
   short colors,colorpattern[8];
   vertice_3d initpos;
   dir initdir,initrot;
   float max_radius,average_radius,object_scale;
} level_object;

typedef struct {
   short flags,tmapindx,color,points;
} facehead;

typedef struct {
   short flags,tmapindx,color,points;
   short indexs[8];
} facedata;

typedef union {
   facehead;
   facedata;
} face;

#pragma pack (pop);            // Restore original alignement

typedef struct {
   object_header head;
   vertice_3d *vertices_p;
   face *faces_p;
} loaded_object;

//-- Function Prototypes ----------------------------------------------------

int Init_3DW(char *name);
void Close_3DW(void);
level *Load_Level(int level);
object_3d *Load_Object(void);
