Unit ALPHA2;  {ALPHA ENGINE - FIRST TOTAL REBUILD}

{This is the first total rebuild of the ALPHA engine. I had to redo some
 things, like the rotations and all the polygon routines, to inprove the
 accuracy, the speed and the maintainability of the engine.
 I believe this can not be done much faster (but I'm proven wrong all the
 time...). This engine can shift A LOT of polygons per frame.

 So far, the following features are in:
 - Perfect hierarchical objects
 - VESA support for 320x400 and 640x400 (LFB)
 - Fast PORTAL based indoor and object engine
 - Dynamic lighting
 - Perspective texturemapping (with all the latest speed tricks)
 - Fast and accurate phong shading
 - Environment mapping
 - Classic texturemapping (less accurate, but faster)
 - Build-in support for random mirror objects (via envmapping)
 - Color mixing tables for almost true-color image quality by PITBULL
 - Transparent texturemaps (50/50 mixing with screen buffer)
 - Fast gourad shading
 - Gourad shaded perspective correct texturemapping
 - Fast plain polygon rendering
 - Highly accurate rotations
 - High performance
 - Protected mode code
 - Perfect z-clipping
 - Full 2D clipping
 - Polygons with 3-6 vertices, easily extendable
 - Rock solid design
 - Hidden surface removal
 - Advanced Tree Sorting
 - Support for 3D Studio .ASC and binary files (.3DS) by PITBULL
 - Support for the .PLG and .GEM file formats by PITBULL
 - PCX texturefile support by PITBULL

 To be added (already done in the first version of ALPHA):
 - Gourad shaded perspective correct texturemapping
 - Phong shaded perspective correct texturemapping
 - Bump mapping (the twisted phong type), both plain and textured
 - Mirror surface rendering (real mirrors)

 To be added (still on my wish list):
 - Mipmapping for better looking distant objects
 - 32K, 64K and 16M modes (true color) support
 - Random texture loading by dynamic palette reorganization
 - 32bit coord records to decrease cache misses
 - Transparent phong-maps
 - Smoke effects (via transparent/zoomed sprite spheres)
 - Collision detection

 Dropped ideas:
 - BSP support via external compiler (JALUIT, Marchini!)
   I think portals are better, faster and easier to implement
   but maybe the last reason is the most important one... :)
 - Polygons with gaps (invisible color 0)

 Known bugs:
 - Outlines are shaky again
 - Clockwise cannot check polygons >200 units
 - Objects completely suck again...

 This source is freeware. You may use it in your own programs.
 Use the TMT Protected mode Pascal compiler (get it from
 www.tmt.com) to compile the sources.
 You don't have to put my name on the screen, just send me a mail to
 say you like the stuff. My address is Jacco.Bikker@QuestIntl.Com.
 Please spread this source freely. Also visit my homepage @ URL:

 www.geocities.com/CapeCanaveral/5402

 Big thanks to:
 - Sander Kouwenhoven: FPU/Optimizing/386 code/Texture interpolator
                       and a big lot of other things...
 - Chris Lambert:      For the TMT Pascal tip.
 - Andrea Marchini:    For coding the Jaluit BSP compiler.
 - Ben Chauveau:       For the VESA unit, that I ripped apart :)
 - Everyone who visited my page, and showed interest for the 3D
   engine! Keep coming, keep writing, this stuff will be amazing!
}

INTERFACE

uses crt,graphics,polygons,struct,dos;

      {Polygon types}
const TEXTURED=1;     {Standard perspective correct texturemapped polygon}
      TRTEXT=14;      {Transparent perspective correct texture}
      PLAIN=2;        {Single colored polygon}
      GOURAD=7;       {Gourad shaded polygon}
      PHONG=8;        {Phong shaded polygon}
      GOURADTX=5;     {Gourad shaded textured polygon}
      ENVMAP=15;      {Environment-mapped polygon}
      FASTTEXT=16;    {Old fashioned texturemapper}
      PORTAL=19;      {Portal polygon}
      REVPROTAL=20;   {Internal: Reversed Portal Polygon)}
      FASTTRTX=17;    {Transparent low-quality texturemapper}

      SMOOTHED=3;     {Smoothed texturemapped polygon (not implemented yet)}
      SHADETEX=4;     {Color mixed textured polygon (not implemented yet)}
      PHONGTX=6;      {Phong shaded textured polygon (not implemented yet)}
      BUMP=9;         {Bump mapped polygon (not implemented yet)}
      BUMPTX=10;      {Bump mapped textured polygon (not implemented yet)}
      MIRROR=11;      {Mirror surface (not implemented yet)}
      SUN=12;         {Lens flare causing object (not implemented yet)}
      TRANSP=13;      {Transparent plain polygon (not implemented yet)}
      SPRITE=18;      {3D sprite (single-coord polygon) (not finished yet)}

var _treetop: longint;                           {next free address in tree}
    _count: longint;
    buffer: array[0..1023] of byte;

procedure CreateCoord(x,y,z: single);
function  FindCoord(x,y,z: single): longint;
procedure Init3DEngine;
procedure SetRotation(n: longint; rx,ry,rz: longint);
procedure SetPosition(n: longint; px,py,pz: single);
function  CreateObject: longint;
procedure parent;
procedure Render(rx,rz,ry: longint; px,py,pz: single);
procedure EnvRender(rz,rx,ry: longint; px,py,pz: single);
procedure Diagnostics;
procedure DrawTree(n: dword);
procedure CalcVertexNormals;
procedure CalcBasePlanes;
procedure CreateSprite(p1: longint);
procedure CreateBase(p1,p2,p3: longint);
procedure PMapBase(pol,bas: longint);
procedure CreatePolygon(p1,p2,p3,p4,tp,cl: longint);
procedure Create6Plane(p1,p2,p3,p4,p5,p6,tp,cl: longint);
procedure Create5Plane(p1,p2,p3,p4,p5,tp,cl: longint);
procedure CreateTriangle(p1,p2,p3,tp,cl: longint);
procedure PolygonColor(n,col: longint);
procedure SetLight(x,y,z: longint);
procedure DoubleSided(n: Longint);
procedure SingleSided(n: Longint);
procedure PolygonType(n: longint; tp: longint);
procedure CenterObject;
function  LastPolygon: longint;
function  LastCoord: longint;
procedure ForceMipMap;
procedure ZClipping(depth: single);
procedure CalcFrameRate;
procedure Sector(s: longint);
procedure CheckGuards;
procedure DefineSector(n: longint; x1,y1,z1,x2,y2,z2: longint);
function  GetSector(x,y,z: longint): longint;


IMPLEMENTATION


{  _______________________________________________________________________
  /                                                                      /
 / DefineSector: Define bounding box for sector                         /
/______________________________________________________________________/ }

procedure DefineSector(n: longint; x1,y1,z1,x2,y2,z2: longint);
var h,i: longint;
begin
   if (n<1) or (n>MAXSECTR) then Error('Bad sector number','DefineSector');
   if x1>x2 then begin h:=x1; x1:=x2; x2:=h end;
   if y1>y2 then begin h:=y1; y1:=y2; y2:=h end;
   if z1>z2 then begin h:=z1; z1:=z2; z2:=h end;
   _sx1[n]:=x1;
   _sy1[n]:=y1;
   _sz1[n]:=z1;
   _sx2[n]:=x2;
   _sy2[n]:=y2;
   _sz2[n]:=z2;
   i:=n;
   if n>_lsd then _lsd:=n;
end;


{  _______________________________________________________________________
  /                                                                      /
 / GetSector: Find out in wich sector a coordinate is                   /
/______________________________________________________________________/ }

function GetSector(x,y,z: longint): longint;
var s,i: longint;
begin
   asm
      mov [s],0
      mov esi,0
      mov eax,[x]
      mov ebx,[y]
      mov ecx,[z]
   @testloop:
      cmp eax,[_sx1+esi*4]
      jnge @other
      cmp eax,[_sx2+esi*4]
      jge @other
      cmp ebx,[_sy1+esi*4]
      jnge @other
      cmp ebx,[_sy2+esi*4]
      jge @other
      cmp ecx,[_sz1+esi*4]
      jnge @other
      cmp ecx,[_sz2+esi*4]
      jge @other
      mov [s],esi
      inc [s]
      jmp @ready
   @other:
      inc esi
      cmp esi,[_lsd]
      jb  @testloop
   @ready:
   end;
   GetSector:=s;
end;


{  _______________________________________________________________________
  /                                                                      /
 / ResetWindow: Initialize viewing window                               /
/______________________________________________________________________/ }

procedure ResetWindow;
var i,x1,y1: longint;
begin
   for i:=0 to YSIZE-1 do begin
      _cll[i]:=0;
      _clr[i]:=XSIZE-1;
   end;
   _clu:=64;
   _clb:=(YSIZE-1)*64;
   _cbl:=0;
   _cbr:=(XSIZE-1)*64;
   for i:=0 to 20000 do begin
      x1:=round(XCENTR*cos(i*3.1415/2000))+XCENTR;
      y1:=round(YCENTR*sin(i*3.1415/2000))+YCENTR;
      if x1<0 then x1:=0;
      if x1>=XSIZE then x1:=XSIZE-1;
      if (y1>0) and (y1<YSIZE) then begin
         if (x1<XCENTR) then _cll[y1]:=x1 else _clr[y1]:=x1;
      end;
   end;
end;


{  _______________________________________________________________________
  /                                                                      /
 / TestWindow: Test for the portal routines                             /
/______________________________________________________________________/ }

procedure TestWindow;
var i: longint;
begin
   for i:=0 to 99 do begin
      _cll[i]:=50-(i div 2);
      _clr[i]:=269+(i div 2);
   end;
   for i:=0 to 99 do begin
      _cll[i+100]:=i div 2;
      _clr[i+100]:=319-(i div 2);
   end;
   _clu:=20*64;
   _clb:=(YSIZE-20)*64;
end;


{  _______________________________________________________________________
  /                                                                      /
 / CalcFrameRate: Calculate the number of frames used for rendering     /
/______________________________________________________________________/ }

procedure CalcFrameRate;
var ho,mo,so,s1o: word;
    elaps       : longint;
    fps         : single;
begin
   ho:=_hh;
   mo:=_mm;
   so:=_ss;
   s1o:=_s100;
   GetTime(_hh,_mm,_ss,_s100);
   elaps:=(_s100+100*_ss+6000*_mm)-(s1o+100*so+6000*mo);
   if elaps<01 then elaps:=01;
   if elaps>99 then elaps:=99;
   fps:=((100/elaps)+(31*_lfps)) / 32;
   _lfps:=fps;
   WriteNum((round(fps) div 10) mod 10,0,0);
   WriteNum( round(fps)         mod 10,6,0);
end;


{  _______________________________________________________________________
  /                                                                      /
 / ZClipping: Change the location of the clipping plane                 /
/______________________________________________________________________/ }

procedure ZClipping(depth: single);
begin
   if (depth<10) or (depth>10000) then Error('Bad clipping plane','ZClipping');
   _zmin:=depth
end;


{  _______________________________________________________________________
  /                                                                      /
 / LastPolygon: Return number of last defined polygon                   /
/______________________________________________________________________/ }

function LastPolygon: longint;
begin LastPolygon:=_polies end;


{  _______________________________________________________________________
  /                                                                      /
 / LastCoord: Return number of last defined coordinate                  /
/______________________________________________________________________/ }

function LastCoord: longint;
begin LastCoord:=_coords end;


{  _______________________________________________________________________
  /                                                                      /
 / PhongMap: Create a phong light map                                   /
/______________________________________________________________________/ }

procedure PhongMap;
var x,y: longint;
    v1,v2,v3,v4: single;
    w: dword;
begin
   for x:=0 to 127 do begin
      v1:=sin(x*3.1415/256);
      v1:=v1*v1;
      v4:=v1*v1;
      v4:=v4*v4;
      v4:=v4*v4;
      for y:=0 to 127 do begin
         v3:=sin(y*3.1415/256);
         v3:=v3*v3;
         v2:=147*(v1*v3);
         v3:=v3*v3;
         v3:=v3*v3;
         v3:=v3*v3;
         w:=round(v2+147*(v4*v3));
         if w>255 then w:=255;
         if w<40 then w:=40;
         w:=w div 2;
         mem[pmap+x+y*256]:=w;
         mem[pmap+(255-x)+y*256]:=w;
         mem[pmap+(255-x)+(255-y)*256]:=w;
         mem[pmap+x+(255-y)*256]:=w;
      end;
   end;
end;


{  _______________________________________________________________________
  /                                                                      /
 / ShowViewWindow: Show the outlines of the current window              /
/______________________________________________________________________/ }

procedure ShowViewWindow;
var i: longint;
begin
   if _clb>_clu then for i:=_clu div 64 to (_clb div 64)-1 do begin
      if (_cll[i]>0) and (_cll[i]<320) then mem[buf+_cll[i]+i*320]:=255;
      if (_clr[i]>0) and (_clr[i]<320) then mem[buf+_clr[i]+i*320]:=255;
   end;
end;


{  _______________________________________________________________________
  /                                                                      /
 / PolygonType: Alter the type of a single polygon                      /
/______________________________________________________________________/ }

procedure PolygonType(n: longint; tp: longint);
begin
   if n>_polies then Error('Polygon not defined','PolygonType');
   if (tp=0) or (tp>13) then Error('Undefined polygontype','PolygonType');
   _pt[n]:=tp;
end;


{  _______________________________________________________________________
  /                                                                      /
 / TreeSort: Create a tree containing all the polygons                  /
/______________________________________________________________________/ }

procedure TreeSort(adr: dword; sect: longint); Assembler;
var cpoly,loopc,curad,count: longint;
    xlow,ylow,xhi,yhi,zlow,zhi: longint;
asm
   mov [_count],0
   {Calculate address in sector array}
   mov ebx,[sect]
   mov eax,MAXSECSR
   dec ebx
   mov ecx,[_nsr+ebx*4]
   mov [loopc],ecx
   imul ebx
   mov [curad],eax
   {Create a root node for the tree}
   mov eax,[adr]
   mov dword ptr [_Tree+eax],0
   mov dword ptr [_Tree+eax+4],0
   mov dword ptr [_Tree+eax+8],0
   add eax,16
@forloop:
   {Retrieve polygonnumber from sector array}
   mov ebx,[curad]                                  {EBX=Recnr in _sec}
   mov ecx,[_sec+ebx*4]
   dec ecx                                          {ECX=polynr from _sec}
   mov ebx,ecx
   shl ebx,2
   add ebx,ecx
   add ebx,ecx
   shl ebx,2                                        {EBX=polynr*24}
   {Test for visibility}
   push ebx
   mov edi,[_pp+ebx]
   mov edx,[_xs+edi*4-4]
   mov [xlow],edx
   mov [xhi],edx
   mov edx,[_ys+edi*4-4]
   mov [yhi],edx
   mov [ylow],edx
   mov edx,[_rz+edi*4-4]
   mov [zlow],edx
   mov [zhi],edx
   mov edx,[_pnv+ecx*4]
   dec edx
   mov [count],edx
@coordloop:
   add ebx,4
   mov edi,[_pp+ebx]
   mov edx,[_xs+edi*4-4]
   cmp edx,[xlow]
   jge @noxlower
   mov [xlow],edx
@noxlower:
   cmp edx,[xhi]
   jnge @noxhigher
   mov [xhi],edx
@noxhigher:
   mov edx,[_ys+edi*4-4]
   cmp edx,[ylow]
   jge @noylower
   mov [ylow],edx
@noylower:
   cmp edx,[yhi]
   jnge @noyhigher
   mov [yhi],edx
@noyhigher:
   mov edx,[_rz+edi*4-4]
   cmp edx,[zlow]
   jge @nozlower
   mov [zlow],edx
@nozlower:
   cmp edx,[zhi]
   jnge @nozhigher
   mov [zhi],edx
@nozhigher:
   dec [count]
   jnz @coordloop
   pop ebx
   {Test for clockwise or double sided}
   mov edi,[_pp+ebx]
   cmp dword ptr [_pd+ecx*4],1
   jz @skipclockwise
   push esi
   push eax
   mov eax,[xlow]
   cmp eax,[_cbr]
   jge @noinsert
   mov eax,[ylow]
   cmp eax,[_clb]
   jge @noinsert
   mov eax,[xhi]
   cmp eax, [_cbl]
   jnge @noinsert
   mov eax,[yhi]
   cmp eax,[_clu]
   jnge @noinsert
   mov edx,[_zmin]
   cmp [zhi],edx
   jge @noinsert
   mov esi,[_pp+ebx+4]
   mov edx,[_pp+ebx+8]
   mov eax,[_ys+esi*4-4]
   sub eax,[_ys+edi*4-4]
   sar eax,6
   mov edx,[_xs+edx*4-4]
   sub edx,[_xs+edi*4-4]
   sar edx,6
   imul edx
   mov [_left],eax
   mov edx,[_pp+ebx+8]
   mov eax,[_ys+edx*4-4]
   sub eax,[_ys+edi*4-4]
   sar eax,6
   mov edx,[_xs+esi*4-4]
   sub edx,[_xs+edi*4-4]
   sar edx,6
   imul edx
   cmp [_left],eax
   jg  @noinsert
   pop eax
   pop esi
@skipclockwise:
   {determine depth of current polygon}
   mov edx,[_rz+edi*4-4]
   mov edi,[_pp+ebx+8]
   add edx,[_rz+edi*4-4]   {d:=_rz[_pp[i,1]]+_rz[_pp[i,3]];}
   {Create a node for this polygon and insert it in the tree}
@normalinsert:
   mov dword ptr [_Tree+eax],edx        {depth}
   mov dword ptr [_Tree+eax+4],0        {left branch}
   mov dword ptr [_Tree+eax+8],0        {right branch}
   mov dword ptr [_Tree+eax+12],ecx     {polygon number}
   mov [_Treetop],eax
   mov esi,eax                    {addres of inserted leaf}
   mov edi,[adr]
@while:
   cmp edx,[_Tree+edi]
   jnge @elsebranch
   cmp dword ptr [_Tree+edi+4],0
   jnz @ifelse
   mov [_Tree+edi+4],esi
   jmp @inserted
@ifelse:
   mov edi,[_Tree+edi+4]
   jmp @while
@elsebranch:
   cmp dword ptr [_Tree+edi+8],0
   jnz @elseelse
   mov [_Tree+edi+8],esi
   jmp @inserted
@elseelse:
   mov edi,[_Tree+edi+8]
   jmp @while
@noinsert:
   pop eax
   pop esi
   inc [curad]
   dec [loopc]
   jnz @forloop
   jmp @ready
@inserted:
   add eax,16
   inc [curad]
   inc [_count]
   dec [loopc]
   jnz @forloop
@ready:
end;


{  _______________________________________________________________________
  /                                                                      /
 /  OnScreen: Check whether a polygon is (partially) on screen          /
/______________________________________________________________________/ }

function OnScreen(poly: longint): boolean;
var i,maxx,maxy,minx,miny: longint;
begin
   minx:=100000;
   miny:=100000;
   maxx:=-100000;
   maxy:=-100000;
   for i:=1 to _pnv[poly] do begin
      if _xs[_pp[poly,i]]<minx then minx:=_xs[_pp[poly,i]];
      if _xs[_pp[poly,i]]>maxx then maxx:=_xs[_pp[poly,i]];
      if _ys[_pp[poly,i]]<miny then miny:=_ys[_pp[poly,i]];
      if _ys[_pp[poly,i]]>maxy then maxy:=_ys[_pp[poly,i]];
   end;
   if (minx>(XSIZE*64)) or (miny>(YSIZE*64))
   or (maxx<0) or (maxy<0) then OnScreen:=False else OnScreen:=True;
end;


{  _______________________________________________________________________
  /                                                                      /
 /  DrawPortalPolygon: Create new viewing window                        /
/______________________________________________________________________/ }

procedure DrawPortalPolygon(poly: longint);
var x3,y3,x4,y4,t: single;
    z3,z4,_ylow,_yhigh: longint;
    i,_tn,oldcrds,p1,p2: longint;
    x1,y1,x2,y2,ixh,ixl: longint;
    tx,ty,tz: single;
    treebase,oldsect,oldtop,oldbot: longint;
    minx,maxx,oldl,oldr: longint;
begin
   poly:=poly+1;
   if (_dr[poly]=0) and OnScreen(poly) then begin
      _dr[poly]:=1;
      inc(_depth);
      if _depth>8 then Error('Too many portals in portals','DrawPortalPolygon');
      {Store old view window}
      for i:=0 to YSIZE do begin
         oldwinl[_depth,i]:=_cll[i];
         oldwinr[_depth,i]:=_clr[i];
      end;
      oldtop:=_clu;
      oldbot:=_clb;
      {z-clipping: Make sure that the polygon is not partially behind viewer}
      _tn:=0;
      oldcrds:=_coords;
      for i:=1 to _pnv[poly] do begin
         {Determine begin and end of one of the polygon's edges}
         p1:=_pp[poly,i];
         z3:=_rz[p1];
         if i<_pnv[poly] then p2:=_pp[poly,i+1] else p2:=_pp[poly,1];
         z4:=_rz[p2];
         {Case 1: Both in, output end coord}
         if (z3>50) and (z4>50) then begin
            inc(_tn);
            _tmp[_tn]:=p2
         end else if (z3>50) and (z4<=50) then begin
         {Case 2: Going in: Output clipped coord}
            x3:=_rx[p1];
            y3:=_ry[p1];
            x4:=_rx[p2];
            y4:=_ry[p2];
            t:=(50-_frz[p1])/(_frz[p2]-_frz[p1]);
            inc(_coords);
            tx:=x3+(x4-x3)*t;
            ty:=y3+(y4-y3)*t;
            _xs[_coords]:=round((tx*DIST*64)/50)+XCENTR*64;
            _ys[_coords]:=round((ty*DIST*64)/50)+YCENTR*64;
            inc(_tn);
            _tmp[_tn]:=_coords;
         end else if (z3<=50) and (z4>50) then begin
            {Case 3: Coming out: Output clipped coord and end coord}
            x3:=_rx[p1];
            y3:=_ry[p1];
            x4:=_rx[p2];
            y4:=_ry[p2];
            t:=(50-_frz[p1])/(_frz[p2]-_frz[p1]);
            inc(_coords);
            tx:=x3+(x4-x3)*t;
            ty:=y3+(y4-y3)*t;
            _xs[_coords]:=round((tx*DIST*64)/50)+XCENTR*64;
            _ys[_coords]:=round((ty*DIST*64)/50)+YCENTR*64;
            inc(_tn,2);
            _tmp[_tn-1]:=_coords;
            _tmp[_tn]:=p2;
         end;
      end;  {_tmp is array with coords for polygon, _tn is number of vertices}
      _coords:=oldcrds;
      minx:=_xs[_tmp[1]];
      maxx:=minx;
      oldl:=_cbl;
      oldr:=_cbr;
      for i:=2 to _tn do begin
         if _xs[_tmp[i]]<minx then minx:=_xs[_tmp[i]];
         if _xs[_tmp[i]]>maxx then maxx:=_xs[_tmp[i]];
      end;
      _cbl:=minx;
      _cbr:=maxx;
      asm
         mov ebx,[_tn]
         cmp ebx,3
         jb  @noportal
         dec ebx
         mov esi,dword ptr [_tmp]
         mov edi,[_ys+esi*4-4]
         mov ecx,edi
      @findloop:
         mov esi,[_tmp+ebx*4]
         mov eax,[_ys+esi*4-4]
         cmp eax,edi
         jge @nolowupdate
         mov edi,eax
      @nolowupdate:
         cmp eax,ecx
         jnge @nohiupdate
         mov ecx,eax
      @nohiupdate:
         dec ebx
         jnz @findloop
         cmp edi,[_clu]
         jge @nozeroclip
         mov edi,[_clu]
      @nozeroclip:
         cmp ecx,[_clu]
         jnge @noportal
         cmp edi,[_clb]
         jge @noportal
         cmp ecx,[_clb]
         jnge @nobotclip
         mov ecx,[_clb]
      @nobotclip:
         sar ecx,6
         sar edi,6
         cmp ecx,edi
         jbe @noportal
         mov [_ylow],edi
         mov [_yhigh],ecx
         sub ecx,edi
         xor eax,eax
         mov ebx,XSIZE-1
      @clearloop:
         mov [_edgeb+edi*4],ebx
         mov [_edgee+edi*4],eax
         inc edi
         dec ecx
         jnz @clearloop
         mov [i],0
      @foriloop:
         inc [i]
         {Get x1/y1/x2/y2, wrap around for last vertex}
         mov esi,[i]
         cmp esi,[_tn]
         ja  @endofiloop
         mov edi,[_tmp+4*esi-4]
         jz  @1
         mov esi,[_tmp+4*esi]
         jmp @n
      @1: mov esi,dword ptr [_tmp]
      @n: mov edx,[_xs+4*edi-4]  {hehe not a single AGI stall}
         mov eax,[_ys+4*edi-4]
         mov ebx,[_xs+4*esi-4]
         mov ecx,[_ys+4*esi-4]
         mov [x1],edx
         mov [y1],eax
         mov [x2],ebx
         mov [y2],ecx
         cmp eax,ecx
         jz  @endofline
         jnge @noswap
         mov ecx,[y1]
         mov eax,[y2]
         mov edx,[x2]
         mov ebx,[x1]
         mov [x1],edx
         mov [x2],ebx        {registers: eax = y1  }
         mov [y1],eax                {   ebx = x2  }
         mov [y2],ecx                {   ecx = y2  }
      @noswap:                       {   edx = x1  }
         cmp eax,[_clb]
         jge @endofline
         cmp ecx,[_clu]
         jnge @endofline
         cmp eax,ecx
         jge @endofline      {if (y1<YSIZE) and (y2>0) and (y2>y1)}
         cmp eax,[_clu]
         jge @notopclipping  {if y1<0 then begin}
         sub ecx,[y1]
         xor eax,-1
         add eax,[_clu]
         sal eax,8
         cdq; idiv ecx
         mov esi,[x2]
         sub esi,[x1]
         mov ecx,[_clu]
         mov [y1],ecx
         imul esi
         sar eax,8
         add [x1],eax
         mov ecx,[y2]
      @notopclipping:
         {Fixed point line drawing}
         sub ecx,[y1]        {ecx = delta-y}
         jz  @endofline
         mov eax,[x2]
         sub eax,[x1]        {eax = delta-x}
         sal eax,10
         cdq
         idiv ecx            {eax = delta-x/delta-y, f.p. 10bit}
         mov esi,[y1]        {pointer in outline table}
         sar esi,6
         mov ebx,[y2]
         cmp ebx,[_clb]
         jnge @drawline      {clipping to bottom of screen needed?}
         sub ebx,[_clb]
         sub ecx,ebx         {then do it}
      @drawline:
         shr ecx,6
         inc ecx
         or  ecx,ecx
         jz  @endofline
         mov ebx,[x1]
         sal ebx,(10-6)
      @pixels:
         mov edx,ebx
         sar edx,10
         cmp edx,[_edgeb+esi*4]
         jge @nobupdate
         mov [_edgeb+esi*4],edx
      @nobupdate:
         cmp edx,[_edgee+esi*4]
         jnge @noeupdate
         mov [_edgee+esi*4],edx
      @noeupdate:
         add ebx,eax
         inc esi
         dec ecx
         jnz @pixels
      @endofline:
         jmp @foriloop
      @endofiloop:
         {Copy outlines to custom view window array}
         mov edi,[_ylow]
         shl edi,2
         mov ecx,[_yhigh]
         sub ecx,[_ylow]
      @cloop:
         mov eax,[_edgeb+edi]
         cmp eax,[_cll+edi]
         jge @noleftclip1
         mov eax,[_cll+edi]
      @noleftclip1:
         cmp eax,[_clr+edi]
         jnge @norightclip1
         mov eax,[_clr+edi]
      @norightclip1:
         mov ebx,[_edgee+edi]
         cmp ebx,[_cll+edi]
         jge @noleftclip2
         mov ebx,[_cll+edi]
      @noleftclip2:
         cmp ebx,[_clr+edi]
         jnge @norightclip2
         mov ebx,[_clr+edi]
      @norightclip2:
         mov [_cll+edi],eax
         mov [_clr+edi],ebx
         add edi,4
         dec ecx
         jnz @cloop
         mov eax,[_ylow]
         shl eax,6
         mov ebx,[_yhigh]
         shl ebx,6
         mov [_clu],eax
         mov [_clb],ebx
         jmp @endofpoly
      @noportal:
         mov [_clu],0
         mov [_clb],0
      @endofpoly:
      end;
      {Fill the portal}
      if _clb>_clu then begin
         _TreeTop:=_TreeTop+16;
         TreeBase:=_TreeTop;
         oldsect:=_sector;
         _sector:=_pc[poly];
         TreeSort(TreeBase,_sector);
         asm
            mov edi,[TreeBase]
            cmp dword ptr [_Tree+edi+4],0
            jz  @noleftbranch
            push edi
            mov eax,[_Tree+edi+4]
            push eax
            Call DrawTree
            pop edi
         @noleftbranch:
            cmp dword ptr [_Tree+edi+8],0
            jz  @norightbranch
            mov eax,[_Tree+edi+8]
            push eax
            Call DrawTree
         @norightbranch:
         end;
         {Restore old view window}
         for i:=0 to YSIZE do begin
            _cll[i]:=oldwinl[_depth,i];
            _clr[i]:=oldwinr[_depth,i];
         end;
         _sector:=oldsect;
      end;
      _clu:=oldtop;
      _clb:=oldbot;
      _cbl:=oldl;
      _cbr:=oldr;
      dec(_depth);
   end;
end;


{  _______________________________________________________________________
  /                                                                      /
 / DrawTree: Draw the polygon tree                                      /
/______________________________________________________________________/ }

procedure DrawTree(n: dword); Assembler;
asm
   mov edi,[n]
   cmp dword ptr [_Tree+edi+4],0
   jz  @skipleft
   push edi
   push dword ptr [_Tree+edi+4]
   call DrawTree
   pop edi
@skipleft:
   push edi
   mov ecx,[_Tree+edi+12]
   cmp [_pt+ecx*4],PLAIN
   jnz @noplain
   push ecx
   call DrawPlainPolygon
   jmp @skiprest
@noplain:
   cmp [_pt+ecx*4],PHONG
   jnz @nophong
   push ecx
   call DrawPhongPolygon
   jmp @skiprest
@nophong:
   cmp [_pt+ecx*4],FASTTEXT
   jnz @nofasttext
   push ecx
   call DrawFastTPolygon
   jmp @skiprest
@nofasttext:
   cmp [_pt+ecx*4],GOURAD
   jnz @nogourad
   push ecx
   call DrawGouradPolygon
   jmp @skiprest
@nogourad:
   cmp [_pt+ecx*4],GOURADTX
   jnz @nogouradtx
   push ecx
   call DrawGPTextPolygon
   jmp @skiprest
@nogouradTx:
   cmp [_pt+ecx*4],TEXTURED
   jnz @notextured
   push ecx
   call DrawPTextPolygon
   jmp @skiprest
@notextured:
   cmp [_pt+ecx*4],FASTTRTX
   jnz @nofasttrtx
   push ecx
   call DrawFastTrTPolygon
   jmp @skiprest
@nofasttrtx:
   cmp [_pt+ecx*4],PORTAL
   jnz @noportal
   push ecx
   call DrawPortalPolygon
   jmp @skiprest
@noportal:
   cmp [_pt+ecx*4],PHONGTX
   jnz @nophongtx
   push ecx
   call DrawPPTextPolygon       {The Ultimate Polygon :) }
   jmp @skiprest
@nophongtx:
   cmp [_pt+ecx*4],TRTEXT
   jnz @notrtext
   push ecx
   call DrawTPTextPolygon
   jmp @skiprest
@notrtext:
   cmp [_pt+ecx*4],ENVMAP
   jnz @noenvmap
   push ecx
   call DrawEnvPolygon
   jmp @skiprest
@noenvmap:
   cmp [_pt+ecx*4],SPRITE
   jnz @nosprite
   push ecx
   call DrawSprite
   jmp @skiprest
@nosprite:
@skiprest:
   pop edi
   cmp dword ptr [_Tree+edi+8],0
   jz  @skipright
   push dword ptr [_Tree+edi+8]
   call DrawTree
@skipright:
end;


{  _______________________________________________________________________
  /                                                                      /
 /  RotMatrix: Create a rotation/Translation matrix                     /
/______________________________________________________________________/ }

procedure RotMatrix(rx,ry,rz: longint; px,py,pz: single);
var t1,t2: single;
begin {Matrix definition}
   if (rx<0) or (rx>2879) or (ry<0) or (ry>2879) or (rz<0) or (rz>2879) then
      error('Illegal coord passed to renderer','RotMatrix');
   t1:=_cs[rz]*_sn[ry];
   t2:=_sn[rz]*_sn[ry];
   tmp1[0,0]:= t2*_sn[rx]+_cs[rz]*_cs[rx];
   tmp1[0,1]:= _cs[ry]*_sn[rx];
   tmp1[0,2]:= _sn[rz]*_cs[rx]-t1*_sn[rx];
   tmp1[0,3]:=px;
   tmp1[1,0]:= t2*_cs[rx]-_cs[rz]*_sn[rx];
   tmp1[1,1]:= _cs[ry]*_cs[rx];
   tmp1[1,2]:= -t1*_cs[rx]-_sn[rz]*_sn[rx];
   tmp1[1,3]:=py;
   tmp1[2,0]:=-_sn[rz]*_cs[ry];
   tmp1[2,1]:= _sn[ry];
   tmp1[2,2]:= _cs[rz]*_cs[ry];
   tmp1[2,3]:=pz;
   tmp1[3,0]:=0;
   tmp1[3,1]:=0;
   tmp1[3,2]:=0;
   tmp1[3,3]:=1;
end;


{  _______________________________________________________________________
  /                                                                      /
 /  LightMatrix: Construct a matrix for the vertex normals              /
/______________________________________________________________________/ }

procedure LightMatrix(rx,ry,rz: longint; px,py,pz: single);
var t1,t2: single;
    x,y: longint;
begin {Matrix definition}
   if (rx<0) or (rx>2879) or (ry<0) or (ry>2879) or (rz<0) or (rz>2879) then
      error('Illegal coord passed to renderer','LightMatrix');
   t1:=_cs[rz]*_sn[ry];
   t2:=_sn[rz]*_sn[ry];
   tmp2[0,0]:= t2*_sn[rx]+_cs[rz]*_cs[rx];
   tmp2[0,1]:= _cs[ry]*_sn[rx];
   tmp2[0,2]:= _sn[rz]*_cs[rx]-t1*_sn[rx];
   tmp2[0,3]:=px;
   tmp2[1,0]:= t2*_cs[rx]-_cs[rz]*_sn[rx];
   tmp2[1,1]:= _cs[ry]*_cs[rx];
   tmp2[1,2]:= -t1*_cs[rx]-_sn[rz]*_sn[rx];
   tmp2[1,3]:=py;
   tmp2[2,0]:=-_sn[rz]*_cs[ry];
   tmp2[2,1]:= _sn[ry];
   tmp2[2,2]:= _cs[rz]*_cs[ry];
   tmp2[2,3]:=pz;
   tmp2[3,0]:=0;
   tmp2[3,1]:=0;
   tmp2[3,2]:=0;
   tmp2[3,3]:=1;
   for y:=0 to 3 do for x:=0 to 3 do
   {Calculate position x,y: multiply row y with column x}
   tmpl[x,y]:=tmp2[0,y]*tmp1[x,0]+tmp2[1,y]*tmp1[x,1]+
              tmp2[2,y]*tmp1[x,2]+tmp2[3,y]*tmp1[x,3];
end;


{  _______________________________________________________________________
  /                                                                      /
 / Diagnostics: Print some basic information on the current scene       /
/______________________________________________________________________/ }

procedure Diagnostics;
var i,n: longint;
    a: char;
begin
   TextBackground(BLACK);
   ClrScr;
   TextBackground(RED);
   Write(' ALPHA Engine V2.0                         Copyright January 1997 ''The Phantom'' ');
   TextBackground(BLACK);
   TextColor(WHITE); Write('Coords:           ');
   TextColor(LIGHTGREEN); Writeln(_coords:6);
   TextColor(WHITE); Write('Planes:           ');
   TextColor(LIGHTGREEN); Writeln(_polies:6);
   _sector:=1;
   for i:=2 to MAXSECTR do if _nsr[i]>0 then _sector:=i;
   TextColor(WHITE); Write('Sectors:          ');
   TextColor(LIGHTGREEN); Writeln(_lsd:6);
   for i:=1 to _sector do begin
      TextColor(WHITE); Write('   Polygons in sector ',i,': ');
      TextColor(LIGHTGREEN); Writeln(_nsr[i]);
   end;
   TextColor(WHITE); Write('Base polygons:    ');
   TextColor(LIGHTGREEN); Writeln(_base:6);
   TextColor(WHITE); Write('Objects in world: ');
   TextColor(LIGHTGREEN); Writeln(world.nobj:6);
   TextColor(WHITE); Write('Coords in world:  ');
   TextColor(LIGHTGREEN); Writeln(world.npnt:6);
   for i:=1 to world.nobj do begin
      TextColor(WHITE);
      write('   Sub-Objects in object ',i:2,':   ');
      TextColor(LIGHTGREEN); Writeln(thing[i].nobj:3);
      if thing[i].nobj>0 then for n:=1 to thing[i].nobj do begin
         TextColor(WHITE);
         Writeln(', subobject 1 (object ',thing[i].obj[n],') contains ');
         TextColor(LIGHTGREEN); Write(thing[thing[i].obj[n]].npnt);
         TextColor(WHITE); Writeln(' coords');
      end;
      TextColor(WHITE);
      write('   Coords in object ',i:2,':        ');
      TextColor(LIGHTGREEN); Writeln(thing[i].npnt:3);
   end;
   TextColor(WHITE); Write('Size of world record:     ');
   TextColor(LIGHTGREEN); Writeln(sizeof(world):6,' bytes');
   TextColor(WHITE); Write('Size of object record:    ');
   TextColor(LIGHTGREEN); Writeln(sizeof(thing[1]):6,' bytes');
   TextColor(WHITE); Write('Size of sort tree record: ');
   TextColor(LIGHTGREEN); Writeln(sizeof(_tree[1]):6,' bytes');
   if _warn>0 then begin
      n:=_warn;
      if n>10 then n:=10;
      TextColor(RED);
      for i:=1 to n do Writeln(wrn[i]);
      TextColor(WHITE);
      if _warn>10 then Writeln('(more warnings reported, only first 10 printed)');
   end;
   TextColor(BLUE);
   TextBackground(BLACK);
   Write('');
   TextBackground(LIGHTBLUE);
   TextColor(WHITE);
   Write('  This engine and it''s sources are freeware and you are encouraged to spread    ');
   Write('  them freely. Please contact the author if you have any questions, and visit   ');
   Write('  my homepage for the latest info and updates:                                  ');
   TextColor(YELLOW);
   Write('  www.geocities.com/CapeCanaveral/5402     Jacco.Bikker@QuestIntl.Com           ');
   TextColor(BLACK);
   TextBackground(BLUE);
   Write('');
   TextColor(WHITE);
   TextBackground(BLACK);
   asm
      mov ax,0ee40h
      int 31h
   end;
   while keypressed do a:=readkey;
end;


{  _______________________________________________________________________
  /                                                                      /
 /  CreateCoord: Define a 3D coordinate                                 /
/______________________________________________________________________/ }

procedure CreateCoord(x,y,z: single);
begin
   inc(_coords);
   if _coords>MAXCOORD then Error('Too many coordinates','CreateCoord');
   _px[_coords]:=x;
   _py[_coords]:=y;
   _pz[_coords]:=z;
   {add a pointer to this coord to the world or the current object}
   if _obj=0 then begin
      inc(world.npnt);
      world.pnt[world.npnt]:=_coords;
   end else begin
      inc(thing[_obj].npnt);
      if thing[_obj].npnt>=MXSUBPNT then Error('Too many coordinates in object','CreateCoord');
      thing[_obj].pnt[thing[_obj].npnt]:=_coords;
   end;
end;


{  _______________________________________________________________________
  /                                                                      /
 / SampleBump: Create a sample bumpmap texture                          /
/______________________________________________________________________/ }

procedure SetLight(x,y,z: longint);
begin
   _lrx:=x mod 2880;
   _lry:=y mod 2880;
   _lrz:=z mod 2880;
end;


{  _______________________________________________________________________
  /                                                                      /
 /  FindCoord: Find coordinate or create new one                        /
/______________________________________________________________________/ }

function FindCoord(x,y,z: single): longint;
var i,cp: longint;
begin
   cp:=0;
   for i:=1 to _coords do if ((_px[i]=x) and (_py[i]=y) and (_pz[i]=z))
   then begin cp:=i; i:=_coords end;
   if cp=0 then begin
      inc(_coords);
      if _coords>MAXCOORD then Error('Too many coordinates','FindCoord');
      _px[_coords]:=x;
      _py[_coords]:=y;
      _pz[_coords]:=z;
      cp:=_coords;
      {add a pointer to this coord to the world or the current object}
      if _obj=0 then begin
         inc(world.npnt);
         world.pnt[world.npnt]:=cp;
      end else begin
         inc(thing[_obj].npnt);
         if thing[_obj].npnt>MXSUBPNT then Error('Too many coordinates in object','CreateCoord');
         thing[_obj].pnt[thing[_obj].npnt]:=cp;
      end;
   end;
   FindCoord:=cp;
end;


{  _______________________________________________________________________
  /                                                                      /
 /  MulMatrix: Concatenate two matrices                                 /
/______________________________________________________________________/ }

procedure MulMatrix;
var tmp3: matrix;
    x,y: longint;
begin {Multiply matrix 1 and 2, target is matrix 1}
   for y:=0 to 3 do for x:=0 to 3 do
   {Calculate position x,y: multiply row y with column x}
   tmp3[x,y]:=tmp1[0,y]*tmp2[x,0]+tmp1[1,y]*tmp2[x,1]+
              tmp1[2,y]*tmp2[x,2]+tmp1[3,y]*tmp2[x,3];
   tmp1:=tmp3;
end;


{  _______________________________________________________________________
  /                                                                      /
 /  Sector: Set active sector for rendering and world-defining          /
/______________________________________________________________________/ }

procedure Sector(s: longint);
begin
   if (s<1) or (s>MAXSECTR) then Error('Bad sector specified','Sector');
   _sector:=s
end;


{  _______________________________________________________________________
  /                                                                      /
 /  CalcMixTable: Construct a table with optimal color mixing           /
/______________________________________________________________________/ }

procedure CalcMixTable;
var x,y,i,o: longint;
    rn,gn,bn: longint;
    rd,gd,bd: longint;
    sd,d: single;
    quad: array[-64..64] of longint;
    buf: array[0..255] of byte;
    p: pointer;
    f: file;
begin
   for x:=-64 to 64 do quad[x]:=x*x;
   for x:=0 to 255 do for y:=x to 255 do begin
      rn:=(_r[x]+_r[y]) div 2;
      gn:=(_g[x]+_g[y]) div 2;
      bn:=(_b[x]+_b[y]) div 2;
      sd:=100000;
      o:=0;
      for i:=0 to 255 do begin
         rd:=quad[_r[i]-rn];
         gd:=quad[_g[i]-gn];
         bd:=quad[_b[i]-bn];
         d:=sqrt(rd+gd+bd);
         if d<sd then begin
            sd:=d;
            o:=i;
         end;
      end;
      if y<200 then mem[scrn+x+y*320]:=o;
      if x<200 then mem[scrn+y+x*320]:=o;
      mem[pmap+y+x*256]:=o;
      mem[pmap+x+y*256]:=o;
      if keypressed then Halt;
   end;
   p:=@buf;
   Assign(f,'mtable.rgb');
   Rewrite(f,1);
   for i:=0 to 255 do begin
      for x:=0 to 255 do buf[x]:=mem[pmap+i*256+x];
      BlockWrite(f,p^,256);
   end;
   Close(f);
end;


{  _______________________________________________________________________
  /                                                                      /
 /  CalcLightTable: Calculate optimal light mixing table                /
/______________________________________________________________________/ }

procedure CalcLightTable;
var x,y,i,o: longint;
    rn,gn,bn: longint;
    rd,gd,bd: longint;
    sd,d: single;
    quad: array[-64..64] of longint;
    buf: array[0..255] of byte;
    p: pointer;
    f: file;
begin
   for x:=-64 to 64 do quad[x]:=x*x;
   for x:=0 to 255 do for y:=0 to 127 do begin
      if y<64 then begin
         rn:=(_r[x]*y) div 64;
         gn:=(_g[x]*y) div 64;
         bn:=(_b[x]*y) div 64;
      end else begin
         rn:=(_r[x]*(127-y)+64*(y-63)) div 64;
         gn:=(_g[x]*(127-y)+64*(y-63)) div 64;
         bn:=(_b[x]*(127-y)+64*(y-63)) div 64;
      end;
      sd:=1000;
      o:=0;
      for i:=0 to 255 do begin
         rd:=quad[_r[i]-rn];
         gd:=quad[_g[i]-gn];
         bd:=quad[_b[i]-bn];
         d:=sqrt(rd+gd+bd);
         if d<sd then begin
            sd:=d;
            o:=i;
         end;
      end;
      if y<200 then mem[scrn+x+y*320]:=o;
      mem[pmap+x+y*256]:=o;
      if keypressed then Halt;
   end;
   p:=@buf;
   Assign(f,'ltable.rgb');
   Rewrite(f,1);
   for i:=0 to 255 do begin
      for x:=0 to 255 do buf[x]:=mem[pmap+i*256+x];
      BlockWrite(f,p^,256);
   end;
   Close(f);
end;


{  _______________________________________________________________________
  /                                                                      /
 /  CalcMipMap: Calculate smoothed version of texture                   /
/______________________________________________________________________/ }

procedure CalcMipMap(as,ad: dword);
var x,y,w1,w2: longint;
begin
   for x:=0 to 63 do for y:=0 to 63 do begin
      w1:=mem[mix+mem[as+x+x+y shl 9]+mem[as+x+x+(y shl 9)+256] shl 8];
      w2:=mem[mix+mem[as+1+x+x+y shl 9]+mem[as+x+x+1+(y shl 9)+256] shl 8];
      mem[ad+x+x+y shl 9]:=mem[mix+w1+w2 shl 8];
      mem[ad+x+x+y shl 9+1]:=mem[ad+x+x+y shl 9];
      mem[ad+x+x+y shl 9+256]:=mem[ad+x+x+y shl 9];
      mem[ad+x+x+y shl 9+257]:=mem[ad+x+x+y shl 9];
   end;
end;


{  _______________________________________________________________________
  /                                                                      /
 /  Init3DEngine: Setup the basic engine data                           /
/______________________________________________________________________/ }

procedure Init3DEngine;
var i,l: longint;
    f: file;
    p: pointer;
    a: dword;
    x,y: longint;
begin
   TextBackGround(BLACK); ClrScr;
   TextBackground(RED); TextColor(WHITE);
   Write('ALPHA 2 3D ENGINE COPYRIGHT 1997 BY THE PHANTOM - Initialization                ');
   TextBackground(BLACK); Writeln(' Initializing mouse');
   if not initmouse then error('Mouse driver not installed','Init3DEngine');
   Writeln(' Initializing FPU');
   asm finit end;
   SetFPUPrecision(1);
   SetFPURounding(3);
   InitBuffers;
   Writeln(' Generating Phong lightmap');
   PhongMap;
   Writeln(' Internal engine data setup');
   _Coords:=0;
   _Obj:=0;
   _lObj:=0;
   _Polies:=0;
   _warn:=0;
   _lfps:=70;
   _base:=0;
   _lsd:=0;
   _guard1:=$99999999;
   _guard2:=$99999999;
   _guard3:=$99999999;
   _guard4:=$99999999;
   _guard5:=$99999999;
   memd[tex+$40800]:=$99999999;
   Writeln('   Portal data setup');
   _sector:=1;
   for i:=1 to MAXSECTR do _nsr[i]:=0;
   ZClipping(ZMIN);
   Writeln('   Precalculating tables');
   for i:=0 to 2879 do begin
      _sn[i]:=sin(i*3.1415926534/1440);
      _cs[i]:=cos(i*3.1415926534/1440)
   end;
   for i:=0 to 1000 do _qdiv[i]:=(i-500)/DIST;
   world.npnt:=0;
   world.nobj:=0;
   Writeln(' Reading light remap tables from disk');
   Assign(f,'ltable.rgb');
   Reset(f,1);
   p:=@buffer;
   a:=mix;
   for i:=0 to 63 do begin
      BlockRead(f,p^,1000);
      for l:=0 to 1023 do mem[a+l]:=buffer[l];
      inc(a,1000);
   end;
   Close(f);
   for x:=0 to 255 do for y:=0 to 255 do lit[x+y*256]:=mem[mix+y+x*256];
   Writeln(' Reading color mixing tables from disk');
   Assign(f,'mtable.rgb');
   Reset(f,1);
   a:=mix;
   for i:=0 to 63 do begin
      BlockRead(f,p^,1024);
      for l:=0 to 1023 do mem[a+l]:=buffer[l];
      inc(a,1024);
   end;
   Close(f);
   Writeln(' Reading textures from disk');
   Writeln(' Initializing view window');
   ResetWindow;
   ClearBuffer;
   Writeln(' Starting engine');
   if HIRES=0 then Init320x200;
   if HIRES=1 then Init320x400;
   if HIRES=2 then VesaInit(320,400);
   if HIRES=3 then VesaInit(640,400);
   LoadPCX('tiles.pcx');
   for i:=0 to 199 do for l:=0 to 255 do mem[env+i*256+l]:=mem[tex+32+i*320+l];
   for i:=199 to 255 do for l:=0 to 255 do mem[env+i*256+l]:=mem[tex+1000];
   if RECALC=1 then begin
      CalcLightTable;
      CalcMixTable;
   end;
   _tba[1]:=tex+256*256;
   _tba[2]:=tex+256*256+128;
   _tba[3]:=tex+384*256;
   _tba[4]:=tex+384*256+128;
   _tba[5]:=tex+512*256;
   _tba[6]:=tex+512*256+128;
   _tba[7]:=tex+640*256;
   _tba[8]:=tex+640*256+128;
   _tba[9]:=tex+768*256;
   _tba[10]:=tex+768*256+128;
   _tba[11]:=tex+896*256;
   _tba[12]:=tex+896*256+128;
   _tba[13]:=tex;
   _tba[14]:=tex+128;
   _tba[15]:=tex+256*128;
   _tba[16]:=tex+256*128+128;
end;


{  _______________________________________________________________________
  /                                                                      /
 / ForceMipmap: Set all polygon texturenumbers to MIPMAPS               /
/______________________________________________________________________/ }

procedure ForceMipMap;
var i: longint;
begin
   for i:=1 to _polies do
      if ((_pt[i]=TEXTURED) or (_pt[i]=TRTEXT) or (_pt[i]=PhongTx)
      or (_pt[i]=FASTTEXT) or (_pt[i]=FASTTRTX)) then _pc[i]:=_pc[i]+8;
end;


{  _______________________________________________________________________
  /                                                                      /
 / SetRotation: Rotate a (sub)object relative to its parent             /
/______________________________________________________________________/ }

procedure SetRotation(n: longint; rx,ry,rz: longint);
begin
   if n>_lobj then Error('Object does not exist','SetRotation');
   Thing[n].rx:=rx;
   Thing[n].ry:=ry;
   Thing[n].rz:=rz;
end;


{  _______________________________________________________________________
  /                                                                      /
 / Setposition: Alter the relative position of a (sub)object            /
/______________________________________________________________________/ }

procedure SetPosition(n: longint; px,py,pz: single);
begin
   if n>_lobj then Error('Object does not exist','SetPosition');
   Thing[n].ox:=px;
   Thing[n].oy:=py;
   Thing[n].oz:=pz;
end;


{  _______________________________________________________________________
  /                                                                      /
 / CreateObject: Create a new object in the world or in an object       /
/______________________________________________________________________/ }

function CreateObject: longint;
begin
   if _lobj=MAXTHING then Error('Too many (sub)objects defined','CreateObject')
   else begin
      inc(_lobj);             {Create a new object}
      thing[_lobj].nobj:=0;
      thing[_lobj].npnt:=0;
      thing[_lobj].parent:=_obj;
      SetRotation(_lobj,0,0,0);
      SetPosition(_lobj,0,0,0);
   end;
   if _obj=0 then begin
      inc(world.nobj);        {increase object counter for world}
      if world.nobj>MAXENT then Error('Too many objects in world','CreateObject');
      world.obj[world.nobj]:=_lobj;
   end else begin
      inc(thing[_obj].nobj);  {increase subobject counter for object}
      if thing[_obj].nobj>MXSUBENT then Error('Too many subobjects in object','CreateObject');
      thing[_obj].obj[thing[_obj].nobj]:=_lobj;
   end;
   _obj:=_lobj;               {current object becomes just created one}
   CreateObject:=_lobj;       {return a pointer to this object}
end;


{  _______________________________________________________________________
  /                                                                      /
 / Parent: Return to parent after editing object                        /
/______________________________________________________________________/ }

procedure parent;
begin
   if _obj=0 then Error('World does not have a parent','Parent');
   _obj:=thing[_obj].parent;
end;


{  _______________________________________________________________________
  /                                                                      /
 / RotateObject: Rotate an object and its subobjects                    /
/______________________________________________________________________/ }

procedure RotateObject(n: longint; pz,py,px: single);
var i,l: longint;
    bck,bckl: matrix;
    a: dword;
    fdist,f,f64,fhi,fhx,_rfz,n64,nx,ny,nz: single;
begin
   {Calculate relative position and direction}
   tmp2:=tmp1;         {store the world matrix in tmp2}
   RotMatrix(thing[n].rx,thing[n].ry,thing[n].rz,px,py,pz);
   MulMatrix;          {concatenate the matrices}
   LightMatrix(_lrx,_lry,_lrz,0,0,0);
   {Rotate the subobjects}
   if thing[n].nobj>0 then begin
      bck:=tmp1;       {don't loose the matrix for the current object}
      bckl:=tmpl;
      for i:=1 to thing[n].nobj do begin
         RotateObject(thing[n].obj[i],nx,ny,nz);
         tmp1:=bck;    {restore after it has been changed by the subobject}
         tmpl:=bckl;
      end;
   end;
   {Rotate coordinates in this object}
   i:=thing[n].npnt;
   a:=(n-1)*sizeof(objtree);
   f:=DIST;
   f64:=DIST*64;
   if HIRES=0 then fhi:=1 else fhi:=2;
   if HIRES=3 then fhx:=2 else fhx:=1;
   n64:=64;
   fdist:=200;
   asm
      mov ecx,[i]
      or  ecx,ecx
      jz  @endofobject
      lea esi,thing
      add esi,[a]
      align 4
   @ploop:
      mov edi,[esi]
      add esi,4
      dec edi
      shl edi,2
      fld   [_px+edi]
      fadd  dword ptr [tmp1+12]
      fmul  dword ptr [tmp1+0]
      fld   [_px+edi]
      fadd  dword ptr [tmp1+12]
      fmul  dword ptr [tmp1+16]
      fld   [_px+edi]
      fadd  dword ptr [tmp1+12]
      fmul  dword ptr [tmp1+32]
      fld   [_py+edi]
      fadd  dword ptr [tmp1+28]
      fmul  dword ptr [tmp1+4]
      fld   [_py+edi]
      fadd  dword ptr [tmp1+28]
      fmul  dword ptr [tmp1+20]
      fld   [_py+edi]
      fadd  dword ptr [tmp1+28]
      fmul  dword ptr [tmp1+36]
      fxch  st(2)
      faddp st(5),st(0)
      faddp st(3),st(0)
      faddp st(1),st(0)
      fld   [_pz+edi]
      fadd  dword ptr [tmp1+44]
      fmul  dword ptr [tmp1+8]
      fld   [_pz+edi]
      fadd  dword ptr [tmp1+44]
      fmul  dword ptr [tmp1+24]
      fld   [_pz+edi]
      fadd  dword ptr [tmp1+44]
      fmul  dword ptr [tmp1+40]
      fxch  st(2)
      faddp st(5),st(0)
      faddp st(3),st(0)
      faddp st(1),st(0)
      fxch  st(2)
      fxch  st(1)
      fxch  st(2)
      fxch  st(1)
      fmul  [fhx]
      fstp  [_rx+edi]
      fadd  [fdist]
      fist  [_rz+edi]
      fstp  [_frz+edi]
      fmul  [fhi]
      fst   [_ry+edi]
      fmul  [f64]
      fld   [_frz+edi]
      fdivp
      fistp [_ys+edi]
      add   [_ys+edi],YCENTR*64
      fld   [_rx+edi]
      fmul  [f64]
      fld   [_frz+edi]
      fdivp
      fistp [_xs+edi]
      add   [_xs+edi],XCENTR*64
      {Rotate vertex normal}
      fld   [_nx+edi]
      fmul  dword ptr [tmpl+0]
      fld   [_nx+edi]
      fmul  dword ptr [tmpl+16]
      fld   [_nx+edi]
      fmul  dword ptr [tmpl+32]
      fld   [_ny+edi]
      fmul  dword ptr [tmpl+4]
      fld   [_ny+edi]
      fmul  dword ptr [tmpl+20]
      fld   [_ny+edi]
      fmul  dword ptr [tmpl+36]
      fxch  st(2)
      faddp st(5),st(0)
      faddp st(3),st(0)
      faddp st(1),st(0)
      fld   [_nz+edi]
      fmul  dword ptr [tmpl+8]
      fld   [_nz+edi]
      fmul  dword ptr [tmpl+24]
      fld   [_nz+edi]
      fmul  dword ptr [tmpl+40]
      fxch  st(2)
      faddp st(5),st(0)
      faddp st(3),st(0)
      faddp st(1),st(0)
      fxch  st(2)
      fxch  st(1)
      fxch  st(2)
      fxch  st(1)
      fmul  [n64]
      fistp [_vx+edi]
      fmul  [n64]
      fistp [_vz+edi]
      fmul  [n64]
      fistp [_vy+edi]
      add [_vx+edi],128*64    {note: even phong requires only}
      add [_vy+edi],128*64    {two out of three values!}
      add [_vz+edi],128*64    {Must be optimized.}
      dec ecx
      jnz @ploop
   @endofobject:
   end;
end;


{  _______________________________________________________________________
  /                                                                      /
 / RotateCoords: Rotate the world and objects (messed-up QUAKE code:)   /
/______________________________________________________________________/ }

procedure RotateCoords(px,py,pz: single); {rotate coordinates in world system}
var f,fhi,fhx,f64,n64,fdist: single;
    bck,bckl: matrix;
begin
   f:=DIST;
   f64:=DIST*64;
   if HIRES=0 then fhi:=1 else fhi:=2;
   if HIRES=3 then fhx:=2 else fhx:=1;
   n64:=64;
   fdist:=200;
   asm
      mov ecx,[world.npnt]
      or  ecx,ecx
      jz  @endofworld
      xor esi,esi
      align 4
   @ploop:
      mov edi,[world.pnt+esi]
      add esi,4
      dec edi
      shl edi,2
      fld   [_px+edi]
      fadd  dword ptr [tmp1+12]
      fmul  dword ptr [tmp1+0]
      fld   [_px+edi]
      fadd  dword ptr [tmp1+12]
      fmul  dword ptr [tmp1+16]
      fld   [_px+edi]
      fadd  dword ptr [tmp1+12]
      fmul  dword ptr [tmp1+32]
      fld   [_py+edi]
      fadd  dword ptr [tmp1+28]
      fmul  dword ptr [tmp1+4]
      fld   [_py+edi]
      fadd  dword ptr [tmp1+28]
      fmul  dword ptr [tmp1+20]
      fld   [_py+edi]
      fadd  dword ptr [tmp1+28]
      fmul  dword ptr [tmp1+36]
      fxch  st(2)
      faddp st(5),st(0)
      faddp st(3),st(0)
      faddp st(1),st(0)
      fld   [_pz+edi]
      fadd  dword ptr [tmp1+44]
      fmul  dword ptr [tmp1+8]
      fld   [_pz+edi]
      fadd  dword ptr [tmp1+44]
      fmul  dword ptr [tmp1+24]
      fld   [_pz+edi]
      fadd  dword ptr [tmp1+44]
      fmul  dword ptr [tmp1+40]
      fxch  st(2)
      faddp st(5),st(0)
      faddp st(3),st(0)
      faddp st(1),st(0)
      fxch  st(2)
      fxch  st(1)
      fxch  st(2)
      fxch  st(1)
      fmul  [fhx]
      fstp  [_rx+edi]
      fadd  [fdist]
      fist  [_rz+edi]
      fstp  [_frz+edi]
      fmul  [fhi]
      fst   [_ry+edi]
      fmul  [f64]
      fld   [_frz+edi]
      fdivp
      fistp [_ys+edi]
      add   [_ys+edi],YCENTR*64
      fld   [_rx+edi]
      fmul  [f64]
      fld   [_frz+edi]
      fdivp
      fistp [_xs+edi]
      add   [_xs+edi],XCENTR*64
      {Rotate vertex normal}
      fld   [_nx+edi]
      fmul  dword ptr [tmpl+0]
      fld   [_nx+edi]
      fmul  dword ptr [tmpl+16]
      fld   [_nx+edi]
      fmul  dword ptr [tmpl+32]
      fld   [_ny+edi]
      fmul  dword ptr [tmpl+4]
      fld   [_ny+edi]
      fmul  dword ptr [tmpl+20]
      fld   [_ny+edi]
      fmul  dword ptr [tmpl+36]
      fxch  st(2)
      faddp st(5),st(0)
      faddp st(3),st(0)
      faddp st(1),st(0)
      fld   [_nz+edi]
      fmul  dword ptr [tmpl+8]
      fld   [_nz+edi]
      fmul  dword ptr [tmpl+24]
      fld   [_nz+edi]
      fmul  dword ptr [tmpl+40]
      fxch  st(2)
      faddp st(5),st(0)
      faddp st(3),st(0)
      faddp st(1),st(0)
      fxch  st(2)
      fxch  st(1)
      fxch  st(2)
      fxch  st(1)
      fmul  [n64]
      fistp [_vx+edi]
      fmul  [n64]
      fistp [_vz+edi]
      fmul  [n64]
      fistp [_vy+edi]
      add [_vx+edi],128*64       {note: even phong requires only}
      add [_vy+edi],128*64       {two out of three values!}
      add [_vz+edi],128*64       {Must be optimized.}
      dec ecx
      jnz @ploop
   @endofworld:
   end;                           {Don't forget the px/py/pz stuff...}
   {process objects in world}
   if world.nobj>0 then begin
      bck:=tmp1;
      bckl:=tmpl;
      for _i:=1 to world.nobj do begin
         RotateObject(world.obj[_i],px,py,pz);
         tmp1:=bck;
         tmpl:=bckl;
      end;
   end;
end;


{  _______________________________________________________________________
  /                                                                      /
 / SingleSided: Mark a polygon as single sided                          /
/______________________________________________________________________/ }

procedure SingleSided(n: Longint);
begin
   if n>_polies then Error('Polygon does not exist','SingleSided');
   if n<1 then Error('Bad polygon number','SingleSided');
   _pd[n]:=0;
end;


{  _______________________________________________________________________
  /                                                                      /
 / DoubleSided: Mark a polygon as double sided                          /
/______________________________________________________________________/ }

procedure DoubleSided(n: Longint);
begin
   if n>_polies then Error('Polygon does not exist','DoubleSided');
   if n<1 then Error('Bad polygon number','DoubleSided');
   _pd[n]:=1;
end;


{  _______________________________________________________________________
  /                                                                      /
 / CalcNormal: Calculate the polygon's normal (for last polygon)        /
/______________________________________________________________________/ }

procedure CalcNormal;
var rx1,ry1,rz1,rx2,ry2,rz2: single;
    n: longint;
begin
   {Calculate normalvector for last created polygon (from PASPHONG)}
   if (_pnv[_polies]<5) then n:=3 else n:=4;
   rx1:=_px[_pp[_polies,2]]-_px[_pp[_polies,1]];
   ry1:=_py[_pp[_polies,2]]-_py[_pp[_polies,1]];
   rz1:=_pz[_pp[_polies,2]]-_pz[_pp[_polies,1]];
   rx2:=_px[_pp[_polies,n]]-_px[_pp[_polies,1]];
   ry2:=_py[_pp[_polies,n]]-_py[_pp[_polies,1]];
   rz2:=_pz[_pp[_polies,n]]-_pz[_pp[_polies,1]];
   _pn[_polies,1]:=ry1*rz2-ry2*rz1;
   _pn[_polies,2]:=rz1*rx2-rz2*rx1;
   _pn[_polies,3]:=rx1*ry2-rx2*ry1;
   _pd[_polies]:=0;  {Not doublesided}
end;


{  _______________________________________________________________________
  /                                                                      /
 / CenterObject: Make sure the object is in the center of the scene     /
/______________________________________________________________________/ }

procedure CenterObject;
var ax1,ay1,az1,ax2,ay2,az2: single;
    i: longint;
begin
   ax1:=10000;
   ax2:=-10000;
   ay1:=ax1;
   ay2:=ax2;
   az1:=ax1;
   az2:=ax2;
   for i:=1 to _coords do begin
      if _px[i]<ax1 then ax1:=_px[i];
      if _px[i]>ax2 then ax2:=_px[i];
      if _py[i]<ay1 then ay1:=_py[i];
      if _py[i]>ay2 then ay2:=_py[i];
      if _pz[i]<az1 then az1:=_pz[i];
      if _pz[i]>az2 then az2:=_pz[i];
   end;
   ax1:=(ax2-ax1)/2+ax1;
   ay1:=(ay2-ay1)/2+ay1;
   az1:=(az2-az1)/2+az1;
   for i:=1 to _coords do begin
      _px[i]:=_px[i]-ax1;
      _py[i]:=_py[i]-ay1;
      _pz[i]:=_pz[i]-az1;
   end;
end;


{  _______________________________________________________________________
  /                                                                      /
 / CheckSize: See if polygon size is suspicious (debugging)             /
/______________________________________________________________________/ }

procedure CheckSize(n: longint);
var p1,p2,p3: longint;
    sz1,sz2,x1,y1,x2,y2,x3,y3,z1,z2,z3: single;
begin
   if (n<1) or (n>_polies) or (n>MAXSIDES) then
      error('Illegal side number slipped through','CheckSize');
   p1:=_pp[n,1];
   p2:=_pp[n,2];
   p3:=_pp[n,3];
   if (p1<1) or (p1>_coords) then error('Poly '+n2s(n)+' - Illigal coord ('+n2s(p1)+' slipped through','CheckSize');
   if (p2<1) or (p2>_coords) then error('Poly '+n2s(n)+' - Illigal coord ('+n2s(p2)+' slipped through','CheckSize');
   if (p3<1) or (p3>_coords) then error('Poly '+n2s(n)+' - Illigal coord ('+n2s(p3)+' slipped through','CheckSize');
   x1:=_px[p1];
   x2:=_px[p2];
   x3:=_px[p3];
   y1:=_py[p1];
   y2:=_py[p2];
   y3:=_py[p3];
   z1:=_pz[p1];
   z2:=_pz[p2];
   z3:=_pz[p3];
   sz1:=sqrt((x2-x1)*(x2-x1)+(y2-y1)*(y2-y1)+(z2-z1)*(z2-z1));
   sz2:=sqrt((x3-x1)*(x3-x1)+(y3-y1)*(y3-y1)+(z3-z1)*(z3-z1));
   if (sz1>PMSIZE) or (sz2>PMSIZE) then begin
      Warning('Suspicious big polygon ('+n2s(n)+') skipped','CheckSize');
      dec(_polies);
   end;
end;


{  _______________________________________________________________________
  /                                                                      /
 / PMapBase: Connect a PMap to a fake/base polygon                      /
/______________________________________________________________________/ }

procedure PMapBase(pol,bas: longint);
begin _pfb[pol]:=bas end;


{  _______________________________________________________________________
  /                                                                      /
 / CreateBase: Define a fake/base polygon for perspective maps          /
/______________________________________________________________________/ }

procedure CreateBase(p1,p2,p3: longint);
begin
   inc(_base);
   _fp1[_base]:=p1;
   _fp2[_base]:=p2;
   _fp3[_base]:=p3;
end;


{  _______________________________________________________________________
  /                                                                      /
 / CreateTriangle: Define a triangle                                    /
/______________________________________________________________________/ }

procedure CreateTriangle(p1,p2,p3,tp,cl: longint);
var oldsect: longint;
begin
   inc(_polies);
   if (tp<1) or (tp>19) then Error('Polygon type does not exist','CreateTriangle');
   if _polies>MAXSIDES then Error('Too many polygons/sprites','CreateTriangle');
   if (p1<1) or (p1>_coords) then Error('Bad coord ('+n2s(p1)+') for new polygon '+n2s(_polies),'CreateTriangle');
   if (p2<1) or (p2>_coords) then Error('Bad coord ('+n2s(p2)+') for new polygon '+n2s(_polies),'CreateTriangle');
   if (p3<1) or (p3>_coords) then Error('Bad coord ('+n2s(p3)+') for new polygon '+n2s(_polies),'CreateTriangle');
   _pp[_polies,1]:=p1;
   _pp[_polies,2]:=p2;
   _pp[_polies,3]:=p3;
   if tp=20 then _pt[_polies]:=19 else _pt[_polies]:=tp;
   _pc[_polies]:=cl;
   _pnv[_polies]:=3;
   CalcNormal;
   inc(_nsr[_sector]);
   if _nsr[_sector]>MAXSECSR then Error('Too many polygons in sector','CreateTriangle');
   _sec[_sector,_nsr[_sector]]:=_polies;
   if tp=PORTAL then begin
      oldsect:=_sector;
      _sector:=cl;
      CreateTriangle(p3,p2,p1,tp+1,oldsect);
      _sector:=oldsect;
   end;
end;


{  _______________________________________________________________________
  /                                                                      /
 / CreateSprite: Create a 3D sprite                                     /
/______________________________________________________________________/ }

procedure CreateSprite(p1: longint);
begin
   _polies:=_polies+1;
   if _polies>MAXSIDES then Error('Too many polygons/sprites','CreateSprite');
   if (p1=0) or (p1>MAXCOORD)
      then Error('Bad coordinate number for new sprite ('+n2s(p1)+')','CreateSprite');
   _pp[_polies,1]:=p1;
   _pt[_polies]:=SPRITE;
   _pc[_polies]:=0;
   _pnv[_polies]:=1;
   _pn[_polies,1]:=0.5;
   _pn[_polies,2]:=0.5;
   _pn[_polies,3]:=0.5;
   _pd[_polies]:=1;  {Doublesided}
   inc(_nsr[_sector]);
   if _nsr[_sector]>MAXSECSR then Error('Too many polygons in sector','CreateSprite');
   _sec[_sector,_nsr[_sector]]:=_polies;
end;


{  _______________________________________________________________________
  /                                                                      /
 / CreatePolygon: Construct one polyon                                  /
/______________________________________________________________________/ }

procedure CreatePolygon(p1,p2,p3,p4,tp,cl: longint);
var oldsect: longint;
begin
   _polies:=_polies+1;
   if (tp<1) or (tp>20) then Error('Polygon type does not exist','CreatePolygon');
   if _polies>MAXSIDES then Error('Too many polygons/sprites','CreatePolygon');
   if (p1<1) or (p1>_coords) then Error('Bad coord ('+n2s(p1)+') for new polygon '+n2s(_polies),'CreatePolygon');
   if (p2<1) or (p2>_coords) then Error('Bad coord ('+n2s(p2)+') for new polygon '+n2s(_polies),'CreatePolygon');
   if (p3<1) or (p3>_coords) then Error('Bad coord ('+n2s(p3)+') for new polygon '+n2s(_polies),'CreatePolygon');
   if (p4<1) or (p4>_coords) then Error('Bad coord ('+n2s(p4)+') for new polygon '+n2s(_polies),'CreatePolygon');
   _pp[_polies,1]:=p1;
   _pp[_polies,2]:=p2;
   _pp[_polies,3]:=p3;
   _pp[_polies,4]:=p4;
   if tp=20 then _pt[_polies]:=19 else _pt[_polies]:=tp;
   _pc[_polies]:=cl;
   _pnv[_polies]:=4;
   CalcNormal;
   inc(_nsr[_sector]);
   if _nsr[_sector]>MAXSECSR then Error('Too many polygons in sector','CreatePolygon');
   _sec[_sector,_nsr[_sector]]:=_polies;
   if tp=PORTAL then begin
      oldsect:=_sector;
      _sector:=cl;
      CreatePolygon(p4,p3,p2,p1,20,oldsect);
      _sector:=oldsect;
   end;
end;


{  _______________________________________________________________________
  /                                                                      /
 / Create5Plane: Construct one polyon with 5 vertices                   /
/______________________________________________________________________/ }

procedure Create5Plane(p1,p2,p3,p4,p5,tp,cl: longint);
var oldsect: longint;
begin
   inc(_polies);
   if (tp<1) or (tp>20) then Error('Polygon type does not exist','Create5Plane');
   if _polies>MAXSIDES then Error('Too many polygons/sprites','Create5Plane');
   if (p1<1) or (p1>_coords) then Error('Bad coord ('+n2s(p1)+') for new polygon '+n2s(_polies),'Create5Plane');
   if (p2<1) or (p2>_coords) then Error('Bad coord ('+n2s(p2)+') for new polygon '+n2s(_polies),'Create5Plane');
   if (p3<1) or (p3>_coords) then Error('Bad coord ('+n2s(p3)+') for new polygon '+n2s(_polies),'Create5Plane');
   if (p4<1) or (p4>_coords) then Error('Bad coord ('+n2s(p4)+') for new polygon '+n2s(_polies),'Create5Plane');
   if (p5<1) or (p5>_coords) then Error('Bad coord ('+n2s(p5)+') for new polygon '+n2s(_polies),'Create5Plane');
   _pp[_polies,1]:=p1;
   _pp[_polies,2]:=p2;
   _pp[_polies,3]:=p3;
   _pp[_polies,4]:=p4;
   _pp[_polies,5]:=p5;
   if tp=20 then _pt[_polies]:=19 else _pt[_polies]:=tp;
   _pc[_polies]:=cl;
   _pnv[_polies]:=5;
   CalcNormal;
   inc(_nsr[_sector]);
   if _nsr[_sector]>MAXSECSR then Error('Too many polygons in sector','Create5Plane');
   _sec[_sector,_nsr[_sector]]:=_polies;
   if tp=PORTAL then begin
      oldsect:=_sector;
      _sector:=cl;
      Create5Plane(p5,p4,p3,p2,p1,20,oldsect);
      _sector:=oldsect;
   end;
end;


{  _______________________________________________________________________
  /                                                                      /
 / Create6Plane: Construct one polyon with 6 vertices                   /
/______________________________________________________________________/ }

procedure Create6Plane(p1,p2,p3,p4,p5,p6,tp,cl: longint);
var oldsect: longint;
begin
   inc(_polies);
   if (tp<1) or (tp>20) then Error('Polygon type does not exist','Create6Plane');
   if _polies>MAXSIDES then Error('Too many polygons/sprites','Create6Plane');
   if (p1<1) or (p1>_coords) then Error('Bad coord ('+n2s(p1)+') for new polygon '+n2s(_polies),'Create6Plane');
   if (p2<1) or (p2>_coords) then Error('Bad coord ('+n2s(p2)+') for new polygon '+n2s(_polies),'Create6Plane');
   if (p3<1) or (p3>_coords) then Error('Bad coord ('+n2s(p3)+') for new polygon '+n2s(_polies),'Create6Plane');
   if (p4<1) or (p4>_coords) then Error('Bad coord ('+n2s(p4)+') for new polygon '+n2s(_polies),'Create6Plane');
   if (p5<1) or (p5>_coords) then Error('Bad coord ('+n2s(p5)+') for new polygon '+n2s(_polies),'Create6Plane');
   if (p6<1) or (p6>_coords) then Error('Bad coord ('+n2s(p6)+') for new polygon '+n2s(_polies),'Create6Plane');
   _pp[_polies,1]:=p1;
   _pp[_polies,2]:=p2;
   _pp[_polies,3]:=p3;
   _pp[_polies,4]:=p4;
   _pp[_polies,5]:=p5;
   _pp[_polies,6]:=p6;
   if tp=20 then _pt[_polies]:=19 else _pt[_polies]:=tp;
   _pc[_polies]:=cl;
   _pnv[_polies]:=6;
   CalcNormal;
   inc(_nsr[_sector]);
   if _nsr[_sector]>MAXSECSR then Error('Too many polygons in sector','Create6Plane');
   _sec[_sector,_nsr[_sector]]:=_polies;
   if tp=PORTAL then begin
      oldsect:=_sector;
      _sector:=cl;
      Create6Plane(p6,p5,p4,p3,p2,p1,20,oldsect);
      _sector:=oldsect;
   end;
end;


{  _______________________________________________________________________
  /                                                                      /
 / PolygonColor: Alter the color of a single polygon                    /
/______________________________________________________________________/ }

procedure PolygonColor(n,col: longint);
begin
   if n>_polies then Error('Polygon not defined','PolygonColor');
   _pc[n]:=col;
end;


{  _______________________________________________________________________
  /                                                                      /
 /  CalcBasePlanes: Determine the unique set of base planes             /
/______________________________________________________________________/ }

procedure CalcBasePlanes;
var i,p,l: longint;
    A,B,C,D: single;
    mx: single;
begin
   for i:=1 to _polies do
   if (_pt[i]=TEXTURED) or (_pt[i]=TRTEXT)
   or (_pt[i]=PHONGTX) or (_pt[i]=GOURADTX) then begin
      {Determine plane equation of polygon}
      A:=_pn[i,1];
      B:=_pn[i,2];
      C:=_pn[i,3];
      {Normalize plane equation}
      mx:=0;
      if abs(A)>abs(mx) then mx:=A;
      if abs(B)>abs(mx) then mx:=B;
      if abs(C)>abs(mx) then mx:=C;
      A:=A/mx;
      B:=B/mx;
      C:=C/mx;
      D:=-(A*_px[_pp[i,1]]+B*_py[_pp[i,1]]+C*_pz[_pp[i,1]]);
      p:=0;
      {Check set of base polies for lookalike}
      if _base>0 then for l:=1 to _base do begin
         if (abs((_fa[l]/A)-1)<0.01) and
         (abs((_fb[l]/B)-1)<0.01) and
         (abs((_fc[l]/C)-1)<0.01) and
         (abs((_fd[l]/D)-1)<0.01) then begin
            p:=l;
            l:=_base;
         end;
      end;
      {If not found, then create new base polygon}
      if p=0 then begin
         inc(_base);
         if _base>MAXFAKE then Error('Too many base polygons','CalcBasePlanes');
         _fp1[_base]:=_pp[i,1];
         _fp2[_base]:=_pp[i,2];
         _fp3[_base]:=_pp[i,4];
         _fa[_base]:=A;
         _fb[_base]:=B;
         _fc[_base]:=C;
         _fd[_base]:=D;
         p:=_base;
      end;
      {Connect current polygon to this base polygon}
      _pfb[i]:=p;
   end;
end;


{  _______________________________________________________________________
  /                                                                      /
 / CalcVertexNormals: Calculate the normal for each vertex              /
/______________________________________________________________________/ }

procedure CalcVertexNormals;  {Got this from PASPHONG}
var j,k: longint;
    len: single;
begin
   {Some tests on the datastructures}
   CheckGuards;
   if (_polies<1) or (_polies>MAXSIDES) or
      (_coords<1) or (_coords>MAXCOORD)
      then error('Data integrity error - Bad nr of coords or polies','CalcVertexNormals');
   for j:=1 to _coords do begin
      _rz[j]:=0;
      _rx[j]:=0;
      _ry[j]:=0;
      _frz[j]:=0;
   end;
   for j:=1 to _polies do begin
      for k:=1 to _pnv[j] do begin
         if (_pp[j,k]<1) or (_pp[j,k]>_coords) then
            Error('Data integrity error - '+n2s(_pp[j,k])+' is a bad coord','CalcVertexNormals');
         _rx[_pp[j,k]]:=_rx[_pp[j,k]]+_pn[j,1];
         _ry[_pp[j,k]]:=_ry[_pp[j,k]]+_pn[j,2];
         _frz[_pp[j,k]]:=_frz[_pp[j,k]]+_pn[j,3];
         inc(_rz[_pp[j,k]]);
      end;
   end;
   for j:=1 to _coords do begin
      if _rz[j]>0 then begin
         len:=sqrt(_rx[j]*_rx[j]+_ry[j]*_ry[j]+_frz[j]*_frz[j]);
         _nx[j]:=_rx[j]*127/len;
         _ny[j]:=_ry[j]*127/len;
         _nz[j]:=_frz[j]*127/len
      end else Warning('Unused coord ('+n2s(j)+') detected','CalcVertexNormals');
   end;
   Sector(1);
   CalcBasePlanes;
end;


{  _______________________________________________________________________
  /                                                                      /
 /  CopyBufToPhong: Make the current screenbuffer the environment map   /
/______________________________________________________________________/ }

procedure CopyBufToEnv; Assembler;
asm
   {Clear 28 lines of the envmap}
   mov ecx,28*64
   mov edi,[env]
   xor eax,eax
   rep stosd
   {Copy 200 lines from the screenbuffer}
   mov edi,28*256
   add edi,[env]
   mov ebx,200
   mov edx,0
@lloop:
   mov ecx,64
   mov esi,[buf]
   add esi,edx
   rep movsd
   add edx,320
   dec ebx
   jnz @lloop
   {Clear another 28 lines of the envmap}
   mov edi,[env]
   add edi,228*256
   mov ecx,28*64
   xor eax,eax
   rep stosd
end;


{  _______________________________________________________________________
  /                                                                      /
 /  CheckGuards: Check for array boundary violations                    /
/______________________________________________________________________/ }

procedure CheckGuards;
begin
   if _Guard1<>$99999999 then Error('Guard 1 has been killed','CheckGuards');
   if _Guard2<>$99999999 then Error('Guard 2 has been killed','CheckGuards');
   if _Guard3<>$99999999 then Error('Guard 3 has been killed','CheckGuards');
   if _Guard4<>$99999999 then Error('Guard 4 has been killed','CheckGuards');
   if _Guard5<>$99999999 then Error('Guard 5 has been killed','CheckGuards');
   {if memd[tex+$40800]<>$99999999 then Error('Tex Guard collapsed','CheckGuards');}
end;


{  _______________________________________________________________________
  /                                                                      /
 /  EnvRender: Rendering pipeline for creating dynamic environment map  /
/______________________________________________________________________/ }

procedure EnvRender(rz,rx,ry: longint; px,py,pz: single);
var x,y: longint;
begin
   {Clear the display buffer}
   {ClearBuffer;}
   {Setup the rotation/translation matrix for the world}
   RotMatrix(rx mod 2880,ry mod 2880,rz mod 2880,px,py,pz);
   LightMatrix(_lrx,_lry,_lrz,0,0,0);
   {Rotate the coordinates}
   RotateCoords;
   {Sort the polygons}
   TreeSort(0,1);
   {Draw the sorted polygons}
   If _Tree[1].lbr>0 then DrawTree(_Tree[1].lbr);
   if _Tree[1].rbr>0 then DrawTree(_Tree[1].rbr);
   {Copy the screenbuffer to the Env-Map}
   CopyBuftoEnv;
end;


{  _______________________________________________________________________
  /                                                                      /
 /  DrawCrossHairs: Draw crosshairs over the rendered image             /
/______________________________________________________________________/ }

procedure DrawCrossHairs;
var i,c1: longint;
    a: dword;
    h,v,ad: longint;
begin
   if hires=0 then h:=1 else h:=2;
   if hires=3 then begin v:=2; ad:=640 end else begin v:=1; ad:=320 end;
   c1:=FindColor(0,63,0)*256;
   a:=buf+XCENTR+(10*h)*ad;
   for i:=0 to (180*h) do begin
      mem[a]:=mem[mix+mem[a]+c1];
      inc(a,ad);
   end;
   a:=buf+10+(100*h)*ad;
   for i:=0 to 300*v do begin
      mem[a]:=mem[mix+mem[a]+c1];
      inc(a);
   end;
   c1:=FindColor(63,0,0);
   mem[buf+XCENTR+(100*h)*ad]:=c1;
end;


{  _______________________________________________________________________
  /                                                                      /
 /  Render: Rendering pipeline                                          /
/______________________________________________________________________/ }

procedure Render(rx,rz,ry: longint; px,py,pz: single);
var i,x,y: longint;
begin
   {Clear the display buffer (not needed for indoor: Full screen is updated)}
   {ClearBuffer;}
   {Setup the rotation/translation matrix for the world}
   RotMatrix(rx mod 2880,ry mod 2880,rz mod 2880,px,py,pz);
   LightMatrix(_lrx,_lry,_lrz,0,0,0);
   {Rotate the coordinates}
   RotateCoords(px,py,pz);
   {Sort the polygons}
   asm
      fld [px]
      fistp [x]
      fld [pz]
      fistp [y]
   end;
   _sector:=GetSector(-x,0,-y);
   if _sector=0 then _sector:=1;
   if _sector>_lsd then Error('Illegal sector reference ('+n2s(_sector)+')'
                       +chr(10)+chr(13)+'for coord ('+n2s(-x)+','+n2s(0)+','
                               +n2s(-y)+')','Render');
   TreeSort(0,_sector);
   {Clear the perspective fake/base flags}
   if _base>0 then for i:=1 to _base do _fclc[i]:=0;
   for i:=1 to _polies do _dr[i]:=0;
   {Draw the sorted polygons}
   _depth:=0;
   if _Tree[1].lbr>0 then DrawTree(_Tree[1].lbr);
   if _Tree[1].rbr>0 then DrawTree(_Tree[1].rbr);
   {Display the screenbuffer}
   CalcFrameRate;
   DrawCrossHairs;
   CheckGuards;
   if SPEEDTST=1 then Palette(0,63,0,0);
   WaitVbl;
   if SPEEDTST=1 then Palette(0,0,0,0);
   if HIRES=0 then ShowBuffer;
   if HIRES=1 then ShowHiBuf;
   if (HIRES=2) or (HIRES=3) then ShowVESA;
end;

end.
{ Mail: Jacco.Bikker@QuestIntl.com
  URL:  www.geocities.com/CapeCanaveral/5402
  Snailmail: Jacco Bikker, De Tip 14, 8061BX Hasselt, NETHERLANDS
  phone: 038-4774491
  May 1st I will move from Quest (Unilever) to Lost Boys Interactive,
  so my mail account will probably change. Keep an eye on my page,
  I will put the new address on the page.}
