#ifndef _BALDEF_H
#define _BALDEF_H


#include <stdlib.h>
#include <string.h>
#include <math.h>

#ifdef _UNIX
#include <unistd.h>
#include <fcntl.h>
#endif





typedef  unsigned int     BUINT;
typedef  unsigned short   BWORD;
typedef  unsigned char    BBYTE;
typedef  int              BBOOL;
typedef  double           BREAL;
typedef  unsigned char    BCOLOR;



#define  BTRUE      1
#define  BFALSE     0


#define  BNULL      0
#define  BVOID      (-1)
#define  BPI        3.141592659f


#define  BLIMMIN   -99999.9f
#define  BLIMMAX    99999.9f



struct BRGB {
    unsigned char r;
    unsigned char g;
    unsigned char b;
};



#define  BFRONT     0
#define  BBACK      1
#define  BFACE      int
#define  BOTHER(x)  (1-(x))



#ifndef  _DEBUG
#define  BASSERT(exp)  BNULL
#else
void     BAssert (BBOOL, const char *, int) ;
#define  BASSERT(exp)  BAssert((BBOOL)(exp), __FILE__, __LINE__)
#endif





void BMessage   (const char *);


#ifdef _UNIX
void BShowError (const char * prefix = BNULL);
#else
#define BShowError(prefix)
#endif



#define  BEPSILON   0.001f


inline BREAL BABS   (BREAL x)           {return x>=0.0f ? x : -x;}
inline int   BABS   (int   x)           {return x>=0 ? x : -x;}
inline BREAL BSQRT  (BREAL x)           {return (BREAL) sqrt (x);}
inline BREAL BCOS   (BREAL x)           {return (BREAL) cos  (x);}
inline BREAL BSIN   (BREAL x)           {return (BREAL) sin  (x);}
inline BREAL BTAN   (BREAL x)           {return (BREAL) tan  (x);}
inline BREAL BDIST  (BREAL x, BREAL y)  {return BSQRT (x*x + y*y);}
inline BBOOL BNEAR  (BREAL x, BREAL y)  {return BABS (x-y) < BEPSILON;}
inline BREAL BMAX   (BREAL x, BREAL y)  {return x>y ? x : y;}
inline BREAL BMIN   (BREAL x, BREAL y)  {return x<y ? x : y;}
inline int   BMAX   (int   x, int   y)  {return x>y ? x : y;}
inline int   BMIN   (int   x, int   y)  {return x<y ? x : y;}
inline BREAL BFLOOR (BREAL x)           {return (BREAL) floor (x);}
inline BREAL BCEIL  (BREAL x)           {return (BREAL) ceil  (x);}




inline void BMEMCOPY (void * dest, const void * source, int count) {
#ifdef _UNIX
    memcpy (dest, source, count);
#endif
#ifdef _WINDOWS
    memcpy (dest, source, count);
#endif
}




#ifdef _BIGENDIAN

inline short BREORDER (short x)
    {return ((x & 0x00FF) << 8) | (x >> 8);}

inline int   BREORDER (int x) {
    return ((x & 0x000000FF) << 24) |
           ((x & 0x0000FF00) <<  8) |
           ((x & 0x00FF0000) >>  8) |
           ((x & 0xFF000000) >> 24) ;
}

#else

inline short BREORDER (short x)   {return x;}
inline int   BREORDER (int   x)   {return x;}

#endif




#define  BDIV8(x)   ((x)>>3)
#define  BDIV4(x)   ((x)>>2)
#define  BDIV2(x)   ((x)>>1)
#define  BMUL8(x)   ((x)<<3)
#define  BMUL4(x)   ((x)<<2)
#define  BMUL2(x)   ((x)<<1)
#define  BMOD8(x)   ((x)&0x0007)
#define  BMOD4(x)   ((x)&0x0003)
#define  BMOD2(x)   ((x)&0x0001)
#define  BINT8(x)   ((x)&0xFFF8)
#define  BINT4(x)   ((x)&0xFFFC)
#define  BINT2(x)   ((x)&0xFFFE)


#define  BHINIBBLE(ch)       ((BUCHAR) ((ch) >> 4))
#define  BLONIBBLE(ch)       ((BUCHAR) ((ch) & 0x0F))
#define  BHIBYTEC(word)      ((BUCHAR) ((word) >> 8))
#define  BLOBYTEC(word)      ((BUCHAR) ((word) & 0x00FF))
#define  BHIWORDC(dword)     ((BUSHORT) ((dword) >> 16))
#define  BLOWORDC(dword)     ((BUSHORT) ((dword) & 0x0000FFFF))


#define  BHIBYTE(word)       (*((BUCHAR*) & (word) + 1))
#define  BLOBYTE(word)       ((BUCHAR) (word))
#define  BHIWORD(dword)      (*((BUSHORT*) & (dword) + 1))
#define  BLOWORD(dword)      ((BUSHORT) (dword))


#define  BMAKELONG(lo,hi)    ((BUINT) ((hi)<<16|(lo)))
#define  BMAKEWORD(lo,hi)    ((BUSHORT) ((hi)<<8|(lo)))




#endif

