#include "Copyright.h"


#ifdef _UNIX
#include <errno.h>
#include <stdarg.h>
#include <stdio.h>
#endif

#ifdef _WINDOWS
#include <windows.h>
#include <wchar.h>
#endif


#include "BalDef.h"



#ifdef _UNIX
#define  BSPRINTF   sprintf
#endif

#ifdef _WINDOWS
#define  BSPRINTF   wsprintf
#endif



#ifdef _DEBUG
void BAssert (BBOOL exp, const char * file, int line) {
    if (! exp) {
        char buffer [256];

        BSPRINTF (buffer, "Assertion failed: File %s line %d.", file, line);
        BMessage (buffer);
        abort () ;
        }
}
#endif





void BMessage (const char * message) {
#ifdef _WINDOWS
    MessageBox (BNULL, message, "Baltimore Message", MB_ICONEXCLAMATION);
#endif

#ifdef _UNIX
    write (2, message, strlen (message));
    write (2, "\n", 1);
#endif
}








#ifdef _UNIX


static const struct {
    int  num ;
    char var [16] ;
    char str [64] ;
} BErrorIndex [] = {
    {0,          "----",      "No error"} ,
    {EPERM,      "EPERM",     "Not super-user"} ,
    {ENOENT,     "ENOENT",    "No such file or directory"} ,
    {ESRCH,      "ESRCH",     "No such process"} ,
    {EINTR,      "EINTR",     "Interrupted system call"} ,
    {EIO,        "EIO",       "I/O error"} ,
    {ENXIO,      "ENXIO",     "No such device or address"} ,
    {E2BIG,      "E2BIG",     "Arg list too long"} ,
    {ENOEXEC,    "ENOEXEC",   "Exec format error"} ,
    {EBADF,      "EBADF",     "Bad file number"} ,
    {ECHILD,     "ECHILD",    "No children"} ,
    {EAGAIN,     "EAGAIN",    "No more processes"} ,
    {ENOMEM,     "ENOMEM",    "Not enough core"},
    {EACCES,     "EACCES",    "Permission denied"} ,
    {EFAULT,     "EFAULT",    "Bad address"} ,
    {ENOTBLK,    "ENOTBLK",   "Block device required"} ,
    {EBUSY,      "EBUSY",     "Mount device busy"} ,
    {EEXIST,     "EEXIST",    "File exists"} ,
    {EXDEV,      "EXDEV",     "Cross-device link"} ,
    {ENODEV,     "ENODEV",    "No such device"} ,
    {ENOTDIR,    "ENOTDIR",   "Not a directory"} ,
    {EISDIR,     "EISDIR",    "Is a directory"} ,
    {EINVAL,     "EINVAL",    "Invalid argument"} ,
    {ENFILE,     "ENFILE",    "File table overflow"} ,
    {EMFILE,     "EMFILE",    "Too many open files"} ,
    {ENOTTY,     "ENOTTY",    "Not a typewriter"} ,
    {ETXTBSY,    "ETXTBSY",   "Text file busy"} ,
    {EFBIG,      "EFBIG",     "File too large"} ,
    {ENOSPC,     "ENOSPC",    "No space left on device"} ,
    {ESPIPE,     "ESPIPE",    "Illegal seek"} ,
    {EROFS,      "ESOFS",     "Read only file system"} ,
    {EMLINK,     "EMLINK",    "Too many links"} ,
    {EPIPE,      "EPIPE",     "Broken pipe"} ,
    {EDOM,       "EDOM",      "Math arg out of domain of func"} ,
    {ERANGE,     "ERANGE",    "Math result not representable"} ,
    {ENOMSG,     "ENOMSG",    "No message of desired type"} ,
    {EIDRM,      "EIDRM",     "Identifier removed"} ,
    {EDEADLK,    "EDEADLK",   "Deadlock condition"} ,
    {ENOLCK,     "ENOLCK",    "No record locks available"} ,
    {9999,       "",          ""}
} ;





void BShowError (const char * prefix) {
    BBOOL explanable = BFALSE;
    int   ptr = 0;
    char  buffer [256];

    if (prefix != BNULL) {
        strcpy (buffer, prefix);
        ptr = strlen (buffer);
        buffer [ptr++] = ' ';
        }

    for (register ushort i=0; BErrorIndex[i].num!=9999; i++)
        if (BErrorIndex[i].num == errno)
            {
            explanable = BTRUE;
            ptr += BSPRINTF (buffer+ptr, "%d %s %s.",
                             errno,
                             BErrorIndex[i].var,
                             BErrorIndex[i].str);
            }

    if (! explanable)
        ptr += BSPRINTF (buffer+ptr, "  No error entry.");
    buffer [ptr] = 0;
    BMessage (buffer);
}



#endif
