/*---------------------------------------------------------------------------
                        THE CHAOS ENGINE - CHAOS.C

                 A 3D Engine By Lars Andersson 1997-1997
---------------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>

#include "chaos.h"
#include "3dengine.h"
#include "math.h"
#include "keyboard.h"
#include "timer.h"
#include "3dw.h"
#include "gfx.h"
#include "text.h"

extern int polygons_drawn, ysize, xsize;
extern char *palette,*gfxbuffer, **textures_pp;

char *message_p;
int mode=1,ticks,total_time,message_ticks,message_x,cubesnurr=0,spd=0;
float shipspeed=0,pushspeed=0;
dir ship_dir;
vertice_3d ship_pos;
vector_3d push_vector;
object_3d *active,*rocket;
level *levinfo;

//-- MAIN -------------------------------------------------------------------

main(int argv, char *argc[])
{
int i,t,error,gfxmode,frames=0;
char file[32];

printf("\n");
printf(" The Chaos-Engine 1.0  Lars Andersson 1996-1997\n");
printf("\n");

if (!Init_Chaos()) Stop("Init Error.",1);
if (argv>1) {
   if (!Init_3DW(argc[1])) Stop("",1);
}
else {
   if (!Init_3DW("chaos.3dw")) Stop("",1);
}

if (NULL==(levinfo=Load_Level(0))) Stop("",1);

puts("Alright! I like thoose messages. They are soo cool ;)");
puts("HitMe!");
while(!KeyHit());

if (!Set_GfxMode(MODE13)) {
   TextMode();
   Stop("Not enough memory for framebuffer.",1);
}

ship_pos=Get_ViewPoint();
ship_dir=Get_ViewDirection();
total_time=0;
Reset_MicroTimer();
ticks=TICKS_PER_FRAME;
message_ticks=MESSAGE_DELAY;
rocket=Object_Address(0);
Message("TCE 1.0 - PRESS 'I' FOR HELP");
do {
   if (TestKey(TAB_KEY)) Activate_Object();
   if (TestKey(F5_KEY)) Set_GfxMode(MODE13);
   if (TestKey(F6_KEY)) Set_GfxMode(MODEX320X240);
   if (TestKey(F7_KEY)) Set_GfxMode(MODEX320X400);

   if (TestKey(PLUS_KEY)) Increase_GfxWindowSize();
   if (TestKey(MINUS_KEY)) Decrease_GfxWindowSize();

   if (TestKey(I_KEY)) Info();

   Set_ViewPoint(&ship_pos);
   Set_ViewDirection(&ship_dir);

   Make_Camera_Matrix();

   Remove_Hidden_Objects();
   Rotate_Objects_3d();
   Local_2_World();
   World_2_Camera();
   Project_Vertices();
   ZClip_Objects();
   Remove_BackFaces();

   Make_PolyList();
   Sort_PolyList();
   Draw_PolyList();

   if (TestKey(F12_KEY)) Save_TGA();

   PutStringXY(0,ysize-8,63,"FPS:");
   PutInt(127,TICKS_PER_SECOND/ticks);

   if (message_ticks<MESSAGE_DELAY) PutStringXY(message_x,1,63,message_p);

   Display_Frame();
   frames++;

   while((TICKS_PER_FRAME/3)>Elapsed_Ticks());
   ticks=Elapsed_Ticks();
   total_time+=ticks;
   message_ticks+=ticks;

   while(keys[P_KEY]);
   Reset_MicroTimer();

   Animate_Objects();
   Control_Ship();

} while(!Key(ESC_KEY));

TextMode();
printf("\nStatistics...            \n");
printf("-\n");
printf("  Elapsed Time(s)     : %-16.1f\n", (float)total_time/TICKS_PER_SECOND);
printf("  Frames Drawn        : %-16d\n", frames);
printf("  Polygons Per Frame  : %-16d\n", polygons_drawn/frames);
printf("  Frames Per Second   : %-16.1f\n", (float)frames/((float)total_time/TICKS_PER_SECOND));
printf("-\n\n");
printf("Check out my 3d programming homepage at 'www.algonet.se/~lars_and/'\nfor the latest news and updates. ");
printf("You will also find the full C/ASM\nsource code and some good programming links and downloads.\n\n");
printf("Lars Andersson - lars_and@algonet.se\n");

Stop("", 0);
}

/*---------------------------------------------------------------------------
            Finds the closest object and makes it the active one
---------------------------------------------------------------------------*/

void Activate_Object(void)
{
object_3d *obj;
float dist,mindist=100000;
int i;

active=NULL;
for (i=0; i<levinfo->objects; i++) {
   obj=Object_Address(i);
   if (obj->flags & OBJ_VISIBLE) {
      dist=Vertice_Distance(&(ship_pos), &(obj->world));
      if (dist<mindist) {
         active=obj;
         mindist=dist;
      }
   }
}
}

/*---------------------------------------------------------------------------
                            Animates the objects
---------------------------------------------------------------------------*/

void Animate_Objects(void)
{
int i;
object_3d *obj_p;

if (pushspeed>0) {
   active->world.x+=(push_vector.x*pushspeed*ticks)/TICKS_PER_FRAME;
   active->world.y+=(push_vector.y*pushspeed*ticks)/TICKS_PER_FRAME;
   active->world.z+=(push_vector.z*pushspeed*ticks)/TICKS_PER_FRAME;
   pushspeed*=0.9;
   if (pushspeed<0.1) pushspeed=0;
}

if (active!=NULL) {
   if (Key(SPACE_KEY)) {
      Message("PUSH!");
      pushspeed=10+10/Vertice_Distance(&(ship_pos), &(active->world));
      push_vector=active->world;
      push_vector.x-=ship_pos.x;
      push_vector.y-=ship_pos.y;
      push_vector.z-=ship_pos.z;
      Normalize_Vector(&push_vector);
   }

   if (Key(RSHIFT_KEY)) {
      if (Key(F1_KEY)) active->angle.x-=(10*ticks)/TICKS_PER_FRAME;
      if (Key(F2_KEY)) active->angle.y-=(10*ticks)/TICKS_PER_FRAME;
      if (Key(F3_KEY)) active->angle.z-=(10*ticks)/TICKS_PER_FRAME;
   }
   else {
      if (Key(F1_KEY)) active->angle.x+=(10*ticks)/TICKS_PER_FRAME;
      if (Key(F2_KEY)) active->angle.y+=(10*ticks)/TICKS_PER_FRAME;
      if (Key(F3_KEY)) active->angle.z+=(10*ticks)/TICKS_PER_FRAME;
   }
}
for (i=0; i<levinfo->objects; i++) {
   obj_p=Object_Address(i);
   if (obj_p!=active) {
      obj_p->angle.x+=(obj_p->rotation.x*ticks)/TICKS_PER_FRAME;
      obj_p->angle.y+=(obj_p->rotation.y*ticks)/TICKS_PER_FRAME;
      obj_p->angle.z+=(obj_p->rotation.z*ticks)/TICKS_PER_FRAME;
   }
}
}

/*---------------------------------------------------------------------------
                         Controls the ship movement
---------------------------------------------------------------------------*/

void Control_Ship(void)
{
float movespeed, lookspeed;

static int rocket_dist;
static char rocket_on=FALSE,afterburner=FALSE;

lookspeed=ticks*15/TICKS_PER_FRAME;
if (Key(CAPSLOCK_KEY) && message_ticks>MESSAGE_DELAY) {
   afterburner=!afterburner;
   if (afterburner) Message("AFTERBURNER ACTIVATED");
   else Message("AFTERBURNER DEACTIVATED");
}
if (afterburner) movespeed=ticks*20/TICKS_PER_FRAME;
else movespeed=ticks*4/TICKS_PER_FRAME;


// Update viewangles

if (!Key(ALT_KEY)) {
   if (Key(LEFT_KEY)) {
      ship_dir.y-=lookspeed;
   }
   if (Key(RIGHT_KEY)) {
      ship_dir.y+=lookspeed;
   }
/*
   if (Key(UP_KEY)) {
      ship_dir.x-=lookspeed;
   }
   if (Key(DOWN_KEY)) {
      ship_dir.x+=lookspeed;
   }
*/
}

else {
   if (Key(LEFT_KEY)) {
      ship_pos.x-=movespeed*Cos(ship_dir.y);
      ship_pos.z+=movespeed*Sin(ship_dir.y);
   }
   if (Key(RIGHT_KEY)) {
      ship_pos.x+=movespeed*Cos(ship_dir.y);
      ship_pos.z-=movespeed*Sin(ship_dir.y);
   }

   if (Key(UP_KEY)) ship_pos.y+=movespeed;
   if (Key(DOWN_KEY)) ship_pos.y-=movespeed;
}

ship_dir.x&=SCMASK;
ship_dir.y&=SCMASK;
ship_dir.z&=SCMASK;

if (Key(A_KEY)) {
   shipspeed=movespeed;
}
if (Key(Z_KEY)) {
   shipspeed=-movespeed;
}
ship_pos.x+=shipspeed*Sin(ship_dir.y)*Cos(ship_dir.x);
ship_pos.z+=shipspeed*Cos(ship_dir.y)*Cos(ship_dir.x);
shipspeed*=0.9;


if (rocket_on) {
   rocket->world.x+=(32*Sin(rocket->angle.y)*ticks)/TICKS_PER_FRAME;
   rocket->world.z+=(32*Cos(rocket->angle.y)*ticks)/TICKS_PER_FRAME;
   if (10000<(Vertice_Distance(&ship_pos,&(rocket->world)))) {
      Object_Off(rocket);
      rocket_on=FALSE;
   }
}

if (Key(CTRL_KEY)) {
   Message("SHOOOT!");
   rocket->world.x=ship_pos.x - 100*Cos(ship_dir.y);
   rocket->world.y=ship_pos.y - 32;
   rocket->world.z=ship_pos.z + 100*Sin(ship_dir.y);
   rocket->angle=ship_dir;

   Object_On(rocket);
   rocket_dist=0;
   rocket_on=TRUE;
}
}

/*---------------------------------------------------------------------------
                   Show basic key combinations and POV,FOV
---------------------------------------------------------------------------*/

void Info(void)
{
int winsize;

winsize=Get_GfxWindowSize();
Set_GfxWindowSize(0);
Project_Vertices();
Draw_PolyList();
PutStringXY(80,16,180,"BASIC KEY COMBINATIONS...");
PutStringXY(50,35,62,"A            -  MOVE FORWARD");
PutStringXY(50,45,60,"Z            -  MOVE BACKWARD");
PutStringXY(50,55,58,"LEFT ARROW   -  TURN LEFT");
PutStringXY(50,65,56,"RIGHT ARROW  -  TURN RIGHT");
PutStringXY(50,75,54,"ALT ARROW    -  STRAFE");
PutStringXY(50,85,52,"CAPS-LOCK    -  AFTERBURNER ON/OFF");
PutStringXY(50,95,50,"TAB          -  HOOK OBJECT");
PutStringXY(50,105,48,"F1/F2/F3     -  ROTATE OBJECT");
PutStringXY(50,115,46,"SPACE        -  PUSH OBJECT");
PutStringXY(50,125,44,"ESC          -  QUIT");

PutStringXY(20,ysize-40,63,"POV:");

PutInt(127, (int)(ship_pos.x));
PutChar(63, ',');
PutInt(127, (int)(ship_pos.y));
PutChar(63, ',');
PutInt(127, (int)(ship_pos.z));

PutStringXY(160,ysize-40,63,"FOV:");

PutInt(127, ship_dir.x);
PutChar(63, ',');
PutInt(127, ship_dir.y);
PutChar(63, ',');
PutInt(127, ship_dir.z);

PutStringXY(70,ysize-20,255,"PRESS ANY KEY TO CONTINUE.");

Display_Frame();
while(!KeyHit());

Set_GfxWindowSize(winsize);
Reset_MicroTimer();
}

// Show a message

void Message(char *mess)
{
message_p=mess;
message_ticks=NULL;
message_x=(xsize-strlen(mess)*FONTXSIZE)>>1;
}

// Initialize engine 

int Init_Chaos(void)
{
Init_MicroTimer();
Install_KeyHandler();
if (Init_3DEngine());
return(TRUE);
}

// Close down

void Quit(void)
{
Close_3DEngine();
Remove_MicroTimer();
Remove_KeyHandler();
}

// Exit Chaos

void Stop(char *mess, int kod)
{
Quit();
Close_Level();
Close_3DW();
cputs(mess);
exit(kod);
}
