#include <math.h>
#include <malloc.h>
#include "fmath.h"

int *fsint,*fcost,*facost;

void addmatrix(matrix a,matrix b,matrix r){
   int x,y;
   for(x=0;x<3;x++)
       for(y=0;y<3;y++)
           r[x][y]=a[x][y]+b[x][y];
}

void submatrix(matrix a,matrix b,matrix r){
   int x,y;
   for(x=0;x<3;x++)
       for(y=0;y<3;y++)
           r[x][y]=a[x][y]-b[x][y];
}

void copymatrix(matrix a,matrix r){
   int x,y;
   for(x=0;x<3;x++)
       for(y=0;y<3;y++)
           r[x][y]=a[x][y];
}
     
void mulmatrix(matrix a,matrix b,matrix r){
   int x,y,z;
   for(x=0;x<3;x++)
       for(y=0;y<3;y++){
           r[y][x] =fixmul(a[0][x],b[y][0]);
           r[y][x]+=fixmul(a[1][x],b[y][1]);
           r[y][x]+=fixmul(a[2][x],b[y][2]);
       }    
}

void mulmatrixvec(matrix a,vect b,vect r){
   int x,y;
   r[0] =fixmul(a[0][0],b[0]);
   r[0]+=fixmul(a[1][0],b[1]);
   r[0]+=fixmul(a[2][0],b[2]);
   r[1] =fixmul(a[0][1],b[0]);
   r[1]+=fixmul(a[1][1],b[1]);
   r[1]+=fixmul(a[2][1],b[2]);
   r[2] =fixmul(a[0][2],b[0]);
   r[2]+=fixmul(a[1][2],b[1]);
   r[2]+=fixmul(a[2][2],b[2]);
}

void scalevector(vect a,int b,vect r){
   r[0]=fixmul(a[0],b);
   r[1]=fixmul(a[1],b);
   r[2]=fixmul(a[2],b);
}

int vectordot(vect a,vect b){
   return fixmul(a[0],b[0])+fixmul(a[1],b[1])+fixmul(a[2],b[2]);
}

int vectorlength(vect v){
   int i,lng;
   i=fixmul(v[0],v[0])+fixmul(v[1],v[1])+fixmul(v[2],v[2]);
//   if (i) lng=(sqrt(i)*256); else lng=65536; //256=sqrt(65536)
   if (i) lng=fixsqrt(i,sqrt_table); else lng=65536;
   return(lng);
}

void vectornormalize(vect v){
   int lng;
   lng=vectorlength(v);
   v[0]=fixmuldiv(v[0],65536,lng);
   v[1]=fixmuldiv(v[1],65536,lng);
   v[2]=fixmuldiv(v[2],65536,lng);
}

void vectorcross(vect a,vect b,vect r){
   r[0]=fixmul(a[1],b[2])-fixmul(a[2],b[1]);
   r[1]=fixmul(a[2],b[0])-fixmul(a[0],b[2]);
   r[2]=fixmul(a[0],b[1])-fixmul(a[1],b[0]);
}

void orthonormalizematrix(matrix a){
   vect temp,scale;
// normalise le 1er vecteur
   vectornormalize(a[0]);
// met le 2eme vecteur perpendiculaire au 1er
   scalevector(a[0],vectordot(a[1],a[0]),temp);
   subvector(a[1],temp,a[1]);
// normalise le 2eme vecteur
   vectornormalize(a[1]);
// le 3eme vecteur est calcul (produit crois)
   vectorcross(a[0],a[1],a[2]);
}

void addvector(vect a,vect b,vect r){
   r[0]=a[0]+b[0];
   r[1]=a[1]+b[1];
   r[2]=a[2]+b[2];
}

void subvector(vect a,vect b,vect r){
   r[0]=a[0]-b[0];
   r[1]=a[1]-b[1];
   r[2]=a[2]-b[2];
}

void copyvector(vect a,vect r){
   r[0]=a[0];
   r[1]=a[1];
   r[2]=a[2];
}

void rotatevectors(vect v1,vect v2,vect u1,vect u2,int s,int c){
   vect t1,t2,g1,g2;
   scalevector(v1,c,t1);
   scalevector(v2,s,t2);
   scalevector(v1,s,g1);
   scalevector(v2,c,g2);
   subvector(t1,t2,u1);
   addvector(g1,g2,u2);
}

void rotatebase(matrix m,int axis,int s,int c){
   int a,b;
   a=axis+1;
   if(a>=3) a=0;
   b=a+1;
   if(b>=3) b=0;
   rotatevectors(m[a],m[b],m[a],m[b],s,c);
}

void scalematrix(matrix a,int b,matrix r){
   int x,y;
   for(x=0;x<3;x++)
       for(y=0;y<3;y++)
           r[x][y]=fixmul(a[x][y],b);
}

void setmatrix(matrix r,int a1,int a2,int a3,int b1,int b2,int b3,int c1,int c2,int c3){
   r[0][0]=a1; r[0][1]=a2; r[0][2]=a3;
   r[1][0]=b1; r[1][1]=b2; r[1][2]=b3;
   r[2][0]=c1; r[2][1]=c2; r[2][2]=c3;
}

void setaffine(affine *r,int a1,int a2,int a3,int b1,int b2,int b3,int c1,int c2,int c3,int x,int y,int z){
   setmatrix(r->matrice,a1,a2,a3,b1,b2,b3,c1,c2,c3);
   r->vector[0]=x;   
   r->vector[1]=y;   
   r->vector[2]=z;   
}

void printmatrix(matrix r){
   printf("| %10i,%10i,%10i |\n",r[0][0],r[0][1],r[0][2]);
   printf("| %10i,%10i,%10i |\n",r[1][0],r[1][1],r[1][2]);
   printf("| %10i,%10i,%10i |\n",r[2][0],r[2][1],r[2][2]);
}

void printvector(vect r){
   int i;
   i=fixmul(r[0],r[0])+fixmul(r[1],r[1])+fixmul(r[2],r[2]);
   printf("(%10i,%10i,%10i) length=%10i\n",r[0],r[1],r[2],i);
}

void printaffine(affine r){
   printmatrix(r.matrice);
   printvector(r.vector);
}

void affine_xform(affine a,vect i,vect j){
   mulmatrixvec(a.matrice,i,j);
   addvector(j,a.vector,j);
}

void fbuild_sincos(void){
   int i;
   if (fsint!=NULL) return;
   fsint=(int *)malloc(1024*4);
   fcost=(int *)malloc(1024*4);
   facost=(int *)malloc(1024*4);
   if (fsint==NULL||fcost==NULL||facost==NULL){
       printf("FMATH.C: out of memory in fbuild_sincos()\n");
       exit(1);
   }
   for (i=0;i<1024;i++){
      fsint[i]=floor(sin(i/(1024/2.0)*pi)*(float)65536+.5);
      fcost[i]=floor(cos(i/(1024/2.0)*pi)*(float)65536+.5);
      facost[i]=floor((acos((i-(1024/2))/(1024/2.0))*512/pi)*(float)65536+.5);
   }   
}

short sqrt_table[512];

void fbuild_sqrt(void){
  float curfloat;
  unsigned long *curfloatint;
  unsigned int count;

  curfloatint=(unsigned long *)&curfloat;
  for (count=0;count<0x100;count++){
    *curfloatint=0;
    *curfloatint=(count<<15)|(127<<23);
    curfloat=sqrt(curfloat);
    sqrt_table[count]=(*curfloatint>>15)&0xff;
    *curfloatint=0;
    *curfloatint=(count<<15)|(128<<23);
    curfloat=sqrt(curfloat);
    sqrt_table[count+0x100]=(*curfloatint>>15)&0xff;
  }
}

void setvector(vect v,int x,int y,int z){
   v[0]=x; v[1]=y; v[2]=z;
}

void init_fmath(void){
    fbuild_sincos();
    fbuild_sqrt();
}

//void printinfo_fmath(void){
//   printf("FMATH v%1i.%02i compiled on %s at %s\n",_FMA_VERHI,_FMA_VERLO,WCC_DATE,WCC_TIME);
//}
