/*
    16.16 fixed-point matrices & vectors math functions
    (c)1996 by Deks/incamasters aka Jean-Francois Dube
*/

#ifndef _FMATH_H
#define _FMATH_H

#define pi 3.14159265358
#define sinus(a) fsint[a&1023]
#define cosin(a) fcost[a&1023]
#define arcos(a) facost[a&1023]

typedef int vect[3];                    // vecteur (x,y,z)
typedef vect matrix[3];                 // matrice 3x3
typedef struct{
    matrix matrice;
    vect vector;
} affine;                               // transformation affine (3x3+vecteur)

extern int *fsint,*fcost,*facost;
extern short sqrt_table[512];

// stuff matrices
void addmatrix(matrix a,matrix b,matrix r);
void submatrix(matrix a,matrix b,matrix r);
void copymatrix(matrix a,matrix r);
void mulmatrix(matrix a,matrix b,matrix r);
void mulmatrixvec(matrix a,vect b,vect r);
void orthonormalizematrix(matrix a);
void scalematrix(matrix a,int b,matrix r);
void setmatrix(matrix r,int a1,int a2,int a3,int b1,int b2,int b3,int c1,int c2,int c3);
void setaffine(affine *r,int a1,int a2,int a3,int b1,int b2,int b3,int c1,int c2,int c3,int x,int y,int z);
void affine_xform(affine a,vect i,vect j);
void printmatrix(matrix r);

// stuff vecteurs
void scalevector(vect a,int b,vect r);
int vectordot(vect a,vect b);
int vectorlength(vect v);
void vectornormalize(vect v);
void vectorcross(vect a,vect b,vect r);
void addvector(vect a,vect b,vect r);
void subvector(vect a,vect b,vect r);
void copyvector(vect a,vect r);
void rotatevectors(vect v1,vect v2,vect u1,vect u2,int s,int c);
void rotatebase(matrix m,int axis,int s,int c);
void printvector(vect r);
void setvector(vect v,int x,int y,int z);

// stuff millanchousse
void fbuild_sincos(void);
void fbuild_sqrt(void);
void init_fmath(void);
int fixsqrt(int a,short *sqrtbl);
int fixmul(int a,int b);                // (a*b)>>16 sur 64 bit
int fixdiv(int a,int b);                // (a/b)>>16 sur 64 bit
int fixmuldiv(int a,int b,int c);       // ((a*b)/c)>>16 sur 64 bit
void printinfo_fmath(void);

#pragma aux fixsqrt=\
    "bsr ecx,eax"\
    "jz @@done"\
    "ror eax,cl"\
    "sub ecx,16"\
    "shr eax,24"\
    "mov edx,ecx"\
    "and edx,1"\
    "shl edx,8"\
    "or eax,edx"\
    "sar ecx,1"\
    "mov eax,[ebx+eax*2]"\
    "and eax,0ffffh"\
    "or eax,100h"\
    "add ecx,8"\
    "js @@shiftright"\
    "shl eax,cl"\
    "jmp @@done"\
    "@@shiftright:"\
    "neg cl"\
    "shr eax,cl"\
    "@@done:"\
    value [eax] parm [eax] [ebx] modify [eax ebx ecx edx];

#pragma aux fixmul=\
    "imul edx"\
    "shl edx,16"\
    "shr eax,16"\
    "or eax,edx"\
    parm [eax] [edx] value [eax];
    
#pragma aux fixdiv=\
    "mov eax,edx"\
    "sar edx,16"\
    "shl eax,16"\
    "idiv ecx"\
    parm [edx] [ecx] value [eax] modify [eax];

#pragma aux fixmuldiv=\
    "imul edx"\
    "idiv ecx"\
    parm [eax] [edx] [ecx] value [eax];

#endif
