#include "Copyright.h"
#include "File.h"





BFile::BFile () : 

#ifdef _UNIX
    m_fd       (-1), 
#endif

#ifdef _WINDOWS
    m_handle   (INVALID_HANDLE_VALUE),
#endif

    m_canread  (BFALSE),
    m_canwrite (BFALSE),
    m_rmaxsize (2048),
    m_rbuffer  (new BBYTE [m_rmaxsize]),
    m_rbufptr  (0),
    m_rbufsize (0),
    m_wmaxsize (2048),
    m_wbuffer  (new BBYTE [m_wmaxsize]),
    m_wbufsize (0)   
    {}





BFile::~BFile () {
    if (m_canread || m_canwrite) 
        Close ();

    delete [] m_rbuffer;
    return;
}





#ifdef _WINDOWS

BBOOL BFile::GetBuffer () {
    BASSERT (m_canread);
    BASSERT (m_rbufptr == m_rbufsize);
    
    DWORD count;
    if (ReadFile (m_handle, m_rbuffer, m_rmaxsize, & count, NULL) == FALSE)
        return BFALSE;
    m_rbufsize = count;
    m_rbufptr  = 0;
    return BTRUE;
}



BBOOL BFile::PutBuffer () {
    BASSERT (m_canwrite);
    BASSERT (m_wbufsize > 0 && m_wbufsize <= m_wbufsize);
    
    DWORD count;
    if (WriteFile (m_handle, m_wbuffer, m_wbufsize, & count, NULL) == FALSE)
        return BFALSE;
    if (count != (DWORD) m_wbufsize)
        return BFALSE;
    m_wbufsize = 0;
    return BTRUE;
}




BBOOL BFile::Open (const char * name) {
    BASSERT (! m_canread && ! m_canwrite);

    m_handle = CreateFile (name, GENERIC_READ, 0, NULL, 
                           OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, 0);
    if (m_handle == INVALID_HANDLE_VALUE)
        return BFALSE;
    
    m_canread  = BTRUE;
    m_rbufptr  = 0;
    m_rbufsize = 0;
    return BTRUE;
}




BBOOL BFile::Create (const char * name) {
    BASSERT (! m_canread && ! m_canwrite);
    
    m_handle = CreateFile (name, GENERIC_WRITE, 0, NULL, 
                           CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, 0);
    if (m_handle == INVALID_HANDLE_VALUE)
        return BFALSE;
    
    m_canwrite = BTRUE;
    m_wbufsize = 0;
    return BTRUE;
}




void  BFile::Close () {
    BASSERT (m_canread || m_canwrite);
    
    if (m_canwrite && m_wbufsize > 0)
        PutBuffer ();
    
    CloseHandle (m_handle);
    m_canread  = BFALSE;
    m_canwrite = BFALSE;
    return;
}


#endif



#ifdef _UNIX

BBOOL BFile::GetBuffer () {
    BASSERT (m_canread);
    BASSERT (m_rbufptr == m_rbufsize);
    
    int count = read (m_fd, m_rbuffer, m_rmaxsize);
    if (count == -1)
        return BFALSE;
    m_rbufsize = count;
    m_rbufptr  = 0;
    return BTRUE;
}



BBOOL BFile::PutBuffer () {
    BASSERT (m_canwrite);
    BASSERT (m_wbufsize > 0 && m_wbufsize <= m_wbufsize);
    
    int count = write (m_fd, m_wbuffer, m_wbufsize);
    if (count != m_wbufsize)
        return BFALSE;
    m_wbufsize = 0;
    return BTRUE;
}




BBOOL BFile::Open (const char * name) {
    BASSERT (! m_canread && ! m_canwrite);

    m_fd = open (name, O_RDONLY);
    if (m_fd < 0)
        return BFALSE;
    
    m_canread  = BTRUE;
    m_rbufptr  = 0;
    m_rbufsize = 0;
    return BTRUE;
}




BBOOL BFile::Create (const char * name) {
    BASSERT (! m_canread && ! m_canwrite);
    
    m_fd = open (name, O_WRONLY | O_CREAT | O_TRUNC, 0644);
    if (m_fd < 0)
        return BFALSE;
    
    m_canwrite = BTRUE;
    m_wbufsize = 0;
    return BTRUE;
}




void  BFile::Close () {
    BASSERT (m_canread || m_canwrite);
    
    if (m_canwrite && m_wbufsize > 0)
        PutBuffer ();
    
    close (m_fd);
    m_canread  = BFALSE;
    m_canwrite = BFALSE;
    return;
}


#endif





int BFile::Read (void * buffer, int len) {
    BASSERT (m_canread);

    int total = 0;
    while (len-total > m_rbufsize-m_rbufptr) {
        BMEMCOPY ((char *)buffer+total, m_rbuffer+m_rbufptr, m_rbufsize-m_rbufptr);
        total += m_rbufsize - m_rbufptr;
        m_rbufptr = m_rbufsize;
        if (GetBuffer () == BFALSE) {
            BASSERT (total < len);
            return total;
            }
        }
    
    BMEMCOPY ((char *)buffer+total, m_rbuffer+m_rbufptr, len-total);
    m_rbufptr += len - total;
    return len;
}




int BFile::Write (const void * buffer, int len) {
    BASSERT (m_canwrite);
    
    const BBYTE * data = (const BBYTE *) buffer;
    if (len <= m_wmaxsize - m_wbufsize) {
        BMEMCOPY (m_wbuffer + m_wbufsize, data, len);
        m_wbufsize += len;
        return len;
        }

    int remain = len;
    BMEMCOPY (m_wbuffer+m_wbufsize, data, m_wmaxsize-m_wbufsize);
    data += m_wmaxsize-m_wbufsize;
    len  -= m_wmaxsize-m_wbufsize;
    do {
        if (PutBuffer () == BFALSE)
            return remain-len;
        int count = BMIN (len, m_wmaxsize);
        BMEMCOPY (m_wbuffer, data, count);
        data += count;
        len  -= count;
        m_wbufsize = count;
    } while (m_wbufsize == m_wmaxsize);

    return remain;
}


