/*---------------------------------------------------------------------------
                        THE CHAOS ENGINE - GFX.C

                         Misc graphics routines
---------------------------------------------------------------------------*/

#include <stdio.h>

#include "linear.h"
#include "3dengine.h"
#include "math.h"
#include "gfx.h"
#include "miscgfx.h"

//-- Variables --------------------------------------------------------------

gfxmode gfxmodes[MODES]={320,200,0,GFXMODE_MODE13,
                         320,240,0,GFXMODE_MODEX,
                         320,400,0,GFXMODE_MODEX};

float scrsizes[5]={1, 0.8, 0.6, 0.4, 0.2};

char *gfxbuffer;

int bottom_clip, right_clip;
int gfx_mode,xres,yres,xsize,ysize,xcenter,ycenter,vmemdest;
int bytes_per_frame,gfxbufsize;
int backfill,ymultab[400];

float pmulx,pmuly,pmulx_zclip,pmuly_zclip,hfov_delta,vfov_delta;

rgb palette[256];

void (*Display)(void);

//-- External Variables -----------------------------------------------------

extern int edge1[400],ymultab[400];
extern vector_3d light_source;
extern int ambience,spd;

//-- Pragmas ----------------------------------------------------------------

#pragma aux SetMode13h=\
                       "mov   ax, 13h",\
                       "int   10h",\
                       modify [eax];

//-- Initialize some gfx stuff ----------------------------------------------

void Init_Gfx(void)
{
int i;
backfill=0x09090909;
gfxbuffer=NULL;
for (i=0; i<400; i++) edge1[i]=MAXLONG;
Init_Poly();
puts("Gfx Engine Initialized...");
}

//-- Shut down the gfx stuff ------------------------------------------------

void Close_Gfx(void)
{
if (gfxbuffer!=NULL) free(gfxbuffer);
}

//-- Set video mode ---------------------------------------------------------

int Set_GfxMode(char mode)
{
// Init gfxmode constants.

gfx_mode=mode;
xres=gfxmodes[mode].sx;
yres=gfxmodes[mode].sy;
Init_GfxConstants();

// Free previous framebuffer.

if (gfxbuffer!=NULL) {
   free(gfxbuffer);
}

// Allocate memory for new framebuffer.

if (NULL==(gfxbuffer=(char *)malloc(xres*yres, 1))) {
   printf("Gfx Memory Error\n");
   while(!KeyHit());
   return(NULL);
}

ClearBuffer(0);
WaitRetrace();

if (gfxmodes[mode].flags & GFXMODE_MODE13) {
   Display=DisplayBuffer_Mode13;
   SetMode13h();
}
if (gfxmodes[mode].flags & GFXMODE_MODEX) {
   Display=DisplayBuffer_Tweaked;
   SetTweakedMode(mode-1);
}
SetPalette((rgb *)palette);
return(TRUE);
}

//-- Increase window size ---------------------------------------------------

void Increase_GfxWindowSize(void)
{
if (gfxmodes[gfx_mode].size>0) {
   gfxmodes[gfx_mode].size-=1;
   Init_GfxConstants();
}
}

//-- Decrease window size ---------------------------------------------------

void Decrease_GfxWindowSize(void)
{
if (gfxmodes[gfx_mode].size<4) {
   ClearBuffer(0);
   WaitRetrace();
   Display_Frame();
   gfxmodes[gfx_mode].size+=1;
   Init_GfxConstants();
}
}

//-- Set gfxwindow size -----------------------------------------------------

void Set_GfxWindowSize(int size)
{
if (gfxmodes[gfx_mode].size<size) {
   ClearBuffer(0);
   WaitRetrace();
   Display_Frame();
}
gfxmodes[gfx_mode].size=size;
Init_GfxConstants();
}

//-- Returns the gfxwindow size ---------------------------------------------

int Get_GfxWindowSize(void)
{
return(gfxmodes[gfx_mode].size);
}

//-- Initialize the gfxmode constants ---------------------------------------

void Init_GfxConstants(void)
{
int i;

xsize=(int)((float)xres*scrsizes[gfxmodes[gfx_mode].size]); xcenter=xsize>>1;
ysize=(int)((float)yres*scrsizes[gfxmodes[gfx_mode].size]); ycenter=ysize>>1;
bytes_per_frame=xsize*ysize;
right_clip=xsize-1;
bottom_clip=ysize-1;
gfxbufsize=xres*yres;
vmemdest=(xres*((yres-ysize)>>1))+((xres-xsize)>>1);
pmulx=xsize*VIEWER_DISTANCE/320;
pmuly=ysize*VIEWER_DISTANCE/240;
pmulx_zclip=pmulx/NEAR_CLIP;
pmuly_zclip=pmuly/NEAR_CLIP;
hfov_delta=(float)xcenter/(VIEWER_DISTANCE*scrsizes[gfxmodes[gfx_mode].size]);
vfov_delta=(float)ycenter*1.20/(VIEWER_DISTANCE*scrsizes[gfxmodes[gfx_mode].size]);

// Init Y offsets table
for (i=0; i<yres; i++) ymultab[i]=xsize*i;

}

//-- Displays the framebuffer -----------------------------------------------

void Display_Frame(void)
{
Display();
}

//-- Saves the framebuffer as a TGA file ------------------------------------

int Save_TGA(void)
{
static int tganum=0;
int i;
char tmpcol,numbuff[8],tgafile[16]="chaos";
tga_header header;
FILE *tga_file, *pal_file;

header.id_length=0;
header.color_map_type=header.image_type=1;
header.first_color_map_entry=0;
header.color_map_length=256;
header.image_width=xsize;
header.image_height=ysize;
header.image_x_origin=0;
header.image_y_origin=0;

header.bits_per_pixel=8;
header.color_map_entry_size=24;
header.image_descriptor_bits=0;

itoa(tganum++, numbuff, 10);
strcat(tgafile, numbuff);
strcat(tgafile, ".tga");
if (NULL==(tga_file=fopen(tgafile,"wb"))) return(1);
pal_file=fopen("chaos.pal", "wb");
fwrite(&header, sizeof(tga_header), 1, tga_file);
for (i=0; i<256; i++) {
   tmpcol=palette[i].b<<2;
   fwrite(&tmpcol, 1, 1, tga_file);
   fwrite(&tmpcol, 1, 1, pal_file);
   tmpcol=palette[i].g<<2;
   fwrite(&tmpcol, 1, 1, tga_file);
   fwrite(&tmpcol, 1, 1, pal_file);
   tmpcol=palette[i].r<<2;
   fwrite(&tmpcol, 1, 1, tga_file);
   fwrite(&tmpcol, 1, 1, pal_file);
}
for (i=ysize-1; i>=0; i--) {
   fwrite(&(gfxbuffer[ymultab[i]]), xsize, 1, tga_file);
}
fclose(tga_file);
fclose(pal_file);
Message("TGA SAVED");
return(NULL);
}

