Unit Graphics;  {Supporting Functions for the new ALPHA engine}

INTERFACE

uses crt,struct,dos;

Const   UseLFB  = $4000;

Type TVbeInfoBlock = Record
     VbeSignature     :Array[1..4] of Char;   { VBE Signature }
     VbeVersion       :Word;                  { VBE Version   }
     OemStringPtr     :DWord;                 { Pointer to OEM String }
     Capabilities     :Array[1..4] of Byte;   { Capabilities of graphics cont. }
     VideoModePtr     :Pointer;               { Pointer to Video Mode List }
     TotalMemory      :Word;                  { Number of 64kb memory blocks }
                                              { Added for VBE 2.0 }
     OemSoftwareRev   :Word;                  { VBE implementation Software revision }
     OemVendorNamePtr :DWord;                 { Pointer to Vendor Name String }
     OemProductNamePtr:DWord;                 { Pointer to Product Name String }
     OemProductRevPtr :DWord;                 { Pointer to Product Revision String }
     _Reserved_       :Array[1..222] of Byte;
     OemData          :Array[1..256] of Byte; { Data Area for OEM Strings }
     End;

Type TModeInfoBlock = Record

     { Mandatory information for all VBE revisions }
     ModeAttributes  : Word;    { mode attributes           }
     WinAAttributes  : Byte;    { window A attributes       }
     WinBAttributes  : Byte;    { window B attributes       }
     WinGranularity  : Word;    { window granularity        }
     WinSize         : Word;    { window size               }
     WinASegment     : Word;    { window A start segment    }
     WinBSegment     : Word;    { window B start segment    }
     WinFuncPtr      : Pointer; { pointer to window function}
     BytesPerScanLine: Word;    { bytes per scan line       }

     { Mandatory information for VBE 1.2 and above }
     XResolution       : Word;  { horizontal resolution in pixels or chars}
     YResolution       : Word;  { vertical resolution in pixels or chars  }
     XCharSize         : Byte;  { character cell width in pixels }
     YCharSize         : Byte;  { character cell height in pixels}
     NumberOfPlanes    : Byte;  { number of memory planes        }
     BitsPerPixel      : Byte;  { bits per pixel                 }
     NumberOfBanks     : Byte;  { number of banks                }
     MemoryModel       : Byte;  { memory model type              }
     BankSize          : Byte;  { bank size in KB                }
     NumberOfImagePages: Byte;  { number of images               }
     _Reserved         : Byte;  { reserved for page function     }

     { Direct Color fields (required for direct/6 and YUV/7 memory models) }
     RedMaskSize        : Byte; { size of direct color red mask in bits}
     RedFieldPosition   : Byte; { bit position of lsb of red mask      }
     GreenMaskSize      : Byte; { size of direct color green mask in bits}
     GreenFieldPosition : Byte; { bit position of lsb of green mask       }
     BlueMaskSize       : Byte; { size of direct color blue mask in bits   }
     BlueFieldPosition  : Byte; { bit position of lsb of blue mask          }
     RsvdMaskSize       : Byte; { size of direct color reserved mask in bits }
     RsvdFieldPosition  : Byte; { bit position of lsb of reserved mask}
     DirectColorModeInfo: Byte; { direct color mode attributes         }

     { Mandatory information for VBE 2.0 and above }
     PhysBasePtr         : DWord;      { physical address for flat frame buffer]}
     OffScreenMemOffset  : DWord;      { pointer to start of off screen memory    }
     OffScreenMemSize    : Word;       { amount of off screen memory in 1k units  }
     __Reserved          : Array[1..206] of Byte; {remainder of ModeInfoBlock}
     End;

Var VbeInfoBlock       : ^TvbeInfoBlock;
    ModeInfoBlock      : ^TModeInfoBlock;

    Real_ES            : Word;
    TotalMem           : Word;
    VBE_LFB            : DWord;

    buffer: array[0..771] of byte;
    scrnptr: dword;
    readptr: dword;
    f: file;
    pixel  : byte;
    len    : byte;
    i: longint;
    _mrx,_mry: integer;


function  FindColor(cr,cg,cb: Integer): Byte;
procedure ClearBuffer;
procedure ShowScreen;
procedure WaitVbl;
procedure Palette(n,r,g,b: byte);
procedure WaitForKey;
procedure LoadScreen(fname: string; p: pointer);
procedure Init320x200;
procedure Init320x400;
procedure Error(s: string; p: string);
procedure Abort(n: longint);
procedure Warning(s: string; p: string);
procedure SetFPUPrecision(Precision: word);
procedure SetFPURounding(RCMode: word);
procedure ShowBuffer;
procedure ShowHiBuf;
procedure EchoBuffer;
procedure CloseScreen;
function  n2s(n: longint): string;
function  s2n(s: string): longint;
function  upper(s: string): string;
function  InitBuffers: boolean;
procedure WriteNum(n,x,y: longint);
procedure LoadPCX(fname: string);
function  InitMouse: Boolean;
function  LeftButton: Boolean;
function  RightButton: Boolean;
procedure MouseMove;
procedure Displine(offset: word);
procedure Status(a,b: longint);
Procedure VesaInit(Length,Width : Word);
procedure ShowVesa;


IMPLEMENTATION


Procedure GetBuffer; assembler;
asm
   mov ax,$0EE02
   int $31
   mov Real_ES,ax
end;

Function VBE_Set_Mode(Mode:WORD): Boolean;
var i: word;
Begin asm
   mov eax,0
   mov ebx,0
   mov ecx,0
   mov edx,0
   mov bx,[mode]
   mov ax,$4F02
   int 10h
   mov [i],ax
end;
VBE_Set_Mode:=(i=$004F);
End;

Function VBE_Set_Linear(VesaMode: Word): Byte;
Var Tmp: DWord;
Begin
   VBE_Set_Linear := 0;
   If not VBE_Set_Mode(VesaMode) then Begin
      VBE_Set_Linear:=1;
      Exit;
   End;
   { Map Video Memory }
   TMP := ModeInfoBlock^.PhysBasePtr;
   asm
      mov ebx,TMP
      mov eax,-1
      mov cx,bx
      shr ebx,16
      mov si,(1024/64)
      xor edi,edi
      mov ax,0800h        { map physical memory}
      int 31h
      jc  @ModeNotGood
      shl ebx,16
      mov bx,cx
      mov eax,ebx         { eax = linear address}
      mov TMP,eax
      @ModeNotGood:
   end;
   VBE_LFB := TMP;
End;

Function VBE_Init_Mode(VesaMode: Word): Byte;
Var Regs: Registers;
Begin
   VBE_Init_Mode:=0;
   Regs.AX:=$4F01;
   Regs.CX:=VesaMode;
   Regs.DI:=0;
   Regs.ES:=Real_ES;
   Intr($10,Regs);
   ModeInfoBlock:=Ptr(Real_ES,0);
   If (Regs.AX<>$004F) then Begin
      VBE_Init_Mode:=1;
      Exit;
   End;
End;

Function VBE_Get_Mode(Width,Height,Bbp : Word) : Word;
Var i : Word;
Begin
VBE_Get_Mode := 0;
For i:=$100 to $200 do Begin
   If VBE_Init_Mode(I)=0 then Begin
      If (ModeInfoBlock^.XResolution=Width)
      and (ModeInfoBlock^.YResolution=Height)
      and (ModeInfoBlock^.BitsPerPixel=bbp)
      then Begin
         VBE_Get_Mode:=I;
         Exit;
      End;
   End;
   End;
End;

Function VBE_Init: Byte;
Var Regs: Registers;
    I,J: Byte;
Begin
   VBE_Init:= 0;
   GetBuffer;
   { Prepare }
   VBEInfoBlock:=Ptr(Real_ES,0);
   With VBEINfoBlock^ do Begin
        VBESignature[1]:='V'; VBESignature[2]:='B'; VBESignature[3]:='E'; VBESignature[4]:='2'; End;
   { Get SuperVGA Information }
   Regs.AX:=$4F00;
   Regs.DI:=0;
   Regs.ES:=Real_ES;
   Intr($10,Regs);
   { Use Information }
   If Regs.AX<>$004F then Begin VBE_Init:=1; exit; end;
   With VBEINfoBlock^ do Begin
      If not ((VBESignature[1]='V') and (VBESignature[2]='E') and (VBESignature[3]='S') and (VBESignature[4]='A')) then Begin
         VBE_Init:=2;
         Exit;
      End;
      If (VBEVersion shr 8<2) then Begin
         VBE_Init:=3;
         Exit;
      End;
      TotalMem:= TotalMemory;
      If TotalMem<(1024/64) then Begin
         VBE_Init:=4;
         Exit;
      End;
   End;
End;

Procedure VesaInit(Length,Width : Word);
Var Mode : Word;
Begin
   Writeln('TMT Vesa Support (c) 1997 Ben Chauveau');
   Case VBE_Init of
     0 : Begin Writeln(' VESA 2.0 detected'); End;
     1 : Begin Writeln('VESA Error : VESA not available (function $00 is not supported) '); Halt; End;
     2 : Begin Writeln('VESA Error : Invalid Block (VESA Signature not found) '); Halt; End;
     3 : Begin Writeln('VESA Error : VESA 2.0 not detected (Try to load UNIVBE.EXE) '); Halt; End;
     4 : Begin Writeln('VESA Error : You Need At Least 1 Mb of Video Memory'); Halt; End;
   End;
   Mode := VBE_Get_Mode(Length,Width,8);
   If Mode=0 then Begin Writeln('VESA Error : ',Length,'*',Width,'*256 not supported'); Halt; End;
   Case VBE_Init_Mode(Mode+USELFB) of
     0 : Begin Writeln(' Mode ',Length,'*',Width,'*','256 detected (=',Mode,')'); End;
     1 : Begin Writeln('VESA Error : Mode ',Length,'*',Width,'*256 with LFB not supported (Try to Load UNIVBE.EXE) '); Halt; End;
   End;
   Delay(100);
   Case VBE_Set_Linear(Mode+USELFB) of
     1 : Begin asm mov ax,3; int 10h; end; Writeln('VESA Error : Cannot initialise Mode ',Length,'*',Width,'*256 with LFB (Try to load UNIVBE.EXE) '); Halt; End;
     2 : Begin asm mov ax,3; int 10h; end; Writeln('VESA Error : Cannot remap memory (INT $31,Fct $0800 Error) '); Halt; End;
   End;
End;

procedure Status(a,b: longint);
begin write(fl,chr(a)+n2s(b)+'| ') end;

procedure Displine(offset: word);
{Sets the display start address to the specified address}
begin
   Port[$3d4]:=$C;
   Port[$3d5]:=offset shr 8;
   Port[$3d4]:=$D;
   Port[$3d5]:=offset and $ff;
end;

procedure LoadPCX(fname: string);
   function loadPCXBuffer(p: pointer):byte;
   begin
      loadPCXBuffer:=buffer[readptr];
      inc(readptr);
      if readptr=256 then
      begin
         blockread(f,p^,256);
         readptr:=0;
      end;
   end;
   procedure drawnextpix(pixel: byte);assembler;
   asm
      mov edi,[tex]          -- offset src
      mov ebx,[scrnptr]
      inc [scrnptr]
      mov [edi+ebx],al
   end;
var x: word;
    p: pointer;
    pal: array[0..256*3] of byte;
begin
   p:=@buffer;
   assign(f,fname);
   reset(f,1);
   seek(f,128);                  -- skip header
   readptr:=255;
   scrnptr:=0;
   pixel:=loadPCXbuffer(@buffer);
   while scrnptr<(256*1024) do begin
      pixel:=loadPCXbuffer(@buffer);
      if (pixel and 192)=192 then
      begin
         len:=(pixel and 63);
         pixel:=loadPCXbuffer(@buffer);
         for x:=1 to len do drawnextpix(pixel);
      end else drawnextpix(pixel);
   end;
   seek(f,filesize(f)-768);
   p:=@buffer;
   blockread(f,p^,768);
   close(f);
   for scrnptr:=0 to 767 do buffer[scrnptr]:=buffer[scrnptr] shr 2;
   for i:=0 to 255 do Palette(i,buffer[i*3],buffer[i*3+01],buffer[i*3+02]);
end;

function InitMouse: Boolean; Assembler;
asm
   mov  ax,0
   int  33h
end;

function LeftButton: Boolean; Assembler;
asm
   mov  ax,3
   int  33h
   mov  ax,0
   test bx,1
   jz   @not_pressed
   mov  ax,$FFFF
@not_pressed:
end;

function RightButton: Boolean; Assembler;
asm
   mov  ax,3
   int  33h
   mov  ax,0
   test bx,2
   jz   @not_pressed
   mov  ax,$FFFF
@not_pressed:
end;

procedure MouseMove;
begin
   asm
      mov ax,$B
      int 33h
      add [_mrx],cx
      add [_mry],dx
   end;
   if _mrx<0 then _mrx:=2879;
   if _mry<0 then _mry:=2879;
   if _mrx>2879 then _mrx:=0;
   if _mry>2879 then _mry:=0;
end;

procedure WriteNum(n,x,y: longint);
var a,ad: dword;
begin
   a:=buf+x+y*320;
   if hires=3 then ad:=640 else ad:=320;
   if n=0 then begin
      memd[a]:=$FAFAFAFA;
      memd[a+ad]:=$FA0000FA;
      memd[a+ad*2]:=$FA0000FA;
      memd[a+ad*3]:=$FA0000FA;
      memd[a+ad*4]:=$FAFAFAFA;
   end;
   if n=1 then begin
      memd[a]     :=$00FA0000;
      memd[a+ad]  :=$00FA0000;
      memd[a+ad*2]:=$00FA0000;
      memd[a+ad*3]:=$00FA0000;
      memd[a+ad*4]:=$00FA0000;
   end;
   if n=2 then begin
      memd[a]     :=$FAFAFAFA;
      memd[a+ad]  :=$FA000000;
      memd[a+ad*2]:=$FAFAFAFA;
      memd[a+ad*3]:=$000000FA;
      memd[a+ad*4]:=$FAFAFAFA;
   end;
   if n=3 then begin
      memd[a]     :=$FAFAFAFA;
      memd[a+ad]  :=$FA000000;
      memd[a+ad*2]:=$FAFAFAFA;
      memd[a+ad*3]:=$FA000000;
      memd[a+ad*4]:=$FAFAFAFA;
   end;
   if n=4 then begin
      memd[a]     :=$FA0000FA;
      memd[a+ad]  :=$FA0000FA;
      memd[a+ad*2]:=$FAFAFAFA;
      memd[a+ad*3]:=$FA000000;
      memd[a+ad*4]:=$FA000000;
   end;
   if n=5 then begin
      memd[a]     :=$FAFAFAFA;
      memd[a+ad]  :=$000000FA;
      memd[a+ad*2]:=$FAFAFAFA;
      memd[a+ad*3]:=$FA000000;
      memd[a+ad*4]:=$FAFAFAFA;
   end;
   if n=6 then begin
      memd[a]     :=$FAFAFAFA;
      memd[a+ad]  :=$000000FA;
      memd[a+ad*2]:=$FAFAFAFA;
      memd[a+ad*3]:=$FA0000FA;
      memd[a+ad*4]:=$FAFAFAFA;
   end;
   if n=7 then begin
      memd[a]     :=$FAFAFAFA;
      memd[a+ad]  :=$FA000000;
      memd[a+ad*2]:=$FA000000;
      memd[a+ad*3]:=$FA000000;
      memd[a+ad*4]:=$FA000000;
   end;
   if n=8 then begin
      memd[a]     :=$FAFAFAFA;
      memd[a+ad]  :=$FA0000FA;
      memd[a+ad*2]:=$FAFAFAFA;
      memd[a+ad*3]:=$FA0000FA;
      memd[a+ad*4]:=$FAFAFAFA;
   end;
   if n=9 then begin
      memd[a]     :=$FAFAFAFA;
      memd[a+ad]  :=$FA0000FA;
      memd[a+ad*2]:=$FAFAFAFA;
      memd[a+ad*3]:=$FA000000;
      memd[a+ad*4]:=$FAFAFAFA;
   end;
end;

procedure InitBuffers;
var err: longint;
begin
   err:=0;
   asm
      {Reserve memory for phong-map}
      mov ax,0ee42h  -- fnk number
      mov edx,10000h -- 64 Kb memory
      int 31h        -- get 64 Kb of memory
      mov [pmap],edx
      adc [err],0
      {Reserve memory for screen buffer}
      mov ax,0ee42h  -- fnk number
      mov edx,41000h -- 128 Kb memory
      int 31h        -- get 256 Kb of memory (max. 640x400)
      mov [buf],edx
      adc [err],0
      {Reserve some texture memory plus mipmaps}
      mov ax,0ee42h  -- fnk number
      mov edx,41000h -- 512 Kb memory
      int 31h        -- get 512 Kb of memory
      mov [tex],edx
      adc [err],0
      {Reserve memory for color mixing table buffer}
      mov ax,0ee42h  -- fnk number
      mov edx,10000h -- 64 Kb memory
      int 31h        -- get 64 Kb of memory
      mov [mix],edx
      adc [err],0
      {Reserve memory for an environment map}
      mov ax,0ee42h  -- fnk number
      mov edx,10000h -- 64 Kb memory
      int 31h        -- get 64 Kb of memory
      mov [env],edx
      adc [err],0
   end;
   if err>0 then Error('Couldn''t allocate memory','InitBuffers');
end;

function FindColor(cr,cg,cb: Integer): Byte;
{Finds the closest match (color number) for a given r/g/b color}
var l,i: Integer;
    n: longint;
begin
   l:=10000;
   for i:=0 to 255 do begin
      n:=(cr-_r[i])*(cr-_r[i])+(cg-_g[i])*(cg-_g[i])+(cb-_b[i])*(cb-_b[i]);
      if n<l then begin
         FindColor:=i;
         l:=n;
      end;
   end;
end;

procedure SetFPUPrecision(Precision: word);
{precision values:  1  -  24 bits (SINGLE)   }
{                   2  -  53 bits (DOUBLE)   }
{                   3  -  64 bits (EXTENDED) }
var CtrlWord: word;
begin {SetPrecisionCtrl}
   if Precision=1 then Precision := 0;
   Precision := Precision shl 8; { make mask for PC field in ctrl word}
   asm
      fstcw    [CtrlWord]        { store NDP control word }
      mov      AX, [CtrlWord]    { load control word into CPU }
      and      AX, 0FCFFh        { mask out precision control field }
      or       AX, [Precision]   { set desired precision in PC field }
      mov      [CtrlWord], AX    { store new control word }
      fldcw    [CtrlWord]        { set new precision control in NDP }
   end;
end;

procedure SetFPURounding(RCMode: word);
{0  -  Round to nearest (default)             }
{1  -  Round down (towards negative infinity) }
{2  -  Round up (towards positive infinity)   }
{3  -  Chop (truncate, round towards zero)    }
var CtrlWord: word;
begin
   RCMode := RCMode shl 10;  { make mask for RC field in control word}
   asm
      fstcw [CtrlWord]       { store NDP control word }
      mov   ax,[CtrlWord]    { load control word into CPU }
      and   ax,0F3FFh        { mask out rounding control field }
      or    ax,[RCMode]      { set desired precision in RC field }
      mov   [CtrlWord],ax    { store new control word }
      fldcw [CtrlWord]       { set new rounding control in NDP }
   end;
end;

function n2s(n: longint): string;
var s,t: string;
begin
   s:='';
   if n<0 then t:='-' else t:='';
   while abs(n)>0 do begin
      s:=chr(ord('0')+abs(n) mod 10)+s;
      n:=n div 10;
   end;
   if s='' then s:='0';
   s:=t+s;
   n2s:=s;
end;

function upper(s: string): string;
var i: longint;
begin
   for i:=1 to length(s) do if (s[i]>='a') and (s[i]<='z') then s[i]:=chr(ord(s[i])-32);
   upper:=s;
end;

function s2n(s: string): longint;
var t,i,ng: longint;
begin
   t:=0;
   ng:=1;
   for i:=1 to length(s) do begin
      if (s[i]>='0') and (s[i]<='9') then t:=10*t+(ord(s[i])-ord('0'));
      if s[i]='-' then ng:=-1;
   end;
   s2n:=t*ng;
end;

procedure ClearBuffer; Assembler;
asm {Clears array 'buf'}
   align 4
   mov edi,[buf]
   mov ecx,BUFSIZE/4
   mov eax,0
   rep stosd
end;

procedure ShowBuffer; Assembler;
asm {Dumps array 'buf' to the screen}
   mov esi,[buf]
   mov edi,[scrn]
   mov ecx,BUFSIZE/4
   rep movsd
end;

procedure ShowVesa; Assembler;
asm {Dumps array 'buf' to the screen}
   mov esi,[buf]
   mov edi,VBE_LFB
   mov ecx,BUFSIZE/4
   rep movsd
end;

procedure EchoBuffer; Assembler;
asm
   mov edi,[buf]
   mov ecx,64000
   xor eax,eax
   xor ebx,ebx
@ploop:
   mov al,byte ptr [edi]
   mov bl,byte ptr [edi+1]
   shr ebx,1
   add eax,ebx
   shr eax,1
   mov byte ptr [edi],al
   inc edi
   dec ecx
   jnz @ploop
end;

procedure WaitVbl; assembler;
asm
   cli
   nop
   mov dx,3DAh
@1: in al,dx
   and al,08h
   jnz @1
@2: in al,dx
   and al,08h
   jz @2
   sti
end;

procedure Palette(n,r,g,b: byte);
begin
   _r[n]:=r;
   _g[n]:=g;
   _b[n]:=b;
   Port[$3c8]:=n;
   Port[$3C9]:=r;
   Port[$3c9]:=g;
   Port[$3c9]:=b
end;

procedure WaitForKey;
var a: char;
begin
   while keypressed do a:=readkey;
   repeat until keypressed;
   while keypressed do a:=readkey
end;

procedure LoadScreen(fname: string; p: pointer);
var f: file;
begin
   Assign(f,fname);
   Reset(f,1);
   BlockRead(f,p^,9);
   BlockRead(f,p^,64000+256*3);
   Close(f)
end;

procedure ShowScreen;
var a: dword;
begin
   for a:=0 to 63999 do mem[scrn+a]:=mem[buf+a];
end;

procedure Init320x200;
var i,b: longint;
begin
   asm mov ax,13h; int 10h end;
   scrn:=_zero+$A0000;
end;

procedure Init320x400; Assembler;
{initializes a 320x400 screen}
asm
   mov edi,$a0000
   add edi,_zero
   mov [scrn],edi
   mov ax,$13
   int 10h
   mov dx,$3C4
   mov ax,$604
   out dx,ax
   mov ax,$F02
   out dx,ax
   mov dx,$3D4
   mov ax,$14
   out dx,ax
   mov ax,$E317
   out dx,ax
   mov ax,$4009
   out dx,ax
   {Clear buffer}
   mov edx,$3c4
   mov eax,$F02  {All four planes at once}
   out dx,ax
   mov ecx,8192
   xor eax,eax
   rep stosd
end;

procedure ShowHiBuf; Assembler;
{displays one memory bank on-screen}
var x,f: longint;
asm
   mov [f],4
@floop:
   mov edx,$3c4
   mov eax,$102
   mov ecx,[f]
   dec ecx
   shl ah,cl
   out dx,ax
   xor ebx,ebx
   mov edx,[scrn]
@xloop:
   mov edi,edx
   mov ecx,200
   xor esi,esi
   mov esi,ebx
   shl esi,3
   add esi,[f]
   dec esi
   add esi,[buf]
@yloop:
   mov al,[esi]
   mov ah,[esi+4]
   mov [edi],ax
   mov al,[esi+320]
   mov ah,[esi+324]
   mov [edi+80],ax
   add edi,160
   add esi,640
   dec ecx
   jnz @yloop
   add edx,2
   inc ebx
   cmp ebx,40
   jb  @xloop
   dec [f]
   jnz @floop
end;

procedure CloseScreen;
begin asm mov ax,3; int 10h end end;

procedure Error(s: string; p: string);
begin
   CloseScreen;
   Writeln('Error: '+s+' (in '+p+')');
   Halt;
end;

procedure Abort(n: longint);
begin
   CloseScreen;
   Writeln('Program halted...');
   Writeln('Data: '+n2s(n));
   Halt;
end;

procedure Warning(s: string; p: string);
begin
   inc(_warn);
   if _warn<11 then wrn[_warn]:='Warning: '+s+' (in '+p+')';
end;

end.
