#ifndef  _LINE_H
#define  _LINE_H


#ifndef  _BALDEF_H
#include "BalDef.h"
#endif

#ifndef  _POINT_H
#include "Point.h"
#endif

#ifndef  _VECTOR_H
#include "Vector.h"
#endif




class BMap;
class BRengine;
class BPoint;
class BSector;
class BTexture;




class BLine {
public:
    unsigned int     m_timestamp;
    const int        m_index;
    const BPoint   * m_start;
    const BPoint   * m_end;
    const BVector    m_vector;
    
    
    struct {
        const BSector  * m_sector;    // Which sector.
        const BTexture * m_utext;     // Upper texture number.
        const BTexture * m_mtext;     // Middle texture number.
        const BTexture * m_ltext;     // Lower texture number.
        BREAL            m_xshift;    // Horitontal texture offset.
        BREAL            m_yshift;    // Vertical texture offset.
    } m_face [2];



public:
    BLine (int index, const BPoint * start, const BPoint * end);



    inline BREAL Length () const;


    inline void CrossTest  (
        const BPoint  & point,
        const BVector & vector,
        BREAL   iproduct,
        BREAL & alpha, 
        BREAL & beta) const;


    const BSector * ThruTest (
        BFACE face, 
        BREAL & top, 
        BREAL & bottom) const;
};






inline BREAL BLine::Length () const {
    return BSQRT ((m_start->x-m_end->x) * (m_start->x-m_end->x) +
                  (m_start->y-m_end->y) * (m_start->y-m_end->y));
}
        




inline void BLine::CrossTest (
    const BPoint  & point,
    const BVector & vector,
    BREAL   iproduct,
    BREAL & alpha, 
    BREAL & beta) const {
    
    const BREAL vx = vector.x;
    const BREAL vy = vector.y;
    const BREAL ex = point.x;
    const BREAL ey = point.y;
    const BREAL px = m_start->x;
    const BREAL py = m_start->y;

    alpha = (vx*(py-ey) - vy*(px-ex)) / iproduct;
    beta  = ((py-ey) * m_vector.x - (px-ex) * m_vector.y) / iproduct;
}



#endif
