/*---------------------------------------------------------------------------
                        THE CHAOS ENGINE - MATH.C

                            3D math routines
---------------------------------------------------------------------------*/



#include <stdlib.h>
#include <math.h>

#include "3dengine.h"
#include "math.h"

//-- External variables -----------------------------------------------------

extern object_3d *objects_p[MAX_OBJECTS];
extern int objects_in_list;
extern dir view_dir;
extern vertice_3d view_point;

//-- Local variables --------------------------------------------------------

float camera_matrix[3][3];
float sinus[SCMASK+1], cosinus[SCMASK+1];


//-- Init Math tables etc... ------------------------------------------------

void Init_Math(void)
{
int i;
for (i=0; i<=SCMASK; i++) {
   sinus[i]=sin(i*TOSC);
//   cprintf("Sin(%d)=%f\n\r",i,sinus[i]);
//   while(!KeyHit());
   cosinus[i]=cos(i*TOSC);
}
puts("Lockup tables generated...");
}

//-- Sinus ------------------------------------------------------------------

float Sin(int angle)
{
return(sinus[angle]);
}

//-- Cosinus ----------------------------------------------------------------

float Cos(int angle)
{
return(cosinus[angle]);
}

/*---------------------------------------------------------------------------
                            Rotate_Objects_3d
---------------------------------------------------------------------------*/

void Rotate_Objects_3d(void)
{
int i,o,xyz;
float xyz_matrix[3][3];
float xsin, xcos, ysin, ycos, zsin, zcos;
float sxsz, sxcz, szcx, cxcz;
dir rot;
vertice_3d *vp, tmp;
object_3d *obj;

// Loop through every object in list and rotate it.

for (o=0; o<objects_in_list; o++) {
   obj=objects_p[o];
   if ((obj->flags & OBJ_VISIBLE) && (obj->flags & OBJ_ON)) {
      vp=obj->vertices_p;

      rot.x=obj->angle.x - obj->oldangle.x;
      rot.y=obj->angle.y - obj->oldangle.y;
      rot.z=obj->angle.z - obj->oldangle.z;

      obj->oldangle=obj->angle;

      if (rot.x) xyz=1; else xyz=0;
      if (rot.y) xyz+=2;
      if (rot.z) xyz+=4;

      rot.x &= SCMASK;
      rot.y &= SCMASK;
      rot.z &= SCMASK;

      // Do X rotation only
      if (xyz==1) {
         xsin=Sin(rot.x);
         xcos=Cos(rot.x);
         for (i=0; i<(obj->vertices+obj->normals); i++) {
            ysin = vp[i].y * xcos - vp[i].z * xsin;
            vp[i].z = vp[i].y * xsin + vp[i].z * xcos;
            vp[i].y = ysin;
         }
         continue;
      }

      // Do Y rotation only
      if (xyz==2) {
         ysin=Sin(rot.y);
         ycos=Cos(rot.y);
         for (i=0; i<(obj->vertices+obj->normals); i++) {
            xsin = vp[i].z * ysin + vp[i].x * ycos;
            vp[i].z = vp[i].z * ycos - vp[i].x * ysin;
            vp[i].x = xsin;
         }
         continue;
      }

      // Do Z rotation only
      if (xyz==4) {
         zsin=Sin(rot.z);
         zcos=Cos(rot.z);
         for (i=0; i<(obj->vertices+obj->normals); i++) {
            xsin = vp[i].x * zcos - vp[i].y * zsin;
            vp[i].y = vp[i].x * zsin + vp[i].y * zcos;
            vp[i].x = xsin;
         }
         continue;
      }

      // Do xyz rotation
      if (xyz) {
         // Get SIN/COS values

         xsin=Sin(rot.x);
         xcos=Cos(rot.x);
         ysin=Sin(rot.y);
         ycos=Cos(rot.y);
         zsin=Sin(rot.z);
         zcos=Cos(rot.z);

         // Calculate rotation-matrix

         sxsz=xsin * zsin;
         sxcz=xsin * zcos;
         szcx=zsin * xcos;
         cxcz=xcos * zcos;

         xyz_matrix[0][0]=ycos * zcos;
         xyz_matrix[1][0]=ycos * zsin;
         xyz_matrix[2][0]=-ysin;

         xyz_matrix[0][1]=ysin * sxcz - szcx;
         xyz_matrix[1][1]=ysin * sxsz + cxcz;
         xyz_matrix[2][1]=xsin * ycos;

         xyz_matrix[0][2]=ysin * cxcz + sxsz;
         xyz_matrix[1][2]=ysin * szcx - sxcz;
         xyz_matrix[2][2]=xcos * ycos;

         // Rotate all vertices

         for (i=0; i<(obj->vertices+obj->normals); i++) {
            tmp.x = (vp[i].x * xyz_matrix[0][0]) + (vp[i].y * xyz_matrix[0][1]) + (vp[i].z * xyz_matrix[0][2]);
            tmp.y = (vp[i].x * xyz_matrix[1][0]) + (vp[i].y * xyz_matrix[1][1]) + (vp[i].z * xyz_matrix[1][2]);
            tmp.z = (vp[i].x * xyz_matrix[2][0]) + (vp[i].y * xyz_matrix[2][1]) + (vp[i].z * xyz_matrix[2][2]);
            vp[i]=tmp;
         }

      }
   }
}
}

/*---------------------------------------------------------------------------
                       Local_2_World transformation
---------------------------------------------------------------------------*/

void Local_2_World(void)
{
int i,o;
object_3d *obj;
for (o=0; o<objects_in_list; o++) {
   obj=objects_p[o];
   if ((obj->flags & OBJ_VISIBLE) && (obj->flags & OBJ_ON)) {
      for (i=0; i<obj->vertices; i++) {
         obj->transformed_p[i].x=obj->vertices_p[i].x+obj->world.x;
         obj->transformed_p[i].y=obj->vertices_p[i].y+obj->world.y;
         obj->transformed_p[i].z=obj->vertices_p[i].z+obj->world.z;
      }
   }
}
}

/*---------------------------------------------------------------------------
                    Make the camera transformation matrix
---------------------------------------------------------------------------*/

void Make_Camera_Matrix(void)
{
float xsin, xcos, ysin, ycos, zsin, zcos;
float sxsz, sxcz, szcx, cxcz;
dir angle;

/* Updatera vinklarna */

angle.x=(-view_dir.x) & SCMASK;
angle.y=(-view_dir.y) & SCMASK;
angle.z=(-view_dir.z) & SCMASK;

/* Hmta SIN/COS vrden */

xsin=Sin(angle.x);
xcos=Cos(angle.x);
ysin=Sin(angle.y);
ycos=Cos(angle.y);
zsin=Sin(angle.z);
zcos=Cos(angle.z);

/* Berkna rotations-matrisen */

sxsz=xsin * zsin;
sxcz=xsin * zcos;
szcx=zsin * xcos;
cxcz=xcos * zcos;

camera_matrix[0][0]=ycos * zcos;
camera_matrix[1][0]=ycos * zsin;
camera_matrix[2][0]=-ysin;

camera_matrix[0][1]=ysin * sxcz - szcx;
camera_matrix[1][1]=ysin * sxsz + cxcz;
camera_matrix[2][1]=xsin * ycos;

camera_matrix[0][2]=ysin * cxcz + sxsz;
camera_matrix[1][2]=ysin * szcx - sxcz;
camera_matrix[2][2]=xcos * ycos;

}

/*---------------------------------------------------------------------------
                       World_2_Camers transformation
---------------------------------------------------------------------------*/

void World_2_Camera(void)
{
int i,o;
vertice_3d *tp, tmp;
object_3d *obj;

for (o=0; o<objects_in_list; o++) {
   obj=objects_p[o];
   if ((obj->flags & OBJ_VISIBLE) && (obj->flags & OBJ_ON)) {

      tp=obj->transformed_p;

      for (i=0; i<obj->vertices; i++) {
         tp[i].x-=view_point.x;
         tp[i].y-=view_point.y;
         tp[i].z-=view_point.z;

         tmp.x = (tp[i].x * camera_matrix[0][0]) + (tp[i].y * camera_matrix[0][1])
            + (tp[i].z * camera_matrix[0][2]);
         tmp.y = (tp[i].x * camera_matrix[1][0]) + (tp[i].y * camera_matrix[1][1])
            + (tp[i].z * camera_matrix[1][2]);
         tmp.z = (tp[i].x * camera_matrix[2][0]) + (tp[i].y * camera_matrix[2][1])
            + (tp[i].z * camera_matrix[2][2]);
         obj->transformed_p[i]=tmp;

      }
   }
}
}

//-- Dot Product ------------------------------------------------------------

float Dot_Product(vector_3d *u, vector_3d *v)
{
return((u->x * v->x) + (u->y * v->y) + (u->z * v->z));
}

//-- Cross Product ----------------------------------------------------------

void Cross_Product(vector_3d *u, vector_3d *v, vector_3d *normal)
{
normal->x=(u->y * v->z) - (u->z * v->y);
normal->y=(u->z * v->x) - (u->x * v->z);
normal->z=(u->x * v->y) - (u->y * v->x);
}

//-- Vector Magnitude -------------------------------------------------------

float Vector_Magnitude(vector_3d *u)
{
return(sqrt(u->x * u->x + u->y * u->y + u->z * u->z));
}


//-- Vector Distance --------------------------------------------------------

float Vertice_Distance(vertice_3d *u, vertice_3d *v)
{
double tmp;
tmp=(u->x-v->x)*(u->x-v->x) + (u->y-v->y)*(u->y-v->y) + (u->z-v->z)*(u->z-v->z);
return((float)sqrt(tmp));
}

//-- Normalize Vector -------------------------------------------------------

void Normalize_Vector(vector_3d *u)
{
float length;
length=Vector_Magnitude(u);
if (length!=NULL) {
   u->x/=length;
   u->y/=length;
   u->z/=length;
}
}

