;Ŀ
;                       -- Misc Graphics Routines --                        
;                                                                           
;


        .486P
        .MODEL FLAT
	.DATA

        EXTRN _xres : DWORD, _bytes_per_frame : DWORD, _backfill : DWORD
        EXTRN _edge1 : DWORD, _edge2 : DWORD, _xsize : DWORD, _ysize : DWORD
        EXTRN _bottom_clip : DWORD, _right_clip : DWORD
        EXTRN _gfxbuffer : DWORD, _gfxbufsize : DWORD, _vmemdest : DWORD

        PUBLIC SetTweakedMode_, TextMode_, WaitRetrace_, SetPalette_, SetColor_
        PUBLIC DisplayBuffer_Tweaked_

SC_INDEX          EQU   03c4h   ;Sequence Controller Index
SC_DATA           EQU   03c5h   ;Sequence kontroller data
CRTC_INDEX        EQU   03d4h   ;CRT Controller Index
MISC_OUTPUT       EQU   03c2h   ;Miscellaneous Output register

MAP_MASK          EQU   2

yloop             dd    ?


X320Y240 label word
    db      0e3h    ; dot clock
    db      10      ; Number of CRTC Registers to update
    dw      00d06h  ; vertical total
    dw      03e07h  ; overflow (bit 8 of vertical counts)
    dw      04109h  ; cell height (2 to double-scan)
    dw      0ea10h  ; v sync start
    dw      0ac11h  ; v sync end and protect cr0-cr7
    dw      0df12h  ; vertical displayed
    dw      00014h  ; turn off dword mode
    dw      0e715h  ; v blank start
    dw      00616h  ; v blank end
    dw      0e317h  ; turn on byte mode
    dw      320     ; width
    dw      240     ; height

X320Y400 label word
    db      0e3h    ; dot clock
    db      03      ; Number of CRTC Registers to update
    dw      04009h  ; cell height
    dw      00014h  ; turn off dword mode
    dw      0e317h  ; turn on byte mode
    dw      320     ; width
    dw      400     ; height

ModeTable label dword    ; Mode X tweak table
    dd      offset X320Y240
    dd      offset X320Y400

        .CODE

;----------------------------------------------------------------------------
;                        Tweaked mode graphic routines
;----------------------------------------------------------------------------

SetTweakedMode_ PROC NEAR
        pushad

        push  eax

        mov   ax, 13h              ; let the BIOS set standard 256-color
        int   10h                  ; mode (320x200 linear)
    
        mov    dx, 3dah
@WaitRet:
        in     al, dx
        test   al, 8
        jz     @WaitRet


        mov   dx, SC_INDEX
        mov   ax, 2101h
        out   dx, ax
        mov   ax, 0604h
        out   dx, ax               ; disable chain4 mode
        mov   ax, 0100h
        out   dx, ax               ; synchronous reset while setting Misc
                                   ; Output for safety, even though clock
                                   ; unchanged
        pop   ebx
        add   ebx, ebx
        add   ebx, ebx
        mov   esi, [ebx + ModeTable]
        lodsb

        or    al, al
        jz    @DontSetDot
        mov   dx, MISC_OUTPUT
        out   dx, al               ; select the dot clock and Horiz scanning rate
@DontSetDot:
        mov   dx, SC_INDEX
        mov   ax, 0300h
        out   dx, ax               ; undo reset (restart sequencer)

        mov   edx, CRTC_INDEX      ; reprogram the CRT Controller
        mov   al, 11h              ; VSync End reg contains register write
        out   dx, al               ; protect bit
        inc   edx                  ; CRT Controller Data register
        in    al, dx               ; get current VSync End register setting
        and   al, 07fh             ; remove write protect on various
        out   dx, al               ; CRTC registers
        dec   edx                  ; CRT Controller Index
        cld
        xor   ecx, ecx
        lodsb
        mov   cl,al

@SetCRTParmsLoop:
        lodsw                       ; get the next CRT Index/Data pair
        out   dx,ax               ; set the next CRT Index/Data pair
        dec   cl
        jnz   @SetCRTParmsLoop

        mov   dx, SC_INDEX
        mov   ax, 0f02h
        out   dx, ax               ; enable writes to all four planes

        xor   eax, eax             ; now clear all display memory, 8 pixels
        mov   edi, 0a0000h         ; at a time
        mov   ecx, 16000           ; # of dwords in display memory
        rep   stosd               ; clear all of display memory

    ;  Mode X is set, now set the required logical page width.

        mov   ecx, [_xres]
        shr   ecx, 3                ; divide by 8
        mov   dx, CRTC_INDEX
        mov   al, 13h
        mov   ah, cl
        out   dx, ax

        mov    dx, 3dah
@WaitRet2:
        in     al, dx
        test   al, 8
        jz     @WaitRet2

        mov    dx, SC_INDEX
        mov    ax, 0101h
        out    dx, ax

 

        popad
        ret
SetTweakedMode_ ENDP


DisplayBuffer_Tweaked_ PROC NEAR
        pushad

        mov    edi, 0a0000h
        mov    esi, [_gfxbuffer]
        mov    eax, [_vmemdest]
        shr    eax, 2
        add    edi, eax

        mov    ebx, [_xsize]
        shr    ebx, 4
        mov    eax, [_ysize]
        mov    [yloop], eax
@YLoop:
        mov    ecx, 3
@PLoop:
        mov    dx, SC_INDEX                 ; Turn on all bits
        mov    al, MAP_MASK
        mov    ah, 1
        shl    ah, cl
        out    dx, ax
        mov    ebp, esi
        add    esi, ecx

        mov    bh, bl
        mov    edx, edi
@XLoop:
        mov    ah, [esi+12]
        mov    al, [esi+8]
        shl    eax, 16
        mov    ah, [esi+4]
        mov    al, [esi]
        add    esi, 16

        mov    [edi], eax
        add    edi, 4
        dec    bh

        jnz    @XLoop

        mov    edi, edx
        mov    esi, ebp

        dec    ecx
        jge    @PLoop

        add    esi, [_xsize]
        add    edi, 80

        dec    [yloop]
        jnz    @YLoop

        mov    eax, [_backfill]
        mov    edi, [_gfxbuffer]
        mov    ecx, [_gfxbufsize]
        shr    ecx, 2
        rep    stosd


        popad
        ret
DisplayBuffer_Tweaked_ ENDP



TextMode_ PROC NEAR
        push  eax

        mov   ax, 3h
        int   10h

        pop   eax
        ret
TextMode_ ENDP


SetColor_ PROC
        push  ebx
        mov   ebx, edx
        mov   dx, 03c8h
        out   dx, al
        inc   edx

        rol   ebx, 24
        mov   al, bh
        out   dx, al
        mov   al, bl
        out   dx, al
        shr   ebx, 16
        mov   al, bh
        out   dx, al
        pop   ebx
        ret
SetColor_ ENDP


SetPalette_ PROC NEAR
        push  ecx
        push  edx
        push  esi
        mov   esi, eax

        xor   al, al
        mov   dx, 3c8h                      ; Address till PEL AWM
        out   dx, al
        inc   dx
        mov   ecx, 768
        rep   outsb                         ; Stt hela paletten

        pop   esi
        pop   edx
        pop   ecx
        ret
SetPalette_ ENDP


WaitRetrace_ PROC NEAR
        push  eax
        push  edx
        mov   dx, 03dah
Wait1:
        in    al, dx
        test  al, 8
        jnz   Wait1
Wait2:
        in    al, dx
        test  al, 8
        jz    Wait2
        pop   edx
        pop   eax
        ret
WaitRetrace_ ENDP


END


