/*---------------------------------------------------------------------------
                        THE CHAOS ENGINE - POLY.C

                         Polygon drawing routines
---------------------------------------------------------------------------*/

#include <stdio.h>

#include "gfx.h"
#include "3dengine.h"
#include "linear.h"
#include "poly.h"

#define TEX_SIZE 63

//-- External variables -----------------------------------------------------

extern int bottom_clip, right_clip, ymultab[400];
extern int res,yres,xsize,ysize,xcenter,ycenter;
extern float pmulx,pmuly,pmulx_zclip,pmuly_zclip;

extern char *gfxbuffer, **textures_pp;

//-- Local variables --------------------------------------------------------

int edge1[400], edge2[400], color1[400], color2[400], *edge_p, *color_p;
pix tex1[400], tex2[400], *tex_p;
int ymin, ymax;

pix tri_mapping[3]={0,0, 0,TEX_SIZE, TEX_SIZE,TEX_SIZE};
pix quad_mapping[4]={0,0, 0,TEX_SIZE, TEX_SIZE,TEX_SIZE, TEX_SIZE,0};
pix *poly_cornermappings[5];

//-- Pragmas ----------------------------------------------------------------

void GLine(char *ptr, int xd, int startc, int ck);

#pragma aux GLine="@GLoop:              "\
                  "   inc    edi        "\
                  "   add    ebx, eax   "\
                  "   dec    ecx        "\
                  "   mov    [edi], bh  "\
                  "   jnz    @GLoop     "\
                  parm[edi] [ecx] [ebx] [eax]\
                  modify[edi ecx ebx eax];

void TLine(char *scrptr, char *texptr, int xd, int xs, int ys, int xk, int yk);

#pragma aux TLine="@TLoop:              "\
                  "   inc    edi        "\
                  "   add    ebx, eax   "\
                  "   dec    ecx        "\
                  "   mov    [edi], bh  "\
                  "   jnz    @TLoop     "\
                  parm[edi] [esi] [ebp] [eax] [ebx] [ecx] [edx]\
                  modify[edi ecx ebx eax];



//-- Init poly drawing functions --------------------------------------------


void Init_Poly(void)
{
poly_cornermappings[0]=&(tri_mapping[0]);
poly_cornermappings[1]=&(quad_mapping[0]);
}

/*---------------------------------------------------------------------------
           Triangle drawing functions - No Z-clipping supported
---------------------------------------------------------------------------*/


void Draw_Gouraud_Triangle(poly_3d *poly_p)
{
int i,indx,y1,y2,min,mid,max,l,r,k1,k2;
object_3d *obj;
gpix corners[3];

ymin=MAXLONG; ymax=-MAXLONG;
obj=(object_3d *)(poly_p->homeobj_p);

indx=poly_p->pindxs[0];
corners[0].x=obj->projected_p[indx].x;
corners[0].y=obj->projected_p[indx].y;
corners[0].c=poly_p->color+obj->vertex_colors_p[poly_p->pindxs[0]];

indx=poly_p->pindxs[1];
corners[1].x=obj->projected_p[indx].x;
corners[1].y=obj->projected_p[indx].y;
corners[1].c=poly_p->color+obj->vertex_colors_p[poly_p->pindxs[1]];

indx=poly_p->pindxs[2];
corners[2].x=obj->projected_p[indx].x;
corners[2].y=obj->projected_p[indx].y;
corners[2].c=poly_p->color+obj->vertex_colors_p[poly_p->pindxs[2]];

// Find top point
if (corners[0].y<corners[1].y) min=0; else min=1;
if (corners[2].y<corners[min].y) min=2;
// Find bottom
if (corners[0].y>corners[1].y) max=0; else max=1;
if (corners[2].y>corners[max].y) max=2;

mid=3-min-max;

// Scan longest edge
edge_p=&edge1; color_p=&color1;
ScanTriEdge_G(corners[min].x, corners[min].y, corners[min].c, corners[max].x, corners[max].y, corners[max].c);

// Scan the other two edges
edge_p=&edge2; color_p=&color2;
ScanTriEdge_G(corners[min].x, corners[min].y, corners[min].c, corners[mid].x, corners[mid].y, corners[mid].c);
ScanTriEdge_G(corners[mid].x, corners[mid].y, corners[mid].c, corners[max].x, corners[max].y, corners[max].c);

if ((ymin<=bottom_clip) & (ymax>=NULL)) DrawHLines_Gris(ymin, ymax+1);

}


void Draw_Texture_Triangle(poly_3d *poly_p)
{
int i,indx,min,mid,max;
object_3d *obj;
tpix corners[3];

ymin=MAXLONG; ymax=-MAXLONG;
obj=(object_3d *)(poly_p->homeobj_p);

indx=poly_p->pindxs[0];
corners[0].x=obj->projected_p[indx].x;
corners[0].y=obj->projected_p[indx].y;
corners[0].tx=tri_mapping[0].x;
corners[0].ty=tri_mapping[0].y;

indx=poly_p->pindxs[1];
corners[1].x=obj->projected_p[indx].x;
corners[1].y=obj->projected_p[indx].y;
corners[1].tx=tri_mapping[1].x;
corners[1].ty=tri_mapping[1].y;



indx=poly_p->pindxs[2];
corners[2].x=obj->projected_p[indx].x;
corners[2].y=obj->projected_p[indx].y;
corners[2].tx=tri_mapping[2].x;
corners[2].ty=tri_mapping[2].y;


// Find top point
if (corners[0].y<corners[1].y) min=0; else min=1;
if (corners[2].y<corners[min].y) min=2;
// Find bottom
if (corners[0].y>corners[1].y) max=0; else max=1;
if (corners[2].y>corners[max].y) max=2;

mid=3-min-max;

// Scan longest edge
edge_p=&edge1; color_p=&color1; tex_p=&tex1;
ScanTriEdge_T(corners[min].x, corners[min].y, corners[min].tx, corners[min].ty,
   corners[max].x, corners[max].y, corners[max].tx, corners[max].ty);

// Scan the other two edges
edge_p=&edge2; color_p=&color2; tex_p=&tex2;
ScanTriEdge_T(corners[min].x, corners[min].y, corners[min].tx, corners[min].ty,
   corners[mid].x, corners[mid].y, corners[mid].tx, corners[mid].ty);
ScanTriEdge_T(corners[mid].x, corners[mid].y, corners[mid].tx, corners[mid].ty,
   corners[max].x, corners[max].y, corners[max].tx, corners[max].ty);

if ((ymin<=bottom_clip) & (ymax>=NULL)) DrawHLines_T(ymin, ymax+1, textures_pp[poly_p->tmapindx]);
}

/*---------------------------------------------------------------------------
                     Polygon functions with Z-clipping
---------------------------------------------------------------------------*/

void Draw_Solid_Polygon(poly_3d *poly_p)
{
int i,p1,p2,p,first,pts,indx;
float x,y,zdelta,xk,yk;
object_3d *obj;
short *i_p;
vertice_3d *t_p;
pix corners[9];

ymin=MAXLONG; ymax=-MAXLONG;
pts=poly_p->points;
obj=(object_3d *)(poly_p->homeobj_p);
i_p=poly_p->pindxs;

// Do we have to zclip the polygon?

if (!(poly_p->flags & POLY_CLIPPED)) {       // Nop
   for (p=0; p<pts; p++) {
      corners[p].x=obj->projected_p[i_p[p]].x;
      corners[p].y=obj->projected_p[i_p[p]].y;
   }
   pts--;
}
else {                                       // Yes we do.
   t_p=obj->transformed_p;
   poly_p->flags&=~POLY_CLIPPED;

   // Get first point on the right side of near clipping plane.
   p=p1=0;
   while (t_p[i_p[p1]].z<=NEAR_CLIP) {
      p1++;
      if (p1>8) {
//         Display_Error("CLIP ERROR : CAN'T FIND VALID POINT!");
         return;
      }
   }
   first=p1;

   // Add points until we meet one behind NEAR_CLIP
   do {
      corners[p++]=obj->projected_p[i_p[p1]];
      p2=p1;
      p1++;
      if (p1==pts) p1=0;
      if (p>9) {
//         Display_Error("CLIP ERROR : CAN'T FIND INVALID POINT.");
         return;
      }
 
   } while (t_p[i_p[p1]].z>NEAR_CLIP);

   // Now p1 and p2 are on different sides of NEAR_CLIP thus we calculate
   // were the line p1-p2 intersects the near clipping plane.


   zdelta=(NEAR_CLIP - t_p[i_p[p1]].z) / (t_p[i_p[p2]].z - t_p[i_p[p1]].z);
   x=t_p[i_p[p1]].x + ((t_p[i_p[p2]].x - t_p[i_p[p1]].x) * zdelta);
   y=t_p[i_p[p1]].y + ((t_p[i_p[p2]].y - t_p[i_p[p1]].y) * zdelta);
   corners[p].x=xcenter+(int)(x*pmulx_zclip);
   corners[p].y=ycenter-(int)(y*pmuly_zclip);
   p++;

   // Skip points until we are on the right side of clipping plane.
   do {
      p2=p1;
      p1++;
      if (p1==pts) p1=0;
      if (p1==(first+1)) {
//         Display_Error("CLIP ERROR : CAN'T GET OUT OF SKIPPING LOOP.");
         return;
      }
 
   } while (t_p[i_p[p1]].z<=NEAR_CLIP);


   zdelta=(NEAR_CLIP - t_p[i_p[p1]].z) / (t_p[i_p[p2]].z - t_p[i_p[p1]].z);

   x=t_p[i_p[p1]].x + ((t_p[i_p[p2]].x - t_p[i_p[p1]].x) * zdelta);
   y=t_p[i_p[p1]].y + ((t_p[i_p[p2]].y - t_p[i_p[p1]].y) * zdelta);
   corners[p].x=xcenter+(int)(x*pmulx_zclip);
   corners[p].y=ycenter-(int)(y*pmuly_zclip);
   p++;


   // Now add points until we meet the point we started with.
   while (p1!=first) {
      corners[p++]=obj->projected_p[i_p[p1]];
      p1++;
      if (p1==pts) p1=0;
      if (p>8) {
//         Display_Error("CLIP ERROR : P TOO BIG IN FIND-START LOOP.");
         return;
      }
   }
   pts=p-1;
}
for (i=0; i<pts; i++)
   ScanEdge_S(corners[i].x, corners[i].y, corners[i+1].x, corners[i+1].y);
ScanEdge_S(corners[pts].x, corners[pts].y, corners[0].x, corners[0].y);
if ((ymin<=bottom_clip) & (ymax>=NULL)) DrawHLines_S(poly_p->shade, ymin, ymax+1);
}

//-- Draws a gouraud shaded polygon -----------------------------------------

void Draw_Gouraud_Polygon(poly_3d *poly_p)
{
int i,p,p1,p2,first,pts,indx;
float x,y,zdelta;
short *i_p;
vertice_3d *t_p;
object_3d *obj;
gpix corners[9];

ymin=MAXLONG; ymax=-MAXLONG;
pts=poly_p->points;
obj=(object_3d *)(poly_p->homeobj_p);
i_p=poly_p->pindxs;

if (!(poly_p->flags & POLY_CLIPPED)) {
   for (p=0; p<pts; p++) {
      corners[p].x=obj->projected_p[i_p[p]].x;
      corners[p].y=obj->projected_p[i_p[p]].y;
      corners[p].c=poly_p->color+obj->vertex_colors_p[poly_p->pindxs[p]];

   }
   pts--;
}
else {
   t_p=obj->transformed_p;
   poly_p->flags&=~POLY_CLIPPED;

   // Get first point on the right side of near clipping plane.
   p=p1=0;
   while (t_p[i_p[p1]].z<=NEAR_CLIP) p1++;
   first=p1;

   // Add points until we meet one behind NEAR_CLIP
   do {
      corners[p].x=obj->projected_p[i_p[p1]].x;
      corners[p].y=obj->projected_p[i_p[p1]].y;
      corners[p++].c=poly_p->color+obj->vertex_colors_p[i_p[p1]];
      p2=p1++;
      if (p1==pts) p1=0;
   } while (t_p[i_p[p1]].z>NEAR_CLIP);

   // Now p1 and p2 are on different sides of NEAR_CLIP thus we calculate
   // were the line p1-p2 intersects the near clipping plane.

   zdelta=(NEAR_CLIP - t_p[i_p[p1]].z) / (t_p[i_p[p2]].z - t_p[i_p[p1]].z);
   x=t_p[i_p[p1]].x + ((t_p[i_p[p2]].x - t_p[i_p[p1]].x) * zdelta);
   y=t_p[i_p[p1]].y + ((t_p[i_p[p2]].y - t_p[i_p[p1]].y) * zdelta);
   corners[p].x=xcenter+(int)(x*pmulx_zclip);
   corners[p].y=ycenter-(int)(y*pmuly_zclip);
   corners[p++].c=poly_p->color + obj->vertex_colors_p[i_p[p1]] +
      (int)((float)(obj->vertex_colors_p[i_p[p2]] -
      obj->vertex_colors_p[i_p[p1]]) * zdelta);
 
   // Skip points until we are on the right side of clipping plane.
   do {
      p2=p1++;
      if (p1==pts) p1=0;
   } while (t_p[i_p[p1]].z<=NEAR_CLIP);

   zdelta=(NEAR_CLIP - t_p[i_p[p1]].z) / (t_p[i_p[p2]].z - t_p[i_p[p1]].z);
   x=t_p[i_p[p1]].x + ((t_p[i_p[p2]].x - t_p[i_p[p1]].x) * zdelta);
   y=t_p[i_p[p1]].y + ((t_p[i_p[p2]].y - t_p[i_p[p1]].y) * zdelta);
   corners[p].x=xcenter+(int)(x*pmulx_zclip);
   corners[p].y=ycenter-(int)(y*pmuly_zclip);
   corners[p++].c=poly_p->color + obj->vertex_colors_p[i_p[p1]] +
      (int)((float)(obj->vertex_colors_p[i_p[p2]] -
      obj->vertex_colors_p[i_p[p1]]) * zdelta);

   // Now add points until we meet the point we started with.
   while (p1!=first) {
      corners[p].x=obj->projected_p[i_p[p1]].x;
      corners[p].y=obj->projected_p[i_p[p1]].y;
      corners[p++].c=poly_p->color + obj->vertex_colors_p[i_p[p1]];
      if (++p1==pts) p1=0;
   }
   pts=p-1;
}

for (i=0; i<pts; i++)
   ScanEdge_G(corners[i].x, corners[i].y, corners[i].c, corners[i+1].x, corners[i+1].y, corners[i+1].c);

ScanEdge_G(corners[pts].x, corners[pts].y, corners[pts].c, corners[0].x, corners[0].y, corners[0].c);
if ((ymin<=bottom_clip) & (ymax>=NULL)) DrawHLines_Gris(ymin, ymax+1);
}

//-- Draws a texture mapped unshaded polygon --------------------------------

void Draw_Texture_Polygon(poly_3d *poly_p)
{
int i,p1,p2,p,first,pts,indx;
float x,y,zdelta,xk,yk;
object_3d *obj;
short *i_p;
vertice_3d *t_p;
pix *mapping;
tpix corners[9];

ymin=MAXLONG; ymax=-MAXLONG;
pts=poly_p->points;
mapping=poly_cornermappings[pts-3];

obj=(object_3d *)(poly_p->homeobj_p);
i_p=poly_p->pindxs;

// Do we have to zclip the polygon?

if (!(poly_p->flags & POLY_CLIPPED)) {          // Nop
   for (p=0; p<pts; p++) {
      corners[p].x=obj->projected_p[i_p[p]].x;
      corners[p].y=obj->projected_p[i_p[p]].y;
      corners[p].tx=mapping[p].x;
      corners[p].ty=mapping[p].y;
   }
   pts--;
}
else {
   t_p=obj->transformed_p;
   poly_p->flags&=~POLY_CLIPPED;

   // Get first point on the right side of near clipping plane.
   p=p1=0;
   while (t_p[i_p[p1]].z<=NEAR_CLIP) p1++;
   first=p1;

   // Add points until we meet one behind NEAR_CLIP
   do {
      corners[p].x=obj->projected_p[i_p[p1]].x;
      corners[p].y=obj->projected_p[i_p[p1]].y;
      corners[p].tx=mapping[p1].x;
      corners[p++].ty=mapping[p1].y;
      p2=p1++;
      if (p1==pts) p1=0;
   } while (t_p[i_p[p1]].z>NEAR_CLIP);

   // Now p1 and p2 are on different sides of NEAR_CLIP thus we calculate
   // were the line p1-p2 intersects the near clipping plane.

   zdelta=(NEAR_CLIP - t_p[i_p[p1]].z) / (t_p[i_p[p2]].z - t_p[i_p[p1]].z);
   x=t_p[i_p[p1]].x + ((t_p[i_p[p2]].x - t_p[i_p[p1]].x) * zdelta);
   y=t_p[i_p[p1]].y + ((t_p[i_p[p2]].y - t_p[i_p[p1]].y) * zdelta);
   corners[p].x=xcenter+(int)(x*pmulx_zclip);
   corners[p].y=ycenter-(int)(y*pmuly_zclip);
   corners[p].tx=mapping[p1].x + (int)((float)(mapping[p2].x - mapping[p1].x) * zdelta);
   corners[p++].ty=mapping[p1].y + (int)((float)(mapping[p2].y - mapping[p1].y) * zdelta);

   // Skip points until we are on the right side of clipping plane.
   do {
      p2=p1++;
      if (p1==pts) p1=0;
   } while (t_p[i_p[p1]].z<=NEAR_CLIP);

   zdelta=(NEAR_CLIP - t_p[i_p[p1]].z) / (t_p[i_p[p2]].z - t_p[i_p[p1]].z);
   x=t_p[i_p[p1]].x + ((t_p[i_p[p2]].x - t_p[i_p[p1]].x) * zdelta);
   y=t_p[i_p[p1]].y + ((t_p[i_p[p2]].y - t_p[i_p[p1]].y) * zdelta);
   corners[p].x=xcenter+(int)(x*pmulx_zclip);
   corners[p].y=ycenter-(int)(y*pmuly_zclip);
   corners[p].tx=mapping[p1].x + (int)((float)(mapping[p2].x - mapping[p1].x) * zdelta);
   corners[p++].ty=mapping[p1].y + (int)((float)(mapping[p2].y - mapping[p1].y) * zdelta);


   // Now add points until we meet the point we started with.
   while (p1!=first) {
      corners[p].x=obj->projected_p[i_p[p1]].x;
      corners[p].y=obj->projected_p[i_p[p1]].y;
      corners[p].tx=mapping[p1].x;
      corners[p++].ty=mapping[p1].y;
      if (++p1==pts) p1=0;
   }
   pts=p-1;
}

for (i=0; i<pts; i++)
   ScanEdge_T(corners[i].x, corners[i].y, corners[i].tx, corners[i].ty,
      corners[i+1].x, corners[i+1].y, corners[i+1].tx, corners[i+1].ty);

ScanEdge_T(corners[pts].x, corners[pts].y, corners[pts].tx, corners[pts].ty,
   corners[0].x, corners[0].y, corners[0].tx, corners[0].ty);

if ((ymin<=bottom_clip) & (ymax>=NULL)) DrawHLines_T(ymin, ymax+1, textures_pp[poly_p->tmapindx]);
}


/*---------------------------------------------------------------------------
                           Edge-scanning functions
----------------------------------------------------------------------------*/


// Solid polygon edge scanning

void ScanEdge_S(int x1, int y1, int x2, int y2)
{
int k,i;
if (y1!=y2) {

   if (y1>y2) {                // r y1 > y2?
      i=x2;                    // Ja, byt koordinater.
      x2=x1;
      x1=i;

      i=y2;
      y2=y1;
      y1=i;
   }

   k=((x2-x1)<<10) / (y2-y1);      // Rkna ut k-vrdet.

   if (y2>bottom_clip) y2=bottom_clip;

   if (y1<NULL) {
      x2<<=10;
      x2-=k>>1;
      y1++;
      y1=NULL;
      for (i=y2; i>=y1; i--) {
         if (edge1[i]==MAXLONG) edge1[i]=x2; else edge2[i]=x2;
         x2-=k;
      }
   } else {
      x1<<=10;
      x1+=k>>1;
      y1++;
      for (i=y1; i<=y2; i++) {
         if (edge1[i]==MAXLONG) edge1[i]=x1; else edge2[i]=x1;
         x1+=k;
      }
   }

   if (y1<ymin) ymin=y1;
   if (y2>ymax) ymax=y2;
}
}

// Gouraud shaded polygon edge scanning

void ScanEdge_G(int x1, int y1, int c1, int x2, int y2, int c2)
{
int i, xk,ck;
if (y1!=y2) {

   if (y1>y2) {                // r y1 > y2?
      i=x2;                    // Ja, byt koordinater.
      x2=x1;
      x1=i;

      i=y2;
      y2=y1;
      y1=i;

      i=c2;
      c2=c1;
      c1=i;
   }

   xk=((x2-x1)<<16) / (y2-y1);      // Rkna ut k-vrdet.
   ck=((c2-c1)<<16) / (y2-y1);
   y1++;
   x1=x1<<16;
   c1=c1<<16;
   c1+=ck>>1;
   x1+=xk>>1;

   if (y1<NULL) {
      x1+=(NULL-y1)*xk;
      c1+=(NULL-y1)*ck;
      y1=NULL;
   }
   if (y2>bottom_clip) y2=bottom_clip;
   if (y1<ymin) ymin=y1;
   if (y2>ymax) ymax=y2;

   for (i=y1; i<=y2; i++) {
      if (edge1[i]==MAXLONG) {
         edge1[i]=x1;
         color1[i]=c1;
      } else {
         edge2[i]=x1;
         color2[i]=c1;
      }
      x1+=xk;
      c1+=ck;
   }
}
}

// Texture Mapped polygon edge scanning

void ScanEdge_T(int x1, int y1, int tx1, int ty1, int x2, int y2, int tx2, int ty2)
{
int i, xk,txk,tyk;
if (y1!=y2) {

   if (y1>y2) {                // r y1 > y2?
      i=x2;                    // Ja, byt koordinater.
      x2=x1;
      x1=i;

      i=y2;
      y2=y1;
      y1=i;

      i=tx2;
      tx2=tx1;
      tx1=i;

      i=ty2;
      ty2=ty1;
      ty1=i;
   }

   xk=((x2-x1)<<16) / (y2-y1);      // Rkna ut k-vrdet.
   txk=((tx2-tx1)<<16) / (y2-y1);
   tyk=((ty2-ty1)<<16) / (y2-y1);
   y1++;
   x1<<=16;
   tx1<<=16;
   ty1<<=16;
   x1+=xk>>1;

   if (y1<NULL) {
      x1+=(NULL-y1)*xk;
      tx1+=(NULL-y1)*txk;
      ty1+=(NULL-y1)*tyk;
      y1=NULL;
   }
   if (y2>bottom_clip) y2=bottom_clip;
   if (y1<ymin) ymin=y1;
   if (y2>ymax) ymax=y2;

   for (i=y1; i<=y2; i++) {
      if (edge1[i]==MAXLONG) {
         edge1[i]=x1;
         tex1[i].x=tx1;
         tex1[i].y=ty1;
      } else {
         edge2[i]=x1;
         tex2[i].x=tx1;
         tex2[i].y=ty1;
      }
      x1+=xk;
      tx1+=txk;
      ty1+=tyk;
   }
}
}

// Gouraud shaded triangle edge scanning

void ScanTriEdge_G(int x1, int y1, int c1, int x2, int y2, int c2)
{
int i, xk,ck;

if (y1!=y2) {

   xk=((x2-x1)<<16) / (y2-y1);      // Rkna ut k-vrdet.
   ck=((c2-c1)<<16) / (y2-y1);
   y1++;
   x1=x1<<16;
   c1=c1<<16;
   c1+=ck>>1;
   x1+=xk>>1;

   if (y1<NULL) {
      x1+=(NULL-y1)*xk;
      c1+=(NULL-y1)*ck;
      y1=NULL;
   }

   if (y2>bottom_clip) y2=bottom_clip;
   if (y1<ymin) ymin=y1;
   if (y2>ymax) ymax=y2;


   for (i=y1; i<=y2; i++) {
      edge_p[i]=x1;
      x1+=xk;
      color_p[i]=c1;
      c1+=ck;
   }
}
}

// Texture mapped triangle edge scanning

void ScanTriEdge_T(int x1, int y1, int tx1, int ty1, int x2, int y2, int tx2, int ty2)
{
int i,xk,txk,tyk;

if (y1!=y2) {

   xk=((x2-x1)<<16) / (y2-y1);        // Rkna ut k-vrdet.
   txk=((tx2-tx1)<<16) / (y2-y1);     // Rkna ut k-vrdet.
   tyk=((ty2-ty1)<<16) / (y2-y1);;    // Rkna ut k-vrdet.

   x1<<=16;
   tx1<<=16;
   ty1<<=16;

   y1++;
   x1+=xk>>1;

   if (y1<NULL) {
      x1+=(NULL-y1)*xk;
      tx1+=(NULL-y1)*txk;
      ty1+=(NULL-y1)*tyk;
      y1=NULL;
   }

   if (y2>bottom_clip) y2=bottom_clip;
   if (y1<ymin) ymin=y1;
   if (y2>ymax) ymax=y2;

   for (i=y1; i<=y2; i++) {
      edge_p[i]=x1;
      x1+=xk;
      tex_p[i].x=tx1;
      tx1+=txk;
      tex_p[i].y=ty1;
      ty1+=tyk;
   }
}
}

// Draw Gouraud lines in edge buffer

void DrawHLines_Gris(int starty, int endy)
{
char *sptr;
register int y,x1,x2,c1,c2,ck,dx;

sptr=gfxbuffer+ymultab[starty]-xsize;
for (y=starty; y<endy; y++) {
   x1=edge1[y];
   x2=edge2[y];
   if (x1<x2) {
      c1=color1[y];
      c2=color2[y];
   }
   else {
      x1=edge2[y];
      x2=edge1[y];
      c1=color2[y];
      c2=color1[y];
   }
   edge1[y]=MAXLONG;
   ck=((c2-c1)<<8) / (x2-x1+128000);
   c1>>=8;
   x2>>=16;
 
   if (x1<0) {
      c1+=((0-x1)*ck)>>16;
      x1=0;
   }
   x1>>=16;

   if (x2>right_clip) x2=right_clip;
   dx=(x2-x1+1);
   sptr+=xsize;
   if (dx<1) continue;

   GLine(sptr+x1-1,dx,c1,ck);
}
}

// Draw texture lines in edge buffer

void DrawHLines_T(int starty, int endy, char *texture_p)
{
char *sptr;
int x,y,x1,x2,xdelta,tx1,tx2,ty1,ty2,txk,tyk,tmp;

for (y=starty; y<endy; y++) {
   x1=edge1[y];
   x2=edge2[y];
   edge1[y]=MAXLONG;
   if (x1<x2) {
      tx1=tex1[y].x;
      tx2=tex2[y].x;
      ty1=tex1[y].y;
      ty2=tex2[y].y;

   } else {
      tmp=x1;
      x1=x2;
      x2=tmp;
      tx1=tex2[y].x;
      tx2=tex1[y].x;
      ty1=tex2[y].y;
      ty2=tex1[y].y;
   }
   xdelta=(x2-x1+65536)>>8;
   txk=(((tx2-tx1)<<8) / xdelta);
   tyk=(((ty2-ty1)<<8) / xdelta);

   if (x1<0) {
      tx1+=(((0-x1)>>8)*txk)>>8;
      ty1+=(((0-x1)>>8)*tyk)>>8;
      x1=0;
   }
   x1>>=16;
   x2>>=16;
   if (x2>right_clip) x2=right_clip;
   sptr=gfxbuffer+ymultab[y]+x1;
   for (x=x1; x<=x2; x++) {
      *sptr++=texture_p[(ty1>>8 & 0xff00)+(tx1>>16)];
      tx1+=txk;
      ty1+=tyk;
   }
}
}

