UNIT POLYGONS;

INTERFACE

uses struct,graphics;

procedure DrawSprite(n: longint);
procedure DrawPhongPolygon(poly: longint);    {portal-prepared}
procedure DrawPlainPolygon(poly: longint);    {portal-prepared}
procedure DrawEnvPolygon(poly: longint);      {portal-prepared}
procedure DrawPTextPolygon(poly: longint);    {portal-prepared}
procedure DrawTPTextPolygon(poly: longint);   {portal-prepared}
procedure DrawGPTextPolygon(poly: longint);   {portal-prepared}
procedure DrawPPTextPolygon(poly: longint);   {portal-prepared}
procedure DrawGouradPolygon(poly: longint);   {portal-prepared}
procedure DrawFastTPolygon(poly: longint);    {portal-prepared}
procedure DrawFastTrTPolygon(poly: longint);  {portal-prepared}


IMPLEMENTATION


{  _______________________________________________________________________
  /                                                                      /
 / CalcBasePoly: Calculates the base polygon for a perspective map      /
/______________________________________________________________________/ }

procedure CalcBasePoly(bs: longint);
var u: longint;
    store: single;
asm
   mov [u],256*128
   fild [u]
   fstp [store]
   mov edi,[bs]
   mov eax,[_fp1+edi]
   fld [_rx+eax*4-4]; fstp [_fx0+edi]
   fld [_ry+eax*4-4]; fstp [_fy0+edi]
   fld [_frz+eax*4-4]; fstp [_fz0+edi]
   mov eax,[_fp2+edi]
   fld [_rx+eax*4-4]; fsub [_fx0+edi]; fstp [_fVx+edi]
   fld [_ry+eax*4-4]; fsub [_fy0+edi]; fstp [_fVy+edi]
   fld [_frz+eax*4-4]; fsub [_fz0+edi]; fstp [_fVz+edi]
   mov eax,[_fp3+edi]
   fld [_fx0+edi]; fsub [_rx+eax*4-4]; fstp [_fUx+edi]
   fld [_fy0+edi]; fsub [_ry+eax*4-4]; fstp [_fUy+edi]
   fld [_fz0+edi]; fsub [_frz+eax*4-4]; fstp [_fUz+edi]
   fld [_fx0+edi]; fmul [store]; fstp [_fx0+edi]
   fld [_fy0+edi]; fmul [store]; fstp [_fy0+edi]
   fld [_fz0+edi]; fmul [store]; fstp [_fz0+edi]
   fld [_fVz+edi]; fmul [_fy0+edi]; fld [_fVy+edi]
   fmul [_fz0+edi]; fsubp; fstp [_fui+edi]
   fld [_fVx+edi]; fmul [_fz0+edi]; fld [_fVz+edi]
   fmul [_fx0+edi]; fsubp; fstp [_fuj+edi]
   fld [_fUy+edi]; fmul [_fz0+edi]; fld [_fUz+edi]
   fmul [_fy0+edi]; fsubp; fstp [_fvi+edi]
   fld [_fUz+edi]; fmul [_fx0+edi]; fld [_fUx+edi]
   fmul [_fz0+edi]; fsubp; fstp [_fvj+edi]
   fld [_fVy+edi]; fmul [_fUz+edi]; fld [_fVz+edi]
   fmul [_fUy+edi]; fsubp; fstp [_fzi+edi]
   fld [_fVz+edi]; fmul [_fUx+edi]; fld [_fVx+edi]
   fmul [_fUz+edi]; fsubp; fstp [_fzj+edi]
   {Calculate deltas for each scanline}
   fld dword [_qdiv+501*4]; fmul [_fuj+edi]; fstp [_fxxd+edi]
   fld dword [_qdiv+501*4]; fmul [_fvj+edi]; fstp [_fyyd+edi]
   fld dword [_qdiv+501*4]; fmul [_fzj+edi]; fstp [_fzzd+edi]
   {Calculate deltas for full subspan}
   fld dword [_qdiv+(500+TSPAN)*4]; fmul [_fui+edi]; fstp [_fxxp+edi]
   fld dword [_qdiv+(500+TSPAN)*4]; fmul [_fvi+edi]; fstp [_fyyp+edi]
   fld dword [_qdiv+(500+TSPAN)*4]; fmul [_fzi+edi]; fstp [_fzzp+edi]
   {Mark the base polygon as 'calculated'}
   mov dword ptr [_fclc+edi],1
end;


{  _______________________________________________________________________
  /                                                                      /
 / DrawSprite: Draws a scaled sprite                                    /
/______________________________________________________________________/ }

procedure DrawSprite(n: longint);
var x1,y1,x2,y2,sx1,sy1,sx2,sy2: longint;
    dsx,dsy,h,v,crd: longint;
begin
   {A sprite is a polygon with only one coordinate}
   crd:=_pp[n+1,1];
   x1:=(_xs[crd]-10) div 64;
   y1:=(_ys[crd]-10) div 64;
   x2:=x1+20;
   y2:=y1+20;
   {determine size of sprite}
   x2:=round(((_rx[crd]+SPRSIZE)*DIST)/_zmin)+XCENTR;
   y2:=round(((_ry[crd]+SPRSIZE)*DIST)/_zmin)+YCENTR;
   x1:=x1-(x2-x1);
   y1:=y1-(y2-y1);
   {Calculate deltas}
   if (x2>x1) and (y2>y1) then begin
      dsx:=(64*256) div (x2-x1);
      dsy:=(64*256) div (y2-y1);
      {Clip the sprite (assume source picture of 64x64 pixels)}
      sx1:=0;
      sy1:=0;
      sx2:=63*256;
      sy2:=63*256;
      if x1<0 then begin
         sx1:=sx1+((-x1*64*256) div (x2-x1));
         x1:=0;
      end;
      if y1<0 then begin
         sy1:=sy1+((-y1*64*256) div (y2-y1));
         y1:=0;
      end;
      if x2>=XSIZE then x2:=XSIZE-1;
      if y2>=YSIZE then y2:=YSIZE-1;
      {Draw the sprite, if it's still on the screen}
      if (x2>y1) and (y2>y1) then begin
         for v:=y1 to y2 do begin
            for h:=x1 to x2 do begin
               mem[buf+h+v*320]:=255; -- mem[tex+(sx1 div 256)+(sy1 div 256)*64];
               sx1:=sx1+dsx;
               sy1:=sy1+dsy;
            end;
         end;
      end;
   end;
end;


{  _______________________________________________________________________
  /                                                                      /
 /  DrawFastTPolygon: Fast classic texturemapping                       /
/______________________________________________________________________/ }

procedure DrawFastTPolygon(poly: longint);
var x3,y3,x4,y4,t: single;
    z3,z4,_ylow,_yhigh: longint;
    i,l,_tn,oldcrds,pp1,pp2,p1,p2: longint;
    x1,y1,x2,y2,ixh,ixl: longint;
    txt,nxt,px1,py1,px2,py2: longint;
    pdelta,ploop,ladr,dpx,dpy,dlx: longint;
    tx,ty: single;
    _fx,_fy: array[1..7] of longint;
    ba: longint;
begin
   inc(poly);
   txt:=_pc[poly];
   _fx[3]:=0;            _fy[3]:=128*64;
   _fx[2]:=128*64; _fy[2]:=128*64;
   _fx[1]:=128*64; _fy[1]:=0;
   _fx[4]:=0;            _fy[4]:=0;
   ba:=_tba[txt];
   {z-clipping: Make sure that the polygon is not partially behind viewer}
   _tn:=0;
   nxt:=5;
   oldcrds:=_coords;
   for i:=1 to _pnv[poly] do begin
      {Determine begin and end of one of the polygon's edges}
      p1:=_pp[poly,i];
      pp1:=i;
      z3:=_rz[p1];
      if i<_pnv[poly] then begin
         p2:=_pp[poly,i+1];
         pp2:=i+1;
      end else begin
         p2:=_pp[poly,1];
         pp2:=1;
      end;
      z4:=_rz[p2];
      {Case 1: Both in, output end coord}
      if (z3>_zmin) and (z4>_zmin) then begin
         inc(_tn);
         _tmp[_tn]:=p2;
         _ttp[_tn]:=pp2;
      end else if (z3>_zmin) and (z4<=_zmin) then begin
      {Case 2: Going in: Output clipped coord}
         x3:=_rx[p1];
         y3:=_ry[p1];
         x4:=_rx[p2];
         y4:=_ry[p2];
         px1:=_fx[pp1];
         py1:=_fy[pp1];
         px2:=_fx[pp2];
         py2:=_fy[pp2];
         t:=(_zmin-_frz[p1])/(_frz[p2]-_frz[p1]);
         inc(_coords);
         tx:=x3+(x4-x3)*t;                 {Clipping of polygon outline}
         ty:=y3+(y4-y3)*t;
         _fx[nxt]:=px1+round((px2-px1)*t); {Clipping of texture}
         _fy[nxt]:=py1+round((py2-py1)*t);
         _xs[_coords]:=round((tx*DIST*64)/_zmin)+XCENTR*64;
         _ys[_coords]:=round((ty*DIST*64)/_zmin)+YCENTR*64;
         inc(_tn);
         inc(nxt);
         _tmp[_tn]:=_coords;
         _ttp[_tn]:=nxt-1;
      end else if (z3<=_zmin) and (z4>_zmin) then begin
         {Case 3: Coming out: Output clipped coord and end coord}
         x3:=_rx[p1];
         y3:=_ry[p1];
         x4:=_rx[p2];
         y4:=_ry[p2];
         px1:=_fx[pp1];
         py1:=_fy[pp1];
         px2:=_fx[pp2];
         py2:=_fy[pp2];
         t:=(_zmin-_frz[p1])/(_frz[p2]-_frz[p1]);
         inc(_coords);
         tx:=x3+(x4-x3)*t;                 {Clipping of polygon outline}
         ty:=y3+(y4-y3)*t;
         _fx[nxt]:=px1+round((px2-px1)*t); {Clipping of texture}
         _fy[nxt]:=py1+round((py2-py1)*t);
         _xs[_coords]:=round((tx*DIST*64)/_zmin)+XCENTR*64;
         _ys[_coords]:=round((ty*DIST*64)/_zmin)+YCENTR*64;
         _tmp[_tn+1]:=_coords;
         _ttp[_tn+1]:=nxt;
         inc(_tn,2);
         inc(nxt);
         _tmp[_tn]:=p2;
         _ttp[_tn]:=pp2;
      end;
   end;                {_tmp is array with coords for polygon, }
   _coords:=oldcrds;   {_tn is number of vertices, }
   asm                 {_ttp is array with texturecoord numbers }
      mov ebx,[_tn]
      cmp ebx,3
      jb  @endofpoly
      dec ebx
      mov esi,dword ptr [_tmp]
      mov edi,[_ys+esi*4-4]
      mov ecx,edi
   @findloop:
      mov esi,[_tmp+ebx*4]
      mov eax,[_ys+esi*4-4]
      cmp eax,edi
      jge @nolowupdate
      mov edi,eax
   @nolowupdate:
      cmp eax,ecx
      jnge @nohiupdate
      mov ecx,eax
   @nohiupdate:
      dec ebx
      jnz @findloop
      cmp edi,[_clu]
      jge @nozeroclip
      mov edi,[_clu]
   @nozeroclip:
      cmp ecx,[_clu]
      jnge @endofpoly
      cmp edi,[_clb]
      jge @endofpoly
      cmp ecx,[_clb]
      jnge @nobotclip
      mov ecx,[_clb]
   @nobotclip:
      sar ecx,6
      sar edi,6
      cmp ecx,edi
      jbe @endofpoly
      mov [_ylow],edi
      mov [_yhigh],ecx
      sub ecx,edi
      xor eax,eax
      mov ebx,(XSIZE-1)*1024
   @clearloop:
      mov [_edgeb+edi*4],ebx
      mov [_edgee+edi*4],eax
      inc edi
      dec ecx
      jnz @clearloop
      mov [i],0
   @foriloop:
      inc [i]
      {Get x1/y1/x2/y2, wrap around for last vertex}
      mov esi,[i]
      cmp esi,[_tn]
      ja  @endofiloop
      mov edi,[_ttp+4*esi-4]
      jz  @1
      mov esi,[_ttp+4*esi]
      jmp @n
   @1:mov esi,dword ptr [_ttp]
   @n:mov eax,[_fx+4*edi-4]
      mov ebx,[_fy+4*edi-4]
      mov ecx,[_fx+4*esi-4]
      mov edx,[_fy+4*esi-4]
      mov [px1],eax
      mov [py1],ebx
      mov [px2],ecx
      mov [py2],edx
      mov esi,[i]
      mov edi,[_tmp+4*esi-4]
      jz  @2
      mov esi,[_tmp+4*esi]
      jmp @m
   @2:mov esi,dword ptr [_tmp]
   @m:mov edx,[_xs+4*edi-4]
      mov eax,[_ys+4*edi-4]
      mov ebx,[_xs+4*esi-4]
      mov ecx,[_ys+4*esi-4]
      mov [x1],edx
      mov [y1],eax
      mov [x2],ebx
      mov [y2],ecx
      cmp eax,ecx
      jz  @endofline
      jnge @noswap
      mov eax,[px1]
      mov ebx,[px2]
      mov ecx,[py1]
      mov edx,[py2]
      mov [px1],ebx
      mov [px2],eax
      mov [py1],edx
      mov [py2],ecx
      mov ecx,[y1]
      mov eax,[y2]
      mov edx,[x2]
      mov ebx,[x1]
      mov [x1],edx
      mov [x2],ebx        {registers: eax = y1  }
      mov [y1],eax                {   ebx = x2  }
      mov [y2],ecx                {   ecx = y2  }
   @noswap:                       {   edx = x1  }
      cmp eax,[_clb]
      jge @endofline
      cmp ecx,[_clu]
      jnge @endofline
      cmp eax,ecx
      jge @endofline      {if (y1<YSIZE) and (y2>0) and (y2>y1)}
      cmp eax,[_clu]
      jge @notopclipping  {if y1<0 then begin}
      sub ecx,[y1]
      xor eax,-1
      add eax,[_clu]
      sal eax,8
      cdq; idiv ecx
      mov ecx,eax
      mov esi,[_clu]
      mov [y1],esi
      mov esi,[x2]        {t:=-y1/(y2-y1)}
      sub esi,[x1]        {x1:=x1+(x2-x1)*t}
      imul esi
      sar eax,8
      add [x1],eax
      mov eax,ecx         {Clipping of texture line}
      mov esi,[px2]
      sub esi,[px1]
      imul esi
      sar eax,8
      add [px1],eax
      mov eax,ecx
      mov esi,[py2]
      sub esi,[py1]
      imul esi
      sar eax,8
      add [py1],eax
      mov ecx,[y2]
   @notopclipping:
      {Fixed point line drawing}
      sub ecx,[y1]        {ecx = delta-y}
      jz  @endofline
      mov eax,[px2]
      sub eax,[px1]
      sal eax,8
      cdq
      idiv ecx
      mov [dpx],eax       {dpx=texture delta-x, f.p. 10bit}
      mov eax,[py2]
      sub eax,[py1]
      sal eax,8
      cdq
      idiv ecx
      mov [dpy],eax       {dpy=texture delta-y, f.p. 10bit}
      mov eax,[x2]
      sub eax,[x1]        {eax = delta-x}
      sal eax,10
      cdq
      idiv ecx
      mov [dlx],eax       {dlx = delta-x/delta-y, f.p. 10bit}
      mov esi,[y1]        {pointer in outline table}
      sar esi,6
      mov ebx,[y2]
      cmp ebx,[_clb]
      jnge @drawline      {clipping to bottom of screen needed?}
      sub ebx,[_clb]
      sub ecx,ebx         {then do it}
   @drawline:             {line is OK, draw it}
      shr ecx,6
      inc ecx
      or  ecx,ecx
      jz  @endofline
      mov ebx,[x1]        {line start x for this vertex}
      sal ebx,4
      mov edi,[px1]       {texture start x}
      sal edi,2
      mov eax,[py1]       {texture start y}
      sal eax,2
   @pixels:
      mov edx,ebx
      cmp edx,[_edgeb+esi*4]
      jge @nobupdate
      mov [_edgeb+esi*4],edx
      mov [_edgxb+esi*4],edi
      mov [_edgyb+esi*4],eax
   @nobupdate:
      cmp edx,[_edgee+esi*4]
      jnge @noeupdate
      mov [_edgee+esi*4],edx
      mov [_edgxe+esi*4],edi
      mov [_edgye+esi*4],eax
   @noeupdate:
      add ebx,[dlx]     {interpolate line slope}
      add edi,[dpx]     {interpolate texture x component}
      add eax,[dpy]     {interpolate texture y component}
      inc esi
      dec ecx
      jnz @pixels
   @endofline:
      jmp @foriloop
   @endofiloop:
      {Fill the outlines}
      mov edi,[_ylow]
      sub [_yhigh],edi
      mov esi,edi
      shl esi,6
      lea esi,[esi+4*esi]
      add esi,[buf]
      mov [ladr],esi
   @spanloop:
      mov edi,[_ylow]
      mov eax,[_edgee+edi*4]
      mov ebx,eax
      sar eax,10
      cmp eax,[_cll+edi*4]
      jnge @skipspan
      mov esi,[_edgeb+edi*4]
      sub ebx,esi
      sar esi,10
      cmp esi,[_clr+edi*4]
      jge @skipspan
      {Calc texture deltas before clipping}
      or  ebx,ebx
      jz  @skipspan
      mov [dpx],ebx
      mov ecx,[_edgyb+edi*4]
      mov eax,[_edgye+edi*4]
      sub eax,ecx
      sal eax,10
      cdq
      idiv ebx
      mov ebx,eax
      shl ecx,16
      mov cx,word ptr [_edgxb+edi*4]
      mov eax,[_edgxe+edi*4]
      sub eax,[_edgxb+edi*4]
      sal eax,10
      cdq
      idiv [dpx]
      rol ebx,16
      mov bx,ax
      mov [pdelta],ebx
      cmp esi,[_cll+edi*4]
      jge @noleftclip
      {Clip Phong line}
      mov eax,esi
      xor eax,-1
      add eax,[_cll+edi*4]
      mov esi,[_edgee+edi*4]
      sub esi,[_edgeb+edi*4]
      sar esi,10
      or  esi,esi
      jz  @escapefromshit
      sal eax,8
      cdq
      idiv esi
      mov ebx,eax
      mov esi,[_edgxe+edi*4]
      sub esi,[_edgxb+edi*4]
      imul esi
      sar eax,8
      add [_edgxb+edi*4],eax
      mov eax,ebx
      mov esi,[_edgye+edi*4]
      sub esi,[_edgyb+edi*4]
      imul esi
      sar eax,8
      add [_edgyb+edi*4],eax
   @escapefromshit:
      mov esi,[_cll+edi*4]
   @noleftclip:
      mov ecx,[_edgee+edi*4]
      sar ecx,10
      cmp ecx,[_clr+edi*4]
      jnge @norightclip
      mov ecx,[_clr+edi*4]
   @norightclip:
      sub ecx,esi
      or  ecx,ecx
      jz  @skipspan
      add esi,[ladr]
      mov [ploop],ecx
      mov ecx,[_edgyb+edi*4]
      shl ecx,16
      mov cx,word ptr [_edgxb+edi*4]
      mov edi,[ba]
   @ploop:
      shr [ploop],1
      jnc @nobyte
      mov eax,ecx
      shr eax,16
      mov al,ch
      and eax,$7F7F
      add ecx,[pdelta]
      inc esi
      mov dl,byte ptr [edi+eax]
      mov [esi-1],dl
   @nobyte:
      shr [ploop],1
      jnc @noword
      mov eax,ecx
      shr eax,16
      mov al,ch
      and eax,$7F7F
      add ecx,[pdelta]
      mov dl,byte ptr [edi+eax]
      mov eax,ecx
      shr eax,16
      mov al,ch
      and eax,$7F7F
      add ecx,[pdelta]
      mov dh,byte ptr [edi+eax]
      mov [esi],dx
      add esi,2
   @noword:
      cmp [ploop],0
      jz  @skipspan
      align 4
   @dwords:
      mov eax,ecx
      shr eax,16
      mov al,ch
      and eax,$7F7F
      add ecx,[pdelta]
      mov dl,byte ptr [edi+eax]
      mov eax,ecx
      shr eax,16
      mov al,ch
      and eax,$7F7F
      add ecx,[pdelta]
      mov dh,byte ptr [edi+eax]
      shl edx,16
      mov eax,ecx
      shr eax,16
      mov al,ch
      and eax,$7F7F
      add ecx,[pdelta]
      mov dl,byte ptr [edi+eax]
      mov eax,ecx
      shr eax,16
      mov al,ch
      and eax,$7F7F
      add ecx,[pdelta]
      add esi,4
      mov dh,byte ptr [edi+eax]
      rol edx,16
      mov [esi-4],edx
      dec [ploop]
      jnz @dwords
   @skipspan:
      add [ladr],320
      inc [_ylow]
      dec [_yhigh]
      jnz @spanloop
   @endofpoly:
   end;
end;


{  _______________________________________________________________________
  /                                                                      /
 /  DrawFastTrTPolygon: Fast classic transparent texturemapping         /
/______________________________________________________________________/ }

procedure DrawFastTrTPolygon(poly: longint);
var x3,y3,x4,y4,t: single;
    z3,z4,_ylow,_yhigh: longint;
    i,l,_tn,oldcrds,pp1,pp2,p1,p2: longint;
    x1,y1,x2,y2,ixh,ixl: longint;
    txt,nxt,px1,py1,px2,py2: longint;
    pdelta,ploop,ladr,dpx,dpy,dlx: longint;
    tx,ty: single;
    _fx,_fy: array[1..7] of longint;
    ba: longint;
begin
   poly:=poly+1;
   txt:=_pc[poly];
   ba:=_tba[txt];
   _fx[3]:=0;      _fy[3]:=128*64;
   _fx[2]:=128*64; _fy[2]:=128*64;
   _fx[1]:=128*64; _fy[1]:=0;
   _fx[4]:=0;      _fy[4]:=0;
   {z-clipping: Make sure that the polygon is not partially behind viewer}
   _tn:=0;
   nxt:=5;
   oldcrds:=_coords;
   for i:=1 to _pnv[poly] do begin
      {Determine begin and end of one of the polygon's edges}
      p1:=_pp[poly,i];
      pp1:=i;
      z3:=_rz[p1];
      if i<_pnv[poly] then begin
         p2:=_pp[poly,i+1];
         pp2:=i+1;
      end else begin
         p2:=_pp[poly,1];
         pp2:=1;
      end;
      z4:=_rz[p2];
      {Case 1: Both in, output end coord}
      if (z3>_zmin) and (z4>_zmin) then begin
         inc(_tn);
         _tmp[_tn]:=p2;
         _ttp[_tn]:=pp2;
      end else if (z3>_zmin) and (z4<=_zmin) then begin
      {Case 2: Going in: Output clipped coord}
         x3:=_rx[p1];
         y3:=_ry[p1];
         x4:=_rx[p2];
         y4:=_ry[p2];
         px1:=_fx[pp1];
         py1:=_fy[pp1];
         px2:=_fx[pp2];
         py2:=_fy[pp2];
         t:=(_zmin-_frz[p1])/(_frz[p2]-_frz[p1]);
         inc(_coords);
         tx:=x3+(x4-x3)*t;                 {Clipping of polygon outline}
         ty:=y3+(y4-y3)*t;
         _fx[nxt]:=px1+round((px2-px1)*t); {Clipping of texture}
         _fy[nxt]:=py1+round((py2-py1)*t);
         _xs[_coords]:=round((tx*DIST*64)/_zmin)+XCENTR*64;
         _ys[_coords]:=round((ty*DIST*64)/_zmin)+YCENTR*64;
         inc(_tn);
         inc(nxt);
         _tmp[_tn]:=_coords;
         _ttp[_tn]:=nxt-1;
      end else if (z3<=_zmin) and (z4>_zmin) then begin
         {Case 3: Coming out: Output clipped coord and end coord}
         x3:=_rx[p1];
         y3:=_ry[p1];
         x4:=_rx[p2];
         y4:=_ry[p2];
         px1:=_fx[pp1];
         py1:=_fy[pp1];
         px2:=_fx[pp2];
         py2:=_fy[pp2];
         t:=(_zmin-_frz[p1])/(_frz[p2]-_frz[p1]);
         inc(_coords);
         tx:=x3+(x4-x3)*t;                 {Clipping of polygon outline}
         ty:=y3+(y4-y3)*t;
         _fx[nxt]:=px1+round((px2-px1)*t); {Clipping of texture}
         _fy[nxt]:=py1+round((py2-py1)*t);
         _xs[_coords]:=round((tx*DIST*64)/_zmin)+XCENTR*64;
         _ys[_coords]:=round((ty*DIST*64)/_zmin)+YCENTR*64;
         _tmp[_tn+1]:=_coords;
         _ttp[_tn+1]:=nxt;
         inc(_tn,2);
         inc(nxt);
         _tmp[_tn]:=p2;
         _ttp[_tn]:=pp2;
      end;
   end;                {_tmp is array with coords for polygon, }
   _coords:=oldcrds;   {_tn is number of vertices, }
   asm                 {_ttp is array with texturecoord numbers }
      mov ebx,[_tn]
      cmp ebx,3
      jb  @endofpoly
      dec ebx
      mov esi,dword ptr [_tmp]
      mov edi,[_ys+esi*4-4]
      mov ecx,edi
   @findloop:
      mov esi,[_tmp+ebx*4]
      mov eax,[_ys+esi*4-4]
      cmp eax,edi
      jge @nolowupdate
      mov edi,eax
   @nolowupdate:
      cmp eax,ecx
      jnge @nohiupdate
      mov ecx,eax
   @nohiupdate:
      dec ebx
      jnz @findloop
      cmp edi,[_clu]
      jge @nozeroclip
      mov edi,[_clu]
   @nozeroclip:
      cmp ecx,[_clu]
      jnge @endofpoly
      cmp edi,[_clb]
      jge @endofpoly
      cmp ecx,[_clb]
      jnge @nobotclip
      mov ecx,[_clb]
   @nobotclip:
      sar ecx,6
      sar edi,6
      cmp ecx,edi
      jbe @endofpoly
      mov [_ylow],edi
      mov [_yhigh],ecx
      sub ecx,edi
      xor eax,eax
      mov ebx,(XSIZE-1)*1024
   @clearloop:
      mov [_edgeb+edi*4],ebx
      mov [_edgee+edi*4],eax
      inc edi
      dec ecx
      jnz @clearloop
      mov [i],0
   @foriloop:
      inc [i]
      {Get x1/y1/x2/y2, wrap around for last vertex}
      mov esi,[i]
      cmp esi,[_tn]
      ja  @endofiloop
      mov edi,[_ttp+4*esi-4]
      jz  @1
      mov esi,[_ttp+4*esi]
      jmp @n
   @1:mov esi,dword ptr [_ttp]
   @n:mov eax,[_fx+4*edi-4]
      mov ebx,[_fy+4*edi-4]
      mov ecx,[_fx+4*esi-4]
      mov edx,[_fy+4*esi-4]
      mov [px1],eax
      mov [py1],ebx
      mov [px2],ecx
      mov [py2],edx
      mov esi,[i]
      mov edi,[_tmp+4*esi-4]
      jz  @2
      mov esi,[_tmp+4*esi]
      jmp @m
   @2:mov esi,dword ptr [_tmp]
   @m:mov edx,[_xs+4*edi-4]
      mov eax,[_ys+4*edi-4]
      mov ebx,[_xs+4*esi-4]
      mov ecx,[_ys+4*esi-4]
      mov [x1],edx
      mov [y1],eax
      mov [x2],ebx
      mov [y2],ecx
      cmp eax,ecx
      jz  @endofline
      jnge @noswap
      mov eax,[px1]
      mov ebx,[px2]
      mov ecx,[py1]
      mov edx,[py2]
      mov [px1],ebx
      mov [px2],eax
      mov [py1],edx
      mov [py2],ecx
      mov ecx,[y1]
      mov eax,[y2]
      mov edx,[x2]
      mov ebx,[x1]
      mov [x1],edx
      mov [x2],ebx        {registers: eax = y1  }
      mov [y1],eax                {   ebx = x2  }
      mov [y2],ecx                {   ecx = y2  }
   @noswap:                       {   edx = x1  }
      cmp eax,[_clb]
      jge @endofline
      cmp ecx,[_clu]
      jnge @endofline
      cmp eax,ecx
      jge @endofline      {if (y1<YSIZE) and (y2>0) and (y2>y1)}
      cmp eax,[_clu]
      jge @notopclipping  {if y1<0 then begin}
      sub ecx,[y1]
      xor eax,-1
      add eax,[_clu]
      sal eax,8
      cdq; idiv ecx
      mov ecx,eax
      mov esi,[_clu]
      mov [y1],esi
      mov esi,[x2]        {t:=-y1/(y2-y1)}
      sub esi,[x1]        {x1:=x1+(x2-x1)*t}
      imul esi
      sar eax,8
      add [x1],eax
      mov eax,ecx         {Clipping of texture line}
      mov esi,[px2]
      sub esi,[px1]
      imul esi
      sar eax,8
      add [px1],eax
      mov eax,ecx
      mov esi,[py2]
      sub esi,[py1]
      imul esi
      sar eax,8
      add [py1],eax
      mov ecx,[y2]
   @notopclipping:
      {Fixed point line drawing}
      sub ecx,[y1]        {ecx = delta-y}
      jz  @endofline
      mov eax,[px2]
      sub eax,[px1]
      sal eax,8
      cdq
      idiv ecx
      mov [dpx],eax       {dpx=texture delta-x, f.p. 10bit}
      mov eax,[py2]
      sub eax,[py1]
      sal eax,8
      cdq
      idiv ecx
      mov [dpy],eax       {dpy=texture delta-y, f.p. 10bit}
      mov eax,[x2]
      sub eax,[x1]        {eax = delta-x}
      sal eax,10
      cdq
      idiv ecx
      mov [dlx],eax       {dlx = delta-x/delta-y, f.p. 10bit}
      mov esi,[y1]        {pointer in outline table}
      sar esi,6
      mov ebx,[y2]
      cmp ebx,[_clb]
      jnge @drawline      {clipping to bottom of screen needed?}
      sub ebx,[_clb]
      sub ecx,ebx         {then do it}
   @drawline:             {line is OK, draw it}
      shr ecx,6
      inc ecx
      or  ecx,ecx
      jz  @endofline
      mov ebx,[x1]        {line start x for this vertex}
      sal ebx,4
      mov edi,[px1]       {texture start x}
      sal edi,2
      mov eax,[py1]       {texture start y}
      sal eax,2
   @pixels:
      mov edx,ebx
      cmp edx,[_edgeb+esi*4]
      jge @nobupdate
      mov [_edgeb+esi*4],edx
      mov [_edgxb+esi*4],edi
      mov [_edgyb+esi*4],eax
   @nobupdate:
      cmp edx,[_edgee+esi*4]
      jnge @noeupdate
      mov [_edgee+esi*4],edx
      mov [_edgxe+esi*4],edi
      mov [_edgye+esi*4],eax
   @noeupdate:
      add ebx,[dlx]     {interpolate line slope}
      add edi,[dpx]     {interpolate texture x component}
      add eax,[dpy]     {interpolate texture y component}
      inc esi
      dec ecx
      jnz @pixels
   @endofline:
      jmp @foriloop
   @endofiloop:
      {Fill the outlines}
      mov edi,[_ylow]
      sub [_yhigh],edi
      mov esi,edi
      shl esi,6
      lea esi,[esi+4*esi]
      add esi,[buf]
      mov [ladr],esi
   @spanloop:
      mov edi,[_ylow]
      mov eax,[_edgee+edi*4]
      mov ebx,eax
      sar eax,10
      cmp eax,[_cll+edi*4]
      jnge @skipspan
      mov esi,[_edgeb+edi*4]
      sub ebx,esi
      sar esi,10
      cmp esi,[_clr+edi*4]
      jge @skipspan
      {Calc texture deltas before clipping}
      or  ebx,ebx
      jz  @skipspan
      mov [dpx],ebx
      mov ecx,[_edgyb+edi*4]
      mov eax,[_edgye+edi*4]
      sub eax,ecx
      sal eax,10
      cdq
      idiv ebx
      mov ebx,eax
      shl ecx,16
      mov cx,word ptr [_edgxb+edi*4]
      mov eax,[_edgxe+edi*4]
      sub eax,[_edgxb+edi*4]
      sal eax,10
      cdq
      idiv [dpx]
      rol ebx,16
      mov bx,ax
      mov [pdelta],ebx
      cmp esi,[_cll+edi*4]
      jge @noleftclip
      {Clip texture line}
      mov eax,esi
      xor eax,-1
      add eax,[_cll+edi*4]
      mov esi,[_edgee+edi*4]
      sub esi,[_edgeb+edi*4]
      sar esi,10
      or  esi,esi
      jz  @escapefromshit
      sal eax,8
      cdq
      idiv esi
      mov ebx,eax
      mov esi,[_edgxe+edi*4]
      sub esi,[_edgxb+edi*4]
      imul esi
      sar eax,8
      add [_edgxb+edi*4],eax
      mov eax,ebx
      mov esi,[_edgye+edi*4]
      sub esi,[_edgyb+edi*4]
      imul esi
      sar eax,8
      add [_edgyb+edi*4],eax
   @escapefromshit:
      mov esi,[_cll+edi*4]
   @noleftclip:
      mov ecx,[_edgee+edi*4]
      sar ecx,10
      cmp ecx,[_clr+edi*4]
      jnge @norightclip
      mov ecx,[_clr+edi*4]
   @norightclip:
      sub ecx,esi
      or  ecx,ecx
      jz  @skipspan
      add esi,[ladr]
      mov [ploop],ecx
      mov ecx,[_edgyb+edi*4]
      shl ecx,16
      mov cx,word ptr [_edgxb+edi*4]
      mov ebx,[mix]
      mov edi,[ba]
      xor edx,edx
   @ploop:
      mov eax,ecx
      shr eax,16
      mov al,ch
      and eax,$7F7F
      add ecx,[pdelta]
      inc esi
      mov dl,byte ptr [edi+eax]
      mov dh,[esi-1]
      mov dl,[ebx+edx]
      mov [esi-1],dl
      dec [ploop]
      jnz @ploop
   @skipspan:
      add [ladr],320
      inc [_ylow]
      dec [_yhigh]
      jnz @spanloop
   @endofpoly:
   end;
end;


{  _______________________________________________________________________
  /                                                                      /
 /  DrawPhongPolygon: Spotlight Polygon Rendering                       /
/______________________________________________________________________/ }

procedure DrawPhongPolygon(poly: longint);
var x3,y3,x4,y4,t: single;
    z3,z4,_ylow,_yhigh: longint;
    i,l,_tn,oldcrds,p1,p2: longint;
    x1,y1,x2,y2,ixh,ixl: longint;
    px1,py1,px2,py2: longint;
    col,pdelta,ploop,ladr,dpx,dpy,dlx: longint;
    tx,ty: single;
begin
   poly:=poly+1;
   {z-clipping: Make sure that the polygon is not partially behind viewer}
   _tn:=0;
   oldcrds:=_coords;
   for i:=1 to _pnv[poly] do begin
      {Determine begin and end of one of the polygon's edges}
      p1:=_pp[poly,i];
      z3:=_rz[p1];
      if i<_pnv[poly] then p2:=_pp[poly,i+1] else p2:=_pp[poly,1];
      z4:=_rz[p2];
      {Case 1: Both in, output end coord}
      if (z3>_zmin) and (z4>_zmin) then begin
         inc(_tn);
         _tmp[_tn]:=p2
      end else if (z3>_zmin) and (z4<=_zmin) then begin
      {Case 2: Going in: Output clipped coord}
         x3:=_rx[p1];
         y3:=_ry[p1];
         x4:=_rx[p2];
         y4:=_ry[p2];
         px1:=_vy[p1];                          {Phong coords: Vertex normals}
         py1:=_vz[p1];
         px2:=_vy[p2];
         py2:=_vz[p2];
         t:=(_zmin-_frz[p1])/(_frz[p2]-_frz[p1]);
         inc(_coords);
         _rx[_coords]:=x3+(x4-x3)*t;            {Clipping of polygon outline}
         _ry[_coords]:=y3+(y4-y3)*t;
         _vy[_coords]:=px1+round((px2-px1)*t);  {Clipping of phong line}
         _vz[_coords]:=py1+round((py2-py1)*t);
         _xs[_coords]:=round((_rx[_coords]*DIST*64)/_zmin)+XCENTR*64;
         _ys[_coords]:=round((_ry[_coords]*DIST*64)/_zmin)+YCENTR*64;
         inc(_tn);
         _tmp[_tn]:=_coords;
      end else if (z3<=_zmin) and (z4>_zmin) then begin
         {Case 3: Coming out: Output clipped coord and end coord}
         x3:=_rx[p1];
         y3:=_ry[p1];
         x4:=_rx[p2];
         y4:=_ry[p2];
         px1:=_vy[p1];
         py1:=_vz[p1];
         px2:=_vy[p2];
         py2:=_vz[p2];
         t:=(_zmin-_frz[p1])/(_frz[p2]-_frz[p1]);
         inc(_coords);
         _rx[_coords]:=x3+(x4-x3)*t;            {Clipping of polygon outline}
         _ry[_coords]:=y3+(y4-y3)*t;
         _vy[_coords]:=px1+round((px2-px1)*t);  {Clipping of Phong line}
         _vz[_coords]:=py1+round((py2-py1)*t);
         _xs[_coords]:=round((_rx[_coords]*DIST*64)/_zmin)+XCENTR*64;
         _ys[_coords]:=round((_ry[_coords]*DIST*64)/_zmin)+YCENTR*64;
         _tmp[_tn+1]:=_coords;
         inc(_tn,2);
         _tmp[_tn]:=p2;
      end;
   end;  {_tmp is array with coords for polygon, _tn is number of vertices}
   _coords:=oldcrds;
   col:=_pc[poly];
   asm
      mov ebx,[_tn]
      cmp ebx,3
      jb  @endofpoly
      dec ebx
      mov esi,dword ptr [_tmp]
      mov edi,[_ys+esi*4-4]
      mov ecx,edi
   @findloop:
      mov esi,[_tmp+ebx*4]
      mov eax,[_ys+esi*4-4]
      cmp eax,edi
      jge @nolowupdate
      mov edi,eax
   @nolowupdate:
      cmp eax,ecx
      jnge @nohiupdate
      mov ecx,eax
   @nohiupdate:
      dec ebx
      jnz @findloop
      cmp edi,[_clu]
      jge @nozeroclip
      mov edi,[_clu]
   @nozeroclip:
      cmp ecx,[_clu]
      jnge @endofpoly
      cmp edi,[_clb]
      jge @endofpoly
      cmp ecx,[_clb]
      jnge @nobotclip
      mov ecx,[_clb]
   @nobotclip:
      sar ecx,6
      sar edi,6
      cmp ecx,edi
      jbe @endofpoly
      mov [_ylow],edi
      mov [_yhigh],ecx
      sub ecx,edi
      xor eax,eax
      mov ebx,(XSIZE-1)*1024
   @clearloop:
      mov [_edgeb+edi*4],ebx
      mov [_edgee+edi*4],eax
      inc edi
      dec ecx
      jnz @clearloop
      mov [i],0
   @foriloop:
      inc [i]
      {Get x1/y1/x2/y2, wrap around for last vertex}
      mov esi,[i]
      cmp esi,[_tn]
      ja  @endofiloop
      mov edi,[_tmp+4*esi-4]
      jz  @1
      mov esi,[_tmp+4*esi]
      jmp @n
   @1: mov esi,dword ptr [_tmp]
   @n: mov eax,[_vy+4*edi-4]
      mov ebx,[_vz+4*edi-4]
      mov ecx,[_vy+4*esi-4]
      mov edx,[_vz+4*esi-4]
      mov [px1],eax
      mov [py1],ebx
      mov [px2],ecx
      mov [py2],edx
      mov edx,[_xs+4*edi-4]  {hehe not a single AGI stall}
      mov eax,[_ys+4*edi-4]
      mov ebx,[_xs+4*esi-4]
      mov ecx,[_ys+4*esi-4]
      mov [x1],edx
      mov [y1],eax
      mov [x2],ebx
      mov [y2],ecx
      cmp eax,ecx
      jz  @endofline
      jnge @noswap
      mov eax,[px1]
      mov ebx,[px2]
      mov ecx,[py1]
      mov edx,[py2]
      mov [px1],ebx
      mov [px2],eax
      mov [py1],edx
      mov [py2],ecx
      mov ecx,[y1]
      mov eax,[y2]
      mov edx,[x2]
      mov ebx,[x1]
      mov [x1],edx
      mov [x2],ebx        {registers: eax = y1  }
      mov [y1],eax                {   ebx = x2  }
      mov [y2],ecx                {   ecx = y2  }
   @noswap:                       {   edx = x1  }
      cmp eax,[_clb]
      jge @endofline
      cmp ecx,[_clu]
      jnge @endofline
      cmp eax,ecx
      jge @endofline      {if (y1<YSIZE) and (y2>0) and (y2>y1)}
      cmp eax,[_clu]
      jge @notopclipping  {if y1<0 then begin}
      sub ecx,[y1]
      xor eax,-1
      add eax,[_clu]
      sal eax,8
      cdq; idiv ecx
      mov ecx,eax
      mov esi,[_clu]
      mov [y1],esi
      mov esi,[x2]        {t:=-y1/(y2-y1)}
      sub esi,[x1]        {x1:=x1+(x2-x1)*t}
      imul esi
      sar eax,8
      add [x1],eax
      mov eax,ecx         {Clipping of phong line}
      mov esi,[px2]
      sub esi,[px1]
      imul esi
      sar eax,8
      add [px1],eax
      mov eax,ecx
      mov esi,[py2]
      sub esi,[py1]
      imul esi
      sar eax,8
      add [py1],eax
      mov ecx,[y2]
   @notopclipping:
      {Fixed point line drawing}
      sub ecx,[y1]        {ecx = delta-y}
      jz  @endofline
      mov eax,[px2]
      sub eax,[px1]
      sal eax,8
      cdq
      idiv ecx
      mov [dpx],eax       {dpx=phong delta-x, f.p. 10bit}
      mov eax,[py2]
      sub eax,[py1]
      sal eax,8
      cdq
      idiv ecx
      mov [dpy],eax       {dpy=phong delta-y, f.p. 10bit}
      mov eax,[x2]
      sub eax,[x1]        {eax = delta-x}
      sal eax,10
      cdq
      idiv ecx
      mov [dlx],eax       {dlx = delta-x/delta-y, f.p. 10bit}
      mov esi,[y1]        {pointer in outline table}
      sar esi,6
      mov ebx,[y2]
      cmp ebx,[_clb]
      jnge @drawline      {clipping to bottom of screen needed?}
      sub ebx,[_clb]
      sub ecx,ebx         {then do it}
   @drawline:             {line is OK, draw it}
      shr ecx,6
      inc ecx
      or  ecx,ecx
      jz  @endofline
      mov ebx,[x1]        {line start x for this vertex}
      sal ebx,10-6
      mov edi,[px1]       {phong start x}
      sal edi,2
      mov eax,[py1]       {phong start y}
      sal eax,2
   @pixels:
      mov edx,ebx
      cmp edx,[_edgeb+esi*4]
      jge @nobupdate
      mov [_edgeb+esi*4],edx
      mov [_edgxb+esi*4],edi
      mov [_edgyb+esi*4],eax
   @nobupdate:
      cmp edx,[_edgee+esi*4]
      jnge @noeupdate
      mov [_edgee+esi*4],edx
      mov [_edgxe+esi*4],edi
      mov [_edgye+esi*4],eax
   @noeupdate:
      add ebx,[dlx]     {interpolate line slope}
      add edi,[dpx]     {interpolate phong x component}
      add eax,[dpy]     {interpolate phong y component}
      inc esi
      dec ecx
      jnz @pixels
   @endofline:
      jmp @foriloop
   @endofiloop:
      {Fill the outlines}
      mov edi,[_ylow]
      sub [_yhigh],edi
      mov esi,edi
      shl esi,6
      lea esi,[esi+4*esi]
      add esi,[buf]
      mov [ladr],esi
   @spanloop:
      mov edi,[_ylow]
      mov eax,[_edgee+edi*4]
      mov ebx,eax
      sar eax,10
      cmp eax,[_cll+edi*4]
      jnge @skipspan
      mov esi,[_edgeb+edi*4]
      sub ebx,esi
      sar esi,10
      cmp esi,[_clr+edi*4]
      jge @skipspan
      {Calc Phong deltas before clipping}
      or  ebx,ebx
      jz  @skipspan
      mov [dpx],ebx
      mov ecx,[_edgyb+edi*4]
      mov eax,[_edgye+edi*4]
      sub eax,ecx
      sal eax,10
      cdq
      idiv ebx
      mov ebx,eax
      shl ecx,16
      mov cx,word ptr [_edgxb+edi*4]
      mov eax,[_edgxe+edi*4]
      sub eax,[_edgxb+edi*4]
      sal eax,10
      cdq
      idiv [dpx]
      rol ebx,16
      mov bx,ax
      mov [pdelta],ebx
      cmp esi,[_cll+edi*4]
      jge @noleftclip
      {Clip Phong line}
      mov eax,esi
      xor eax,-1
      add eax,[_cll+edi*4]
      mov esi,[_edgee+edi*4]
      sub esi,[_edgeb+edi*4]
      sar esi,10
      or  esi,esi
      jz  @escapefromshit
      sal eax,8
      cdq
      idiv esi
      mov ebx,eax
      mov esi,[_edgxe+edi*4]
      sub esi,[_edgxb+edi*4]
      imul esi
      sar eax,8
      add [_edgxb+edi*4],eax
      mov eax,ebx
      mov esi,[_edgye+edi*4]
      sub esi,[_edgyb+edi*4]
      imul esi
      sar eax,8
      add [_edgyb+edi*4],eax
   @escapefromshit:
      mov esi,[_cll+edi*4]
   @noleftclip:
      mov ecx,[_edgee+edi*4]
      sar ecx,10
      cmp ecx,[_clr+edi*4]
      jnge @norightclip
      mov ecx,[_clr+edi*4]
   @norightclip:
      sub ecx,esi
      or  ecx,ecx
      jz  @skipspan
      add esi,[ladr]
      mov [ploop],ecx
      mov ecx,[_edgyb+edi*4]
      shl ecx,16
      mov cx,word ptr [_edgxb+edi*4]
      mov ebx,[col]
      shl ebx,8
      mov edi,[pmap]
   @ploop:             {Start of PHONGCANON (tm) routine}
      shr [ploop],1    {Possible enhancements:}
      jnc @nobyte      {1. Reduce number of DIVS (triangle rendering?)}
      mov eax,ecx      {2. Try to get stuff in cache before using it}
      shr eax,16       {This routine uses the light mixing table}
      mov al,ch
      add ecx,[pdelta]
      mov bl,byte ptr [edi+eax]
      mov dl,byte ptr [lit+ebx]
      mov [esi],dl
      inc esi
   @nobyte:
      shr [ploop],1
      jnc @noword
      mov eax,ecx
      shr eax,16
      mov al,ch
      add ecx,[pdelta]
      mov bl,byte ptr [edi+eax]
      mov dl,byte ptr [lit+ebx]
      mov eax,ecx
      shr eax,16
      mov al,ch
      add ecx,[pdelta]
      mov bl,byte ptr [edi+eax]
      mov dh,byte ptr [lit+ebx]
      mov [esi],dx
      add esi,2
   @noword:
      cmp [ploop],0
      jz  @skipspan
      align 4
   @dwords:
      mov eax,ecx
      shr eax,16
      mov al,ch
      add ecx,[pdelta]
      mov bl,byte ptr [edi+eax]
      mov dl,byte ptr [lit+ebx]
      mov eax,ecx
      shr eax,16
      mov al,ch
      add ecx,[pdelta]
      mov bl,byte ptr [edi+eax]
      mov dh,byte ptr [lit+ebx]
      shl edx,16
      mov eax,ecx
      shr eax,16
      mov al,ch
      add ecx,[pdelta]
      mov bl,byte ptr [edi+eax]
      mov dl,byte ptr [lit+ebx]
      mov eax,ecx
      shr eax,16
      mov al,ch
      add ecx,[pdelta]
      add esi,4
      mov bl,byte ptr [edi+eax]
      mov dh,byte ptr [lit+ebx]
      rol edx,16
      mov [esi-4],edx
      dec [ploop]
      jnz @dwords
   @skipspan:
      add [ladr],320
      inc [_ylow]
      dec [_yhigh]
      jnz @spanloop
   @endofpoly:
   end;
end;


{  _______________________________________________________________________
  /                                                                      /
 /  DrawPlainPolygon: Basic Polygon Rendering                           /
/______________________________________________________________________/ }

procedure DrawPlainPolygon(poly: longint);
var x3,y3,x4,y4,t: single;
    z3,z4,_ylow,_yhigh: longint;
    i,_tn,oldcrds,p1,p2: longint;
    x1,y1,x2,y2,ixh,ixl: longint;
    tx,ty,tz: single;
begin
   poly:=poly+1;
   {z-clipping: Make sure that the polygon is not partially behind viewer}
   _tn:=0;
   oldcrds:=_coords;
   for i:=1 to _pnv[poly] do begin
      {Determine begin and end of one of the polygon's edges}
      p1:=_pp[poly,i];
      z3:=_rz[p1];
      if i<_pnv[poly] then p2:=_pp[poly,i+1] else p2:=_pp[poly,1];
      z4:=_rz[p2];
      {Case 1: Both in, output end coord}
      if (z3>_zmin) and (z4>_zmin) then begin
         inc(_tn);
         _tmp[_tn]:=p2
      end else if (z3>_zmin) and (z4<=_zmin) then begin
      {Case 2: Going in: Output clipped coord}
         x3:=_rx[p1];
         y3:=_ry[p1];
         x4:=_rx[p2];
         y4:=_ry[p2];
         t:=(_zmin-_frz[p1])/(_frz[p2]-_frz[p1]);
         inc(_coords);
         tx:=x3+(x4-x3)*t;
         ty:=y3+(y4-y3)*t;
         _xs[_coords]:=round((tx*DIST*64)/_zmin)+XCENTR*64;
         _ys[_coords]:=round((ty*DIST*64)/_zmin)+YCENTR*64;
         inc(_tn);
         _tmp[_tn]:=_coords;
      end else if (z3<=_zmin) and (z4>_zmin) then begin
         {Case 3: Coming out: Output clipped coord and end coord}
         x3:=_rx[p1];
         y3:=_ry[p1];
         x4:=_rx[p2];
         y4:=_ry[p2];
         t:=(_zmin-_frz[p1])/(_frz[p2]-_frz[p1]);
         inc(_coords);
         tx:=x3+(x4-x3)*t;
         ty:=y3+(y4-y3)*t;
         _xs[_coords]:=round((tx*DIST*64)/_zmin)+XCENTR*64;
         _ys[_coords]:=round((ty*DIST*64)/_zmin)+YCENTR*64;
         inc(_tn,2);
         _tmp[_tn-1]:=_coords;
         _tmp[_tn]:=p2;
      end;
   end;  {_tmp is array with coords for polygon, _tn is number of vertices}
   _coords:=oldcrds;
   asm
      mov ebx,[_tn]
      cmp ebx,3
      jb  @endofpoly
      dec ebx
      mov esi,dword ptr [_tmp]
      mov edi,[_ys+esi*4-4]
      mov ecx,edi
   @findloop:
      mov esi,[_tmp+ebx*4]
      mov eax,[_ys+esi*4-4]
      cmp eax,edi
      jge @nolowupdate
      mov edi,eax
   @nolowupdate:
      cmp eax,ecx
      jnge @nohiupdate
      mov ecx,eax
   @nohiupdate:
      dec ebx
      jnz @findloop
      cmp edi,[_clu]
      jge @nozeroclip
      mov edi,[_clu]
   @nozeroclip:
      cmp ecx,[_clu]
      jnge @endofpoly
      cmp edi,[_clb]
      jge @endofpoly
      cmp ecx,[_clb]
      jnge @nobotclip
      mov ecx,[_clb]
   @nobotclip:
      sar ecx,6
      sar edi,6
      cmp ecx,edi
      jbe @endofpoly
      mov [_ylow],edi
      mov [_yhigh],ecx
      sub ecx,edi
      xor eax,eax
      mov ebx,XSIZE-1
   @clearloop:
      mov [_edgeb+edi*4],ebx
      mov [_edgee+edi*4],eax
      inc edi
      dec ecx
      jnz @clearloop
      mov [i],0
   @foriloop:
      inc [i]
      {Get x1/y1/x2/y2, wrap around for last vertex}
      mov esi,[i]
      cmp esi,[_tn]
      ja  @endofiloop
      mov edi,[_tmp+4*esi-4]
      jz  @1
      mov esi,[_tmp+4*esi]
      jmp @n
   @1: mov esi,dword ptr [_tmp]
   @n: mov edx,[_xs+4*edi-4]  {hehe not a single AGI stall}
      mov eax,[_ys+4*edi-4]
      mov ebx,[_xs+4*esi-4]
      mov ecx,[_ys+4*esi-4]
      mov [x1],edx
      mov [y1],eax
      mov [x2],ebx
      mov [y2],ecx
      cmp eax,ecx
      jz  @endofline
      jnge @noswap
      mov ecx,[y1]
      mov eax,[y2]
      mov edx,[x2]
      mov ebx,[x1]
      mov [x1],edx
      mov [x2],ebx        {registers: eax = y1  }
      mov [y1],eax                {   ebx = x2  }
      mov [y2],ecx                {   ecx = y2  }
   @noswap:                       {   edx = x1  }
      cmp eax,[_clb]
      jge @endofline
      cmp ecx,[_clu]
      jnge @endofline
      cmp eax,ecx
      jge @endofline      {if (y1<YSIZE) and (y2>0) and (y2>y1)}
      cmp eax,[_clu]
      jge @notopclipping  {if y1<0 then begin}
      sub ecx,[y1]
      xor eax,-1
      add eax,[_clu]
      sal eax,8
      cdq; idiv ecx
      mov esi,[x2]
      sub esi,[x1]
      mov ecx,[_clu]
      mov [y1],ecx
      imul esi
      sar eax,8
      add [x1],eax
      mov ecx,[y2]
   @notopclipping:
      {Fixed point line drawing}
      sub ecx,[y1]        {ecx = delta-y}
      jz  @endofline
      mov eax,[x2]
      sub eax,[x1]        {eax = delta-x}
      sal eax,10
      cdq
      idiv ecx            {eax = delta-x/delta-y, f.p. 10bit}
      shr ecx,6
      inc ecx
      mov esi,[y1]        {pointer in outline table}
      sar esi,6
      mov ebx,[y2]
      sar ebx,6
      cmp ebx,[_clb]
      jnge @drawline      {clipping to bottom of screen needed?}
      sub ebx,[_clb]
      sub ecx,ebx         {then do it}
      or  ecx,ecx
      jz  @endofline
   @drawline:             {line is OK, draw it}
      mov ebx,[x1]
      sal ebx,4
   @pixels:
      mov edx,ebx
      sar edx,10
      cmp edx,[_edgeb+esi*4]
      jge @nobupdate
      mov [_edgeb+esi*4],edx
   @nobupdate:
      cmp edx,[_edgee+esi*4]
      jnge @noeupdate
      mov [_edgee+esi*4],edx
   @noeupdate:
      add ebx,eax
      inc esi
      dec ecx
      jnz @pixels
   @endofline:
      jmp @foriloop
   @endofiloop:
      {Fill the outlines}
      mov edi,dword ptr [poly]
      movzx ebx,word ptr [_pc+edi*4-4]
      mov bh,bl
      mov edx,ebx
      shl ebx,16
      mov bx,dx
      mov edi,[_ylow]
      mov edx,[_yhigh]
      sub edx,edi
      mov esi,edi
      shl esi,6
      lea esi,[esi+4*esi]
      add esi,[buf]
   @spanloop:
      mov eax,[_edgeb+edi*4]
      cmp eax,[_clr+edi*4]
      jge @skipspan
      cmp eax,[_cll+edi*4]
      jge @noleftclip
      mov eax,[_cll+edi*4]
   @noleftclip:
      mov ecx,[_edgee+edi*4]
      cmp ecx,[_cll+edi*4]
      jnge @skipspan
      cmp ecx,[_clr+edi*4]
      jnge @norightclip
      mov ecx,[_clr+edi*4]
   @norightclip:
      sub ecx,eax
      or  ecx,ecx
      jz  @skipspan
   @ploop:
      shr ecx,1
      jnc @nobyte
      mov [esi+eax],bl
      inc eax
   @nobyte:
      shr ecx,1
      jnc @nowords
      mov [esi+eax],bx
      add eax,2
   @nowords:
      or  ecx,ecx
      jz  @skipspan
   @dwordloop:
      mov [esi+eax],ebx
      add eax,4
      dec ecx
      jnz @dwordloop
   @skipspan:
      add esi,320
      inc edi
      dec edx
      jnz @spanloop
   @endofpoly:
   end;
end;


{  _______________________________________________________________________
  /                                                                      /
 /  DrawEnvPolygon: Environment mapping                                 /
/______________________________________________________________________/ }

procedure DrawEnvPolygon(poly: longint);
var x3,y3,x4,y4,t: single;
    z3,z4,_ylow,_yhigh: longint;
    i,l,_tn,oldcrds,p1,p2: longint;
    x1,y1,x2,y2,ixh,ixl: longint;
    px1,py1,px2,py2: longint;
    col,pdelta,ploop,ladr,dpx,dpy,dlx: longint;
    tx,ty: single;
begin
   poly:=poly+1;
   {z-clipping: Make sure that the polygon is not partially behind viewer}
   _tn:=0;
   oldcrds:=_coords;
   for i:=1 to _pnv[poly] do begin
      {Determine begin and end of one of the polygon's edges}
      p1:=_pp[poly,i];
      z3:=_rz[p1];
      if i<_pnv[poly] then p2:=_pp[poly,i+1] else p2:=_pp[poly,1];
      z4:=_rz[p2];
      {Case 1: Both in, output end coord}
      if (z3>_zmin) and (z4>_zmin) then begin
         inc(_tn);
         _tmp[_tn]:=p2
      end else if (z3>_zmin) and (z4<=_zmin) then begin
      {Case 2: Going in: Output clipped coord}
         x3:=_rx[p1];
         y3:=_ry[p1];
         x4:=_rx[p2];
         y4:=_ry[p2];
         px1:=_vy[p1];                          {Phong coords: Vertex normals}
         py1:=_vz[p1];
         px2:=_vy[p2];
         py2:=_vz[p2];
         t:=(_zmin-_frz[p1])/(_frz[p2]-_frz[p1]);
         inc(_coords);
         tx:=x3+(x4-x3)*t;            {Clipping of polygon outline}
         ty:=y3+(y4-y3)*t;
         _vy[_coords]:=px1+round((px2-px1)*t);  {Clipping of Env line}
         _vz[_coords]:=py1+round((py2-py1)*t);
         _xs[_coords]:=round((tx*DIST*64)/_zmin)+XCENTR*64;
         _ys[_coords]:=round((ty*DIST*64)/_zmin)+YCENTR*64;
         inc(_tn);
         _tmp[_tn]:=_coords;
      end else if (z3<=_zmin) and (z4>_zmin) then begin
         {Case 3: Coming out: Output clipped coord and end coord}
         x3:=_rx[p1];
         y3:=_ry[p1];
         x4:=_rx[p2];
         y4:=_ry[p2];
         px1:=_vy[p1];
         py1:=_vz[p1];
         px2:=_vy[p2];
         py2:=_vz[p2];
         t:=(_zmin-_frz[p1])/(_frz[p2]-_frz[p1]);
         inc(_coords);
         tx:=x3+(x4-x3)*t;            {Clipping of polygon outline}
         ty:=y3+(y4-y3)*t;
         _vy[_coords]:=px1+round((px2-px1)*t);  {Clipping of Env line}
         _vz[_coords]:=py1+round((py2-py1)*t);
         _xs[_coords]:=round((tx*DIST*64)/_zmin)+XCENTR*64;
         _ys[_coords]:=round((ty*DIST*64)/_zmin)+YCENTR*64;
         _tmp[_tn+1]:=_coords;
         inc(_tn,2);
         _tmp[_tn]:=p2;
      end;
   end;  {_tmp is array with coords for polygon, _tn is number of vertices}
   _coords:=oldcrds;
   col:=_pc[poly];
   asm
      mov ebx,[_tn]
      cmp ebx,3
      jb  @endofpoly
      dec ebx
      mov esi,dword ptr [_tmp]
      mov edi,[_ys+esi*4-4]
      mov ecx,edi
   @findloop:
      mov esi,[_tmp+ebx*4]
      mov eax,[_ys+esi*4-4]
      cmp eax,edi
      jge @nolowupdate
      mov edi,eax
   @nolowupdate:
      cmp eax,ecx
      jnge @nohiupdate
      mov ecx,eax
   @nohiupdate:
      dec ebx
      jnz @findloop
      cmp edi,[_clu]
      jge @nozeroclip
      mov edi,[_clu]
   @nozeroclip:
      cmp ecx,[_clu]
      jnge @endofpoly
      cmp edi,[_clb]
      jge @endofpoly
      cmp ecx,[_clb]
      jnge @nobotclip
      mov ecx,[_clb]
   @nobotclip:
      sar ecx,6
      sar edi,6
      cmp ecx,edi
      jbe @endofpoly
      mov [_ylow],edi
      mov [_yhigh],ecx
      sub ecx,edi
      xor eax,eax
      mov ebx,(XSIZE-1)*1024
   @clearloop:
      mov [_edgeb+edi*4],ebx
      mov [_edgee+edi*4],eax
      inc edi
      dec ecx
      jnz @clearloop
      mov [i],0
   @foriloop:
      inc [i]
      {Get x1/y1/x2/y2, wrap around for last vertex}
      mov esi,[i]
      cmp esi,[_tn]
      ja  @endofiloop
      mov edi,[_tmp+4*esi-4]
      jz  @1
      mov esi,[_tmp+4*esi]
      jmp @n
   @1: mov esi,dword ptr [_tmp]
   @n: mov eax,[_vy+4*edi-4]
      mov ebx,[_vz+4*edi-4]
      mov ecx,[_vy+4*esi-4]
      mov edx,[_vz+4*esi-4]
      mov [px1],eax
      mov [py1],ebx
      mov [px2],ecx
      mov [py2],edx
      mov edx,[_xs+4*edi-4]  {hehe not a single AGI stall}
      mov eax,[_ys+4*edi-4]
      mov ebx,[_xs+4*esi-4]
      mov ecx,[_ys+4*esi-4]
      mov [x1],edx
      mov [y1],eax
      mov [x2],ebx
      mov [y2],ecx
      cmp eax,ecx
      jz  @endofline
      jnge @noswap
      mov eax,[px1]
      mov ebx,[px2]
      mov ecx,[py1]
      mov edx,[py2]
      mov [px1],ebx
      mov [px2],eax
      mov [py1],edx
      mov [py2],ecx
      mov ecx,[y1]
      mov eax,[y2]
      mov edx,[x2]
      mov ebx,[x1]
      mov [x1],edx
      mov [x2],ebx        {registers: eax = y1  }
      mov [y1],eax                {   ebx = x2  }
      mov [y2],ecx                {   ecx = y2  }
   @noswap:                       {   edx = x1  }
      cmp eax,[_clb]
      jge @endofline
      cmp ecx,[_clu]
      jnge @endofline
      cmp eax,ecx
      jge @endofline      {if (y1<YSIZE) and (y2>0) and (y2>y1)}
      cmp eax,[_clu]
      jge @notopclipping  {if y1<0 then begin}
      sub ecx,[y1]
      xor eax,-1
      add eax,[_clu]
      sal eax,8
      cdq; idiv ecx
      mov ecx,eax
      mov esi,[_clu]
      mov [y1],esi
      mov esi,[x2]        {t:=-y1/(y2-y1)}
      sub esi,[x1]        {x1:=x1+(x2-x1)*t}
      imul esi
      sar eax,8
      add [x1],eax
      mov eax,ecx         {Clipping of environment line}
      mov esi,[px2]
      sub esi,[px1]
      imul esi
      sar eax,8
      add [px1],eax
      mov ecx,[y2]
   @notopclipping:
      {Fixed point line drawing}
      sub ecx,[y1]        {ecx = delta-y}
      jz  @endofline
      mov eax,[px2]
      sub eax,[px1]
      sal eax,8
      cdq
      idiv ecx
      mov [dpx],eax       {dpx=environment delta-x, f.p. 10bit}
      mov eax,[py2]
      sub eax,[py1]
      sal eax,8
      cdq
      idiv ecx
      mov [dpy],eax       {dpy=environment delta-y, f.p. 10bit}
      mov eax,[x2]
      sub eax,[x1]        {eax = delta-x}
      sal eax,10
      cdq
      idiv ecx
      mov [dlx],eax       {dlx = delta-x/delta-y, f.p. 10bit}
      mov esi,[y1]        {pointer in outline table}
      sar esi,6
      mov ebx,[y2]
      cmp ebx,[_clb]
      jnge @drawline      {clipping to bottom of screen needed?}
      sub ebx,[_clb]
      sub ecx,ebx         {then do it}
   @drawline:             {line is OK, draw it}
      shr ecx,6
      inc ecx
      or  ecx,ecx
      jz  @endofline
      mov ebx,[x1]        {line start x for this vertex}
      sal ebx,4
      mov edi,[px1]       {environment start x}
      sal edi,2
      mov eax,[py1]       {environment start y}
      sal eax,2
   @pixels:
      mov edx,ebx
      cmp edx,[_edgeb+esi*4]
      jge @nobupdate
      mov [_edgeb+esi*4],edx
      mov [_edgxb+esi*4],edi
      mov [_edgyb+esi*4],eax
   @nobupdate:
      cmp edx,[_edgee+esi*4]
      jnge @noeupdate
      mov [_edgee+esi*4],edx
      mov [_edgxe+esi*4],edi
      mov [_edgye+esi*4],eax
   @noeupdate:
      add ebx,[dlx]     {interpolate line slope}
      add edi,[dpx]     {interpolate environment x component}
      add eax,[dpy]     {interpolate environment y component}
      inc esi
      dec ecx
      jnz @pixels
   @endofline:
      jmp @foriloop
   @endofiloop:
      {Fill the outlines}
      mov edi,[_ylow]
      sub [_yhigh],edi
      mov esi,edi
      shl esi,6
      lea esi,[esi+4*esi]
      add esi,[buf]
      mov [ladr],esi
   @spanloop:
      mov edi,[_ylow]
      mov eax,[_edgee+edi*4]
      mov ebx,eax
      sar eax,10
      cmp eax,[_cll+edi*4]
      jnge @skipspan
      mov esi,[_edgeb+edi*4]
      sub ebx,esi
      sar esi,10
      cmp esi,[_clr+edi*4]
      jge @skipspan
      {Calc Envmap deltas before clipping}
      or  ebx,ebx
      jz  @skipspan
      mov [dpx],ebx
      mov ecx,[_edgyb+edi*4]
      mov eax,[_edgye+edi*4]
      sub eax,ecx
      sal eax,10
      cdq
      idiv ebx
      mov ebx,eax
      shl ecx,16
      mov cx,word ptr [_edgxb+edi*4]
      mov eax,[_edgxe+edi*4]
      sub eax,[_edgxb+edi*4]
      sal eax,10
      cdq
      idiv [dpx]
      rol ebx,16
      mov bx,ax
      mov [pdelta],ebx
      cmp esi,[_cll+edi*4]
      jge @noleftclip
      {Clip Envmap line}
      mov eax,esi
      xor eax,-1
      add eax,[_cll+edi*4]
      mov esi,[_edgee+edi*4]
      sub esi,[_edgeb+edi*4]
      sar esi,10
      or  esi,esi
      jz  @escapefromshit
      sal eax,8
      cdq
      idiv esi
      mov ebx,eax
      mov esi,[_edgxe+edi*4]
      sub esi,[_edgxb+edi*4]
      imul esi
      sar eax,8
      add [_edgxb+edi*4],eax
      mov eax,ebx
      mov esi,[_edgye+edi*4]
      sub esi,[_edgyb+edi*4]
      imul esi
      sar eax,8
      add [_edgyb+edi*4],eax
   @escapefromshit:
      mov esi,[_cll+edi*4]
   @noleftclip:
      mov ecx,[_edgee+edi*4]
      sar ecx,10
      cmp ecx,[_clr+edi*4]
      jnge @norightclip
      mov ecx,[_clr+edi*4]
   @norightclip:
      sub ecx,esi
      or  ecx,ecx
      jz  @skipspan
      add esi,[ladr]
      mov [ploop],ecx
      mov ecx,[_edgyb+edi*4]
      shl ecx,16
      mov cx,word ptr [_edgxb+edi*4]
      mov ebx,[col]
      shl ebx,8
      add ebx,[mix]
      mov edi,[env]
   @ploop:
      shr [ploop],1
      jnc @nobyte
      mov eax,ecx
      shr eax,16
      mov al,ch
      add ecx,[pdelta]
      mov bl,byte ptr [edi+eax]
      mov dl,byte ptr [ebx]
      mov [esi],dl
      inc esi
   @nobyte:
      shr [ploop],1
      jnc @noword
      mov eax,ecx
      shr eax,16
      mov al,ch
      add ecx,[pdelta]
      mov bl,byte ptr [edi+eax]
      mov dl,byte ptr [ebx]
      mov eax,ecx
      shr eax,16
      mov al,ch
      add ecx,[pdelta]
      mov bl,byte ptr [edi+eax]
      mov dh,byte ptr [ebx]
      mov [esi],dx
      add esi,2
   @noword:
      cmp [ploop],0
      jz  @skipspan
      align 4
   @dwords:
      mov eax,ecx
      shr eax,16
      mov al,ch
      add ecx,[pdelta]
      mov bl,byte ptr [edi+eax]
      mov dl,byte ptr [ebx]
      mov eax,ecx
      shr eax,16
      mov al,ch
      add ecx,[pdelta]
      mov bl,byte ptr [edi+eax]
      mov dh,byte ptr [ebx]
      shl edx,16
      mov eax,ecx
      shr eax,16
      mov al,ch
      add ecx,[pdelta]
      mov bl,byte ptr [edi+eax]
      mov dl,byte ptr [ebx]
      mov eax,ecx
      shr eax,16
      mov al,ch
      add ecx,[pdelta]
      add esi,4
      mov bl,byte ptr [edi+eax]
      mov dh,byte ptr [ebx]
      rol edx,16
      mov [esi-4],edx
      dec [ploop]
      jnz @dwords
   @skipspan:
      add [ladr],320
      inc [_ylow]
      dec [_yhigh]
      jnz @spanloop
   @endofpoly:
   end;
end;


{  _______________________________________________________________________
  /                                                                      /
 /  DrawGouradPolygon: Simplified shading                               /
/______________________________________________________________________/ }

procedure DrawGouradPolygon(poly: longint);
var x3,y3,x4,y4,t: single;
    z3,z4,_ylow,_yhigh: longint;
    i,l,_tn,oldcrds,p1,p2: longint;
    x1,y1,x2,y2,ixh,ixl: longint;
    px1,px2: longint;
    col,gloop,ladr,dpx,dpy,dlx: longint;
    tx,ty: single;
begin
   poly:=poly+1;
   {z-clipping: Make sure that the polygon is not partially behind viewer}
   _tn:=0;
   oldcrds:=_coords;
   for i:=1 to _pnv[poly] do begin
      {Determine begin and end of one of the polygon's edges}
      p1:=_pp[poly,i];
      z3:=_rz[p1];
      if i<_pnv[poly] then p2:=_pp[poly,i+1] else p2:=_pp[poly,1];
      z4:=_rz[p2];
      {Case 1: Both in, output end coord}
      if (z3>_zmin) and (z4>_zmin) then begin
         inc(_tn);
         _tmp[_tn]:=p2
      end else if (z3>_zmin) and (z4<=_zmin) then begin
      {Case 2: Going in: Output clipped coord}
         x3:=_rx[p1];
         y3:=_ry[p1];
         x4:=_rx[p2];
         y4:=_ry[p2];
         px1:=_vx[p1];                          {Phong coords: Vertex normals}
         px2:=_vx[p2];
         t:=(_zmin-_frz[p1])/(_frz[p2]-_frz[p1]);
         inc(_coords);
         tx:=x3+(x4-x3)*t;            {Clipping of polygon outline}
         ty:=y3+(y4-y3)*t;
         _vx[_coords]:=px1+round((px2-px1)*t);  {Clipping of phong line}
         _xs[_coords]:=round((tx*DIST*64)/_zmin)+XCENTR*64;
         _ys[_coords]:=round((ty*DIST*64)/_zmin)+YCENTR*64;
         inc(_tn);
         _tmp[_tn]:=_coords;
      end else if (z3<=_zmin) and (z4>_zmin) then begin
         {Case 3: Coming out: Output clipped coord and end coord}
         x3:=_rx[p1];
         y3:=_ry[p1];
         x4:=_rx[p2];
         y4:=_ry[p2];
         px1:=_vx[p1];
         px2:=_vx[p2];
         t:=(_zmin-_frz[p1])/(_frz[p2]-_frz[p1]);
         inc(_coords);
         tx:=x3+(x4-x3)*t;            {Clipping of polygon outline}
         ty:=y3+(y4-y3)*t;
         _vx[_coords]:=px1+round((px2-px1)*t);  {Clipping of Phong line}
         _xs[_coords]:=round((tx*DIST*64)/_zmin)+XCENTR*64;
         _ys[_coords]:=round((ty*DIST*64)/_zmin)+YCENTR*64;
         _tmp[_tn+1]:=_coords;
         inc(_tn,2);
         _tmp[_tn]:=p2;
      end;
   end;  {_tmp is array with coords for polygon, _tn is number of vertices}
   _coords:=oldcrds;
   col:=_pc[poly];
   asm
      mov ebx,[_tn]
      cmp ebx,3
      jb  @endofpoly
      dec ebx
      mov esi,dword ptr [_tmp]
      mov edi,[_ys+esi*4-4]
      mov ecx,edi
   @findloop:
      mov esi,[_tmp+ebx*4]
      mov eax,[_ys+esi*4-4]
      cmp eax,edi
      jge @nolowupdate
      mov edi,eax
   @nolowupdate:
      cmp eax,ecx
      jnge @nohiupdate
      mov ecx,eax
   @nohiupdate:
      dec ebx
      jnz @findloop
      cmp edi,[_clu]
      jge @nozeroclip
      mov edi,[_clu]
   @nozeroclip:
      cmp ecx,[_clu]
      jnge @endofpoly
      cmp edi,[_clb]
      jge @endofpoly
      cmp ecx,[_clb]
      jnge @nobotclip
      mov ecx,[_clb]
   @nobotclip:
      sar ecx,6
      sar edi,6
      cmp ecx,edi
      jbe @endofpoly
      mov [_ylow],edi
      mov [_yhigh],ecx
      sub ecx,edi
      xor eax,eax
      mov ebx,XSIZE-1
   @clearloop:
      mov [_edgeb+edi*4],ebx
      mov [_edgee+edi*4],eax
      inc edi
      dec ecx
      jnz @clearloop
      mov [i],0
   @foriloop:
      inc [i]
      {Get x1/y1/x2/y2, wrap around for last vertex}
      mov esi,[i]
      cmp esi,[_tn]
      ja  @endofiloop
      mov edi,[_tmp+4*esi-4]
      jz  @1
      mov esi,[_tmp+4*esi]
      jmp @n
   @1: mov esi,dword ptr [_tmp]
   @n: mov eax,[_vx+4*edi-4]
      mov ecx,[_vx+4*esi-4]
      add eax,128
      add ecx,128
      sar eax,2
      sar ecx,2
      mov [px1],eax
      mov [px2],ecx
      mov edx,[_xs+4*edi-4]  {hehe not a single AGI stall}
      mov eax,[_ys+4*edi-4]
      mov ebx,[_xs+4*esi-4]
      mov ecx,[_ys+4*esi-4]
      mov [x1],edx
      mov [y1],eax
      mov [x2],ebx
      mov [y2],ecx
      cmp eax,ecx
      jz  @endofline
      jnge @noswap
      mov eax,[px1]
      mov ebx,[px2]
      mov [px1],ebx
      mov [px2],eax
      mov ecx,[y1]
      mov eax,[y2]
      mov edx,[x2]
      mov ebx,[x1]
      mov [x1],edx
      mov [x2],ebx        {registers: eax = y1  }
      mov [y1],eax                {   ebx = x2  }
      mov [y2],ecx                {   ecx = y2  }
   @noswap:                       {   edx = x1  }
      cmp eax,[_clb]
      jge @endofline
      cmp ecx,[_clu]
      jnge @endofline
      cmp eax,ecx
      jge @endofline      {if (y1<YSIZE) and (y2>0) and (y2>y1)}
      cmp eax,[_clu]
      jge @notopclipping  {if y1<0 then begin}
      sub ecx,[y1]
      xor eax,-1
      add eax,[_clu]
      sal eax,8
      cdq; idiv ecx
      mov ecx,eax
      mov esi,[_clu]
      mov [y1],esi
      mov esi,[x2]        {t:=-y1/(y2-y1)}
      sub esi,[x1]        {x1:=x1+(x2-x1)*t}
      imul esi
      sar eax,8
      add [x1],eax
      mov eax,ecx         {Clipping of gourad line}
      mov esi,[px2]
      sub esi,[px1]
      imul esi
      sar eax,8
      add [px1],eax
      mov ecx,[y2]
   @notopclipping:
      {Fixed point line drawing}
      sub ecx,[y1]        {ecx = delta-y}
      jz  @endofline
      mov eax,[px2]
      sub eax,[px1]
      sal eax,8
      cdq
      idiv ecx
      mov [dpx],eax       {dpx=gourad delta-x, f.p. 10bit}
      mov eax,[x2]
      sub eax,[x1]        {eax = delta-x}
      sal eax,10
      cdq
      idiv ecx
      mov [dlx],eax       {dlx = delta-x/delta-y, f.p. 10bit}
      mov esi,[y1]        {pointer in outline table}
      sar esi,6
      mov ebx,[y2]
      cmp ebx,[_clb]
      jnge @drawline      {clipping to bottom of screen needed?}
      sub ebx,[_clb]
      sub ecx,ebx         {then do it}
   @drawline:             {line is OK, draw it}
      shr ecx,6
      inc ecx
      or  ecx,ecx
      jz  @endofline
      mov ebx,[x1]        {line start x for this vertex}
      sal ebx,4
      mov edi,[px1]       {gourad start value}
      sal edi,2
   @pixels:
      mov edx,ebx
      sar edx,10
      cmp edx,[_edgeb+esi*4]
      jge @nobupdate
      mov [_edgeb+esi*4],edx
      mov [_edgxb+esi*4],edi
   @nobupdate:
      cmp edx,[_edgee+esi*4]
      jnge @noeupdate
      mov [_edgee+esi*4],edx
      mov [_edgxe+esi*4],edi
   @noeupdate:
      add ebx,[dlx]     {interpolate line slope}
      add edi,[dpx]     {interpolate gourad value}
      inc esi
      dec ecx
      jnz @pixels
   @endofline:
      jmp @foriloop
   @endofiloop:
      {Fill the outlines}
      mov edi,[_ylow]
      sub [_yhigh],edi
      mov esi,edi
      shl esi,6
      lea esi,[esi+4*esi]
      add esi,[buf]
      mov [ladr],esi
   @spanloop:
      mov edi,[_ylow]
      mov ebx,[_edgeb+edi*4]
      cmp ebx,[_clr+edi*4]
      jge @skipspan
      cmp ebx,[_cll+edi*4]
      jge @noleftclip
      mov ebx,[_cll+edi*4]
   @noleftclip:
      mov ecx,[_edgee+edi*4]
      cmp ecx,[_cll+edi*4]
      jnge @skipspan
      cmp ecx,[_clr+edi*4]
      jnge @norightclip
      mov ecx,[_clr+edi*4]
   @norightclip:
      sub ecx,ebx
      or  ecx,ecx
      jz  @skipspan
      add ebx,[ladr]
      mov [gloop],ecx
      mov ecx,[_edgxb+edi*4]
      mov eax,[_edgxe+edi*4]
      sub eax,[_edgxb+edi*4]
      cdq
      idiv [gloop]
      mov [dpy],eax
      mov eax,[col]
      shl eax,8
   @gloop:          {Start of GOURAD routine}
      shr [gloop],1
      jnc @nobyte
      add ecx,[dpy]
      mov al,ch
      inc ebx
      mov dl,byte ptr[lit+eax]
      mov [ebx-1],dl
   @nobyte:
      shr [gloop],1
      jnc @noword
      add ecx,[dpy]
      mov al,ch
      mov dl,byte ptr[lit+eax]
      add ecx,[dpy]
      add ebx,2
      mov al,ch
      mov dh,byte ptr[lit+eax]
      mov [ebx-2],dx
   @noword:
      cmp [gloop],0
      jz  @skipspan
   @dwords:
      add ecx,[dpy]
      mov al,ch
      mov dl,byte ptr[lit+eax]
      add ecx,[dpy]
      mov al,ch
      mov dh,byte ptr[lit+eax]
      shl edx,16
      add ecx,[dpy]
      mov al,ch
      mov dl,byte ptr[lit+eax]
      add ecx,[dpy]
      add ebx,4
      mov al,ch
      mov dh,byte ptr[lit+eax]
      rol edx,16
      mov [ebx-4],edx
      dec [gloop]
      jnz @dwords
   @skipspan:
      add [ladr],320
      inc [_ylow]
      dec [_yhigh]
      jnz @spanloop
   @endofpoly:
   end;
end;


{  _______________________________________________________________________
  /                                                                      /
 / DrawPTextPolygon: Draw a perspective correct texturemap              /
/______________________________________________________________________/ }

procedure DrawPTextPolygon(poly: longint);
var x3,y3,x4,y4,t: single;
    z3,z4,_ylow,_yhigh: longint;
    i,_tn,oldcrds,p1,p2,pp1,pp2,pp3: longint;
    x1,y1,x2,y2,ixh,ixl: longint;
    tx,ty,tz: single;
    u,v,uo,vo: longint;
    x0,y0,z0: single;
    vx,vy,vz,ux,uy,uz: single;
    ui,uj,vi,vj,zi,zj: single;
    xxs,yys,zzs: single;                 {start per scanline}
    xxp,yyp,zzp: single;                 {delta per sub span}
    xxd,yyd,zzd: single;                 {delta per scanline}
    store,xx,yy,zz: single;
    a: dword;
    du,dv,l,olen,slen,ex: longint;
    ba,bsp: longint;
begin
   poly:=poly+1;
   pp1:=_pp[poly,1];
   pp2:=_pp[poly,2];
   pp3:=_pp[poly,4];
   {z-clipping: Make sure that the polygon is not partially behind viewer}
   _tn:=0;
   oldcrds:=_coords;
   for i:=1 to _pnv[poly] do begin
      {Determine begin and end of one of the polygon's edges}
      p1:=_pp[poly,i];
      z3:=_rz[p1];
      if i<_pnv[poly] then p2:=_pp[poly,i+1] else p2:=_pp[poly,1];
      z4:=_rz[p2];
      {Case 1: Both in, output end coord}
      if (z3>_zmin) and (z4>_zmin) then begin
         inc(_tn);
         _tmp[_tn]:=p2
      end else if (z3>_zmin) and (z4<=_zmin) then begin
      {Case 2: Going in: Output clipped coord}
         x3:=_rx[p1];
         y3:=_ry[p1];
         x4:=_rx[p2];
         y4:=_ry[p2];
         t:=(_zmin-_frz[p1])/(_frz[p2]-_frz[p1]);
         inc(_coords);
         tx:=x3+(x4-x3)*t;
         ty:=y3+(y4-y3)*t;
         _xs[_coords]:=round((tx*DIST*64)/_zmin)+XCENTR*64;
         _ys[_coords]:=round((ty*DIST*64)/_zmin)+YCENTR*64;
         inc(_tn);
         _tmp[_tn]:=_coords;
      end else if (z3<=_zmin) and (z4>_zmin) then begin
         {Case 3: Coming out: Output clipped coord and end coord}
         x3:=_rx[p1];
         y3:=_ry[p1];
         x4:=_rx[p2];
         y4:=_ry[p2];
         t:=(_zmin-_frz[p1])/(_frz[p2]-_frz[p1]);
         inc(_coords);
         tx:=x3+(x4-x3)*t;
         ty:=y3+(y4-y3)*t;
         _xs[_coords]:=round((tx*DIST*64)/_zmin)+XCENTR*64;
         _ys[_coords]:=round((ty*DIST*64)/_zmin)+YCENTR*64;
         inc(_tn,2);
         _tmp[_tn-1]:=_coords;
         _tmp[_tn]:=p2;
      end;
   end;  {_tmp is array with coords for polygon, _tn is number of vertices}
   _coords:=oldcrds;
   ba:=_tba[_pc[poly]];
   bsp:=(_pfb[poly]-1) shl 2;
   asm
      mov ebx,[_tn]
      cmp ebx,3
      jb  @endofpoly
      dec ebx
      mov esi,dword ptr [_tmp]
      mov edi,[_ys+esi*4-4]
      mov ecx,edi
   @findloop:
      mov esi,[_tmp+ebx*4]
      mov eax,[_ys+esi*4-4]
      cmp eax,edi
      jge @nolowupdate
      mov edi,eax
   @nolowupdate:
      cmp eax,ecx
      jnge @nohiupdate
      mov ecx,eax
   @nohiupdate:
      dec ebx
      jnz @findloop
      cmp edi,[_clu]
      jge @nozeroclip
      mov edi,[_clu]
   @nozeroclip:
      cmp ecx,[_clu]
      jnge @endofpoly
      cmp edi,[_clb]
      jge @endofpoly
      cmp ecx,[_clb]
      jnge @nobotclip
      mov ecx,[_clb]
   @nobotclip:
      sar ecx,6
      sar edi,6
      cmp ecx,edi
      jbe @endofpoly
      mov [_ylow],edi
      mov [_yhigh],ecx
      sub ecx,edi
      xor eax,eax
      mov ebx,XSIZE-1
   @clearloop:
      mov [_edgeb+edi*4],ebx
      mov [_edgee+edi*4],eax
      inc edi
      dec ecx
      jnz @clearloop
      mov [i],0
   @foriloop:
      inc [i]
      {Get x1/y1/x2/y2, wrap around for last vertex}
      mov esi,[i]
      cmp esi,[_tn]
      ja  @endofiloop
      mov edi,[_tmp+4*esi-4]
      jz  @1
      mov esi,[_tmp+4*esi]
      jmp @n
   @1: mov esi,dword ptr [_tmp]
   @n: mov edx,[_xs+4*edi-4]  {hehe not a single AGI stall}
      mov eax,[_ys+4*edi-4]
      mov ebx,[_xs+4*esi-4]
      mov ecx,[_ys+4*esi-4]
      mov [x1],edx
      mov [y1],eax
      mov [x2],ebx
      mov [y2],ecx
      cmp eax,ecx
      jz  @endofline
      jnge @noswap
      mov ecx,[y1]
      mov eax,[y2]
      mov edx,[x2]
      mov ebx,[x1]
      mov [x1],edx
      mov [x2],ebx        {registers: eax = y1  }
      mov [y1],eax                {   ebx = x2  }
      mov [y2],ecx                {   ecx = y2  }
   @noswap:                       {   edx = x1  }
      cmp eax,[_clb]
      jge @endofline
      cmp ecx,[_clu]
      jnge @endofline
      cmp eax,ecx
      jge @endofline      {if (y1<YSIZE) and (y2>0) and (y2>y1)}
      cmp eax,[_clu]
      jge @notopclipping  {if y1<0 then begin}
      sub ecx,[y1]
      xor eax,-1
      add eax,[_clu]
      sal eax,8
      cdq; idiv ecx
      mov esi,[_clu]
      mov [y1],esi
      mov esi,[x2]
      sub esi,[x1]
      imul esi
      sar eax,8
      add [x1],eax
      mov ecx,[y2]
   @notopclipping:
      {Fixed point line drawing}
      sub ecx,[y1]        {ecx = delta-y}
      jz  @endofline
      mov eax,[x2]
      sub eax,[x1]        {eax = delta-x}
      sal eax,10
      cdq
      idiv ecx            {eax = delta-x/delta-y, f.p. 10bit}
      mov esi,[y1]        {pointer in outline table}
      sar esi,6
      mov ebx,[y2]
      cmp ebx,[_clb]
      jnge @drawline      {clipping to bottom of screen needed?}
      sub ebx,[_clb]
      sub ecx,ebx         {then do it}
   @drawline:             {line is OK, draw it}
      shr ecx,6
      inc ecx
      or  ecx,ecx
      jz  @endofline
      mov ebx,[x1]
      sal ebx,4
   @pixels:
      mov edx,ebx
      sar edx,10
      cmp edx,[_edgeb+esi*4]
      jge @nobupdate
      mov [_edgeb+esi*4],edx
   @nobupdate:
      cmp edx,[_edgee+esi*4]
      jnge @noeupdate
      mov [_edgee+esi*4],edx
   @noeupdate:
      add ebx,eax
      inc esi
      dec ecx
      jnz @pixels
   @endofline:
      jmp @foriloop
   @endofiloop:
      mov edi,[bsp]
      cmp [_fclc+edi],0
      jnz @alreadycalced
      push edi
      call CalcBasePoly
      mov edi,[bsp]
   @alreadycalced:
      {Calculate start for first line}
      mov eax,[_ylow]
      sub eax,YCENTR
      fld [_qdiv+500*4+eax*4]
      fmul [_fuj+edi]; fld  [_fVy+edi]; fmul [_fx0+edi]; faddp
      fld  [_fVx+edi]; fmul [_fy0+edi]; fsubp; fstp [xxs]
      fld [_qdiv+500*4+eax*4]
      fmul [_fvj+edi]; fld  [_fUx+edi]; fmul [_fy0+edi]; faddp
      fld [_fUy+edi];  fmul [_fx0+edi]; fsubp; fstp [yys]
      fld [_qdiv+500*4+eax*4]
      fmul [_fzj+edi]; fld  [_fVx+edi]; fmul [_fUy+edi]; faddp
      fld [_fVy+edi];  fmul [_fUx+edi]; fsubp; fstp [zzs]
   {-----SPAN LOOP-----------}
   @forloop:
      mov eax,[_ylow]
      cmp eax,[_yhigh]
      jge @endofpoly
      inc [_ylow]
      mov ebx,[_edgeb+eax*4]
      mov ecx,[_edgee+eax*4]
      cmp ebx,[_cll+eax*4]
      jge @noleftclip
      mov ebx,[_cll+eax*4]
   @noleftclip:
      cmp ecx,[_clr+eax*4]
      jng @norightclip
      mov ecx,[_clr+eax*4]
   @norightclip:
      mov [ex],ecx
      mov [x1],ebx
      add ebx,[buf]
      mov edi,XSIZE
      imul edi
      add ebx,eax
      mov [a],ebx
      {Calculate first span before entering while-loop}
      mov eax,[x1]
      cmp eax,[ex]
      jge @endofspan
      mov esi,[bsp]
      fld [_qdiv+(500-XCENTR)*4+eax*4]; fmul [_fzi+esi]; fadd [zzs]; fstp [zz]
      fld [_qdiv+(500-XCENTR)*4+eax*4]; fmul [_fvi+esi]; fadd [yys]; fst [yy]
      fld [_qdiv+(500-XCENTR)*4+eax*4]; fmul [_fui+esi]; fadd [xxs]; fst [xx]
      fdiv [zz]; fistp [uo]
      fdiv [zz]; fistp [vo]
      add [x1],TSPAN
      mov eax,[ex]
      cmp [x1],eax
      jnge @x1islower
   @x1isbigger:
      add eax,TSPAN
      sub eax,[x1]
      mov [slen],eax
      mov eax,[ex]
      mov [x1],eax
      jmp @endofthisshit
   @x1islower:
      mov [slen],TSPAN
   @endofthisshit:
      mov eax,[x1]
      mov esi,[bsp]
      fld [_qdiv+(500-XCENTR)*4+eax*4]; fmul [_fzi+esi]; fadd [zzs]; fstp [zz]
      fld [_qdiv+(500-XCENTR)*4+eax*4]; fmul [_fvi+esi]; fadd [yys]; fst [yy]
      fld [_qdiv+(500-XCENTR)*4+eax*4]; fmul [_fui+esi]; fadd [xxs]; fst [xx]
      fdiv [zz]; fistp [u]
      fdiv [zz]; fistp [v]
      cmp [slen],TSPAN
      jz  @defaultlength
   @nodefault:
      mov eax,[u]
      sub eax,[uo]
      cdq
      idiv [slen]
      mov [du],eax
      mov eax,[v]
      sub eax,[vo]
      cdq
      idiv [slen]
      mov [dv],eax
      jmp @endofdeltas
   @defaultlength:
      mov eax,[u]
      sub eax,[uo]
      sar eax,SPANSHFT
      mov [du],eax
      mov eax,[v]
      sub eax,[vo]
      sar eax,SPANSHFT
      mov [dv],eax
   @endofdeltas:
   {------ START OF DO-WHILE LOOP ---------}
      mov edi,[a]
   @dowhile:
      add [x1],TSPAN
      mov eax,[ex]
      add eax,TSPAN
      cmp [x1],eax
      jge @drawlastspan
   @drawaspan:
      mov eax,[ex]
      cmp eax,[x1]
      jge @nextisfull
   @nextispartial:
      add eax,TSPAN
      sub eax,[x1]
      mov [slen],eax
      mov eax,[ex]
      mov [x1],eax
      jmp @endofdelta
   @nextisfull:
      mov [slen],TSPAN
   @endofdelta:
      mov eax,[x1]
      mov esi,[bsp]
      fld [_qdiv+(500-XCENTR)*4+eax*4]; fmul [_fvi+esi]; fadd [yys]; fstp [yy]
      fld [_qdiv+(500-XCENTR)*4+eax*4]; fmul [_fui+esi]; fadd [xxs]; fst [xx]
      fld [_qdiv+(500-XCENTR)*4+eax*4]; fmul [_fzi+esi]; fadd [zzs]; fst [zz]
      fdivp               {ecx is free here - Use for shading}
      mov eax,[uo]
      mov ebx,[vo]
      add edi,TSPAN
      xor edx,edx
      mov esi,[ba]
      mov dh,bh; mov dl,ah; and edx,$7F7F; add eax,[du]; add ebx,[dv]
      mov dl,[edx+esi]; mov [edi-TSPAN],dl
      mov dh,bh; mov dl,ah; and edx,$7F7F; add eax,[du]; add ebx,[dv]
      mov dl,[edx+esi]; mov [edi-TSPAN+1],dl
      mov dh,bh; mov dl,ah; and edx,$7F7F; add eax,[du]; add ebx,[dv]
      mov dl,[edx+esi]; mov [edi-TSPAN+2],dl
      mov dh,bh; mov dl,ah; and edx,$7F7F; add eax,[du]; add ebx,[dv]
      mov dl,[edx+esi]; mov [edi-TSPAN+3],dl
      mov dh,bh; mov dl,ah; and edx,$7F7F; add eax,[du]; add ebx,[dv]
      mov dl,[edx+esi]; mov [edi-TSPAN+4],dl
      mov dh,bh; mov dl,ah; and edx,$7F7F; add eax,[du]; add ebx,[dv]
      mov dl,[edx+esi]; mov [edi-TSPAN+5],dl
      mov dh,bh; mov dl,ah; and edx,$7F7F; add eax,[du]; add ebx,[dv]
      mov dl,[edx+esi]; mov [edi-TSPAN+6],dl
      mov dh,bh; mov dl,ah; and edx,$7F7F; add eax,[du]; add ebx,[dv]
      mov dl,[edx+esi]; mov [edi-TSPAN+7],dl
      fld [yy]
      fdiv [zz]
      mov dh,bh; mov dl,ah; and edx,$7F7F; add eax,[du]; add ebx,[dv]
      mov dl,[edx+esi]; mov [edi-TSPAN+8],dl
      mov dh,bh; mov dl,ah; and edx,$7F7F; add eax,[du]; add ebx,[dv]
      mov dl,[edx+esi]; mov [edi-TSPAN+9],dl
      mov dh,bh; mov dl,ah; and edx,$7F7F; add eax,[du]; add ebx,[dv]
      mov dl,[edx+esi]; mov [edi-TSPAN+10],dl
      mov dh,bh; mov dl,ah; and edx,$7F7F; add eax,[du]; add ebx,[dv]
      mov dl,[edx+esi]; mov [edi-TSPAN+11],dl
      mov dh,bh; mov dl,ah; and edx,$7F7F; add eax,[du]; add ebx,[dv]
      mov dl,[edx+esi]; mov [edi-TSPAN+12],dl
      mov dh,bh; mov dl,ah; and edx,$7F7F; add eax,[du]; add ebx,[dv]
      mov dl,[edx+esi]; mov [edi-TSPAN+13],dl
      mov dh,bh; mov dl,ah; and edx,$7F7F; add eax,[du]; add ebx,[dv]
      mov dl,[edx+esi]; mov [edi-TSPAN+14],dl
      mov dh,bh; mov dl,ah; and edx,$7F7F; add eax,[du]; add ebx,[dv]
      mov dl,[edx+esi]; mov [edi-TSPAN+15],dl
      add [a],TSPAN
      mov eax,[u]
      mov [uo],eax
      mov eax,[v]
      mov [vo],eax
      fistp [v]
      fistp [u]
      cmp [slen],TSPAN
      jz  @defaultspan
   @nodefaultspan:
      mov eax,[u]
      sub eax,[uo]
      cdq
      idiv [slen]
      mov [du],eax
      mov eax,[v]
      sub eax,[vo]
      cdq
      idiv [slen]
      mov [dv],eax
      jmp @endofdcalc
   @defaultspan:
      mov eax,[u]
      sub eax,[uo]
      sar eax,SPANSHFT
      mov [du],eax
      mov eax,[v]
      sub eax,[vo]
      sar eax,SPANSHFT
      mov [dv],eax
   @endofdcalc:
      jmp @dowhile
   @drawlastspan:
      mov ecx,[slen]
      mov eax,[uo]
      mov ebx,[vo]
      xor edx,edx
      mov esi,[ba]
   @ploop:
      mov dh,bh
      mov dl,ah
      and edx,$7F7F
      add eax,[du]
      add ebx,[dv]
      mov dl,[edx+esi]
      mov [edi],dl
      inc edi
      dec ecx
      jnz @ploop
   @endofspan:
      mov esi,[bsp]
      fld [xxs]; fadd [_fxxd+esi]; fstp [xxs]
      fld [yys]; fadd [_fyyd+esi]; fstp [yys]
      fld [zzs]; fadd [_fzzd+esi]; fstp [zzs]
      jmp @forloop
   @endofpoly:
   end;
end;


{  _______________________________________________________________________
  /                                                                      /
 / DrawTPTextPolygon: Draw a transparent perspective correct texturemap /
/______________________________________________________________________/ }

procedure DrawTPTextPolygon(poly: longint);
var x3,y3,x4,y4,t: single;
    z3,z4,_ylow,_yhigh: longint;
    i,_tn,oldcrds,p1,p2,pp1,pp2,pp3: longint;
    x1,y1,x2,y2,ixh,ixl: longint;
    tx,ty,tz: single;
    u,v,uo,vo: longint;
    x0,y0,z0: single;
    vx,vy,vz,ux,uy,uz: single;
    ui,uj,vi,vj,zi,zj: single;
    xxs,yys,zzs: single;                 {start per scanline}
    xxp,yyp,zzp: single;                 {delta per sub span}
    xxd,yyd,zzd: single;                 {delta per scanline}
    store,xx,yy,zz: single;
    a: dword;
    du,dv,l,olen,slen,ex: longint;
    bsp,ba: longint;
begin
   poly:=poly+1;
   pp1:=_pp[poly,1]; {should be positioned BEFORE z-clipper}
   pp2:=_pp[poly,2];
   pp3:=_pp[poly,4];
   {z-clipping: Make sure that the polygon is not partially behind viewer}
   _tn:=0;
   oldcrds:=_coords;
   for i:=1 to _pnv[poly] do begin
      {Determine begin and end of one of the polygon's edges}
      p1:=_pp[poly,i];
      z3:=_rz[p1];
      if i<_pnv[poly] then p2:=_pp[poly,i+1] else p2:=_pp[poly,1];
      z4:=_rz[p2];
      {Case 1: Both in, output end coord}
      if (z3>_zmin) and (z4>_zmin) then begin
         inc(_tn);
         _tmp[_tn]:=p2
      end else if (z3>_zmin) and (z4<=_zmin) then begin
      {Case 2: Going in: Output clipped coord}
         x3:=_rx[p1];
         y3:=_ry[p1];
         x4:=_rx[p2];
         y4:=_ry[p2];
         t:=(_zmin-_frz[p1])/(_frz[p2]-_frz[p1]);
         inc(_coords);
         tx:=x3+(x4-x3)*t;
         ty:=y3+(y4-y3)*t;
         _xs[_coords]:=round((tx*DIST*64)/_zmin)+XCENTR*64;
         _ys[_coords]:=round((ty*DIST*64)/_zmin)+YCENTR*64;
         inc(_tn);
         _tmp[_tn]:=_coords;
      end else if (z3<=_zmin) and (z4>_zmin) then begin
         {Case 3: Coming out: Output clipped coord and end coord}
         x3:=_rx[p1];
         y3:=_ry[p1];
         x4:=_rx[p2];
         y4:=_ry[p2];
         t:=(_zmin-_frz[p1])/(_frz[p2]-_frz[p1]);
         inc(_coords);
         tx:=x3+(x4-x3)*t;
         ty:=y3+(y4-y3)*t;
         _xs[_coords]:=round((tx*DIST*64)/_zmin)+XCENTR*64;
         _ys[_coords]:=round((ty*DIST*64)/_zmin)+YCENTR*64;
         inc(_tn,2);
         _tmp[_tn-1]:=_coords;
         _tmp[_tn]:=p2;
      end;
   end;  {_tmp is array with coords for polygon, _tn is number of vertices}
   _coords:=oldcrds;
   ba:=_tba[_pc[poly]];
   bsp:=(_pfb[poly]-1) shl 2;
   asm
      mov ebx,[_tn]
      cmp ebx,3
      jb  @endofpoly
      dec ebx
      mov esi,dword ptr [_tmp]
      mov edi,[_ys+esi*4-4]
      mov ecx,edi
   @findloop:
      mov esi,[_tmp+ebx*4]
      mov eax,[_ys+esi*4-4]
      cmp eax,edi
      jge @nolowupdate
      mov edi,eax
   @nolowupdate:
      cmp eax,ecx
      jnge @nohiupdate
      mov ecx,eax
   @nohiupdate:
      dec ebx
      jnz @findloop
      cmp edi,[_clu]
      jge @nozeroclip
      mov edi,[_clu]
   @nozeroclip:
      cmp ecx,[_clu]
      jnge @endofpoly
      cmp edi,[_clb]
      jge @endofpoly
      cmp ecx,[_clb]
      jnge @nobotclip
      mov ecx,[_clb]
   @nobotclip:
      sar ecx,6
      sar edi,6
      cmp ecx,edi
      jbe @endofpoly
      mov [_ylow],edi
      mov [_yhigh],ecx
      sub ecx,edi
      xor eax,eax
      mov ebx,XSIZE-1
   @clearloop:
      mov [_edgeb+edi*4],ebx
      mov [_edgee+edi*4],eax
      inc edi
      dec ecx
      jnz @clearloop
      mov [i],0
   @foriloop:
      inc [i]
      {Get x1/y1/x2/y2, wrap around for last vertex}
      mov esi,[i]
      cmp esi,[_tn]
      ja  @endofiloop
      mov edi,[_tmp+4*esi-4]
      jz  @1
      mov esi,[_tmp+4*esi]
      jmp @n
   @1: mov esi,dword ptr [_tmp]
   @n: mov edx,[_xs+4*edi-4]  {hehe not a single AGI stall}
      mov eax,[_ys+4*edi-4]
      mov ebx,[_xs+4*esi-4]
      mov ecx,[_ys+4*esi-4]
      mov [x1],edx
      mov [y1],eax
      mov [x2],ebx
      mov [y2],ecx
      cmp eax,ecx
      jz  @endofline
      jnge @noswap
      mov ecx,[y1]
      mov eax,[y2]
      mov edx,[x2]
      mov ebx,[x1]
      mov [x1],edx
      mov [x2],ebx        {registers: eax = y1  }
      mov [y1],eax                {   ebx = x2  }
      mov [y2],ecx                {   ecx = y2  }
   @noswap:                       {   edx = x1  }
      cmp eax,[_clb]
      jge @endofline
      cmp ecx,[_clu]
      jnge @endofline
      cmp eax,ecx
      jge @endofline      {if (y1<YSIZE) and (y2>0) and (y2>y1)}
      cmp eax,[_clu]
      jge @notopclipping  {if y1<0 then begin}
      sub ecx,[y1]
      xor eax,-1
      add eax,[_clu]
      sal eax,8
      cdq; idiv ecx
      mov esi,[_clu]
      mov [y1],esi
      mov esi,[x2]
      sub esi,[x1]
      imul esi
      sar eax,8
      add [x1],eax
      mov ecx,[y2]
   @notopclipping:
      {Fixed point line drawing}
      sub ecx,[y1]        {ecx = delta-y}
      jz  @endofline
      mov eax,[x2]
      sub eax,[x1]        {eax = delta-x}
      sal eax,10
      cdq
      idiv ecx            {eax = delta-x/delta-y, f.p. 10bit}
      mov esi,[y1]        {pointer in outline table}
      sar esi,6
      mov ebx,[y2]
      cmp ebx,[_clb]
      jnge @drawline      {clipping to bottom of screen needed?}
      sub ebx,[_clb]
      sub ecx,ebx         {then do it}
   @drawline:             {line is OK, draw it}
      shr ecx,6
      inc ecx
      or  ecx,ecx
      jz  @endofline
      mov ebx,[x1]
      sal ebx,10-6
   @pixels:
      mov edx,ebx
      sar edx,10
      cmp edx,[_edgeb+esi*4]
      jge @nobupdate
      mov [_edgeb+esi*4],edx
   @nobupdate:
      cmp edx,[_edgee+esi*4]
      jnge @noeupdate
      mov [_edgee+esi*4],edx
   @noeupdate:
      add ebx,eax
      inc esi
      dec ecx
      jnz @pixels
   @endofline:
      jmp @foriloop
   @endofiloop:
      mov edi,[bsp]
      cmp [_fclc+edi],0
      jnz @alreadycalced
      push edi
      call CalcBasePoly
      mov edi,[bsp]
   @alreadycalced:
      {Calculate start for first line}
      mov eax,[_ylow]
      sub eax,YCENTR
      fld [_qdiv+500*4+eax*4]
      fmul [_fuj+edi]; fld  [_fVy+edi]; fmul [_fx0+edi]; faddp
      fld  [_fVx+edi]; fmul [_fy0+edi]; fsubp; fstp [xxs]
      fld [_qdiv+500*4+eax*4]
      fmul [_fvj+edi]; fld  [_fUx+edi]; fmul [_fy0+edi]; faddp
      fld [_fUy+edi];  fmul [_fx0+edi]; fsubp; fstp [yys]
      fld [_qdiv+500*4+eax*4]
      fmul [_fzj+edi]; fld  [_fVx+edi]; fmul [_fUy+edi]; faddp
      fld [_fVy+edi];  fmul [_fUx+edi]; fsubp; fstp [zzs]
   {-----SPAN LOOP-----------}
   @forloop:
      mov eax,[_ylow]
      cmp eax,[_yhigh]
      jge @endofpoly
      inc [_ylow]
      mov ebx,[_edgeb+eax*4]
      mov ecx,[_edgee+eax*4]
      cmp ebx,[_cll+eax*4]
      jge @noleftclip
      mov ebx,[_cll+eax*4]
   @noleftclip:
      cmp ecx,[_clr+eax*4]
      jng @norightclip
      mov ecx,[_clr+eax*4]
   @norightclip:
      mov [ex],ecx
      mov [x1],ebx
      add ebx,[buf]
      mov edi,XSIZE
      imul edi
      add ebx,eax
      mov [a],ebx
      {Calculate first span before entering while-loop}
      mov eax,[x1]
      cmp eax,[ex]
      jge @endofspan
      mov esi,[bsp]
      fld [_qdiv+(500-XCENTR)*4+eax*4]; fmul [_fzi+esi]; fadd [zzs]; fstp [zz]
      fld [_qdiv+(500-XCENTR)*4+eax*4]; fmul [_fvi+esi]; fadd [yys]; fst [yy]
      fld [_qdiv+(500-XCENTR)*4+eax*4]; fmul [_fui+esi]; fadd [xxs]; fst [xx]
      fdiv [zz]; fistp [uo]
      fdiv [zz]; fistp [vo]
      add [x1],TSPAN
      mov eax,[ex]
      cmp [x1],eax
      jnge @x1islower
   @x1isbigger:
      add eax,TSPAN
      sub eax,[x1]
      mov [slen],eax
      mov eax,[ex]
      mov esi,[bsp]
      fld [_qdiv+(500-XCENTR)*4+eax*4]; fmul [_fzi+esi]; fadd [zzs]; fstp [zz]
      fld [_qdiv+(500-XCENTR)*4+eax*4]; fmul [_fvi+esi]; fadd [yys]; fst [yy]
      fld [_qdiv+(500-XCENTR)*4+eax*4]; fmul [_fui+esi]; fadd [xxs]; fst [xx]
      jmp @endofthisshit
   @x1islower:
      mov [slen],TSPAN
      mov esi,[bsp]
      fld [zz]; fadd [_fzzp+esi]; fstp [zz]
      fld [yy]; fadd [_fyyp+esi]; fst [yy]
      fld [xx]; fadd [_fxxp+esi]; fst [xx]
   @endofthisshit:
      fdiv [zz]; fistp [u]
      fdiv [zz]; fistp [v]
      cmp [slen],TSPAN
      jz  @defaultlength
   @nodefault:
      mov eax,[u]
      sub eax,[uo]
      cdq
      idiv [slen]
      mov [du],eax
      mov eax,[v]
      sub eax,[vo]
      cdq
      idiv [slen]
      mov [dv],eax
      jmp @endofdeltas
   @defaultlength:
      mov eax,[u]
      sub eax,[uo]
      sar eax,SPANSHFT
      mov [du],eax
      mov eax,[v]
      sub eax,[vo]
      sar eax,SPANSHFT
      mov [dv],eax
   @endofdeltas:
   {------ START OF DO-WHILE LOOP ---------}
      mov edi,[a]
   @dowhile:
      add [x1],TSPAN
      mov eax,[ex]
      add eax,TSPAN
      cmp [x1],eax
      jge @drawlastspan
   @drawaspan:
      mov eax,[ex]
      cmp eax,[x1]
      jge @nextisfull
   @nextispartial:
      add eax,TSPAN
      sub eax,[x1]
      mov [slen],eax
      mov eax,[ex]
      mov [x1],eax
      mov eax,[x1]
      mov esi,[bsp]
      fld [_qdiv+(500-XCENTR)*4+eax*4]; fmul [_fvi+esi]; fadd [yys]; fstp [yy]
      fld [_qdiv+(500-XCENTR)*4+eax*4]; fmul [_fui+esi]; fadd [xxs]; fst [xx]
      fld [_qdiv+(500-XCENTR)*4+eax*4]; fmul [_fzi+esi]; fadd [zzs]; fst [zz]
      jmp @endofdelta
   @nextisfull:
      mov [slen],TSPAN
      mov eax,[x1]
      mov esi,[bsp]
      fld [yy]; fadd [_fyyp+esi]; fstp [yy]
      fld [xx]; fadd [_fxxp+esi]; fst [xx]
      fld [zz]; fadd [_fzzp+esi]; fst [zz]
   @endofdelta:
      fdivp
      mov eax,[uo]
      mov ebx,[vo]
      add edi,TSPAN
      xor edx,edx
      mov esi,[ba]
      mov ecx,[mix]
      mov dh,bh; mov dl,ah; and edx,$7F7F; add eax,[du]; add ebx,[dv]
      mov dl,[edx+esi]; mov dh,[edi-TSPAN]; mov dl,[ecx+edx]; mov [edi-TSPAN],dl
      mov dh,bh; mov dl,ah; and edx,$7F7F; add eax,[du]; add ebx,[dv]
      mov dl,[edx+esi]; mov dh,[edi-TSPAN+1]; mov dl,[ecx+edx]; mov [edi-TSPAN+1],dl
      mov dh,bh; mov dl,ah; and edx,$7F7F; add eax,[du]; add ebx,[dv]
      mov dl,[edx+esi]; mov dh,[edi-TSPAN+2]; mov dl,[ecx+edx]; mov [edi-TSPAN+2],dl
      mov dh,bh; mov dl,ah; and edx,$7F7F; add eax,[du]; add ebx,[dv]
      mov dl,[edx+esi]; mov dh,[edi-TSPAN+3]; mov dl,[ecx+edx]; mov [edi-TSPAN+3],dl
      mov dh,bh; mov dl,ah; and edx,$7F7F; add eax,[du]; add ebx,[dv]
      mov dl,[edx+esi]; mov dh,[edi-TSPAN+4]; mov dl,[ecx+edx]; mov [edi-TSPAN+4],dl
      mov dh,bh; mov dl,ah; and edx,$7F7F; add eax,[du]; add ebx,[dv]
      mov dl,[edx+esi]; mov dh,[edi-TSPAN+5]; mov dl,[ecx+edx]; mov [edi-TSPAN+5],dl
      mov dh,bh; mov dl,ah; and edx,$7F7F; add eax,[du]; add ebx,[dv]
      mov dl,[edx+esi]; mov dh,[edi-TSPAN+6]; mov dl,[ecx+edx]; mov [edi-TSPAN+6],dl
      mov dh,bh; mov dl,ah; and edx,$7F7F; add eax,[du]; add ebx,[dv]
      mov dl,[edx+esi]; mov dh,[edi-TSPAN+7]; mov dl,[ecx+edx]; mov [edi-TSPAN+7],dl
      fld [yy]
      fdiv [zz]
      mov dh,bh; mov dl,ah; and edx,$7F7F; add eax,[du]; add ebx,[dv]
      mov dl,[edx+esi]; mov dh,[edi-TSPAN+8]; mov dl,[ecx+edx]; mov [edi-TSPAN+8],dl
      mov dh,bh; mov dl,ah; and edx,$7F7F; add eax,[du]; add ebx,[dv]
      mov dl,[edx+esi]; mov dh,[edi-TSPAN+9]; mov dl,[ecx+edx]; mov [edi-TSPAN+9],dl
      mov dh,bh; mov dl,ah; and edx,$7F7F; add eax,[du]; add ebx,[dv]
      mov dl,[edx+esi]; mov dh,[edi-TSPAN+10]; mov dl,[ecx+edx]; mov [edi-TSPAN+10],dl
      mov dh,bh; mov dl,ah; and edx,$7F7F; add eax,[du]; add ebx,[dv]
      mov dl,[edx+esi]; mov dh,[edi-TSPAN+11]; mov dl,[ecx+edx]; mov [edi-TSPAN+11],dl
      mov dh,bh; mov dl,ah; and edx,$7F7F; add eax,[du]; add ebx,[dv]
      mov dl,[edx+esi]; mov dh,[edi-TSPAN+12]; mov dl,[ecx+edx]; mov [edi-TSPAN+12],dl
      mov dh,bh; mov dl,ah; and edx,$7F7F; add eax,[du]; add ebx,[dv]
      mov dl,[edx+esi]; mov dh,[edi-TSPAN+13]; mov dl,[ecx+edx]; mov [edi-TSPAN+13],dl
      mov dh,bh; mov dl,ah; and edx,$7F7F; add eax,[du]; add ebx,[dv]
      mov dl,[edx+esi]; mov dh,[edi-TSPAN+14]; mov dl,[ecx+edx]; mov [edi-TSPAN+14],dl
      mov dh,bh; mov dl,ah; and edx,$7F7F; add eax,[du]; add ebx,[dv]
      mov dl,[edx+esi]; mov dh,[edi-TSPAN+15]; mov dl,[ecx+edx]; mov [edi-TSPAN+15],dl
      add [a],TSPAN
      mov eax,[u]
      mov [uo],eax
      mov eax,[v]
      mov [vo],eax
      fistp [v]
      fistp [u]
      cmp [slen],TSPAN
      jz  @defaultspan
   @nodefaultspan:
      mov eax,[u]
      sub eax,[uo]
      cdq
      idiv [slen]
      mov [du],eax
      mov eax,[v]
      sub eax,[vo]
      cdq
      idiv [slen]
      mov [dv],eax
      jmp @endofdcalc
   @defaultspan:
      mov eax,[u]
      sub eax,[uo]
      sar eax,SPANSHFT
      mov [du],eax
      mov eax,[v]
      sub eax,[vo]
      sar eax,SPANSHFT
      mov [dv],eax
   @endofdcalc:
      jmp @dowhile
   @drawlastspan:
      mov ecx,[mix]
      mov eax,[uo]
      mov ebx,[vo]
      xor edx,edx
      mov esi,[ba]
   @ploop:
      mov dh,bh
      mov dl,ah
      and edx,$7F7F
      add eax,[du]
      add ebx,[dv]
      mov dl,[edx+esi]
      mov dh,[edi]
      mov dl,[ecx+edx]
      mov [edi],dl
      inc edi
      dec [slen]
      jnz @ploop
   @endofspan:
      mov esi,[bsp]
      fld [xxs]; fadd [_fxxd+esi]; fstp [xxs]
      fld [yys]; fadd [_fyyd+esi]; fstp [yys]
      fld [zzs]; fadd [_fzzd+esi]; fstp [zzs]
      jmp @forloop
   @endofpoly:
   end;
end;


{  _______________________________________________________________________
  /                                                                      /
 / DrawGPTextPolygon: Draw a gourad shaded perspective texturemap       /
/______________________________________________________________________/ }

procedure DrawGPTextPolygon(poly: longint);
var x3,y3,x4,y4,t: single;
    z3,z4,_ylow,_yhigh: longint;
    i,_tn,oldcrds,p1,p2,pp1,pp2,pp3: longint;
    x1,y1,x2,y2,ixh,ixl: longint;
    tx,ty,tz: single;
    u,v,uo,vo: longint;
    x0,y0,z0: single;
    vx,vy,vz,ux,uy,uz: single;
    ui,uj,vi,vj,zi,zj: single;
    xxs,yys,zzs: single;                 {start per scanline}
    xxp,yyp,zzp: single;                 {delta per sub span}
    xxd,yyd,zzd: single;                 {delta per scanline}
    store,xx,yy,zz: single;
    a: dword;
    du,dv,l,olen,slen,ex: longint;
    ba,bsp: longint;
    gval,pixl,px1,px2,dpx,dlx,gdl,bck: longint;
begin
   poly:=poly+1;
   pp1:=_pp[poly,1];
   pp2:=_pp[poly,2];
   pp3:=_pp[poly,4];
   {z-clipping: Make sure that the polygon is not partially behind viewer}
   _tn:=0;
   oldcrds:=_coords;
   for i:=1 to _pnv[poly] do begin
      {Determine begin and end of one of the polygon's edges}
      p1:=_pp[poly,i];
      z3:=_rz[p1];
      if i<_pnv[poly] then p2:=_pp[poly,i+1] else p2:=_pp[poly,1];
      z4:=_rz[p2];
      {Case 1: Both in, output end coord}
      if (z3>_zmin) and (z4>_zmin) then begin
         inc(_tn);
         _tmp[_tn]:=p2
      end else if (z3>_zmin) and (z4<=_zmin) then begin
      {Case 2: Going in: Output clipped coord}
         x3:=_rx[p1];
         y3:=_ry[p1];
         x4:=_rx[p2];
         y4:=_ry[p2];
         px1:=_vx[p1];                          {Phong coords: Vertex normals}
         px2:=_vx[p2];
         t:=(_zmin-_frz[p1])/(_frz[p2]-_frz[p1]);
         inc(_coords);
         tx:=x3+(x4-x3)*t;            {Clipping of polygon outline}
         ty:=y3+(y4-y3)*t;
         _vx[_coords]:=px1+round((px2-px1)*t);  {Clipping of phong line}
         _xs[_coords]:=round((tx*DIST*64)/_zmin)+XCENTR*64;
         _ys[_coords]:=round((ty*DIST*64)/_zmin)+YCENTR*64;
         inc(_tn);
         _tmp[_tn]:=_coords;
      end else if (z3<=_zmin) and (z4>_zmin) then begin
         {Case 3: Coming out: Output clipped coord and end coord}
         x3:=_rx[p1];
         y3:=_ry[p1];
         x4:=_rx[p2];
         y4:=_ry[p2];
         px1:=_vx[p1];
         px2:=_vx[p2];
         t:=(_zmin-_frz[p1])/(_frz[p2]-_frz[p1]);
         inc(_coords);
         tx:=x3+(x4-x3)*t;            {Clipping of polygon outline}
         ty:=y3+(y4-y3)*t;
         _vx[_coords]:=px1+round((px2-px1)*t);  {Clipping of Phong line}
         _xs[_coords]:=round((tx*DIST*64)/_zmin)+XCENTR*64;
         _ys[_coords]:=round((ty*DIST*64)/_zmin)+YCENTR*64;
         _tmp[_tn+1]:=_coords;
         inc(_tn,2);
         _tmp[_tn]:=p2;
      end;
   end;  {_tmp is array with coords for polygon, _tn is number of vertices}
   _coords:=oldcrds;
   ba:=_tba[_pc[poly]];
   bsp:=(_pfb[poly]-1) shl 2;
   asm
      mov ebx,[_tn]
      cmp ebx,3
      jb  @endofpoly
      dec ebx
      mov esi,dword ptr [_tmp]
      mov edi,[_ys+esi*4-4]
      mov ecx,edi
   @findloop:
      mov esi,[_tmp+ebx*4]
      mov eax,[_ys+esi*4-4]
      cmp eax,edi
      jge @nolowupdate
      mov edi,eax
   @nolowupdate:
      cmp eax,ecx
      jnge @nohiupdate
      mov ecx,eax
   @nohiupdate:
      dec ebx
      jnz @findloop
      cmp edi,[_clu]
      jge @nozeroclip
      mov edi,[_clu]
   @nozeroclip:
      cmp ecx,[_clu]
      jnge @endofpoly
      cmp edi,[_clb]
      jge @endofpoly
      cmp ecx,[_clb]
      jnge @nobotclip
      mov ecx,[_clb]
   @nobotclip:
      sar ecx,6
      sar edi,6
      cmp ecx,edi
      jbe @endofpoly
      mov [_ylow],edi
      mov [_yhigh],ecx
      sub ecx,edi
      xor eax,eax
      mov ebx,XSIZE-1
   @clearloop:
      mov [_edgeb+edi*4],ebx
      mov [_edgee+edi*4],eax
      inc edi
      dec ecx
      jnz @clearloop
      mov [i],0
   @foriloop:
      inc [i]
      {Get x1/y1/x2/y2, wrap around for last vertex}
      mov esi,[i]
      cmp esi,[_tn]
      ja  @endofiloop
      mov edi,[_tmp+4*esi-4]
      jz  @1
      mov esi,[_tmp+4*esi]
      jmp @n
   @1: mov esi,dword ptr [_tmp]
   @n: mov eax,[_vx+4*edi-4]
      mov ecx,[_vx+4*esi-4]
      add eax,128*32
      add ecx,128*32
      sar eax,2
      sar ecx,2
      mov [px1],eax
      mov [px2],ecx
      mov edx,[_xs+4*edi-4]  {hehe not a single AGI stall}
      mov eax,[_ys+4*edi-4]
      mov ebx,[_xs+4*esi-4]
      mov ecx,[_ys+4*esi-4]
      mov [x1],edx
      mov [y1],eax
      mov [x2],ebx
      mov [y2],ecx
      cmp eax,ecx
      jz  @endofline
      jnge @noswap
      mov eax,[px1]
      mov ebx,[px2]
      mov [px1],ebx
      mov [px2],eax
      mov ecx,[y1]
      mov eax,[y2]
      mov edx,[x2]
      mov ebx,[x1]
      mov [x1],edx
      mov [x2],ebx        {registers: eax = y1  }
      mov [y1],eax                {   ebx = x2  }
      mov [y2],ecx                {   ecx = y2  }
   @noswap:                       {   edx = x1  }
      cmp eax,[_clb]
      jge @endofline
      cmp ecx,[_clu]
      jnge @endofline
      cmp eax,ecx
      jge @endofline      {if (y1<YSIZE) and (y2>0) and (y2>y1)}
      cmp eax,[_clu]
      jge @notopclipping  {if y1<0 then begin}
      sub ecx,[y1]
      xor eax,-1
      add eax,[_clu]
      sal eax,8
      cdq; idiv ecx
      mov ecx,eax
      mov esi,[_clu]
      mov [y1],esi
      mov esi,[x2]        {t:=-y1/(y2-y1)}
      sub esi,[x1]        {x1:=x1+(x2-x1)*t}
      imul esi
      sar eax,8
      add [x1],eax
      mov eax,ecx         {Clipping of gourad line}
      mov esi,[px2]
      sub esi,[px1]
      imul esi
      sar eax,8
      add [px1],eax
      mov ecx,[y2]
   @notopclipping:
      {Fixed point line drawing}
      sub ecx,[y1]        {ecx = delta-y}
      jz  @endofline
      mov eax,[px2]
      sub eax,[px1]
      sal eax,8
      cdq
      idiv ecx
      mov [dpx],eax       {dpx=gourad delta-x, f.p. 10bit}
      mov eax,[x2]
      sub eax,[x1]        {eax = delta-x}
      sal eax,10
      cdq
      idiv ecx
      mov [dlx],eax       {dlx = delta-x/delta-y, f.p. 10bit}
      mov esi,[y1]        {pointer in outline table}
      sar esi,6
      mov ebx,[y2]
      cmp ebx,[_clb]
      jnge @drawline      {clipping to bottom of screen needed?}
      sub ebx,[_clb]
      sub ecx,ebx         {then do it}
   @drawline:             {line is OK, draw it}
      shr ecx,6
      inc ecx
      or  ecx,ecx
      jz  @endofline
      mov ebx,[x1]        {line start x for this vertex}
      sal ebx,4
      mov edi,[px1]       {gourad start value}
      sal edi,2
   @pixels:
      mov edx,ebx
      sar edx,10
      cmp edx,[_edgeb+esi*4]
      jge @nobupdate
      mov [_edgeb+esi*4],edx
      mov [_edgxb+esi*4],edi
   @nobupdate:
      cmp edx,[_edgee+esi*4]
      jnge @noeupdate
      mov [_edgee+esi*4],edx
      mov [_edgxe+esi*4],edi
   @noeupdate:
      add ebx,[dlx]     {interpolate line slope}
      add edi,[dpx]     {interpolate gourad value}
      inc esi
      dec ecx
      jnz @pixels
   @endofline:
      jmp @foriloop
   @endofiloop:
      mov edi,[bsp]
      cmp [_fclc+edi],0
      jnz @alreadycalced
      push edi
      call CalcBasePoly
      mov edi,[bsp]
   @alreadycalced:
      {Calculate start for first line}
      mov eax,[_ylow]
      sub eax,YCENTR
      fld [_qdiv+500*4+eax*4]
      fmul [_fuj+edi]; fld  [_fVy+edi]; fmul [_fx0+edi]; faddp
      fld  [_fVx+edi]; fmul [_fy0+edi]; fsubp; fstp [xxs]
      fld [_qdiv+500*4+eax*4]
      fmul [_fvj+edi]; fld  [_fUx+edi]; fmul [_fy0+edi]; faddp
      fld [_fUy+edi];  fmul [_fx0+edi]; fsubp; fstp [yys]
      fld [_qdiv+500*4+eax*4]
      fmul [_fzj+edi]; fld  [_fVx+edi]; fmul [_fUy+edi]; faddp
      fld [_fVy+edi];  fmul [_fUx+edi]; fsubp; fstp [zzs]
   {-----SPAN LOOP-----------}
   @forloop:
      mov edi,[_ylow]
      cmp edi,[_yhigh]
      jge @endofpoly
      inc [_ylow]
      {First check if line is completely out of screen}
      mov eax,[_edgeb+edi*4]
      cmp eax,[_clr+edi*4]
      jge @endofspan
      mov ebx,[_edgee+edi*4]
      cmp ebx,[_cll+edi*4]
      jnge @endofspan
      cmp eax,ebx
      jge @endofspan
      {Calculate Gourad delta's before clipping}
      sub ebx,eax
      mov ecx,[_edgxb+edi*4]
      mov [px1],ecx
      mov eax,[_edgxe+edi*4]
      sub eax,ecx
      cdq
      idiv ebx
      mov [gdl],eax
      mov eax,edi
      mov ebx,[_edgeb+eax*4]
      mov ecx,[_edgee+eax*4]
      cmp ebx,[_cll+eax*4]
      jge @noleftclip
      mov edi,eax
      xor ebx,-1
      add ebx,[_cll+edi*4]
      sal ebx,8
      mov eax,ebx
      mov ebx,[_edgee+edi*4]
      sub ebx,[_edgeb+edi*4]
      cdq
      idiv ebx
      mov ebx,[_edgxe+edi*4]
      sub ebx,[_edgxb+edi*4]
      imul ebx
      sar eax,8
      add [px1],eax
      mov eax,edi
      mov ebx,[_cll+eax*4]
   @noleftclip:
      cmp ecx,[_clr+eax*4]
      jnge @norightclip
      mov ecx,[_clr+eax*4]
   @norightclip:
      mov [ex],ecx
      mov [x1],ebx
      add ebx,[buf]
      mov edi,XSIZE
      imul edi
      add ebx,eax
      mov [a],ebx
      {Calculate first span before entering while-loop}
      mov eax,[x1]
      cmp eax,[ex]
      jge @endofspan
      mov esi,[bsp]
      fld [_qdiv+(500-XCENTR)*4+eax*4]; fmul [_fzi+esi]; fadd [zzs]; fstp [zz]
      fld [_qdiv+(500-XCENTR)*4+eax*4]; fmul [_fvi+esi]; fadd [yys]; fst [yy]
      fld [_qdiv+(500-XCENTR)*4+eax*4]; fmul [_fui+esi]; fadd [xxs]; fst [xx]
      fdiv [zz]; fistp [uo]
      fdiv [zz]; fistp [vo]
      add [x1],TSPAN
      mov eax,[ex]
      cmp [x1],eax
      jnge @x1islower
   @x1isbigger:
      add eax,TSPAN
      sub eax,[x1]
      mov [slen],eax
      mov eax,[ex]
      mov esi,[bsp]
      fld [_qdiv+(500-XCENTR)*4+eax*4]; fmul [_fzi+esi]; fadd [zzs]; fstp [zz]
      fld [_qdiv+(500-XCENTR)*4+eax*4]; fmul [_fvi+esi]; fadd [yys]; fst [yy]
      fld [_qdiv+(500-XCENTR)*4+eax*4]; fmul [_fui+esi]; fadd [xxs]; fst [xx]
      jmp @endofthisshit
   @x1islower:
      mov [slen],TSPAN
      mov esi,[bsp]
      fld [zz]; fadd [_fzzp+esi]; fstp [zz]
      fld [yy]; fadd [_fyyp+esi]; fst [yy]
      fld [xx]; fadd [_fxxp+esi]; fst [xx]
   @endofthisshit:
      fdiv [zz]; fistp [u]
      fdiv [zz]; fistp [v]
      cmp [slen],TSPAN
      jz  @defaultlength
   @nodefault:
      mov eax,[u]
      sub eax,[uo]
      cdq
      idiv [slen]
      mov [du],eax
      mov eax,[v]
      sub eax,[vo]
      cdq
      idiv [slen]
      mov [dv],eax
      jmp @endofdeltas
   @defaultlength:
      mov eax,[u]
      sub eax,[uo]
      sar eax,SPANSHFT
      mov [du],eax
      mov eax,[v]
      sub eax,[vo]
      sar eax,SPANSHFT
      mov [dv],eax
   @endofdeltas:
   {------ START OF DO-WHILE LOOP ---------}
      mov edi,[a]
      mov ecx,[px1]
      mov [gval],ecx
      mov ecx,[gdl]
   @dowhile:
      add [x1],TSPAN
      mov eax,[ex]
      add eax,TSPAN
      cmp [x1],eax
      jge @drawlastspan
   @drawaspan:
      mov eax,[ex]
      cmp eax,[x1]
      jge @nextisfull
   @nextispartial:
      add eax,TSPAN
      sub eax,[x1]
      mov [slen],eax
      mov eax,[ex]
      mov [x1],eax
      mov eax,[x1]
      mov esi,[bsp]
      fld [_qdiv+(500-XCENTR)*4+eax*4]; fmul [_fvi+esi]; fadd [yys]; fstp [yy]
      fld [_qdiv+(500-XCENTR)*4+eax*4]; fmul [_fui+esi]; fadd [xxs]; fst [xx]
      fld [_qdiv+(500-XCENTR)*4+eax*4]; fmul [_fzi+esi]; fadd [zzs]; fst [zz]
      jmp @endofdelta
   @nextisfull:
      mov [slen],TSPAN
      mov eax,[x1]
      mov esi,[bsp]
      fld [yy]; fadd [_fyyp+esi]; fstp [yy]
      fld [xx]; fadd [_fxxp+esi]; fst [xx]
      fld [zz]; fadd [_fzzp+esi]; fst [zz]
   @endofdelta:
      fdivp
      mov eax,[uo]
      mov ebx,[vo]
      add edi,TSPAN
      xor edx,edx
      mov esi,[ba]
      mov dh,bh; mov dl,ah; and edx,$7F7F; add eax,[du]; add ebx,[dv]
      mov dh,[edx+esi]; mov dl,byte ptr [gval+1]; add [gval],ecx
      mov dl,byte ptr[lit+edx]; mov [edi-TSPAN],dl
      mov dh,bh; mov dl,ah; and edx,$7F7F; add eax,[du]; add ebx,[dv]
      mov dh,[edx+esi]; mov dl,byte ptr [gval+1]; add [gval],ecx
      mov dl,byte ptr[lit+edx]; mov [edi-TSPAN+1],dl
      mov dh,bh; mov dl,ah; and edx,$7F7F; add eax,[du]; add ebx,[dv]
      mov dh,[edx+esi]; mov dl,byte ptr [gval+1]; add [gval],ecx
      mov dl,byte ptr[lit+edx]; mov [edi-TSPAN+2],dl
      mov dh,bh; mov dl,ah; and edx,$7F7F; add eax,[du]; add ebx,[dv]
      mov dh,[edx+esi]; mov dl,byte ptr [gval+1]; add [gval],ecx
      mov dl,byte ptr[lit+edx]; mov [edi-TSPAN+3],dl
      mov dh,bh; mov dl,ah; and edx,$7F7F; add eax,[du]; add ebx,[dv]
      mov dh,[edx+esi]; mov dl,byte ptr [gval+1]; add [gval],ecx
      mov dl,byte ptr[lit+edx]; mov [edi-TSPAN+4],dl
      mov dh,bh; mov dl,ah; and edx,$7F7F; add eax,[du]; add ebx,[dv]
      mov dh,[edx+esi]; mov dl,byte ptr [gval+1]; add [gval],ecx
      mov dl,byte ptr[lit+edx]; mov [edi-TSPAN+5],dl
      mov dh,bh; mov dl,ah; and edx,$7F7F; add eax,[du]; add ebx,[dv]
      mov dh,[edx+esi]; mov dl,byte ptr [gval+1]; add [gval],ecx
      mov dl,byte ptr[lit+edx]; mov [edi-TSPAN+6],dl
      mov dh,bh; mov dl,ah; and edx,$7F7F; add eax,[du]; add ebx,[dv]
      mov dh,[edx+esi]; mov dl,byte ptr [gval+1]; add [gval],ecx
      mov dl,byte ptr[lit+edx]; mov [edi-TSPAN+7],dl
      fld [yy]
      fdiv [zz]
      mov dh,bh; mov dl,ah; and edx,$7F7F; add eax,[du]; add ebx,[dv]
      mov dh,[edx+esi]; mov dl,byte ptr [gval+1]; add [gval],ecx
      mov dl,byte ptr[lit+edx]; mov [edi-TSPAN+8],dl
      mov dh,bh; mov dl,ah; and edx,$7F7F; add eax,[du]; add ebx,[dv]
      mov dh,[edx+esi]; mov dl,byte ptr [gval+1]; add [gval],ecx
      mov dl,byte ptr[lit+edx]; mov [edi-TSPAN+9],dl
      mov dh,bh; mov dl,ah; and edx,$7F7F; add eax,[du]; add ebx,[dv]
      mov dh,[edx+esi]; mov dl,byte ptr [gval+1]; add [gval],ecx
      mov dl,byte ptr[lit+edx]; mov [edi-TSPAN+10],dl
      mov dh,bh; mov dl,ah; and edx,$7F7F; add eax,[du]; add ebx,[dv]
      mov dh,[edx+esi]; mov dl,byte ptr [gval+1]; add [gval],ecx
      mov dl,byte ptr[lit+edx]; mov [edi-TSPAN+11],dl
      mov dh,bh; mov dl,ah; and edx,$7F7F; add eax,[du]; add ebx,[dv]
      mov dh,[edx+esi]; mov dl,byte ptr [gval+1]; add [gval],ecx
      mov dl,byte ptr[lit+edx]; mov [edi-TSPAN+12],dl
      mov dh,bh; mov dl,ah; and edx,$7F7F; add eax,[du]; add ebx,[dv]
      mov dh,[edx+esi]; mov dl,byte ptr [gval+1]; add [gval],ecx
      mov dl,byte ptr[lit+edx]; mov [edi-TSPAN+13],dl
      mov dh,bh; mov dl,ah; and edx,$7F7F; add eax,[du]; add ebx,[dv]
      mov dh,[edx+esi]; mov dl,byte ptr [gval+1]; add [gval],ecx
      mov dl,byte ptr[lit+edx]; mov [edi-TSPAN+14],dl
      mov dh,bh; mov dl,ah; and edx,$7F7F; add eax,[du]; add ebx,[dv]
      mov dh,[edx+esi]; mov dl,byte ptr [gval+1]; add [gval],ecx
      mov dl,byte ptr[lit+edx]; mov [edi-TSPAN+15],dl
      add [a],TSPAN
      mov eax,[u]
      mov [uo],eax
      mov eax,[v]
      mov [vo],eax
      fistp [v]
      fistp [u]
      cmp [slen],TSPAN
      jz  @defaultspan
   @nodefaultspan:
      mov eax,[u]
      sub eax,[uo]
      cdq
      idiv [slen]
      mov [du],eax
      mov eax,[v]
      sub eax,[vo]
      cdq
      idiv [slen]
      mov [dv],eax
      jmp @endofdcalc
   @defaultspan:
      mov eax,[u]
      sub eax,[uo]
      sar eax,SPANSHFT
      mov [du],eax
      mov eax,[v]
      sub eax,[vo]
      sar eax,SPANSHFT
      mov [dv],eax
   @endofdcalc:
      jmp @dowhile
   @drawlastspan:
      mov ecx,[slen]
      mov [pixl],ecx
      mov ecx,[gdl]
      mov eax,[uo]
      mov ebx,[vo]
      mov esi,[ba]
   @ploop:
      mov dh,bh
      mov dl,ah
      and edx,$7F7F
      add eax,[du]
      add ebx,[dv]
      mov dh,[edx+esi]
      mov dl,byte ptr [gval+1]
      add [gval],ecx
      inc edi
      mov dl,byte ptr [lit+edx]
      mov [edi-1],dl
      dec [pixl]
      jnz @ploop
   @endofspan:
      mov esi,[bsp]
      fld [xxs]; fadd [_fxxd+esi]; fstp [xxs]
      fld [yys]; fadd [_fyyd+esi]; fstp [yys]
      fld [zzs]; fadd [_fzzd+esi]; fstp [zzs]
      jmp @forloop
   @endofpoly:
   end;
end;


{  _______________________________________________________________________
  /                                                                      /
 / DrawPPTextPolygon: Draw a phong shaded perspective texturemap        /
/______________________________________________________________________/ }

procedure DrawPPTextPolygon(poly: longint);
var x3,y3,x4,y4,t: single;
    z3,z4,_ylow,_yhigh: longint;
    i,_tn,oldcrds,p1,p2,pp1,pp2,pp3: longint;
    x1,y1,x2,y2,ixh,ixl: longint;
    tx,ty,tz: single;
    u,v,uo,vo: longint;
    x0,y0,z0: single;
    vx,vy,vz,ux,uy,uz: single;
    ui,uj,vi,vj,zi,zj: single;
    xxs,yys,zzs: single;                 {start per scanline}
    xxp,yyp,zzp: single;                 {delta per sub span}
    xxd,yyd,zzd: single;                 {delta per scanline}
    store,xx,yy,zz: single;
    a: dword;
    du,dv,l,olen,slen,ex: longint;
    ba,bsp: longint;
    gval,pixl,px1,px2,dpx,dlx,gdl,bck: longint;
    py1,py2,dpy: longint;
begin
   poly:=poly+1;
   pp1:=_pp[poly,1];
   pp2:=_pp[poly,2];
   pp3:=_pp[poly,4];
   {z-clipping: Make sure that the polygon is not partially behind viewer}
   _tn:=0;
   oldcrds:=_coords;
   for i:=1 to _pnv[poly] do begin
      {Determine begin and end of one of the polygon's edges}
      p1:=_pp[poly,i];
      z3:=_rz[p1];
      if i<_pnv[poly] then p2:=_pp[poly,i+1] else p2:=_pp[poly,1];
      z4:=_rz[p2];
      {Case 1: Both in, output end coord}
      if (z3>_zmin) and (z4>_zmin) then begin
         inc(_tn);
         _tmp[_tn]:=p2
      end else if (z3>_zmin) and (z4<=_zmin) then begin
      {Case 2: Going in: Output clipped coord}
         x3:=_rx[p1];
         y3:=_ry[p1];
         x4:=_rx[p2];
         y4:=_ry[p2];
         px1:=_vy[p1];                          {Phong coords: Vertex normals}
         py1:=_vz[p1];
         px2:=_vy[p2];
         py2:=_vz[p2];
         t:=(_zmin-_frz[p1])/(_frz[p2]-_frz[p1]);
         inc(_coords);
         _rx[_coords]:=x3+(x4-x3)*t;            {Clipping of polygon outline}
         _ry[_coords]:=y3+(y4-y3)*t;
         _vy[_coords]:=px1+round((px2-px1)*t);  {Clipping of phong line}
         _vz[_coords]:=py1+round((py2-py1)*t);
         _xs[_coords]:=round((_rx[_coords]*DIST*64)/_zmin)+XCENTR*64;
         _ys[_coords]:=round((_ry[_coords]*DIST*64)/_zmin)+YCENTR*64;
         inc(_tn);
         _tmp[_tn]:=_coords;
      end else if (z3<=_zmin) and (z4>_zmin) then begin
         {Case 3: Coming out: Output clipped coord and end coord}
         x3:=_rx[p1];
         y3:=_ry[p1];
         x4:=_rx[p2];
         y4:=_ry[p2];
         px1:=_vy[p1];
         py1:=_vz[p1];
         px2:=_vy[p2];
         py2:=_vz[p2];
         t:=(_zmin-_frz[p1])/(_frz[p2]-_frz[p1]);
         inc(_coords);
         _rx[_coords]:=x3+(x4-x3)*t;            {Clipping of polygon outline}
         _ry[_coords]:=y3+(y4-y3)*t;
         _vy[_coords]:=px1+round((px2-px1)*t);  {Clipping of Phong line}
         _vz[_coords]:=py1+round((py2-py1)*t);
         _xs[_coords]:=round((_rx[_coords]*DIST*64)/_zmin)+XCENTR*64;
         _ys[_coords]:=round((_ry[_coords]*DIST*64)/_zmin)+YCENTR*64;
         _tmp[_tn+1]:=_coords;
         inc(_tn,2);
         _tmp[_tn]:=p2;
      end;
   end;  {_tmp is array with coords for polygon, _tn is number of vertices}
   _coords:=oldcrds;
   ba:=_tba[_pc[poly]];
   bsp:=(_pfb[poly]-1) shl 2;
   asm
      mov ebx,[_tn]
      cmp ebx,3
      jb  @endofpoly
      dec ebx
      mov esi,dword ptr [_tmp]
      mov edi,[_ys+esi*4-4]
      mov ecx,edi
   @findloop:
      mov esi,[_tmp+ebx*4]
      mov eax,[_ys+esi*4-4]
      cmp eax,edi
      jge @nolowupdate
      mov edi,eax
   @nolowupdate:
      cmp eax,ecx
      jnge @nohiupdate
      mov ecx,eax
   @nohiupdate:
      dec ebx
      jnz @findloop
      cmp edi,[_clu]
      jge @nozeroclip
      mov edi,[_clu]
   @nozeroclip:
      cmp ecx,[_clu]
      jnge @endofpoly
      cmp edi,[_clb]
      jge @endofpoly
      cmp ecx,[_clb]
      jnge @nobotclip
      mov ecx,[_clb]
   @nobotclip:
      sar ecx,6
      sar edi,6
      cmp ecx,edi
      jbe @endofpoly
      mov [_ylow],edi
      mov [_yhigh],ecx
      sub ecx,edi
      xor eax,eax
      mov ebx,(XSIZE-1)*1024
   @clearloop:
      mov [_edgeb+edi*4],ebx
      mov [_edgee+edi*4],eax
      inc edi
      dec ecx
      jnz @clearloop
      mov [i],0
   @foriloop:
      inc [i]
      {Get x1/y1/x2/y2, wrap around for last vertex}
      mov esi,[i]
      cmp esi,[_tn]
      ja  @endofiloop
      mov edi,[_tmp+4*esi-4]
      jz  @1
      mov esi,[_tmp+4*esi]
      jmp @n
   @1: mov esi,dword ptr [_tmp]
   @n: mov eax,[_vy+4*edi-4]
      mov ebx,[_vz+4*edi-4]
      mov ecx,[_vy+4*esi-4]
      mov edx,[_vz+4*esi-4]
      mov [px1],eax
      mov [py1],ebx
      mov [px2],ecx
      mov [py2],edx
      mov edx,[_xs+4*edi-4]  {hehe not a single AGI stall}
      mov eax,[_ys+4*edi-4]
      mov ebx,[_xs+4*esi-4]
      mov ecx,[_ys+4*esi-4]
      mov [x1],edx
      mov [y1],eax
      mov [x2],ebx
      mov [y2],ecx
      cmp eax,ecx
      jz  @endofline
      jnge @noswap
      mov eax,[px1]
      mov ebx,[px2]
      mov ecx,[py1]
      mov edx,[py2]
      mov [px1],ebx
      mov [px2],eax
      mov [py1],edx
      mov [py2],ecx
      mov ecx,[y1]
      mov eax,[y2]
      mov edx,[x2]
      mov ebx,[x1]
      mov [x1],edx
      mov [x2],ebx        {registers: eax = y1  }
      mov [y1],eax                {   ebx = x2  }
      mov [y2],ecx                {   ecx = y2  }
   @noswap:                       {   edx = x1  }
      cmp eax,[_clb]
      jge @endofline
      cmp ecx,[_clu]
      jnge @endofline
      cmp eax,ecx
      jge @endofline      {if (y1<YSIZE) and (y2>0) and (y2>y1)}
      cmp eax,[_clu]
      jge @notopclipping  {if y1<0 then begin}
      sub ecx,[y1]
      xor eax,-1
      add eax,[_clu]
      sal eax,8
      cdq; idiv ecx
      mov ecx,eax
      mov esi,[_clu]
      mov [y1],esi
      mov esi,[x2]        {t:=-y1/(y2-y1)}
      sub esi,[x1]        {x1:=x1+(x2-x1)*t}
      imul esi
      sar eax,8
      add [x1],eax
      mov eax,ecx         {Clipping of phong line}
      mov esi,[px2]
      sub esi,[px1]
      imul esi
      sar eax,8
      add [px1],eax
      mov eax,ecx
      mov esi,[py2]
      sub esi,[py1]
      imul esi
      sar eax,8
      add [py1],eax
      mov ecx,[y2]
   @notopclipping:
      {Fixed point line drawing}
      sub ecx,[y1]        {ecx = delta-y}
      jz  @endofline
      mov eax,[px2]
      sub eax,[px1]
      sal eax,8
      cdq
      idiv ecx
      mov [dpx],eax       {dpx=phong delta-x, f.p. 10bit}
      mov eax,[py2]
      sub eax,[py1]
      sal eax,8
      cdq
      idiv ecx
      mov [dpy],eax       {dpy=phong delta-y, f.p. 10bit}
      mov eax,[x2]
      sub eax,[x1]        {eax = delta-x}
      sal eax,10
      cdq
      idiv ecx
      mov [dlx],eax       {dlx = delta-x/delta-y, f.p. 10bit}
      mov esi,[y1]        {pointer in outline table}
      sar esi,6
      mov ebx,[y2]
      cmp ebx,[_clb]
      jnge @drawline      {clipping to bottom of screen needed?}
      sub ebx,[_clb]
      sub ecx,ebx         {then do it}
   @drawline:             {line is OK, draw it}
      shr ecx,6
      inc ecx
      or  ecx,ecx
      jz  @endofline
      mov ebx,[x1]        {line start x for this vertex}
      sal ebx,4
      mov edi,[px1]       {phong start x}
      sal edi,2
      mov eax,[py1]       {phong start y}
      sal eax,2
   @pixels:
      mov edx,ebx
      cmp edx,[_edgeb+esi*4]
      jge @nobupdate
      mov [_edgeb+esi*4],edx
      mov [_edgxb+esi*4],edi
      mov [_edgyb+esi*4],eax
   @nobupdate:
      cmp edx,[_edgee+esi*4]
      jnge @noeupdate
      mov [_edgee+esi*4],edx
      mov [_edgxe+esi*4],edi
      mov [_edgye+esi*4],eax
   @noeupdate:
      add ebx,[dlx]     {interpolate line slope}
      add edi,[dpx]     {interpolate phong x component}
      add eax,[dpy]     {interpolate phong y component}
      inc esi
      dec ecx
      jnz @pixels
   @endofline:
      jmp @foriloop
   @endofiloop:
      mov edi,[bsp]
      cmp [_fclc+edi],0
      jnz @alreadycalced
      push edi
      call CalcBasePoly
      mov edi,[bsp]
   @alreadycalced:
      {Calculate start for first line}
      mov eax,[_ylow]
      sub eax,YCENTR
      fld [_qdiv+500*4+eax*4]
      fmul [_fuj+edi]; fld  [_fVy+edi]; fmul [_fx0+edi]; faddp
      fld  [_fVx+edi]; fmul [_fy0+edi]; fsubp; fstp [xxs]
      fld [_qdiv+500*4+eax*4]
      fmul [_fvj+edi]; fld  [_fUx+edi]; fmul [_fy0+edi]; faddp
      fld [_fUy+edi];  fmul [_fx0+edi]; fsubp; fstp [yys]
      fld [_qdiv+500*4+eax*4]
      fmul [_fzj+edi]; fld  [_fVx+edi]; fmul [_fUy+edi]; faddp
      fld [_fVy+edi];  fmul [_fUx+edi]; fsubp; fstp [zzs]
   {-----SPAN LOOP-----------}
   @forloop:
      mov edi,[_ylow]
      cmp edi,[_yhigh]
      jge @endofpoly
      inc [_ylow]
      {First check if line is completely out of screen}
      mov eax,[_edgeb+edi*4]
      cmp eax,[_clr+edi*4]
      jge @endofspan
      mov ebx,[_edgee+edi*4]
      cmp ebx,[_cll+edi*4]
      jnge @endofspan
      cmp eax,ebx
      jz  @endofspan
      {Calculate Gourad delta's before clipping}
      sub ebx,eax
      mov ecx,[_edgxb+edi*4]
      mov [px1],ecx
      mov eax,[_edgxe+edi*4]
      sub eax,ecx
      cdq
      idiv ebx
      mov [gdl],eax
      mov eax,edi
      mov ebx,[_edgeb+eax*4]
      mov ecx,[_edgee+eax*4]
      cmp ebx,[_cll+eax*4]
      jge @noleftclip

      mov edi,eax
      xor ebx,-1
      add ebx,[_cll+edi*4]
      sal ebx,8
      mov eax,ebx
      mov ebx,[_edgee+edi*4]
      sub ebx,[_edgeb+edi*4]
      cdq
      idiv ebx
      mov ebx,[_edgxe+edi*4]
      sub ebx,[_edgxb+edi*4]
      imul ebx
      sar eax,8
      add [px1],eax
      mov eax,edi

      mov ebx,[_cll+eax*4]
   @noleftclip:
      cmp ecx,[_clr+eax*4]
      jnge @norightclip
      mov ecx,[_clr+eax*4]
   @norightclip:
      mov [ex],ecx
      mov [x1],ebx
      add ebx,[buf]
      mov edi,XSIZE
      imul edi
      add ebx,eax
      mov [a],ebx
      {Calculate first span before entering while-loop}
      mov eax,[x1]
      cmp eax,[ex]
      jge @endofspan
      mov esi,[bsp]
      fld [_qdiv+(500-XCENTR)*4+eax*4]; fmul [_fzi+esi]; fadd [zzs]; fstp [zz]
      fld [_qdiv+(500-XCENTR)*4+eax*4]; fmul [_fvi+esi]; fadd [yys]; fst [yy]
      fld [_qdiv+(500-XCENTR)*4+eax*4]; fmul [_fui+esi]; fadd [xxs]; fst [xx]
      fdiv [zz]; fistp [uo]
      fdiv [zz]; fistp [vo]
      add [x1],TSPAN
      mov eax,[ex]
      cmp [x1],eax
      jnge @x1islower
   @x1isbigger:
      add eax,TSPAN
      sub eax,[x1]
      mov [slen],eax
      mov eax,[ex]
      mov [x1],eax
      jmp @endofthisshit
   @x1islower:
      mov [slen],TSPAN
   @endofthisshit:
      mov eax,[x1]
      mov esi,[bsp]
      fld [_qdiv+(500-XCENTR)*4+eax*4]; fmul [_fzi+esi]; fadd [zzs]; fstp [zz]
      fld [_qdiv+(500-XCENTR)*4+eax*4]; fmul [_fvi+esi]; fadd [yys]; fst [yy]
      fld [_qdiv+(500-XCENTR)*4+eax*4]; fmul [_fui+esi]; fadd [xxs]; fst [xx]
      fdiv [zz]; fistp [u]
      fdiv [zz]; fistp [v]
      cmp [slen],TSPAN
      jz  @defaultlength
   @nodefault:
      mov eax,[u]
      sub eax,[uo]
      cdq
      idiv [slen]
      mov [du],eax
      mov eax,[v]
      sub eax,[vo]
      cdq
      idiv [slen]
      mov [dv],eax
      jmp @endofdeltas
   @defaultlength:
      mov eax,[u]
      sub eax,[uo]
      sar eax,SPANSHFT
      mov [du],eax
      mov eax,[v]
      sub eax,[vo]
      sar eax,SPANSHFT
      mov [dv],eax
   @endofdeltas:
   {------ START OF DO-WHILE LOOP ---------}
      mov edi,[a]
      mov ecx,[px1]
      mov [gval],ecx
      mov ecx,[gdl]
   @dowhile:
      add [x1],TSPAN
      mov eax,[ex]
      add eax,TSPAN
      cmp [x1],eax
      jge @drawlastspan
   @drawaspan:
      mov eax,[ex]
      cmp eax,[x1]
      jge @nextisfull
   @nextispartial:
      add eax,TSPAN
      sub eax,[x1]
      mov [slen],eax
      mov eax,[ex]
      mov [x1],eax
      jmp @endofdelta
   @nextisfull:
      mov [slen],TSPAN
   @endofdelta:
      mov eax,[x1]
      mov esi,[bsp]
      fld [_qdiv+(500-XCENTR)*4+eax*4]; fmul [_fvi+esi]; fadd [yys]; fstp [yy]
      fld [_qdiv+(500-XCENTR)*4+eax*4]; fmul [_fui+esi]; fadd [xxs]; fst [xx]
      fld [_qdiv+(500-XCENTR)*4+eax*4]; fmul [_fzi+esi]; fadd [zzs]; fst [zz]
      fdivp
      mov eax,[uo]
      mov ebx,[vo]
      add edi,TSPAN
      xor edx,edx
      mov esi,[ba]
      mov dh,bh; mov dl,ah; and edx,$7F7F; add eax,[du]; add ebx,[dv]
      mov dh,[edx+esi]; mov dl,byte ptr [gval+1]; add [gval],ecx
      mov dl,byte ptr[lit+edx]; mov [edi-TSPAN],dl
      mov dh,bh; mov dl,ah; and edx,$7F7F; add eax,[du]; add ebx,[dv]
      mov dh,[edx+esi]; mov dl,byte ptr [gval+1]; add [gval],ecx
      mov dl,byte ptr[lit+edx]; mov [edi-TSPAN+1],dl
      mov dh,bh; mov dl,ah; and edx,$7F7F; add eax,[du]; add ebx,[dv]
      mov dh,[edx+esi]; mov dl,byte ptr [gval+1]; add [gval],ecx
      mov dl,byte ptr[lit+edx]; mov [edi-TSPAN+2],dl
      mov dh,bh; mov dl,ah; and edx,$7F7F; add eax,[du]; add ebx,[dv]
      mov dh,[edx+esi]; mov dl,byte ptr [gval+1]; add [gval],ecx
      mov dl,byte ptr[lit+edx]; mov [edi-TSPAN+3],dl
      mov dh,bh; mov dl,ah; and edx,$7F7F; add eax,[du]; add ebx,[dv]
      mov dh,[edx+esi]; mov dl,byte ptr [gval+1]; add [gval],ecx
      mov dl,byte ptr[lit+edx]; mov [edi-TSPAN+4],dl
      mov dh,bh; mov dl,ah; and edx,$7F7F; add eax,[du]; add ebx,[dv]
      mov dh,[edx+esi]; mov dl,byte ptr [gval+1]; add [gval],ecx
      mov dl,byte ptr[lit+edx]; mov [edi-TSPAN+5],dl
      mov dh,bh; mov dl,ah; and edx,$7F7F; add eax,[du]; add ebx,[dv]
      mov dh,[edx+esi]; mov dl,byte ptr [gval+1]; add [gval],ecx
      mov dl,byte ptr[lit+edx]; mov [edi-TSPAN+6],dl
      mov dh,bh; mov dl,ah; and edx,$7F7F; add eax,[du]; add ebx,[dv]
      mov dh,[edx+esi]; mov dl,byte ptr [gval+1]; add [gval],ecx
      mov dl,byte ptr[lit+edx]; mov [edi-TSPAN+7],dl
      fld [yy]
      fdiv [zz]
      mov dh,bh; mov dl,ah; and edx,$7F7F; add eax,[du]; add ebx,[dv]
      mov dh,[edx+esi]; mov dl,byte ptr [gval+1]; add [gval],ecx
      mov dl,byte ptr[lit+edx]; mov [edi-TSPAN+8],dl
      mov dh,bh; mov dl,ah; and edx,$7F7F; add eax,[du]; add ebx,[dv]
      mov dh,[edx+esi]; mov dl,byte ptr [gval+1]; add [gval],ecx
      mov dl,byte ptr[lit+edx]; mov [edi-TSPAN+9],dl
      mov dh,bh; mov dl,ah; and edx,$7F7F; add eax,[du]; add ebx,[dv]
      mov dh,[edx+esi]; mov dl,byte ptr [gval+1]; add [gval],ecx
      mov dl,byte ptr[lit+edx]; mov [edi-TSPAN+10],dl
      mov dh,bh; mov dl,ah; and edx,$7F7F; add eax,[du]; add ebx,[dv]
      mov dh,[edx+esi]; mov dl,byte ptr [gval+1]; add [gval],ecx
      mov dl,byte ptr[lit+edx]; mov [edi-TSPAN+11],dl
      mov dh,bh; mov dl,ah; and edx,$7F7F; add eax,[du]; add ebx,[dv]
      mov dh,[edx+esi]; mov dl,byte ptr [gval+1]; add [gval],ecx
      mov dl,byte ptr[lit+edx]; mov [edi-TSPAN+12],dl
      mov dh,bh; mov dl,ah; and edx,$7F7F; add eax,[du]; add ebx,[dv]
      mov dh,[edx+esi]; mov dl,byte ptr [gval+1]; add [gval],ecx
      mov dl,byte ptr[lit+edx]; mov [edi-TSPAN+13],dl
      mov dh,bh; mov dl,ah; and edx,$7F7F; add eax,[du]; add ebx,[dv]
      mov dh,[edx+esi]; mov dl,byte ptr [gval+1]; add [gval],ecx
      mov dl,byte ptr[lit+edx]; mov [edi-TSPAN+14],dl
      mov dh,bh; mov dl,ah; and edx,$7F7F; add eax,[du]; add ebx,[dv]
      mov dh,[edx+esi]; mov dl,byte ptr [gval+1]; add [gval],ecx
      mov dl,byte ptr[lit+edx]; mov [edi-TSPAN+15],dl
      add [a],TSPAN
      mov eax,[u]
      mov [uo],eax
      mov eax,[v]
      mov [vo],eax
      fistp [v]
      fistp [u]
      cmp [slen],TSPAN
      jz  @defaultspan
   @nodefaultspan:
      mov eax,[u]
      sub eax,[uo]
      cdq
      idiv [slen]
      mov [du],eax
      mov eax,[v]
      sub eax,[vo]
      cdq
      idiv [slen]
      mov [dv],eax
      jmp @endofdcalc
   @defaultspan:
      mov eax,[u]
      sub eax,[uo]
      sar eax,SPANSHFT
      mov [du],eax
      mov eax,[v]
      sub eax,[vo]
      sar eax,SPANSHFT
      mov [dv],eax
   @endofdcalc:
      jmp @dowhile
   @drawlastspan:
      mov ecx,[slen]
      mov [pixl],ecx
      mov ecx,[gdl]
      mov eax,[uo]
      mov ebx,[vo]
      mov esi,[ba]
   @ploop:
      mov dh,bh
      mov dl,ah
      and edx,$7F7F
      add eax,[du]
      add ebx,[dv]
      mov dh,[edx+esi]
      mov dl,byte ptr [gval+1]
      add [gval],ecx
      inc edi
      mov dl,byte ptr [lit+edx]
      mov [edi-1],dl
      dec [pixl]
      jnz @ploop
   @endofspan:
      mov esi,[bsp]
      fld [xxs]; fadd [_fxxd+esi]; fstp [xxs]
      fld [yys]; fadd [_fyyd+esi]; fstp [yys]
      fld [zzs]; fadd [_fzzd+esi]; fstp [zzs]
      jmp @forloop
   @endofpoly:
   end;
end;




end.
