#include "Copyright.h"
#include "PalMan.h"
#include "Texture.h"




BPalMan::BPalMan () : 
    m_head    (BNULL),
    m_table   (BNULL),
    m_tabsize (0)      {}




BPalMan::~BPalMan () {
    for (BImage * ptr=m_head; ptr!=BNULL;) {
        BImage * next = ptr->m_next;
        delete ptr;
        ptr = next;
        }
}





void BPalMan::AddTexture (BTexture & texture, const BRGB * pal) {
    BImage * image = new BImage ();
    
    image->m_texture = & texture;
    image->m_palsize = 0;
    image->m_next    = m_head;
    m_head = image;
    
    
    for (int row=0; row<texture.m_height; row++)
        for (int col=0; col<texture.m_width; col++) {
            BCOLOR pix = texture.GetPixel (col, row);
            BBOOL  used = BFALSE;

            for (int i=0; i<image->m_palsize; i++)
                if (pal[pix].r == image->m_pal[i].r &&
                    pal[pix].g == image->m_pal[i].g &&
                    pal[pix].b == image->m_pal[i].b) {
                    image->m_pixcount[i] ++;
                    texture.SetPixel (col, row, i);
                    used = BTRUE;
                    break;
                    }

            if (used)
                continue;

            image->m_pal [image->m_palsize].r = pal[pix].r;
            image->m_pal [image->m_palsize].g = pal[pix].g;
            image->m_pal [image->m_palsize].b = pal[pix].b;
            image->m_pixcount [image->m_palsize] = 0;
            image->m_palsize ++;
            }

    m_tabsize += image->m_palsize;
    return;
}





inline static BBOOL Close (int lot, const BRGB & c1, const BRGB & c2) {
    if (BABS ((int) c1.r - (int) c2.r) +
        BABS ((int) c1.g - (int) c2.g) +
        BABS ((int) c1.b - (int) c2.b) <= lot)
        return BTRUE;
    else
        return BFALSE;
}


        


void BPalMan::Optimize () {
    CreateTable   ();



    int i, j;
    int lot;
    int palsize;
    for (lot=1, palsize=0; m_tabsize-palsize>256; lot++) {
        for (i=0; i<m_tabsize; i++)
            if (! m_table[i].m_deleted)
                for (j=0; j<m_tabsize; j++)
                    if (j!=i && ! m_table[j].m_deleted)
                        if (Close (lot, m_table[i].m_color, m_table[j].m_color))
                            if (m_table[i].m_pixcount > m_table[j].m_pixcount) {
                                m_table[j].m_deleted = BTRUE;
                                palsize ++;
                                }
        }
    



    for (palsize=0, i=0; i<m_tabsize; i++)
        if (! m_table[i].m_deleted) {
            m_pal[palsize].r = m_table[i].m_color.r;
            m_pal[palsize].g = m_table[i].m_color.g;
            m_pal[palsize].b = m_table[i].m_color.b;
            palsize ++;
            }
            

    BASSERT (palsize <= 256);


    for (BImage * image=m_head; image!=BNULL; image=image->m_next) {
        BMessage ("Converting image ... ");
        BTexture * texture = image->m_texture;

        for (int row=0; row<texture->m_height; row++)
            for (int col=0; col<texture->m_width; col++) {
                BCOLOR pixel = texture->GetPixel (col, row);
                BASSERT (pixel < image->m_palsize);

                for (lot=1;;lot++) {
                    BBOOL found = BFALSE;
                    for (i=0; i<256; i++)
                        if (Close (lot, image->m_pal[pixel], m_pal[i])) {
                            texture->SetPixel (col, row, i);
                            found = BTRUE;
                            break;
                            }
                    if (found)
                        break;
                    }
                }
        }


    DestroyTable  ();
}

    



void BPalMan::CreateTable () {
    if (m_tabsize == 0)
        return;

    int index = 0;
    m_table = new BEntry [m_tabsize];
    for (BImage * image=m_head; image!=BNULL; image=image->m_next)
        for (int i=0; i<image->m_palsize; i++) {
            m_table[index].m_color.r  = image->m_pal[i].r;
            m_table[index].m_color.g  = image->m_pal[i].g;
            m_table[index].m_color.b  = image->m_pal[i].b;
            m_table[index].m_pixcount = image->m_pixcount[i];
            m_table[index].m_deleted  = BFALSE;
            index ++;
            }
    
    BASSERT (index == m_tabsize);
    return;
}
    



void BPalMan::DestroyTable () {
    delete [] m_table;
    return;
}




