UNIT STRUCT;

{This unit contains the datastructures used by the ALPHA2 engine.
 Splitting the sources this way makes it possible to move the
 polygon routines (the biggest chunk of code) to a separate unit}

INTERFACE


const XSIZE=320;
      HIRES=  0;           {Switch for 200 or 400 lines}
      YSIZE=200;
      XCENTR=XSIZE div 2;
      YCENTR=YSIZE div 2;
      BUFSIZE=XSIZE*YSIZE;
      MAXCOORD=1500;
      DIST    = 500;
      ZMIN    = 170;       {Clipping distance: Avoids z=0 spanning polygons}
      MAXSIDES=1500;       {Maximum number of 3D surfaces}
      MAXTEXT =  20;       {Maximum number of different texturemaps}
      MAXENT  =   8;       {Maximum number of objects in world}
      MXSUBENT=   4;       {Maximum number of subobjects in objects}
      MXSUBPNT= 200;       {Maximum number of coordinates in object}
      MAXTHING=   8;       {Maximum total number of (sub)objects}
      PMSIZE  =1500;       {Size of suspiciously big polygons (debugging)}
      TSPAN   =  16;       {Spanlength for perspective texture mapper}
      SPANSHFT=   4;       {Bitshift for spanlength}
      SPRSIZE =  50;       {Size of the scaled sprites}
      MAXSECTR=   8;       {Maximum number of sectors (portal-separated)}
      MAXSECSR= 250;       {Maximum number of polygons in each sector}
      RECALC  =   0;       {Enable/disable recalculation of mixing tables}
      MAXFAKE = 100;       {Maximum number of fake/base polygons}
      SPEEDTST=   0;       {Switch for palette speed measuring}
      DEBUG   =   1;       {Output status to debugging file}

      {Polygon types}
      TEXTURED=1;     {Standard perspective correct texturemapped polygon}
      PLAIN=2;        {Single colored polygon}
      SMOOTHED=3;     {Smoothed texturemapped polygon}
      SHADETEX=4;     {Color mixed textured polygon}
      GOURADTX=5;     {Gourad shaded textured polygon}
      PHONGTX=6;      {Phong shaded textured polygon}
      GOURAD=7;       {Gourad shaded polygon}
      PHONG=8;        {Phong shaded polygon}
      BUMP=9;         {Bump mapped polygon}
      BUMPTX=10;      {Bump mapped textured polygon}
      MIRROR=11;      {Mirror surface (non-recursive!)}
      SUN=12;         {Lens flare causing object}
      TRANSP=13;      {Transparent plain polygon}
      TRTEXT=14;      {Transparent perspective correct texture}
      ENVMAP=15;      {Environment-mapped polygon}
      FASTTEXT=16;    {Old fashioned texturemapper}
      FASTTRTX=17;    {Transparent low-quality texturemapper}
      SPRITE=18;      {3D sprite (single-coord polygon)}
      PORTAL=19;      {Portal polygon}
      REVPORTAL=20;   {Internal: Reversed Portal Polygon)}

type matrix = array [0..3,0..3] of single;
     vector = array [0..2] of single;

type sortree = record
   key: dword;                       {polygonnumber}
   lbr: dword;                       {left branch pointer}
   rbr: dword;                       {right branch pointer}
   dummy: dword;                     {aligning}
end;

type global  = record
   nobj: longint;                    {number of objects in world}
   npnt: longint;                    {number of static coordinates in world}
   obj: array [1..MAXENT] of longint; {non-stationary objects in world}
   pnt: array [1..MAXCOORD] of longint; {stationary coordinates in world}
end;

type objtree = record
   pnt: array[1..MXSUBPNT] of longint; {coordinates in this object}
   obj: array[1..MXSUBENT] of longint; {pointers to sub-objects in this object}
   rx,ry,rz: longint;                {relative rotation of this object}
   ox,oy,oz: single;                 {relative position of this object}
   nobj: longint;                    {number of sub-ojects in this object}
   npnt: longint;                    {number of coordinates in this object}
   parent: longint;                  {parent of this object}
end;

var _px,_py,_pz: array[1..MAXCOORD] of single;   {Floating point 3d coords}
    _guard1: longint;                            {data guardian}
    _nx,_ny,_nz: array[1..MAXCOORD] of single;   {vertex normals (phong)}
    _rx,_ry,_frz: array[1..MAXCOORD] of single;  {3d coords (rotated)}
    _rz: array[1..MAXCOORD] of longint;          {better use int here}
    _guard2: longint;                            {data guardian}
    _vx,_vy,_vz: array[1..MAXCOORD] of longint;  {rotated vertex normals}
    _sn,_cs: array[0..2880] of single;           {sin/cos tables (float)}
    _cll,_clr: array[0..YSIZE+4] of longint;     {custom clipping window}
    _cbl,_cbr: longint;                          {max. and min. x}
    _guard3: longint;                            {data guardian}
    _clu,_clb: longint;                          {top/bottom of custom window}
    _edgee,_edgeb: array[0..YSIZE+4] of longint; {polygon edge tables}
    _edgxe,_edgxb: array[0..YSIZE+4] of longint; {additional edge data}
    _edgye,_edgyb: array[0..YSIZE+4] of longint; {additional edge data}
    _guard4: longint;                            {data guardian}
    thing: array[1..MAXTHING] of objtree;        {non-stationary entities}
    _sec: array[1..MAXSECTR,1..MAXSECSR] of longint; {portal sectors}
    _sx1,_sy1,_sz1,_sx2,_sy2,_sz2: array[1..MAXSECTR] of longint; {bounding}
    _nsr: array[1..MAXSECTR] of longint;         {number of polies in sector}
    _xs,_ys: array[1..MAXCOORD] of longint;      {2d coordinates and colors}
    _pn: array[1..MAXSIDES,1..3] of single;      {polygon normal (for phong)}
    _pp: array[1..MAXSIDES,1..6] of longint;     {polygon vertices}
    _pt: array[1..MAXSIDES] of longint;          {polygon types}
    _dr: array[1..MAXSIDES] of longint;          {drawn flag}
    _pm: array[1..MAXSIDES] of longint;          {texturemap number for poly}
    _pd: array[1..MAXSIDES] of longint;          {double sided or not}
    _pnv: array[1..MAXSIDES] of longint;         {number of vertices in poly}
    _tmp,_ttp: array[1..8] of longint;           {polygon drawing temp. array}
    _sector: longint;                            {current sector}
    _polies: longint;                            {number of polygons}
    _lsd: longint;                               {last defined sector}
    _obj: longint;                               {current object (definition)}
    _lobj: longint;                              {number of last object}
    _lrx,_lry,_lrz: longint;                     {background light vector}
    _pc: array[1..MAXSIDES] of longint;          {polygon colors}
    _tree: array[1..2*MAXSIDES] of sortree;      {sorting tree}
    _guard5: longint;                            {data guardian}
    _qdiv: array[0..1000] of single;             {div table for pmapper}
    _tba: array[1..MAXTEXT] of dword;            {texturemap base adresses e.a}
    _i: longint;                                 {general purpose}
    tmp1,tmp2,tmpl: matrix;                      {3D matrices}
    _left: longint;                              {internal (tree processing)}
    scrn: dword;                                 {screen base pointer}
    _warn: dword;                                {number of warnings}
    _coords: dword;                              {current number of coords}
    _zmin: single;                               {z-clipping distance}
    _sof1,_sof2,_sad1,_sad2: dword;              {hi-res screen swapping}
    env,mix,pmap,tex,buf: dword;                 {pointers to illegal mem}

    _base: longint;                              {number of fake/base polies}
    _fvx,_fvy,_fvz: array[1..MAXFAKE] of single; {fake/base data block}
    _fux,_fuy,_fuz: array[1..MAXFAKE] of single;
    _fx0,_fy0,_fz0: array[1..MAXFAKE] of single;
    _fui,_fuj,_fvi,_fvj,_fzi,_fzj: array[1..MAXFAKE] of single;
    _fxxd,_fyyd,_fzzd: array[1..MAXFAKE] of single;
    _fxxp,_fyyp,_fzzp: array[1..MAXFAKE] of single;
    _fp1,_fp2,_fp3: array[1..MAXFAKE] of longint;{coords for fake/base poly}
    _fclc: array[1..MAXFAKE] of longint;         {flag (1=processed)}
    _pfb: array[1..MAXSIDES] of longint;         {poly<>base connection}
    _fa,_fb,_fc,_fd: array[1..MAXSIDES] of single;

    lit: array[0..65535] of byte;
    world: global;                               {world record}
    wrn: array[1..10] of string;                 {Warning-messages}
    _r,_g,_b: array[0..255] of byte;             {color palette}
    _hh,_mm,_ss,_s100: word;                     {timing variables}
    _lfps: single;
    fl: text;                                    {status file (debugging)}
    oldwinl,oldwinr: array[1..8,0..YSIZE+4] of longint;
    _depth: longint;                             {portal depth}


IMPLEMENTATION

end.
