#include "Copyright.h"
#include "Walker.h"
#include "Line.h"
#include "Sector.h"






BWalker::BWalker (const BPoint  & pos,
                  const BVector & normal,
                  const BREAL     height,
                  const BSector * sector) :
    m_pos     (pos),
    m_normal  (normal),
    m_height  (height),
    m_sector  (sector)   {
    
    BASSERT (BNEAR (normal.Length (), 1.0f));
    return;
}




BBOOL BWalker::Walk (const BPoint & newpos) {
    return SubWalk (newpos-m_pos, m_sector, 0.0f, BNULL);
}





BBOOL BWalker::SubWalk (const BVector & vector, 
                        const BSector * sector,
                        BREAL minbeta, const BLine * excline) {

    BREAL beta = BLIMMAX;
    BREAL alpha;
    BFACE face;
    const BLine * line = BNULL;
    
    
    for (int i=0; i<sector->m_linecount; i++) {
        const BLine * curline = sector->Line (i);
        if (curline != excline) {
            BREAL curalpha;
            BREAL curbeta;
            BFACE curface  = sector->Face (i);
            BREAL iproduct = vector.y * curline->m_vector.x - 
                             vector.x * curline->m_vector.y ;

            if ((curface == BFRONT && iproduct <= 0) ||
                (curface == BBACK  && iproduct >= 0))
                continue;

            
            curline->CrossTest (m_pos, vector, iproduct, curalpha, curbeta);
            if (curalpha >= 0.0f && curalpha <= 1.0f && curbeta >= minbeta)
                if (curbeta < beta) {
                    line  = curline;
                    face  = curface;
                    alpha = curalpha;
                    beta  = curbeta;
                    }
            }
        }


    if (line == BNULL || beta > 1.0f) {
        m_sector = sector;
        m_pos    = m_pos + vector;
        return BTRUE;
        }


    BREAL top, bottom;
    const BSector * nextsec = line->ThruTest (face, top, bottom);
    
    if (nextsec == BNULL ||
        top-bottom < m_height ||
        bottom-sector->m_floorlevel > 3.0) {
        m_sector = sector;
        m_pos    = *line->m_start + line->m_vector * alpha;
        return BFALSE;
        }
    else
        return SubWalk (vector, nextsec, beta, line);
}


        



void BWalker::Move (const BPoint & newpos, const BSector * newsec) {
    m_pos    = newpos;
    m_sector = newsec;
    return;
}



void BWalker::View (const BVector & newnormal) {
    m_normal = newnormal;
    return;
}


